/* Generated by Cython 3.2.4 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/home/runner/work/_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/eigency/eigency.h",
            "/home/runner/work/_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/eigency/eigency_cpp.h",
            "/home/runner/work/_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/_core/include/numpy/arrayobject.h",
            "/home/runner/work/_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/_core/include/numpy/arrayscalars.h",
            "/home/runner/work/_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/_core/include/numpy/ndarrayobject.h",
            "/home/runner/work/_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/_core/include/numpy/ndarraytypes.h",
            "/home/runner/work/_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/_core/include/numpy/ufuncobject.h",
            "NonlinearTMM/src/cpp/Common.h",
            "NonlinearTMM/src/cpp/Material.h",
            "NonlinearTMM/src/cpp/NonlinearLayer.h",
            "NonlinearTMM/src/cpp/NonlinearTMM.h",
            "NonlinearTMM/src/cpp/SecondOrderNLTMM.h",
            "NonlinearTMM/src/cpp/Waves.h"
        ],
        "extra_compile_args": [
            "-std=c++17",
            "-fopenmp",
            "-msse3"
        ],
        "extra_link_args": [
            "-lgomp",
            "-fopenmp"
        ],
        "include_dirs": [
            "/home/runner/work/_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/eigency",
            "/home/runner/work/_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/_core/include",
            "NonlinearTMM/src/cpp",
            "third_party/eigen"
        ],
        "language": "c++",
        "name": "NonlinearTMM._SecondOrderNLTMMCython",
        "sources": [
            "NonlinearTMM/src/SecondOrderNLTMM.pyx",
            "NonlinearTMM/src/cpp/Material.cpp",
            "NonlinearTMM/src/cpp/NonlinearLayer.cpp",
            "NonlinearTMM/src/cpp/SecondOrderNLTMM.cpp",
            "NonlinearTMM/src/cpp/Common.cpp",
            "NonlinearTMM/src/cpp/NonlinearTMM.cpp",
            "NonlinearTMM/src/cpp/Waves.cpp"
        ]
    },
    "module_name": "NonlinearTMM._SecondOrderNLTMMCython"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API)
  #if !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
  #endif
#elif defined(CYTHON_LIMITED_API)
  #ifdef _MSC_VER
  #pragma message ("Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.")
  #else
  #warning Limited API usage is enabled with 'CYTHON_LIMITED_API' but 'Py_LIMITED_API' does not define a Python target version. Consider setting 'Py_LIMITED_API' instead.
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_2_4"
#define CYTHON_HEX_VERSION 0x030204F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #ifndef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 1
  #endif
  #undef CYTHON_IMMORTAL_CONSTANTS
  #define CYTHON_IMMORTAL_CONSTANTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
  #if defined(CYTHON_IMMORTAL_CONSTANTS) && PY_VERSION_HEX < 0x030C0000
    #undef CYTHON_IMMORTAL_CONSTANTS
    #define CYTHON_IMMORTAL_CONSTANTS 0  // definitely won't work
  #elif !defined(CYTHON_IMMORTAL_CONSTANTS)
    #define CYTHON_IMMORTAL_CONSTANTS (PY_VERSION_HEX >= 0x030C0000 && !CYTHON_USE_MODULE_STATE && CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_COMPRESS_STRINGS
  #define CYTHON_COMPRESS_STRINGS 1
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL)
#endif
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#include <stdint.h>
typedef uintptr_t  __pyx_uintptr_t;
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CppInitCode */
#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(const U& other) const { return *ptr == other; }
    template<typename U> bool operator !=(const U& other) const { return *ptr != other; }
    template<typename U> bool operator==(const __Pyx_FakeReference<U>& other) const { return *ptr == *other.ptr; }
    template<typename U> bool operator!=(const __Pyx_FakeReference<U>& other) const { return *ptr != *other.ptr; }
  private:
    T *ptr;
};

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef Py_TPFLAGS_IMMUTABLETYPE
  #define Py_TPFLAGS_IMMUTABLETYPE (1UL << 8)
#endif
#ifndef Py_TPFLAGS_DISALLOW_INSTANTIATION
  #define Py_TPFLAGS_DISALLOW_INSTANTIATION (1UL << 7)
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if PY_VERSION_HEX < 0x03090000 || (CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000)
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) GraalPyFrame_SetLineNumber((frame), (lineno))
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && PyUnstable_Object_IsUniquelyReferenced(obj)) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#elif CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
enum __Pyx_ReferenceSharing {
  __Pyx_ReferenceSharing_DefinitelyUnique, // We created it so we know it's unshared - no need to check
  __Pyx_ReferenceSharing_OwnStrongReference,
  __Pyx_ReferenceSharing_FunctionArgument,
  __Pyx_ReferenceSharing_SharedReference, // Never trust it to be unshared because it's a global or similar
};
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && PY_VERSION_HEX >= 0x030E0000
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing)\
    (sharing == __Pyx_ReferenceSharing_DefinitelyUnique ? 1 :\
      (sharing == __Pyx_ReferenceSharing_FunctionArgument ? PyUnstable_Object_IsUniqueReferencedTemporary(o) :\
      (sharing == __Pyx_ReferenceSharing_OwnStrongReference ? PyUnstable_Object_IsUniquelyReferenced(o) : 0)))
#elif (CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING) || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)sharing), Py_REFCNT(o) == 1)
#else
#define __Pyx_IS_UNIQUELY_REFERENCED(o, sharing) (((void)o), ((void)sharing), 0)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS && !CYTHON_COMPILING_IN_LIMITED_API && CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) (__Pyx_IS_UNIQUELY_REFERENCED(o, unsafe_shared) ?\
    __Pyx_NewRef(PyList_GET_ITEM(o, i)) : __Pyx_PyList_GetItemRef(o, i))
#else
  #define __Pyx_PyList_GetItemRefFast(o, i, unsafe_shared) __Pyx_PyList_GetItemRef(o, i)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
    #define __PYX_EXTERN_C extern "C++"
#endif

#define __PYX_HAVE__NonlinearTMM___SecondOrderNLTMMCython
#define __PYX_HAVE_API__NonlinearTMM___SecondOrderNLTMMCython
/* Early includes */
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <memory>
#include <string.h>
#include <stdio.h>

    /* Using NumPy API declarations from "numpy/__init__.cython-30.pxd" */
    
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"
#include "eigency_cpp.h"
#include "eigency.h"
#include <utility>

    #if __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600)
    // move should be defined for these versions of MSVC, but __cplusplus isn't set usefully
    #include <type_traits>

    namespace cython_std {
    template <typename T> typename std::remove_reference<T>::type&& move(T& t) noexcept { return std::move(t); }
    template <typename T> typename std::remove_reference<T>::type&& move(T&& t) noexcept { return std::move(t); }
    }

    #endif
    
#include "Common.h"
#include "Material.h"
#include "Waves.h"
#include "NonlinearLayer.h"
#include "NonlinearTMM.h"
#include "SecondOrderNLTMM.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifdef CYTHON_FREETHREADING_COMPATIBLE
#if CYTHON_FREETHREADING_COMPATIBLE
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_NOT_USED
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#else
#define __Pyx_FREETHREADING_COMPATIBLE Py_MOD_GIL_USED
#endif
#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif (defined(_Complex_I) && !defined(_MSC_VER)) || ((defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) && !defined(__STDC_NO_COMPLEX__) && !defined(_MSC_VER))
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "NonlinearTMM/src/SecondOrderNLTMM.pyx",
  "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd",
  "<stringsource>",
  "cpython/type.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Atomics.proto (used by UnpackUnboundCMethod) */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #define __pyx_atomic_pointer_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_nonatomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_nonatomic_ptr_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #define __pyx_nonatomic_ptr_type void*
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    static CYTHON_INLINE int __pyx_atomic_pointer_cmp_exchange(__pyx_atomic_ptr_type* value, __pyx_nonatomic_ptr_type* expected, __pyx_nonatomic_ptr_type desired) {
        __pyx_atomic_ptr_type old = _InterlockedCompareExchangePointer(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif

/* CriticalSectionsDefinition.proto (used by CriticalSections) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_End(cs)
#define __Pyx_PyCriticalSection2_End(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_End PyCriticalSection_End
#define __Pyx_PyCriticalSection2_End PyCriticalSection2_End
#endif

/* CriticalSections.proto (used by ParseKeywordsImpl) */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection_Begin(cs, arg) (void)(cs)
#define __Pyx_PyCriticalSection2_Begin(cs, arg1, arg2) (void)(cs)
#else
#define __Pyx_PyCriticalSection_Begin PyCriticalSection_Begin
#define __Pyx_PyCriticalSection2_Begin PyCriticalSection2_Begin
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* BufferFormatStructs.proto (used by BufferFormatCheck) */
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  const struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  const __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  const __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;

/* IncludeStructmemberH.proto (used by FixUpExtensionType) */
#include <structmember.h>

/* #### Code section: numeric_typedefs ### */

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":744
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
*/
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":745
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
*/
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":746
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * 
*/
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":747
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_uint8      uint8_t
*/
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":749
 * ctypedef npy_int64      int64_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
*/
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":750
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
*/
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":751
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * 
*/
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":752
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_float32    float32_t
*/
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":754
 * ctypedef npy_uint64     uint64_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
*/
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":755
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
*/
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":762
 * ctypedef double complex complex128_t
 * 
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
*/
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":763
 * 
 * ctypedef npy_longlong   longlong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
*/
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":765
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
*/
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":766
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
*/
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":768
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
*/
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":769
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
*/
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":770
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef float complex       cfloat_t
*/
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "eigency/core.pxd":3
 * cimport numpy as np
 * 
 * ctypedef unsigned long u_long             # <<<<<<<<<<<<<<
 * ctypedef long long long_long;
 * ctypedef unsigned long long u_long_long;
*/
typedef unsigned long __pyx_t_7eigency_4core_u_long;

/* "eigency/core.pxd":4
 * 
 * ctypedef unsigned long u_long
 * ctypedef long long long_long;             # <<<<<<<<<<<<<<
 * ctypedef unsigned long long u_long_long;
 * ctypedef long double long_double;
*/
typedef PY_LONG_LONG __pyx_t_7eigency_4core_long_long;

/* "eigency/core.pxd":5
 * ctypedef unsigned long u_long
 * ctypedef long long long_long;
 * ctypedef unsigned long long u_long_long;             # <<<<<<<<<<<<<<
 * ctypedef long double long_double;
 * ctypedef long double complex c_long_double;
*/
typedef unsigned PY_LONG_LONG __pyx_t_7eigency_4core_u_long_long;

/* "eigency/core.pxd":6
 * ctypedef long long long_long;
 * ctypedef unsigned long long u_long_long;
 * ctypedef long double long_double;             # <<<<<<<<<<<<<<
 * ctypedef long double complex c_long_double;
 * 
*/
typedef long double __pyx_t_7eigency_4core_long_double;
/* #### Code section: complex_type_declarations ### */
/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< long double > __pyx_t_long_double_complex;
  #else
    typedef long double _Complex __pyx_t_long_double_complex;
  #endif
#else
    typedef struct { long double real, imag; } __pyx_t_long_double_complex;
#endif
static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double, long double);

/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor;
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material;
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave;
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities;
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM;
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM;
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ;
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX;
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave;
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer;
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM;
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities;
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM;
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM;
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM;
struct __pyx_opt_args_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM__Init;
struct __pyx_opt_args_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM__Init;

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":632
 *             del self._thisptr
 * 
 *     cdef _Init(self, SweepResultNonlinearTMMCpp *ptr, bool needDealloc = True, object parent = None):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         self._needDealloc = needDealloc
*/
struct __pyx_opt_args_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM__Init {
  int __pyx_n;
  bool needDealloc;
  PyObject *parent;
};

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":682
 *             del self._thisptr
 * 
 *     cdef _Init(self, WaveSweepResultNonlinearTMMCpp *ptr, bool needDealloc = True, object parent = None):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         self._needDealloc = needDealloc
*/
struct __pyx_opt_args_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM__Init {
  int __pyx_n;
  bool needDealloc;
  PyObject *parent;
};

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":91
 * #===============================================================================
 * 
 * cdef class _Chi2Tensor:             # <<<<<<<<<<<<<<
 *     """_Chi2Tensor()
 * 
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor {
  PyObject_HEAD
  struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_vtab;
  TMM::Chi2Tensor *_thisptr;
  PyObject *_parent;
};


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":212
 * #===============================================================================
 * 
 * cdef class Material:             # <<<<<<<<<<<<<<
 *     """Material(wls, ns)
 * 
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material {
  PyObject_HEAD
  TMM::Material *_thisptr;
  PyObject *chi2;
};


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":294
 * #===============================================================================
 * 
 * cdef class _Wave:             # <<<<<<<<<<<<<<
 *     """_Wave()
 * 
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave {
  PyObject_HEAD
  struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_vtab;
  TMM::Wave *_thisptr;
  bool _needsDealloc;
};


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":556
 * #===============================================================================
 * 
 * cdef class _Intensities:             # <<<<<<<<<<<<<<
 *     """
 *     Helper class to store intensities for :class:`NonlinearTMM`.
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities {
  PyObject_HEAD
  struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_vtab;
  __pyx_t_double_complex inc;
  __pyx_t_double_complex r;
  __pyx_t_double_complex t;
  double I;
  double R;
  double T;
};


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":594
 * #===============================================================================
 * 
 * cdef class _SweepResultNonlinearTMM:             # <<<<<<<<<<<<<<
 *     """
 *     Helper class to store the result of the ``NonlinearTMM.Sweep`` method.
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM {
  PyObject_HEAD
  struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_vtab;
  TMM::SweepResultNonlinearTMM *_thisptr;
  PyArrayObject *inc;
  PyArrayObject *r;
  PyArrayObject *t;
  PyArrayObject *Ii;
  PyArrayObject *Ir;
  PyArrayObject *It;
  PyArrayObject *Ia;
  PyArrayObject *enh;
  bool _needDealloc;
  PyObject *_parent;
};


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":651
 * #===============================================================================
 * 
 * cdef class _WaveSweepResultNonlinearTMM:             # <<<<<<<<<<<<<<
 *     """
 *     The helper class to store the results of `WaveSweep`.
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM {
  PyObject_HEAD
  struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_vtab;
  TMM::WaveSweepResultNonlinearTMM *_thisptr;
  PyArrayObject *Pi;
  PyArrayObject *Pr;
  PyArrayObject *Pt;
  PyArrayObject *enh;
  PyArrayObject *beamArea;
  bool _needDealloc;
  PyObject *_parent;
};


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":697
 * #===============================================================================
 * 
 * cdef class _FieldsZ:             # <<<<<<<<<<<<<<
 *     """
 *     Helper class to store the results of `GetFields`.
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ {
  PyObject_HEAD
  struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_vtab;
  TMM::FieldsZ *_thisptr;
  PyArrayObject *E;
  PyArrayObject *H;
};


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":732
 * #===============================================================================
 * 
 * cdef class _FieldsZX:             # <<<<<<<<<<<<<<
 *     """
 *     Helper class to store the result of `GetFields2D` and `WaveGetFields2D`.
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX {
  PyObject_HEAD
  struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_vtab;
  TMM::FieldsZX *_thisptr;
  PyArrayObject *Ex;
  PyArrayObject *Ey;
  PyArrayObject *Ez;
  PyArrayObject *Hx;
  PyArrayObject *Hy;
  PyArrayObject *Hz;
};


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":820
 * #===============================================================================
 * 
 * cdef class _HomogeneousWave:             # <<<<<<<<<<<<<<
 *     """
 *     Helper class for `NonlinearLayer` to describe the parameters of the
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave {
  PyObject_HEAD
  struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_vtab;
  TMM::HomogeneousWave *_thisptr;
};


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":874
 * #===============================================================================
 * 
 * cdef class _NonlinearLayer:             # <<<<<<<<<<<<<<
 *     """
 *         Helper class for layer-specific data and methods.
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer {
  PyObject_HEAD
  struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_vtab;
  TMM::NonlinearLayer *_thisptr;
  PyObject *hw;
  PyObject *_parent;
  int _layerNr;
};


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":964
 * #===============================================================================
 * 
 * cdef class NonlinearTMM:             # <<<<<<<<<<<<<<
 *     """NonlinearTMM()
 * 
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM {
  PyObject_HEAD
  struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_vtab;
  TMM::NonlinearTMM *_thisptr;
  PyObject *materialsCache;
  PyObject *layers;
  bool _needDealloc;
  PyObject *_parent;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *wave;
};


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1513
 * #===============================================================================
 * 
 * cdef class _SecondOrderNLIntensities:             # <<<<<<<<<<<<<<
 *     """_SecondOrderNLIntensities()
 *     This is a helper class for `SecondOrderNLTMM`. It stores the results of
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities {
  PyObject_HEAD
  struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_vtab;
  PyObject *P1;
  PyObject *P2;
  PyObject *Gen;
};


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1554
 * #===============================================================================
 * 
 * cdef class _SweepResultSecondOrderNLTMM:             # <<<<<<<<<<<<<<
 *     """_SweepResultSecondOrderNLTMM()
 *     Helper class for `SecondOrderNLTMM`. Stores the results of `Sweep` method.
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM {
  PyObject_HEAD
  struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_vtab;
  TMM::SweepResultSecondOrderNLTMM *_thisptr;
  PyObject *P1;
  PyObject *P2;
  PyObject *Gen;
  PyArrayObject *wlsGen;
  PyArrayObject *betasGen;
};


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1605
 * #===============================================================================
 * 
 * cdef class _WaveSweepResultSecondOrderNLTMM:             # <<<<<<<<<<<<<<
 *     """_WaveSweepResultSecondOrderNLTMM()
 *     Helper class for `SecondOrderNLTMM`. Stores the results of `WaveSweep`
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM {
  PyObject_HEAD
  struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_vtab;
  TMM::WaveSweepResultSecondOrderNLTMM *_thisptr;
  PyObject *P1;
  PyObject *P2;
  PyObject *Gen;
  PyArrayObject *wlsGen;
  PyArrayObject *betasGen;
};


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1658
 * #===============================================================================
 * 
 * cdef class SecondOrderNLTMM:             # <<<<<<<<<<<<<<
 *     """SecondOrderNLTMM(mode)
 * 
*/
struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM {
  PyObject_HEAD
  TMM::SecondOrderNLTMM *_thisptr;
  PyObject *materialsCache;
  PyObject *P1;
  PyObject *P2;
  PyObject *Gen;
};



/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":91
 * #===============================================================================
 * 
 * cdef class _Chi2Tensor:             # <<<<<<<<<<<<<<
 *     """_Chi2Tensor()
 * 
*/

struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor {
  PyObject *(*_Init)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *, TMM::Chi2Tensor *, PyObject *);
};
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor;


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":294
 * #===============================================================================
 * 
 * cdef class _Wave:             # <<<<<<<<<<<<<<
 *     """_Wave()
 * 
*/

struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave {
  PyObject *(*_Init)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *, TMM::Wave *);
};
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave;


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":556
 * #===============================================================================
 * 
 * cdef class _Intensities:             # <<<<<<<<<<<<<<
 *     """
 *     Helper class to store intensities for :class:`NonlinearTMM`.
*/

struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities {
  PyObject *(*_Init)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *, TMM::Intensities *);
};
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities;


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":594
 * #===============================================================================
 * 
 * cdef class _SweepResultNonlinearTMM:             # <<<<<<<<<<<<<<
 *     """
 *     Helper class to store the result of the ``NonlinearTMM.Sweep`` method.
*/

struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM {
  PyObject *(*_Init)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *, TMM::SweepResultNonlinearTMM *, struct __pyx_opt_args_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM__Init *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM;


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":651
 * #===============================================================================
 * 
 * cdef class _WaveSweepResultNonlinearTMM:             # <<<<<<<<<<<<<<
 *     """
 *     The helper class to store the results of `WaveSweep`.
*/

struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM {
  PyObject *(*_Init)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *, TMM::WaveSweepResultNonlinearTMM *, struct __pyx_opt_args_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM__Init *__pyx_optional_args);
};
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM;


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":697
 * #===============================================================================
 * 
 * cdef class _FieldsZ:             # <<<<<<<<<<<<<<
 *     """
 *     Helper class to store the results of `GetFields`.
*/

struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ {
  PyObject *(*_Init)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *, TMM::FieldsZ *);
};
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ;


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":732
 * #===============================================================================
 * 
 * cdef class _FieldsZX:             # <<<<<<<<<<<<<<
 *     """
 *     Helper class to store the result of `GetFields2D` and `WaveGetFields2D`.
*/

struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX {
  PyObject *(*_Init)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *, TMM::FieldsZX *);
};
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX;


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":820
 * #===============================================================================
 * 
 * cdef class _HomogeneousWave:             # <<<<<<<<<<<<<<
 *     """
 *     Helper class for `NonlinearLayer` to describe the parameters of the
*/

struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave {
  PyObject *(*_Init)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *, TMM::HomogeneousWave *);
};
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave;


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":874
 * #===============================================================================
 * 
 * cdef class _NonlinearLayer:             # <<<<<<<<<<<<<<
 *     """
 *         Helper class for layer-specific data and methods.
*/

struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer {
  PyObject *(*_Init)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *, TMM::NonlinearLayer *, int, PyObject *);
};
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer;


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":964
 * #===============================================================================
 * 
 * cdef class NonlinearTMM:             # <<<<<<<<<<<<<<
 *     """NonlinearTMM()
 * 
*/

struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM {
  PyObject *(*_Init)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *, TMM::NonlinearTMM *);
};
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM;


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1513
 * #===============================================================================
 * 
 * cdef class _SecondOrderNLIntensities:             # <<<<<<<<<<<<<<
 *     """_SecondOrderNLIntensities()
 *     This is a helper class for `SecondOrderNLTMM`. It stores the results of
*/

struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities {
  PyObject *(*_Init)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *, TMM::SecondOrderNLIntensities *);
};
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities;


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1554
 * #===============================================================================
 * 
 * cdef class _SweepResultSecondOrderNLTMM:             # <<<<<<<<<<<<<<
 *     """_SweepResultSecondOrderNLTMM()
 *     Helper class for `SecondOrderNLTMM`. Stores the results of `Sweep` method.
*/

struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM {
  PyObject *(*_Init)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *, TMM::SweepResultSecondOrderNLTMM *);
};
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM;


/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1605
 * #===============================================================================
 * 
 * cdef class _WaveSweepResultSecondOrderNLTMM:             # <<<<<<<<<<<<<<
 *     """_WaveSweepResultSecondOrderNLTMM()
 *     Helper class for `SecondOrderNLTMM`. Stores the results of `WaveSweep`
*/

struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM {
  PyObject *(*_Init)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *, TMM::WaveSweepResultSecondOrderNLTMM *);
};
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* GetTopmostException.proto (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto (used by SaveResetException) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyImportError_Check.proto */
#define __Pyx_PyExc_ImportError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_ImportError)

/* PyObjectCall.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyErrFetchRestore.proto (used by RaiseException) */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.export */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* IncludeStringH.proto (used by BytesEquals) */
#include <string.h>

/* BytesEquals.proto (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* PyNotImplementedError_Check.proto */
#define __Pyx_PyExc_NotImplementedError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_NotImplementedError)

/* PyValueError_Check.proto */
#define __Pyx_PyExc_ValueError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_ValueError)

/* DictGetItem.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* TupleAndListFromArray.proto (used by fastcall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* py_dict_items.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* CallCFunction.proto (used by CallUnboundCMethod0) */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* PyObjectCallOneArg.proto (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectGetAttrStr.proto (used by UnpackUnboundCMethod) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* UnpackUnboundCMethod.proto (used by CallUnboundCMethod0) */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod0.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* OwnedDictNext.proto (used by RejectKeywords) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue);
#else
CYTHON_INLINE
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue);
#endif

/* RejectKeywords.export */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds);

/* KeywordStringCheck.proto */
static CYTHON_INLINE int __Pyx_CheckKeywordStrings(const char* function_name, PyObject *kw);

/* py_dict_keys.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d);

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck, unsafe_shared) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, unsafe_shared) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck, int unsafe_shared);

/* py_dict_pop.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* PyObjectCall2Args.proto (used by CallUnboundCMethod1) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* CallUnboundCMethod1.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectGetAttrStrNoError.proto (used by GetBuiltinName) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __Pyx_XNewRef(__pyx_dict_cached_value);\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectVectorCallKwBuilder.proto */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyTypeError_Check.proto */
#define __Pyx_PyExc_TypeError_Check(obj)  __Pyx_TypeCheck(obj, PyExc_TypeError)

/* RaiseDoubleKeywords.proto (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywordsImpl.export */
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name
);
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* ArgTypeTestFunc.export */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))

/* IsLittleEndian.proto (used by BufferFormatCheck) */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto (used by BufferGetAndValidate) */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              const __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
  const __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* IterFinish.proto (used by dict_iter) */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectCallNoArg.proto (used by PyObjectCallMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectGetMethod.proto (used by PyObjectCallMethod0) */
#if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);
#endif

/* PyObjectCallMethod0.proto (used by dict_iter) */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto (used by UnpackTuple2) */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto (used by UnpackItemEndCheck) */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto (used by UnpackTuple2) */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto (used by UnpackTupleError) */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto (used by UnpackTuple2) */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto (used by dict_iter) */
static CYTHON_INLINE int __Pyx_unpack_tuple2(
    PyObject* tuple, PyObject** value1, PyObject** value2, int is_tuple, int has_known_size, int decref_tuple);
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* MoveIfSupported.proto */
#if CYTHON_USE_CPP_STD_MOVE
  #include <utility>
  #define __PYX_STD_MOVE_IF_SUPPORTED(x) std::move(x)
#else
  #define __PYX_STD_MOVE_IF_SUPPORTED(x) x
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* unicode_tailmatch.proto */
static int __Pyx_PyUnicode_Tailmatch(
    PyObject* s, PyObject* substr, Py_ssize_t start, Py_ssize_t end, int direction);

/* PyUnicode_Substring.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Substring(
            PyObject* text, Py_ssize_t start, Py_ssize_t stop);

/* SetStringIndexingError.proto (used by GetItemIntUnicode) */
static void __Pyx_SetStringIndexingError(const char* message, int has_gil);

/* GetItemIntUnicode.proto */
#define __Pyx_GetItemInt_Unicode(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil, unsafe_shared)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Unicode_Fast(o, (Py_ssize_t)i, wraparound, boundscheck, has_gil) :\
    (__Pyx_SetStringIndexingError("string index out of range", has_gil), (Py_UCS4)-1))
static CYTHON_INLINE Py_UCS4 __Pyx_GetItemInt_Unicode_Fast(PyObject* ustring, Py_ssize_t i,
                                                           int wraparound, int boundscheck, int has_gil);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* AllocateExtensionType.proto */
static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final);

/* CallTypeTraverse.proto */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* LimitedApiGetTypeDict.proto (used by SetItemOnTypeDict) */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto (used by FixUpExtensionType) */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* ValidateBasesTuple.proto (used by PyType_Ready) */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto (used by MergeVTables) */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
static int __Pyx_MergeVtables(PyTypeObject *type);

/* DelItemOnTypeDict.proto (used by SetupReduce) */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k);
#define __Pyx_DelItemOnTypeDict(tp, k) __Pyx__DelItemOnTypeDict((PyTypeObject*)tp, k)

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_2_4
#define __PYX_HAVE_RT_ImportType_proto_3_2_4
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_2_4(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_2_4 {
   __Pyx_ImportType_CheckSize_Error_3_2_4 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_2_4 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_2_4 = 2
};
static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size);
#endif

/* HasAttr.proto (used by ImportImpl) */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* ImportImpl.export */
static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level);

/* Import.proto */
static CYTHON_INLINE PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level);

/* dict_setdefault.proto (used by FetchCommonType) */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value);

/* AddModuleRef.proto (used by FetchSharedCythonModule) */
#if ((CYTHON_COMPILING_IN_CPYTHON_FREETHREADING ) ||\
     __PYX_LIMITED_VERSION_HEX < 0x030d0000)
  static PyObject *__Pyx_PyImport_AddModuleRef(const char *name);
#else
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#endif

/* FetchSharedCythonModule.proto (used by FetchCommonType) */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto (used by CommonTypesMetaclass) */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto (used by CythonFunctionShared) */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* PyMethodNew.proto (used by CythonFunctionShared) */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto (used by CythonFunctionShared) */
#if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto (used by CythonFunction) */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
#if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_dict;
#endif
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* pybytes_as_double.proto (used by pyunicode_as_double) */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj);
static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length);
static CYTHON_INLINE double __Pyx_PyBytes_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    as_c_string = PyBytes_AS_STRING(obj);
    size = PyBytes_GET_SIZE(obj);
#else
    if (PyBytes_AsStringAndSize(obj, &as_c_string, &size) < 0) {
        return (double)-1;
    }
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}
static CYTHON_INLINE double __Pyx_PyByteArray_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    as_c_string = PyByteArray_AS_STRING(obj);
    size = PyByteArray_GET_SIZE(obj);
#else
    as_c_string = PyByteArray_AsString(obj);
    if (as_c_string == NULL) {
        return (double)-1;
    }
    size = PyByteArray_Size(obj);
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}

/* pyunicode_as_double.proto */
#if !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
static const char* __Pyx__PyUnicode_AsDouble_Copy(const void* data, const int kind, char* buffer, Py_ssize_t start, Py_ssize_t end) {
    int last_was_punctuation;
    Py_ssize_t i;
    last_was_punctuation = 1;
    for (i=start; i <= end; i++) {
        Py_UCS4 chr = PyUnicode_READ(kind, data, i);
        int is_punctuation = (chr == '_') | (chr == '.');
        *buffer = (char)chr;
        buffer += (chr != '_');
        if (unlikely(chr > 127)) goto parse_failure;
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyUnicode_AsDouble_inf_nan(const void* data, int kind, Py_ssize_t start, Py_ssize_t length) {
    int matches = 1;
    Py_UCS4 chr;
    Py_UCS4 sign = PyUnicode_READ(kind, data, start);
    int is_signed = (sign == '-') | (sign == '+');
    start += is_signed;
    length -= is_signed;
    switch (PyUnicode_READ(kind, data, start)) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'a') | (chr == 'A');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'n') | (chr == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'f') | (chr == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+3);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+4);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+5);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+6);
            matches &= (chr == 't') | (chr == 'T');
            chr = PyUnicode_READ(kind, data, start+7);
            matches &= (chr == 'y') | (chr == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static double __Pyx_PyUnicode_AsDouble_WithSpaces(PyObject *obj) {
    double value;
    const char *last;
    char *end;
    Py_ssize_t start, length = PyUnicode_GET_LENGTH(obj);
    const int kind = PyUnicode_KIND(obj);
    const void* data = PyUnicode_DATA(obj);
    start = 0;
    while (Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, start)))
        start++;
    while (start < length - 1 && Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, length - 1)))
        length--;
    length -= start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyUnicode_AsDouble_inf_nan(data, kind, start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    if (length < 40) {
        char number[40];
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((length + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}
#endif
static CYTHON_INLINE double __Pyx_PyUnicode_AsDouble(PyObject *obj) {
#if !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(__Pyx_PyUnicode_READY(obj) == -1))
        return (double)-1;
    if (likely(PyUnicode_IS_ASCII(obj))) {
        const char *s;
        Py_ssize_t length;
        s = PyUnicode_AsUTF8AndSize(obj, &length);
        return __Pyx__PyBytes_AsDouble(obj, s, length);
    }
    return __Pyx_PyUnicode_AsDouble_WithSpaces(obj);
#else
    return __Pyx_SlowPyString_AsDouble(obj);
#endif
}

/* FloatExceptionCheck.proto */
#define __PYX_CHECK_FLOAT_EXCEPTION(value, error_value)\
    ((error_value) == (error_value) ?\
     (value) == (error_value) :\
     (value) != (value))

/* CLineInTraceback.proto (used by AddTraceback) */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto (used by AddTraceback) */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CppExceptionConversion.proto */
#ifndef __Pyx_CppExn2PyErr
#include <new>
#include <typeinfo>
#include <stdexcept>
#include <ios>
static void __Pyx_CppExn2PyErr() {
  try {
    if (PyErr_Occurred())
      ; // let the latest Python exn pass through and ignore the current one
    else
      throw;
  } catch (const std::bad_alloc& exn) {
    PyErr_SetString(PyExc_MemoryError, exn.what());
  } catch (const std::bad_cast& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::bad_typeid& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::domain_error& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::invalid_argument& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::ios_base::failure& exn) {
    PyErr_SetString(PyExc_IOError, exn.what());
  } catch (const std::out_of_range& exn) {
    PyErr_SetString(PyExc_IndexError, exn.what());
  } catch (const std::overflow_error& exn) {
    PyErr_SetString(PyExc_OverflowError, exn.what());
  } catch (const std::range_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::underflow_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::exception& exn) {
    PyErr_SetString(PyExc_RuntimeError, exn.what());
  }
  catch (...)
  {
    PyErr_SetString(PyExc_RuntimeError, "Unknown exception");
  }
}
#endif

/* ToPy.proto */
#define __pyx_PyComplex_FromComplex(z)\
        PyComplex_FromDoubles((double)__Pyx_CREAL(z),\
                              (double)__Pyx_CIMAG(z))

/* FromPy.proto */
static __pyx_t_double_complex __Pyx_PyComplex_As___pyx_t_double_complex(PyObject*);

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_long__double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_long__double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_long__double(a, b) ((a)-(b))
    #define __Pyx_c_prod_long__double(a, b) ((a)*(b))
    #define __Pyx_c_quot_long__double(a, b) ((a)/(b))
    #define __Pyx_c_neg_long__double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_long__double(z) ((z)==(long double)0)
    #define __Pyx_c_conj_long__double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_long__double(z)     (::std::abs(z))
        #define __Pyx_c_pow_long__double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_long__double(z) ((z)==0)
    #define __Pyx_c_conj_long__double(z)    (conjl(z))
    #if 1
        #define __Pyx_c_abs_long__double(z)     (cabsl(z))
        #define __Pyx_c_pow_long__double(a, b)  (cpowl(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_sum_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_diff_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_prod_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_neg_long__double(__pyx_t_long_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_long__double(__pyx_t_long_double_complex);
    static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_conj_long__double(__pyx_t_long_double_complex);
    #if 1
        static CYTHON_INLINE long double __Pyx_c_abs_long__double(__pyx_t_long_double_complex);
        static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_pow_long__double(__pyx_t_long_double_complex, __pyx_t_long_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__TMM_3a__3a_TMMParam(enum TMM::TMMParam value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE enum TMM::TMMParam __Pyx_PyLong_As_enum__TMM_3a__3a_TMMParam(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
#if __PYX_LIMITED_VERSION_HEX < 0x030b0000
static unsigned long __Pyx_cached_runtime_version = 0;
static void __Pyx_init_runtime_version(void);
#else
#define __Pyx_init_runtime_version()
#endif
static unsigned long __Pyx_get_runtime_version(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* DecompressString.proto */
static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_8itemsize_itemsize(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_9alignment_alignment(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_6fields_fields(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_5names_names(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_ArrayDescr *__pyx_f_5numpy_5dtype_8subarray_subarray(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_uint64 __pyx_f_5numpy_5dtype_5flags_flags(PyArray_Descr *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_7numiter_numiter(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_4size_size(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_5index_index(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_2nd_nd(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_9broadcast_10dimensions_dimensions(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE void **__pyx_f_5numpy_9broadcast_5iters_iters(PyArrayMultiIterObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_self, TMM::Chi2Tensor *__pyx_v_ptr, PyObject *__pyx_v_parent); /* proto*/
static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, TMM::Wave *__pyx_v_ptr); /* proto*/
static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self, TMM::Intensities *__pyx_v_ptr); /* proto*/
static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self, TMM::SweepResultNonlinearTMM *__pyx_v_ptr, struct __pyx_opt_args_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM__Init *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self, TMM::WaveSweepResultNonlinearTMM *__pyx_v_ptr, struct __pyx_opt_args_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM__Init *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_self, TMM::FieldsZ *__pyx_v_ptr); /* proto*/
static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self, TMM::FieldsZX *__pyx_v_ptr); /* proto*/
static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_self, TMM::HomogeneousWave *__pyx_v_ptr); /* proto*/
static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self, TMM::NonlinearLayer *__pyx_v_ptr, int __pyx_v_layerNr, PyObject *__pyx_v_parent); /* proto*/
static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, TMM::NonlinearTMM *__pyx_v_ptr); /* proto*/
static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_self, TMM::SecondOrderNLIntensities *__pyx_v_ptr); /* proto*/
static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self, TMM::SweepResultSecondOrderNLTMM *__pyx_v_ptr); /* proto*/
static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self, TMM::WaveSweepResultSecondOrderNLTMM *__pyx_v_ptr); /* proto*/

/* Module declarations from "libcpp" */

/* Module declarations from "libcpp.memory" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "numpy" */

/* Module declarations from "numpy" */

/* Module declarations from "eigency.core" */

/* Module declarations from "libcpp.utility" */

/* Module declarations from "libcpp.pair" */

/* Module declarations from "CppSecondOrderNLTMM" */

/* Module declarations from "NonlinearTMM._SecondOrderNLTMMCython" */
static PyObject *__pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_paramDict = 0;
static PyObject *__pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_paramDictSecondOrderNLTMM = 0;
static PyObject *__pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_waveParamsSet = 0;
static enum TMM::Polarization __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_PolarizationFromStr(PyObject *); /*proto*/
static enum TMM::WaveDirection __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_WaveDirectionFromStr(PyObject *); /*proto*/
static enum TMM::NonlinearTmmMode __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTmmModeFromStr(PyObject *); /*proto*/
static enum TMM::NonlinearProcess __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearProcessFromStr(PyObject *); /*proto*/
static enum TMM::WaveType __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_WaveTypeFromStr(PyObject *); /*proto*/
static enum TMM::TMMParam __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_TmmParamFromStr(PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
static const __Pyx_TypeInfo __Pyx_TypeInfo_double = { "double", NULL, sizeof(double), { 0 }, 0, 'R', 0, 0 };
static const __Pyx_TypeInfo __Pyx_TypeInfo___pyx_t_double_complex = { "double complex", NULL, sizeof(__pyx_t_double_complex), { 0 }, 0, 'C', 0, 0 };
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "NonlinearTMM._SecondOrderNLTMMCython"
extern int __pyx_module_is_main_NonlinearTMM___SecondOrderNLTMMCython;
int __pyx_module_is_main_NonlinearTMM___SecondOrderNLTMMCython = 0;

/* Implementation of "NonlinearTMM._SecondOrderNLTMMCython" */
/* #### Code section: global_var ### */
/* #### Code section: string_decls ### */
/* #### Code section: decls ### */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor___cinit__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_self); /* proto */
static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_2__dealloc__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_4Update(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_self, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_6GetChi2Tensor(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self, PyArrayObject *__pyx_v_wls, PyArrayObject *__pyx_v_ns); /* proto */
static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_4__call__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self, PyObject *__pyx_v_wl); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_6GetN(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self, PyObject *__pyx_v_wl); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_8IsNonlinear(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_4chi2___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4SetParams(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8waveType___get__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8waveType_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, PyObject *__pyx_v_waveTypeStr); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3pwr___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3pwr_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_10overrideE0___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_10overrideE0_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, bool __pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2E0___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2E0_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2w0___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2w0_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2Ly___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2Ly_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1a___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1a_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12nPointsInteg___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12nPointsInteg_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, int __pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4maxX___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4maxX_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_11dynamicMaxX___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_11dynamicMaxX_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, bool __pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_15dynamicMaxXCoef___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_15dynamicMaxXCoef_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_19dynamicMaxXAddition___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_19dynamicMaxXAddition_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6maxPhi___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6maxPhi_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6xRange___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_5betas___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4phis___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3kxs___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3kzs___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12fieldProfile___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_16expansionCoefsKx___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8beamArea___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities___cinit__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_3inc___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1r___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1t___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1I___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1R___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1T___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self); /* proto */
static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3inc___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1r___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1t___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ii___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ir___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2It___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ia___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3enh___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self); /* proto */
static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pi___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pr___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pt___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_3enh___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_8beamArea___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ___cinit__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_self); /* proto */
static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1E___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1H___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX___cinit__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self); /* proto */
static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2EN___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2HN___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ex___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ey___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ez___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hx___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hy___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hz___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_2GetMainFields(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_self, double __pyx_v_z); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_3kzF___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_2kx___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_2GetIntensity(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self, double __pyx_v_z); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_4GetAbsorbedIntensity(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_6GetSrcIntensity(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1d___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1d_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_2hw___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, bool __pyx_v_initStruct, PyObject *__pyx_v_parent, PyObject *__pyx_v_kwargs); /* proto */
static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4AddLayer(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, double __pyx_v_d, struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_material); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_6SetParams(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_8Solve(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10GetIntensities(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_12Sweep(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_paramStr, PyArrayObject *__pyx_v_values, int __pyx_v_layerNr, double __pyx_v_layerZ, bool __pyx_v_outPwr, bool __pyx_v_outAbs, PyObject *__pyx_v_outEnh); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_14GetFields(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyArrayObject *__pyx_v_zs, PyObject *__pyx_v_dir); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_16GetFields2D(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyArrayObject *__pyx_v_zs, PyArrayObject *__pyx_v_xs, PyObject *__pyx_v_dir); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_18GetAbsorbedIntensity(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_20GetEnhancement(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, int __pyx_v_layerNr, double __pyx_v_z); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_22WaveGetPowerFlows(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, int __pyx_v_layerNr, double __pyx_v_x0, double __pyx_v_x1, double __pyx_v_z); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_24WaveGetEnhancement(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, int __pyx_v_layerNr, double __pyx_v_z); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_26WaveSweep(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_paramStr, PyArrayObject *__pyx_v_values, int __pyx_v_layerNr, double __pyx_v_layerZ, bool __pyx_v_outPwr, bool __pyx_v_outR, bool __pyx_v_outT, bool __pyx_v_outEnh); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_28WaveGetFields2D(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyArrayObject *__pyx_v_zs, PyArrayObject *__pyx_v_xs, PyObject *__pyx_v_dirStr); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2wl___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2wl_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4beta___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4beta_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3pol___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3pol_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_polStr); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2I0___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2I0_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10overrideE0___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10overrideE0_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2E0___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2E0_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4mode___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4mode_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_modeStr); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_14materialsCache___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_6layers___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4wave___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_30__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_32__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities___cinit__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2P1___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2P2___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_3Gen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2P1___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2P2___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_3Gen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_6wlsGen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_8betasGen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2P1___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2P2___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_3Gen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_6wlsGen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_8betasGen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyObject *__pyx_v_mode, PyObject *__pyx_v_kwargs); /* proto */
static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_4SetParams(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_6AddLayer(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, double __pyx_v_d, struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_material); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_8Solve(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_10UpdateGenParams(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_12GetIntensities(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14Sweep(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyObject *__pyx_v_paramStr, PyArrayObject *__pyx_v_valuesP1, PyArrayObject *__pyx_v_valuesP2, int __pyx_v_layerNr, double __pyx_v_layerZ, bool __pyx_v_outPwr, bool __pyx_v_outAbs, bool __pyx_v_outEnh, bool __pyx_v_outP1, bool __pyx_v_outP2, bool __pyx_v_outGen); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_16WaveGetPowerFlows(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, int __pyx_v_layerNr, double __pyx_v_x0, double __pyx_v_x1, double __pyx_v_z); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_18WaveSweep(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyObject *__pyx_v_paramStr, PyArrayObject *__pyx_v_valuesP1, PyArrayObject *__pyx_v_valuesP2, int __pyx_v_layerNr, double __pyx_v_layerZ, bool __pyx_v_outPwr, bool __pyx_v_outR, bool __pyx_v_outT, bool __pyx_v_outEnh, bool __pyx_v_outP1, bool __pyx_v_outP2, bool __pyx_v_outGen); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_20WaveGetFields2D(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyArrayObject *__pyx_v_zs, PyArrayObject *__pyx_v_xs, PyObject *__pyx_v_dirStr); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11deltaWlSpdc___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11deltaWlSpdc_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14solidAngleSpdc___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14solidAngleSpdc_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14deltaThetaSpdc___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self); /* proto */
static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14deltaThetaSpdc_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2P1___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2P2___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_3Gen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython_Material(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyTypeObject *__pyx_ptype_5numpy_dtype;
  PyTypeObject *__pyx_ptype_5numpy_flatiter;
  PyTypeObject *__pyx_ptype_5numpy_broadcast;
  PyTypeObject *__pyx_ptype_5numpy_ndarray;
  PyTypeObject *__pyx_ptype_5numpy_generic;
  PyTypeObject *__pyx_ptype_5numpy_number;
  PyTypeObject *__pyx_ptype_5numpy_integer;
  PyTypeObject *__pyx_ptype_5numpy_signedinteger;
  PyTypeObject *__pyx_ptype_5numpy_unsignedinteger;
  PyTypeObject *__pyx_ptype_5numpy_inexact;
  PyTypeObject *__pyx_ptype_5numpy_floating;
  PyTypeObject *__pyx_ptype_5numpy_complexfloating;
  PyTypeObject *__pyx_ptype_5numpy_flexible;
  PyTypeObject *__pyx_ptype_5numpy_character;
  PyTypeObject *__pyx_ptype_5numpy_ufunc;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_Material;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM;
  PyObject *__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM;
  PyTypeObject *__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_keys;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values;
  double __pyx_k_;
  double __pyx_k__2;
  double __pyx_k__3;
  double __pyx_k__4;
  PyObject *__pyx_tuple[6];
  PyObject *__pyx_codeobj_tab[61];
  PyObject *__pyx_string_tab[271];
  PyObject *__pyx_number_tab[4];
/* #### Code section: module_state_contents ### */
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CythonFunctionShared.module_state_decls */
PyTypeObject *__pyx_CyFunctionType;

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_AddLayer_d_material_Adds_layer_t __pyx_string_tab[0]
#define __pyx_kp_u_Chi2Tensor_Update_line_114 __pyx_string_tab[1]
#define __pyx_kp_u_GetN_wl_Returns_refractive_index __pyx_string_tab[2]
#define __pyx_kp_u_Material_GetN_line_251 __pyx_string_tab[3]
#define __pyx_kp_u_NonlinearTMM_AddLayer_line_1044 __pyx_string_tab[4]
#define __pyx_kp_u_NonlinearTMM_Solve_line_1089 __pyx_string_tab[5]
#define __pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM __pyx_string_tab[6]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[7]
#define __pyx_kp_u_Solve_kwargs_Solves_the_structur __pyx_string_tab[8]
#define __pyx_kp_u_Unknown_kwarg_s __pyx_string_tab[9]
#define __pyx_kp_u_Unknown_polarization __pyx_string_tab[10]
#define __pyx_kp_u_Update_kwargs_This_method_change __pyx_string_tab[11]
#define __pyx_kp_u__5 __pyx_string_tab[12]
#define __pyx_kp_u_add_note __pyx_string_tab[13]
#define __pyx_kp_u_disable __pyx_string_tab[14]
#define __pyx_kp_u_enable __pyx_string_tab[15]
#define __pyx_kp_u_gc __pyx_string_tab[16]
#define __pyx_kp_u_isenabled __pyx_string_tab[17]
#define __pyx_kp_u_no_default___reduce___due_to_non __pyx_string_tab[18]
#define __pyx_kp_u_numpy__core_multiarray_failed_to __pyx_string_tab[19]
#define __pyx_kp_u_numpy__core_umath_failed_to_impo __pyx_string_tab[20]
#define __pyx_kp_u_stringsource __pyx_string_tab[21]
#define __pyx_n_u_AddLayer __pyx_string_tab[22]
#define __pyx_n_u_Chi2Tensor __pyx_string_tab[23]
#define __pyx_n_u_Chi2Tensor_GetChi2Tensor __pyx_string_tab[24]
#define __pyx_n_u_Chi2Tensor_Update __pyx_string_tab[25]
#define __pyx_n_u_Chi2Tensor___reduce_cython __pyx_string_tab[26]
#define __pyx_n_u_Chi2Tensor___setstate_cython __pyx_string_tab[27]
#define __pyx_n_u_E0 __pyx_string_tab[28]
#define __pyx_n_u_FieldsZ __pyx_string_tab[29]
#define __pyx_n_u_FieldsZX __pyx_string_tab[30]
#define __pyx_n_u_FieldsZX___reduce_cython __pyx_string_tab[31]
#define __pyx_n_u_FieldsZX___setstate_cython __pyx_string_tab[32]
#define __pyx_n_u_FieldsZ___reduce_cython __pyx_string_tab[33]
#define __pyx_n_u_FieldsZ___setstate_cython __pyx_string_tab[34]
#define __pyx_n_u_GetAbsorbedIntensity __pyx_string_tab[35]
#define __pyx_n_u_GetChi2Tensor __pyx_string_tab[36]
#define __pyx_n_u_GetEnhancement __pyx_string_tab[37]
#define __pyx_n_u_GetFields __pyx_string_tab[38]
#define __pyx_n_u_GetFields2D __pyx_string_tab[39]
#define __pyx_n_u_GetIntensities __pyx_string_tab[40]
#define __pyx_n_u_GetIntensity __pyx_string_tab[41]
#define __pyx_n_u_GetMainFields __pyx_string_tab[42]
#define __pyx_n_u_GetN __pyx_string_tab[43]
#define __pyx_n_u_GetSrcIntensity __pyx_string_tab[44]
#define __pyx_n_u_HomogeneousWave __pyx_string_tab[45]
#define __pyx_n_u_HomogeneousWave_GetMainFields __pyx_string_tab[46]
#define __pyx_n_u_HomogeneousWave___reduce_cython __pyx_string_tab[47]
#define __pyx_n_u_HomogeneousWave___setstate_cyth __pyx_string_tab[48]
#define __pyx_n_u_I0 __pyx_string_tab[49]
#define __pyx_n_u_Intensities __pyx_string_tab[50]
#define __pyx_n_u_Intensities___reduce_cython __pyx_string_tab[51]
#define __pyx_n_u_Intensities___setstate_cython __pyx_string_tab[52]
#define __pyx_n_u_IsNonlinear __pyx_string_tab[53]
#define __pyx_n_u_Ly __pyx_string_tab[54]
#define __pyx_n_u_Material __pyx_string_tab[55]
#define __pyx_n_u_Material_GetN __pyx_string_tab[56]
#define __pyx_n_u_Material_IsNonlinear __pyx_string_tab[57]
#define __pyx_n_u_Material___reduce_cython __pyx_string_tab[58]
#define __pyx_n_u_Material___setstate_cython __pyx_string_tab[59]
#define __pyx_n_u_NonlinearLayer __pyx_string_tab[60]
#define __pyx_n_u_NonlinearLayer_GetAbsorbedInten __pyx_string_tab[61]
#define __pyx_n_u_NonlinearLayer_GetIntensity __pyx_string_tab[62]
#define __pyx_n_u_NonlinearLayer_GetSrcIntensity __pyx_string_tab[63]
#define __pyx_n_u_NonlinearLayer___reduce_cython __pyx_string_tab[64]
#define __pyx_n_u_NonlinearLayer___setstate_cytho __pyx_string_tab[65]
#define __pyx_n_u_NonlinearTMM __pyx_string_tab[66]
#define __pyx_n_u_NonlinearTMM_AddLayer __pyx_string_tab[67]
#define __pyx_n_u_NonlinearTMM_GetAbsorbedIntensit __pyx_string_tab[68]
#define __pyx_n_u_NonlinearTMM_GetEnhancement __pyx_string_tab[69]
#define __pyx_n_u_NonlinearTMM_GetFields __pyx_string_tab[70]
#define __pyx_n_u_NonlinearTMM_GetFields2D __pyx_string_tab[71]
#define __pyx_n_u_NonlinearTMM_GetIntensities __pyx_string_tab[72]
#define __pyx_n_u_NonlinearTMM_SetParams __pyx_string_tab[73]
#define __pyx_n_u_NonlinearTMM_Solve __pyx_string_tab[74]
#define __pyx_n_u_NonlinearTMM_Sweep __pyx_string_tab[75]
#define __pyx_n_u_NonlinearTMM_WaveGetEnhancement __pyx_string_tab[76]
#define __pyx_n_u_NonlinearTMM_WaveGetFields2D __pyx_string_tab[77]
#define __pyx_n_u_NonlinearTMM_WaveGetPowerFlows __pyx_string_tab[78]
#define __pyx_n_u_NonlinearTMM_WaveSweep __pyx_string_tab[79]
#define __pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy __pyx_string_tab[80]
#define __pyx_n_u_NonlinearTMM___reduce_cython __pyx_string_tab[81]
#define __pyx_n_u_NonlinearTMM___setstate_cython __pyx_string_tab[82]
#define __pyx_n_u_Pyx_PyDict_NextRef __pyx_string_tab[83]
#define __pyx_n_u_SecondOrderNLIntensities __pyx_string_tab[84]
#define __pyx_n_u_SecondOrderNLIntensities___redu __pyx_string_tab[85]
#define __pyx_n_u_SecondOrderNLIntensities___sets __pyx_string_tab[86]
#define __pyx_n_u_SecondOrderNLTMM __pyx_string_tab[87]
#define __pyx_n_u_SecondOrderNLTMM_AddLayer __pyx_string_tab[88]
#define __pyx_n_u_SecondOrderNLTMM_GetIntensities __pyx_string_tab[89]
#define __pyx_n_u_SecondOrderNLTMM_SetParams __pyx_string_tab[90]
#define __pyx_n_u_SecondOrderNLTMM_Solve __pyx_string_tab[91]
#define __pyx_n_u_SecondOrderNLTMM_Sweep __pyx_string_tab[92]
#define __pyx_n_u_SecondOrderNLTMM_UpdateGenParams __pyx_string_tab[93]
#define __pyx_n_u_SecondOrderNLTMM_WaveGetFields2D __pyx_string_tab[94]
#define __pyx_n_u_SecondOrderNLTMM_WaveGetPowerFlo __pyx_string_tab[95]
#define __pyx_n_u_SecondOrderNLTMM_WaveSweep __pyx_string_tab[96]
#define __pyx_n_u_SecondOrderNLTMM___reduce_cython __pyx_string_tab[97]
#define __pyx_n_u_SecondOrderNLTMM___setstate_cyth __pyx_string_tab[98]
#define __pyx_n_u_SetParams __pyx_string_tab[99]
#define __pyx_n_u_Solve __pyx_string_tab[100]
#define __pyx_n_u_Sweep __pyx_string_tab[101]
#define __pyx_n_u_SweepResultNonlinearTMM __pyx_string_tab[102]
#define __pyx_n_u_SweepResultNonlinearTMM___reduc __pyx_string_tab[103]
#define __pyx_n_u_SweepResultNonlinearTMM___setst __pyx_string_tab[104]
#define __pyx_n_u_SweepResultSecondOrderNLTMM __pyx_string_tab[105]
#define __pyx_n_u_SweepResultSecondOrderNLTMM___r __pyx_string_tab[106]
#define __pyx_n_u_SweepResultSecondOrderNLTMM___s __pyx_string_tab[107]
#define __pyx_n_u_Update __pyx_string_tab[108]
#define __pyx_n_u_UpdateGenParams __pyx_string_tab[109]
#define __pyx_n_u_Wave __pyx_string_tab[110]
#define __pyx_n_u_WaveGetEnhancement __pyx_string_tab[111]
#define __pyx_n_u_WaveGetFields2D __pyx_string_tab[112]
#define __pyx_n_u_WaveGetPowerFlows __pyx_string_tab[113]
#define __pyx_n_u_WaveSweep __pyx_string_tab[114]
#define __pyx_n_u_WaveSweepResultNonlinearTMM __pyx_string_tab[115]
#define __pyx_n_u_WaveSweepResultNonlinearTMM___r __pyx_string_tab[116]
#define __pyx_n_u_WaveSweepResultNonlinearTMM___s __pyx_string_tab[117]
#define __pyx_n_u_WaveSweepResultSecondOrderNLTMM __pyx_string_tab[118]
#define __pyx_n_u_WaveSweepResultSecondOrderNLTMM_2 __pyx_string_tab[119]
#define __pyx_n_u_WaveSweepResultSecondOrderNLTMM_3 __pyx_string_tab[120]
#define __pyx_n_u_Wave_SetParams __pyx_string_tab[121]
#define __pyx_n_u_Wave___reduce_cython __pyx_string_tab[122]
#define __pyx_n_u_Wave___setstate_cython __pyx_string_tab[123]
#define __pyx_n_u_a __pyx_string_tab[124]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[125]
#define __pyx_n_u_backward __pyx_string_tab[126]
#define __pyx_n_u_beta __pyx_string_tab[127]
#define __pyx_n_u_chi __pyx_string_tab[128]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[129]
#define __pyx_n_u_d __pyx_string_tab[130]
#define __pyx_n_u_d_2 __pyx_string_tab[131]
#define __pyx_n_u_deltaThetaSpdc __pyx_string_tab[132]
#define __pyx_n_u_deltaWlSpdc __pyx_string_tab[133]
#define __pyx_n_u_dfg __pyx_string_tab[134]
#define __pyx_n_u_dir __pyx_string_tab[135]
#define __pyx_n_u_dirStr __pyx_string_tab[136]
#define __pyx_n_u_direction __pyx_string_tab[137]
#define __pyx_n_u_distinctFields __pyx_string_tab[138]
#define __pyx_n_u_dtype __pyx_string_tab[139]
#define __pyx_n_u_dynamicMaxX __pyx_string_tab[140]
#define __pyx_n_u_dynamicMaxXAddition __pyx_string_tab[141]
#define __pyx_n_u_dynamicMaxXCoef __pyx_string_tab[142]
#define __pyx_n_u_forward __pyx_string_tab[143]
#define __pyx_n_u_func __pyx_string_tab[144]
#define __pyx_n_u_gaussian __pyx_string_tab[145]
#define __pyx_n_u_getstate __pyx_string_tab[146]
#define __pyx_n_u_i __pyx_string_tab[147]
#define __pyx_n_u_i1 __pyx_string_tab[148]
#define __pyx_n_u_i2 __pyx_string_tab[149]
#define __pyx_n_u_i3 __pyx_string_tab[150]
#define __pyx_n_u_imag __pyx_string_tab[151]
#define __pyx_n_u_incident __pyx_string_tab[152]
#define __pyx_n_u_initStruct __pyx_string_tab[153]
#define __pyx_n_u_is_coroutine __pyx_string_tab[154]
#define __pyx_n_u_items __pyx_string_tab[155]
#define __pyx_n_u_j __pyx_string_tab[156]
#define __pyx_n_u_k __pyx_string_tab[157]
#define __pyx_n_u_keys __pyx_string_tab[158]
#define __pyx_n_u_kwargs __pyx_string_tab[159]
#define __pyx_n_u_lastLayer __pyx_string_tab[160]
#define __pyx_n_u_layer __pyx_string_tab[161]
#define __pyx_n_u_layerNr __pyx_string_tab[162]
#define __pyx_n_u_layerZ __pyx_string_tab[163]
#define __pyx_n_u_layerptr __pyx_string_tab[164]
#define __pyx_n_u_main __pyx_string_tab[165]
#define __pyx_n_u_material __pyx_string_tab[166]
#define __pyx_n_u_maxPhi __pyx_string_tab[167]
#define __pyx_n_u_maxX __pyx_string_tab[168]
#define __pyx_n_u_mode __pyx_string_tab[169]
#define __pyx_n_u_module __pyx_string_tab[170]
#define __pyx_n_u_nPointsInteg __pyx_string_tab[171]
#define __pyx_n_u_name __pyx_string_tab[172]
#define __pyx_n_u_name_2 __pyx_string_tab[173]
#define __pyx_n_u_nan __pyx_string_tab[174]
#define __pyx_n_u_nonlinear __pyx_string_tab[175]
#define __pyx_n_u_np __pyx_string_tab[176]
#define __pyx_n_u_ns __pyx_string_tab[177]
#define __pyx_n_u_numpy __pyx_string_tab[178]
#define __pyx_n_u_outAbs __pyx_string_tab[179]
#define __pyx_n_u_outEnh __pyx_string_tab[180]
#define __pyx_n_u_outGen __pyx_string_tab[181]
#define __pyx_n_u_outP1 __pyx_string_tab[182]
#define __pyx_n_u_outP2 __pyx_string_tab[183]
#define __pyx_n_u_outPwr __pyx_string_tab[184]
#define __pyx_n_u_outR __pyx_string_tab[185]
#define __pyx_n_u_outT __pyx_string_tab[186]
#define __pyx_n_u_outmask __pyx_string_tab[187]
#define __pyx_n_u_overrideE0 __pyx_string_tab[188]
#define __pyx_n_u_p __pyx_string_tab[189]
#define __pyx_n_u_param __pyx_string_tab[190]
#define __pyx_n_u_paramLayer __pyx_string_tab[191]
#define __pyx_n_u_paramStr __pyx_string_tab[192]
#define __pyx_n_u_parent __pyx_string_tab[193]
#define __pyx_n_u_phiX __pyx_string_tab[194]
#define __pyx_n_u_phiY __pyx_string_tab[195]
#define __pyx_n_u_phiZ __pyx_string_tab[196]
#define __pyx_n_u_planewave __pyx_string_tab[197]
#define __pyx_n_u_pol __pyx_string_tab[198]
#define __pyx_n_u_pop __pyx_string_tab[199]
#define __pyx_n_u_ps __pyx_string_tab[200]
#define __pyx_n_u_pwr __pyx_string_tab[201]
#define __pyx_n_u_pyx_state __pyx_string_tab[202]
#define __pyx_n_u_pyx_vtable __pyx_string_tab[203]
#define __pyx_n_u_qualname __pyx_string_tab[204]
#define __pyx_n_u_real __pyx_string_tab[205]
#define __pyx_n_u_reduce __pyx_string_tab[206]
#define __pyx_n_u_reduce_cython __pyx_string_tab[207]
#define __pyx_n_u_reduce_ex __pyx_string_tab[208]
#define __pyx_n_u_res __pyx_string_tab[209]
#define __pyx_n_u_resCpp __pyx_string_tab[210]
#define __pyx_n_u_s __pyx_string_tab[211]
#define __pyx_n_u_self __pyx_string_tab[212]
#define __pyx_n_u_set_name __pyx_string_tab[213]
#define __pyx_n_u_setdefault __pyx_string_tab[214]
#define __pyx_n_u_setstate __pyx_string_tab[215]
#define __pyx_n_u_setstate_cython __pyx_string_tab[216]
#define __pyx_n_u_sfg __pyx_string_tab[217]
#define __pyx_n_u_solidAngleSpdc __pyx_string_tab[218]
#define __pyx_n_u_spdc __pyx_string_tab[219]
#define __pyx_n_u_sqrt __pyx_string_tab[220]
#define __pyx_n_u_squeeze __pyx_string_tab[221]
#define __pyx_n_u_startswith __pyx_string_tab[222]
#define __pyx_n_u_test __pyx_string_tab[223]
#define __pyx_n_u_total __pyx_string_tab[224]
#define __pyx_n_u_tukey __pyx_string_tab[225]
#define __pyx_n_u_value __pyx_string_tab[226]
#define __pyx_n_u_values __pyx_string_tab[227]
#define __pyx_n_u_valuesP1 __pyx_string_tab[228]
#define __pyx_n_u_valuesP2 __pyx_string_tab[229]
#define __pyx_n_u_w0 __pyx_string_tab[230]
#define __pyx_n_u_waveType __pyx_string_tab[231]
#define __pyx_n_u_wl __pyx_string_tab[232]
#define __pyx_n_u_wls __pyx_string_tab[233]
#define __pyx_n_u_x0 __pyx_string_tab[234]
#define __pyx_n_u_x1 __pyx_string_tab[235]
#define __pyx_n_u_xs __pyx_string_tab[236]
#define __pyx_n_u_z __pyx_string_tab[237]
#define __pyx_n_u_zeros __pyx_string_tab[238]
#define __pyx_n_u_zs __pyx_string_tab[239]
#define __pyx_kp_b_iso88591_5Q_d_9A_q __pyx_string_tab[240]
#define __pyx_kp_b_iso88591_9_a_SSXXYYllm_d_Qit4q_83a __pyx_string_tab[241]
#define __pyx_kp_b_iso88591_9_a_SSXXYYllm_d_Qit4q_83a_2 __pyx_string_tab[242]
#define __pyx_kp_b_iso88591_9_d_ay_q __pyx_string_tab[243]
#define __pyx_kp_b_iso88591_A0_t9E __pyx_string_tab[244]
#define __pyx_kp_b_iso88591_A6_Jc_IV1 __pyx_string_tab[245]
#define __pyx_kp_b_iso88591_AFG4_aq_Y_d_1HIQeS_eef_iq_6_q __pyx_string_tab[246]
#define __pyx_kp_b_iso88591_AN_E_QfE_A_t3as_Q_AQaq_AQaq_F_aq __pyx_string_tab[247]
#define __pyx_kp_b_iso88591_A_1D_q __pyx_string_tab[248]
#define __pyx_kp_b_iso88591_A_Cy_A_Cy_A_D_Q_O7_1 __pyx_string_tab[249]
#define __pyx_kp_b_iso88591_A_F_6_q_uG1_j_4Cq_1F __pyx_string_tab[250]
#define __pyx_kp_b_iso88591_A_F_6_q_uG1_j_4Cq_1F_2 __pyx_string_tab[251]
#define __pyx_kp_b_iso88591_A_IV1 __pyx_string_tab[252]
#define __pyx_kp_b_iso88591_A_IYas_O7_1_T_c_1_4y_q_V1Jk_G7_1 __pyx_string_tab[253]
#define __pyx_kp_b_iso88591_A_I_Q __pyx_string_tab[254]
#define __pyx_kp_b_iso88591_A_T_l_6_1_q __pyx_string_tab[255]
#define __pyx_kp_b_iso88591_A_V_1_1_q_1_q_1_q_1_q_q_HKq_A_AX __pyx_string_tab[256]
#define __pyx_kp_b_iso88591_A_YoQ_q_6_1_q __pyx_string_tab[257]
#define __pyx_kp_b_iso88591_A_aq_Y_d_1HIQeS_eef_iq_6_q __pyx_string_tab[258]
#define __pyx_kp_b_iso88591_A_b_aq_3d_E_aq_U_1_E_aq_q_3e4y_q __pyx_string_tab[259]
#define __pyx_kp_b_iso88591_A_h_1_1_q_1_q_1_q_1_q_1_q_1_q_1 __pyx_string_tab[260]
#define __pyx_kp_b_iso88591_A_t9L __pyx_string_tab[261]
#define __pyx_kp_b_iso88591_A_t9M __pyx_string_tab[262]
#define __pyx_kp_b_iso88591_A_t9_1 __pyx_string_tab[263]
#define __pyx_kp_b_iso88591_A_t9_1_2 __pyx_string_tab[264]
#define __pyx_kp_b_iso88591_A_t9_A __pyx_string_tab[265]
#define __pyx_kp_b_iso88591_C1_Yj_IQe_OqPUU_ha_6_q __pyx_string_tab[266]
#define __pyx_kp_b_iso88591_Q __pyx_string_tab[267]
#define __pyx_kp_b_iso88591_R_R_e_e_y_y_N_N_b_b_u_u_I_I_J_1 __pyx_string_tab[268]
#define __pyx_kp_b_iso88591_UUhh_P_P_a_N_1_1_q_1_q_1_q_q_HK __pyx_string_tab[269]
#define __pyx_kp_b_iso88591_ffg_Yl_iq_Occddiiqqr_iq_6_q __pyx_string_tab[270]
#define __pyx_float_0_0 __pyx_number_tab[0]
#define __pyx_int_0 __pyx_number_tab[1]
#define __pyx_int_2 __pyx_number_tab[2]
#define __pyx_int_3 __pyx_number_tab[3]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_dtype);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flatiter);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_broadcast);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ndarray);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_generic);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_number);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_integer);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_inexact);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_floating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flexible);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_character);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ufunc);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_Material);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM);
  Py_CLEAR(clear_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM);
  Py_CLEAR(clear_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM);
  for (int i=0; i<6; ++i) { Py_CLEAR(clear_module_state->__pyx_tuple[i]); }
  for (int i=0; i<61; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<271; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<4; ++i) { Py_CLEAR(clear_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_clear_contents ### */
/* CommonTypesMetaclass.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_clear */
Py_CLEAR(clear_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_clear_end ### */
return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_dtype);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flatiter);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_broadcast);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ndarray);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_generic);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_number);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_integer);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_inexact);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_floating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flexible);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_character);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ufunc);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_Material);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM);
  Py_VISIT(traverse_module_state->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM);
  Py_VISIT(traverse_module_state->__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM);
  for (int i=0; i<6; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_tuple[i]); }
  for (int i=0; i<61; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<271; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  for (int i=0; i<4; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_number_tab[i]); }
/* #### Code section: module_state_traverse_contents ### */
/* CommonTypesMetaclass.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CommonTypesMetaclassType);

/* CythonFunctionShared.module_state_traverse */
Py_VISIT(traverse_module_state->__pyx_CyFunctionType);

/* #### Code section: module_state_traverse_end ### */
return 0;
}
#endif
/* #### Code section: module_code ### */

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":243
 *         cdef int type_num
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp itemsize(self) noexcept nogil:
 *             return PyDataType_ELSIZE(self)
*/

static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_8itemsize_itemsize(PyArray_Descr *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":245
 *         @property
 *         cdef inline npy_intp itemsize(self) noexcept nogil:
 *             return PyDataType_ELSIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyDataType_ELSIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":243
 *         cdef int type_num
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp itemsize(self) noexcept nogil:
 *             return PyDataType_ELSIZE(self)
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":247
 *             return PyDataType_ELSIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp alignment(self) noexcept nogil:
 *             return PyDataType_ALIGNMENT(self)
*/

static CYTHON_INLINE npy_intp __pyx_f_5numpy_5dtype_9alignment_alignment(PyArray_Descr *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":249
 *         @property
 *         cdef inline npy_intp alignment(self) noexcept nogil:
 *             return PyDataType_ALIGNMENT(self)             # <<<<<<<<<<<<<<
 * 
 *         # Use fields/names with care as they may be NULL.  You must check
*/
  __pyx_r = PyDataType_ALIGNMENT(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":247
 *             return PyDataType_ELSIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp alignment(self) noexcept nogil:
 *             return PyDataType_ALIGNMENT(self)
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":253
 *         # Use fields/names with care as they may be NULL.  You must check
 *         # for this using PyDataType_HASFIELDS.
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline object fields(self):
 *             return <object>PyDataType_FIELDS(self)
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_6fields_fields(PyArray_Descr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("fields", 0);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":255
 *         @property
 *         cdef inline object fields(self):
 *             return <object>PyDataType_FIELDS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDataType_FIELDS(__pyx_v_self);
  __Pyx_INCREF(((PyObject *)__pyx_t_1));
  __pyx_r = ((PyObject *)__pyx_t_1);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":253
 *         # Use fields/names with care as they may be NULL.  You must check
 *         # for this using PyDataType_HASFIELDS.
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline object fields(self):
 *             return <object>PyDataType_FIELDS(self)
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":257
 *             return <object>PyDataType_FIELDS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline tuple names(self):
 *             return <tuple>PyDataType_NAMES(self)
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_5dtype_5names_names(PyArray_Descr *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  __Pyx_RefNannySetupContext("names", 0);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":259
 *         @property
 *         cdef inline tuple names(self):
 *             return <tuple>PyDataType_NAMES(self)             # <<<<<<<<<<<<<<
 * 
 *         # Use PyDataType_HASSUBARRAY to test whether this field is
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDataType_NAMES(__pyx_v_self);
  __Pyx_INCREF(((PyObject*)__pyx_t_1));
  __pyx_r = ((PyObject*)__pyx_t_1);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":257
 *             return <object>PyDataType_FIELDS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline tuple names(self):
 *             return <tuple>PyDataType_NAMES(self)
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":264
 *         # valid (the pointer can be NULL). Most users should access
 *         # this field via the inline helper method PyDataType_SHAPE.
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:
 *             return PyDataType_SUBARRAY(self)
*/

static CYTHON_INLINE PyArray_ArrayDescr *__pyx_f_5numpy_5dtype_8subarray_subarray(PyArray_Descr *__pyx_v_self) {
  PyArray_ArrayDescr *__pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":266
 *         @property
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:
 *             return PyDataType_SUBARRAY(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyDataType_SUBARRAY(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":264
 *         # valid (the pointer can be NULL). Most users should access
 *         # this field via the inline helper method PyDataType_SHAPE.
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline PyArray_ArrayDescr* subarray(self) noexcept nogil:
 *             return PyDataType_SUBARRAY(self)
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":268
 *             return PyDataType_SUBARRAY(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_uint64 flags(self) noexcept nogil:
 *             """The data types flags."""
*/

static CYTHON_INLINE npy_uint64 __pyx_f_5numpy_5dtype_5flags_flags(PyArray_Descr *__pyx_v_self) {
  npy_uint64 __pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":271
 *         cdef inline npy_uint64 flags(self) noexcept nogil:
 *             """The data types flags."""
 *             return PyDataType_FLAGS(self)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = PyDataType_FLAGS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":268
 *             return PyDataType_SUBARRAY(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_uint64 flags(self) noexcept nogil:
 *             """The data types flags."""
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":280
 *     ctypedef class numpy.broadcast [object PyArrayMultiIterObject, check_size ignore]:
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int numiter(self) noexcept nogil:
 *             """The number of arrays that need to be broadcast to the same shape."""
*/

static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_7numiter_numiter(PyArrayMultiIterObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":283
 *         cdef inline int numiter(self) noexcept nogil:
 *             """The number of arrays that need to be broadcast to the same shape."""
 *             return PyArray_MultiIter_NUMITER(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_MultiIter_NUMITER(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":280
 *     ctypedef class numpy.broadcast [object PyArrayMultiIterObject, check_size ignore]:
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int numiter(self) noexcept nogil:
 *             """The number of arrays that need to be broadcast to the same shape."""
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":285
 *             return PyArray_MultiIter_NUMITER(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """The total broadcasted size."""
*/

static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_4size_size(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":288
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """The total broadcasted size."""
 *             return PyArray_MultiIter_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_MultiIter_SIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":285
 *             return PyArray_MultiIter_NUMITER(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """The total broadcasted size."""
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":290
 *             return PyArray_MultiIter_SIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp index(self) noexcept nogil:
 *             """The current (1-d) index into the broadcasted result."""
*/

static CYTHON_INLINE npy_intp __pyx_f_5numpy_9broadcast_5index_index(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":293
 *         cdef inline npy_intp index(self) noexcept nogil:
 *             """The current (1-d) index into the broadcasted result."""
 *             return PyArray_MultiIter_INDEX(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_MultiIter_INDEX(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":290
 *             return PyArray_MultiIter_SIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp index(self) noexcept nogil:
 *             """The current (1-d) index into the broadcasted result."""
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":295
 *             return PyArray_MultiIter_INDEX(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int nd(self) noexcept nogil:
 *             """The number of dimensions in the broadcasted result."""
*/

static CYTHON_INLINE int __pyx_f_5numpy_9broadcast_2nd_nd(PyArrayMultiIterObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":298
 *         cdef inline int nd(self) noexcept nogil:
 *             """The number of dimensions in the broadcasted result."""
 *             return PyArray_MultiIter_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_MultiIter_NDIM(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":295
 *             return PyArray_MultiIter_INDEX(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int nd(self) noexcept nogil:
 *             """The number of dimensions in the broadcasted result."""
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":300
 *             return PyArray_MultiIter_NDIM(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:
 *             """The shape of the broadcasted result."""
*/

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_9broadcast_10dimensions_dimensions(PyArrayMultiIterObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":303
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:
 *             """The shape of the broadcasted result."""
 *             return PyArray_MultiIter_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_MultiIter_DIMS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":300
 *             return PyArray_MultiIter_NDIM(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp* dimensions(self) noexcept nogil:
 *             """The shape of the broadcasted result."""
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":305
 *             return PyArray_MultiIter_DIMS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline void** iters(self) noexcept nogil:
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
*/

static CYTHON_INLINE void **__pyx_f_5numpy_9broadcast_5iters_iters(PyArrayMultiIterObject *__pyx_v_self) {
  void **__pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":309
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
 *             On return, the iterators are adjusted for broadcasting."""
 *             return PyArray_MultiIter_ITERS(self)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = PyArray_MultiIter_ITERS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":305
 *             return PyArray_MultiIter_DIMS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline void** iters(self) noexcept nogil:
 *             """An array of iterator objects that holds the iterators for the arrays to be broadcast together.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":323
 *         # Instead, we use properties that map to the corresponding C-API functions.
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline PyObject* base(self) noexcept nogil:
 *             """Returns a borrowed reference to the object owning the data/memory.
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self) {
  PyObject *__pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":327
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 *             return PyArray_BASE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_BASE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":323
 *         # Instead, we use properties that map to the corresponding C-API functions.
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline PyObject* base(self) noexcept nogil:
 *             """Returns a borrowed reference to the object owning the data/memory.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":329
 *             return PyArray_BASE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline dtype descr(self):
 *             """Returns an owned reference to the dtype of the array.
*/

static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self) {
  PyArray_Descr *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyArray_Descr *__pyx_t_1;
  __Pyx_RefNannySetupContext("descr", 0);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":333
 *             """Returns an owned reference to the dtype of the array.
 *             """
 *             return <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyArray_DESCR(__pyx_v_self);
  __Pyx_INCREF((PyObject *)((PyArray_Descr *)__pyx_t_1));
  __pyx_r = ((PyArray_Descr *)__pyx_t_1);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":329
 *             return PyArray_BASE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline dtype descr(self):
 *             """Returns an owned reference to the dtype of the array.
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":335
 *             return <dtype>PyArray_DESCR(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int ndim(self) noexcept nogil:
 *             """Returns the number of dimensions in the array.
*/

static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":339
 *             """Returns the number of dimensions in the array.
 *             """
 *             return PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_NDIM(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":335
 *             return <dtype>PyArray_DESCR(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline int ndim(self) noexcept nogil:
 *             """Returns the number of dimensions in the array.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":341
 *             return PyArray_NDIM(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp *shape(self) noexcept nogil:
 *             """Returns a pointer to the dimensions/shape of the array.
*/

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":347
 *             Can return NULL for 0-dimensional arrays.
 *             """
 *             return PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_DIMS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":341
 *             return PyArray_NDIM(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp *shape(self) noexcept nogil:
 *             """Returns a pointer to the dimensions/shape of the array.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":349
 *             return PyArray_DIMS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp *strides(self) noexcept nogil:
 *             """Returns a pointer to the strides of the array.
*/

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":354
 *             The number of elements matches the number of dimensions of the array (ndim).
 *             """
 *             return PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_STRIDES(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":349
 *             return PyArray_DIMS(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp *strides(self) noexcept nogil:
 *             """Returns a pointer to the strides of the array.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":356
 *             return PyArray_STRIDES(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """Returns the total size (in number of elements) of the array.
*/

static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":360
 *             """Returns the total size (in number of elements) of the array.
 *             """
 *             return PyArray_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
*/
  __pyx_r = PyArray_SIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":356
 *             return PyArray_STRIDES(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline npy_intp size(self) noexcept nogil:
 *             """Returns the total size (in number of elements) of the array.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":362
 *             return PyArray_SIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline char* data(self) noexcept nogil:
 *             """The pointer to the data buffer as a char*.
*/

static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self) {
  char *__pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":369
 *             of `PyArray_DATA()` instead, which returns a 'void*'.
 *             """
 *             return PyArray_BYTES(self)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = PyArray_BYTES(__pyx_v_self);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":362
 *             return PyArray_SIZE(self)
 * 
 *         @property             # <<<<<<<<<<<<<<
 *         cdef inline char* data(self) noexcept nogil:
 *             """The pointer to the data buffer as a char*.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":777
 * ctypedef long double complex clongdouble_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":778
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":777
 * ctypedef long double complex clongdouble_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":780
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":781
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":780
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":783
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":784
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":783
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":786
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":787
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":786
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":789
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":790
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":789
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":792
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":793
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
*/
  __pyx_t_1 = PyDataType_HASSUBARRAY(__pyx_v_d);
  if (__pyx_t_1) {

    /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":794
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_f_5numpy_5dtype_8subarray_subarray(__pyx_v_d)->shape;
    __Pyx_INCREF(((PyObject*)__pyx_t_2));
    __pyx_r = ((PyObject*)__pyx_t_2);
    goto __pyx_L0;

    /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":793
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
*/
  }

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":796
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_empty_tuple);
    __pyx_r = __pyx_mstate_global->__pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":792
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":995
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
*/

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":996
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
*/
  Py_INCREF(__pyx_v_base);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":997
 * cdef inline void set_array_base(ndarray arr, object base) except *:
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
*/
  __pyx_t_1 = PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 997, __pyx_L1_error)

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":995
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base) except *:             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("numpy.set_array_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":999
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
*/

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1000
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
*/
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1001
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
*/
  __pyx_t_1 = (__pyx_v_base == NULL);
  if (__pyx_t_1) {

    /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1002
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1001
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
*/
  }

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1003
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":999
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
*/

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1007
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
*/

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1008
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1009
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")
*/
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 1009, __pyx_L3_error)

      /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1008
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1010
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
*/
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_Exception))));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1010, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1011
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy._core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
*/
      __pyx_t_9 = NULL;
      __pyx_t_10 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_mstate_global->__pyx_kp_u_numpy__core_multiarray_failed_to};
        __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ImportError)), __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1011, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1011, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1008
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1007
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1013
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
*/

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1014
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1015
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")
*/
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 1015, __pyx_L3_error)

      /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1014
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1016
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.umath failed to import")
 * 
*/
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_Exception))));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1016, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1017
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
*/
      __pyx_t_9 = NULL;
      __pyx_t_10 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_mstate_global->__pyx_kp_u_numpy__core_umath_failed_to_impo};
        __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ImportError)), __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1017, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1017, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1014
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1013
 *         raise ImportError("numpy._core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1019
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
*/

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1020
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1021
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")
*/
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 1021, __pyx_L3_error)

      /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1020
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1022
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy._core.umath failed to import")
 * 
*/
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(((PyTypeObject*)PyExc_Exception))));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1022, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1023
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy._core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * 
*/
      __pyx_t_9 = NULL;
      __pyx_t_10 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_mstate_global->__pyx_kp_u_numpy__core_umath_failed_to_impo};
        __pyx_t_8 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ImportError)), __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1023, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1023, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1020
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
*/
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1019
 *         raise ImportError("numpy._core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1026
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
*/

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1038
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1026
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1041
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
*/

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1053
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1041
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj) noexcept:             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1056
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
*/

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1063
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1056
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1066
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
*/

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1070
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1066
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1073
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
*/

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1077
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../_temp/setup-uv-cache/builds-v0/.tmpxKX0vD/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1073
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) noexcept nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
*/

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":8
 * #===============================================================================
 * 
 * cdef PolarizationCpp PolarizationFromStr(str polStr):             # <<<<<<<<<<<<<<
 *     if polStr == "p":
 *         return P_POL
*/

static enum TMM::Polarization __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_PolarizationFromStr(PyObject *__pyx_v_polStr) {
  enum TMM::Polarization __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PolarizationFromStr", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":9
 * 
 * cdef PolarizationCpp PolarizationFromStr(str polStr):
 *     if polStr == "p":             # <<<<<<<<<<<<<<
 *         return P_POL
 *     elif polStr == "s":
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_polStr, __pyx_mstate_global->__pyx_n_u_p, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 9, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":10
 * cdef PolarizationCpp PolarizationFromStr(str polStr):
 *     if polStr == "p":
 *         return P_POL             # <<<<<<<<<<<<<<
 *     elif polStr == "s":
 *         return S_POL
*/
    __pyx_r = TMM::Polarization::P_POL;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":9
 * 
 * cdef PolarizationCpp PolarizationFromStr(str polStr):
 *     if polStr == "p":             # <<<<<<<<<<<<<<
 *         return P_POL
 *     elif polStr == "s":
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":11
 *     if polStr == "p":
 *         return P_POL
 *     elif polStr == "s":             # <<<<<<<<<<<<<<
 *         return S_POL
 *     else:
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_polStr, __pyx_mstate_global->__pyx_n_u_s, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 11, __pyx_L1_error)
  if (likely(__pyx_t_1)) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":12
 *         return P_POL
 *     elif polStr == "s":
 *         return S_POL             # <<<<<<<<<<<<<<
 *     else:
 *         raise NotImplementedError()
*/
    __pyx_r = TMM::Polarization::S_POL;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":11
 *     if polStr == "p":
 *         return P_POL
 *     elif polStr == "s":             # <<<<<<<<<<<<<<
 *         return S_POL
 *     else:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":14
 *         return S_POL
 *     else:
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 * cdef WaveDirectionCpp WaveDirectionFromStr(str waveStr):
*/
  /*else*/ {
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_NotImplementedError)), __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 14, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":8
 * #===============================================================================
 * 
 * cdef PolarizationCpp PolarizationFromStr(str polStr):             # <<<<<<<<<<<<<<
 *     if polStr == "p":
 *         return P_POL
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.PolarizationFromStr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = (enum TMM::Polarization) 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":16
 *         raise NotImplementedError()
 * 
 * cdef WaveDirectionCpp WaveDirectionFromStr(str waveStr):             # <<<<<<<<<<<<<<
 *     if waveStr == "forward":
 *         return F
*/

static enum TMM::WaveDirection __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_WaveDirectionFromStr(PyObject *__pyx_v_waveStr) {
  enum TMM::WaveDirection __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("WaveDirectionFromStr", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":17
 * 
 * cdef WaveDirectionCpp WaveDirectionFromStr(str waveStr):
 *     if waveStr == "forward":             # <<<<<<<<<<<<<<
 *         return F
 *     elif waveStr == "backward":
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_waveStr, __pyx_mstate_global->__pyx_n_u_forward, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 17, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":18
 * cdef WaveDirectionCpp WaveDirectionFromStr(str waveStr):
 *     if waveStr == "forward":
 *         return F             # <<<<<<<<<<<<<<
 *     elif waveStr == "backward":
 *         return B
*/
    __pyx_r = TMM::WaveDirection::F;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":17
 * 
 * cdef WaveDirectionCpp WaveDirectionFromStr(str waveStr):
 *     if waveStr == "forward":             # <<<<<<<<<<<<<<
 *         return F
 *     elif waveStr == "backward":
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":19
 *     if waveStr == "forward":
 *         return F
 *     elif waveStr == "backward":             # <<<<<<<<<<<<<<
 *         return B
 *     elif waveStr == "total":
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_waveStr, __pyx_mstate_global->__pyx_n_u_backward, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 19, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":20
 *         return F
 *     elif waveStr == "backward":
 *         return B             # <<<<<<<<<<<<<<
 *     elif waveStr == "total":
 *         return TOT
*/
    __pyx_r = TMM::WaveDirection::B;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":19
 *     if waveStr == "forward":
 *         return F
 *     elif waveStr == "backward":             # <<<<<<<<<<<<<<
 *         return B
 *     elif waveStr == "total":
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":21
 *     elif waveStr == "backward":
 *         return B
 *     elif waveStr == "total":             # <<<<<<<<<<<<<<
 *         return TOT
 *     else:
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_waveStr, __pyx_mstate_global->__pyx_n_u_total, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 21, __pyx_L1_error)
  if (likely(__pyx_t_1)) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":22
 *         return B
 *     elif waveStr == "total":
 *         return TOT             # <<<<<<<<<<<<<<
 *     else:
 *         raise NotImplementedError()
*/
    __pyx_r = TMM::WaveDirection::TOT;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":21
 *     elif waveStr == "backward":
 *         return B
 *     elif waveStr == "total":             # <<<<<<<<<<<<<<
 *         return TOT
 *     else:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":24
 *         return TOT
 *     else:
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 * cdef NonlinearTmmModeCpp NonlinearTmmModeFromStr(str modeStr):
*/
  /*else*/ {
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_NotImplementedError)), __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 24, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":16
 *         raise NotImplementedError()
 * 
 * cdef WaveDirectionCpp WaveDirectionFromStr(str waveStr):             # <<<<<<<<<<<<<<
 *     if waveStr == "forward":
 *         return F
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.WaveDirectionFromStr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = (enum TMM::WaveDirection) 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":26
 *         raise NotImplementedError()
 * 
 * cdef NonlinearTmmModeCpp NonlinearTmmModeFromStr(str modeStr):             # <<<<<<<<<<<<<<
 *     if modeStr == "incident":
 *         return MODE_INCIDENT
*/

static enum TMM::NonlinearTmmMode __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTmmModeFromStr(PyObject *__pyx_v_modeStr) {
  enum TMM::NonlinearTmmMode __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("NonlinearTmmModeFromStr", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":27
 * 
 * cdef NonlinearTmmModeCpp NonlinearTmmModeFromStr(str modeStr):
 *     if modeStr == "incident":             # <<<<<<<<<<<<<<
 *         return MODE_INCIDENT
 *     elif modeStr == "nonlinear":
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_modeStr, __pyx_mstate_global->__pyx_n_u_incident, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 27, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":28
 * cdef NonlinearTmmModeCpp NonlinearTmmModeFromStr(str modeStr):
 *     if modeStr == "incident":
 *         return MODE_INCIDENT             # <<<<<<<<<<<<<<
 *     elif modeStr == "nonlinear":
 *         return MODE_NONLINEAR
*/
    __pyx_r = TMM::NonlinearTmmMode::MODE_INCIDENT;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":27
 * 
 * cdef NonlinearTmmModeCpp NonlinearTmmModeFromStr(str modeStr):
 *     if modeStr == "incident":             # <<<<<<<<<<<<<<
 *         return MODE_INCIDENT
 *     elif modeStr == "nonlinear":
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":29
 *     if modeStr == "incident":
 *         return MODE_INCIDENT
 *     elif modeStr == "nonlinear":             # <<<<<<<<<<<<<<
 *         return MODE_NONLINEAR
 *     else:
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_modeStr, __pyx_mstate_global->__pyx_n_u_nonlinear, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 29, __pyx_L1_error)
  if (likely(__pyx_t_1)) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":30
 *         return MODE_INCIDENT
 *     elif modeStr == "nonlinear":
 *         return MODE_NONLINEAR             # <<<<<<<<<<<<<<
 *     else:
 *         raise ValueError()
*/
    __pyx_r = TMM::NonlinearTmmMode::MODE_NONLINEAR;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":29
 *     if modeStr == "incident":
 *         return MODE_INCIDENT
 *     elif modeStr == "nonlinear":             # <<<<<<<<<<<<<<
 *         return MODE_NONLINEAR
 *     else:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":32
 *         return MODE_NONLINEAR
 *     else:
 *         raise ValueError()             # <<<<<<<<<<<<<<
 * 
 * cdef NonlinearProcessCpp NonlinearProcessFromStr(str processStr):
*/
  /*else*/ {
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 32, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":26
 *         raise NotImplementedError()
 * 
 * cdef NonlinearTmmModeCpp NonlinearTmmModeFromStr(str modeStr):             # <<<<<<<<<<<<<<
 *     if modeStr == "incident":
 *         return MODE_INCIDENT
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTmmModeFromStr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = (enum TMM::NonlinearTmmMode) 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":34
 *         raise ValueError()
 * 
 * cdef NonlinearProcessCpp NonlinearProcessFromStr(str processStr):             # <<<<<<<<<<<<<<
 *     if processStr == "sfg":
 *         return SFG
*/

static enum TMM::NonlinearProcess __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearProcessFromStr(PyObject *__pyx_v_processStr) {
  enum TMM::NonlinearProcess __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("NonlinearProcessFromStr", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":35
 * 
 * cdef NonlinearProcessCpp NonlinearProcessFromStr(str processStr):
 *     if processStr == "sfg":             # <<<<<<<<<<<<<<
 *         return SFG
 *     elif processStr == "dfg":
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_processStr, __pyx_mstate_global->__pyx_n_u_sfg, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 35, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":36
 * cdef NonlinearProcessCpp NonlinearProcessFromStr(str processStr):
 *     if processStr == "sfg":
 *         return SFG             # <<<<<<<<<<<<<<
 *     elif processStr == "dfg":
 *         return DFG
*/
    __pyx_r = TMM::NonlinearProcess::SFG;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":35
 * 
 * cdef NonlinearProcessCpp NonlinearProcessFromStr(str processStr):
 *     if processStr == "sfg":             # <<<<<<<<<<<<<<
 *         return SFG
 *     elif processStr == "dfg":
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":37
 *     if processStr == "sfg":
 *         return SFG
 *     elif processStr == "dfg":             # <<<<<<<<<<<<<<
 *         return DFG
 *     elif processStr == "spdc":
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_processStr, __pyx_mstate_global->__pyx_n_u_dfg, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 37, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":38
 *         return SFG
 *     elif processStr == "dfg":
 *         return DFG             # <<<<<<<<<<<<<<
 *     elif processStr == "spdc":
 *         return SPDC
*/
    __pyx_r = TMM::NonlinearProcess::DFG;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":37
 *     if processStr == "sfg":
 *         return SFG
 *     elif processStr == "dfg":             # <<<<<<<<<<<<<<
 *         return DFG
 *     elif processStr == "spdc":
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":39
 *     elif processStr == "dfg":
 *         return DFG
 *     elif processStr == "spdc":             # <<<<<<<<<<<<<<
 *         return SPDC
 *     else:
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_processStr, __pyx_mstate_global->__pyx_n_u_spdc, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 39, __pyx_L1_error)
  if (likely(__pyx_t_1)) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":40
 *         return DFG
 *     elif processStr == "spdc":
 *         return SPDC             # <<<<<<<<<<<<<<
 *     else:
 *         raise ValueError()
*/
    __pyx_r = TMM::NonlinearProcess::SPDC;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":39
 *     elif processStr == "dfg":
 *         return DFG
 *     elif processStr == "spdc":             # <<<<<<<<<<<<<<
 *         return SPDC
 *     else:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":42
 *         return SPDC
 *     else:
 *         raise ValueError()             # <<<<<<<<<<<<<<
 * 
 * cdef WaveTypeCpp WaveTypeFromStr(str waveTypeStr):
*/
  /*else*/ {
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 42, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":34
 *         raise ValueError()
 * 
 * cdef NonlinearProcessCpp NonlinearProcessFromStr(str processStr):             # <<<<<<<<<<<<<<
 *     if processStr == "sfg":
 *         return SFG
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearProcessFromStr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = (enum TMM::NonlinearProcess) 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":44
 *         raise ValueError()
 * 
 * cdef WaveTypeCpp WaveTypeFromStr(str waveTypeStr):             # <<<<<<<<<<<<<<
 *     if waveTypeStr == "planewave":
 *         return PLANEWAVE
*/

static enum TMM::WaveType __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_WaveTypeFromStr(PyObject *__pyx_v_waveTypeStr) {
  enum TMM::WaveType __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("WaveTypeFromStr", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":45
 * 
 * cdef WaveTypeCpp WaveTypeFromStr(str waveTypeStr):
 *     if waveTypeStr == "planewave":             # <<<<<<<<<<<<<<
 *         return PLANEWAVE
 *     if waveTypeStr == "gaussian":
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_waveTypeStr, __pyx_mstate_global->__pyx_n_u_planewave, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 45, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":46
 * cdef WaveTypeCpp WaveTypeFromStr(str waveTypeStr):
 *     if waveTypeStr == "planewave":
 *         return PLANEWAVE             # <<<<<<<<<<<<<<
 *     if waveTypeStr == "gaussian":
 *         return GAUSSIANWAVE
*/
    __pyx_r = TMM::WaveType::PLANEWAVE;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":45
 * 
 * cdef WaveTypeCpp WaveTypeFromStr(str waveTypeStr):
 *     if waveTypeStr == "planewave":             # <<<<<<<<<<<<<<
 *         return PLANEWAVE
 *     if waveTypeStr == "gaussian":
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":47
 *     if waveTypeStr == "planewave":
 *         return PLANEWAVE
 *     if waveTypeStr == "gaussian":             # <<<<<<<<<<<<<<
 *         return GAUSSIANWAVE
 *     if waveTypeStr == "tukey":
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_waveTypeStr, __pyx_mstate_global->__pyx_n_u_gaussian, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 47, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":48
 *         return PLANEWAVE
 *     if waveTypeStr == "gaussian":
 *         return GAUSSIANWAVE             # <<<<<<<<<<<<<<
 *     if waveTypeStr == "tukey":
 *         return TUKEYWAVE
*/
    __pyx_r = TMM::WaveType::GAUSSIANWAVE;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":47
 *     if waveTypeStr == "planewave":
 *         return PLANEWAVE
 *     if waveTypeStr == "gaussian":             # <<<<<<<<<<<<<<
 *         return GAUSSIANWAVE
 *     if waveTypeStr == "tukey":
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":49
 *     if waveTypeStr == "gaussian":
 *         return GAUSSIANWAVE
 *     if waveTypeStr == "tukey":             # <<<<<<<<<<<<<<
 *         return TUKEYWAVE
 *     if waveTypeStr == "spdc":
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_waveTypeStr, __pyx_mstate_global->__pyx_n_u_tukey, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 49, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":50
 *         return GAUSSIANWAVE
 *     if waveTypeStr == "tukey":
 *         return TUKEYWAVE             # <<<<<<<<<<<<<<
 *     if waveTypeStr == "spdc":
 *         return SPDCWAVE;
*/
    __pyx_r = TMM::WaveType::TUKEYWAVE;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":49
 *     if waveTypeStr == "gaussian":
 *         return GAUSSIANWAVE
 *     if waveTypeStr == "tukey":             # <<<<<<<<<<<<<<
 *         return TUKEYWAVE
 *     if waveTypeStr == "spdc":
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":51
 *     if waveTypeStr == "tukey":
 *         return TUKEYWAVE
 *     if waveTypeStr == "spdc":             # <<<<<<<<<<<<<<
 *         return SPDCWAVE;
 *     else:
*/
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_waveTypeStr, __pyx_mstate_global->__pyx_n_u_spdc, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 51, __pyx_L1_error)
  if (likely(__pyx_t_1)) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":52
 *         return TUKEYWAVE
 *     if waveTypeStr == "spdc":
 *         return SPDCWAVE;             # <<<<<<<<<<<<<<
 *     else:
 *         raise NotImplementedError()
*/
    __pyx_r = TMM::WaveType::SPDCWAVE;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":51
 *     if waveTypeStr == "tukey":
 *         return TUKEYWAVE
 *     if waveTypeStr == "spdc":             # <<<<<<<<<<<<<<
 *         return SPDCWAVE;
 *     else:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":54
 *         return SPDCWAVE;
 *     else:
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 * cdef paramDict = {"wl": PARAM_WL, \
*/
  /*else*/ {
    __pyx_t_3 = NULL;
    __pyx_t_4 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_NotImplementedError)), __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 54, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 54, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":44
 *         raise ValueError()
 * 
 * cdef WaveTypeCpp WaveTypeFromStr(str waveTypeStr):             # <<<<<<<<<<<<<<
 *     if waveTypeStr == "planewave":
 *         return PLANEWAVE
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.WaveTypeFromStr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = (enum TMM::WaveType) 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":84
 * 
 * 
 * cdef TMMParamCpp TmmParamFromStr(str paramStr):             # <<<<<<<<<<<<<<
 *     return paramDict[paramStr]
 * 
*/

static enum TMM::TMMParam __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_TmmParamFromStr(PyObject *__pyx_v_paramStr) {
  enum TMM::TMMParam __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  enum TMM::TMMParam __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("TmmParamFromStr", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":85
 * 
 * cdef TMMParamCpp TmmParamFromStr(str paramStr):
 *     return paramDict[paramStr]             # <<<<<<<<<<<<<<
 * 
 * #===============================================================================
*/
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_paramDict, __pyx_v_paramStr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((enum TMM::TMMParam)__Pyx_PyLong_As_enum__TMM_3a__3a_TMMParam(__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":84
 * 
 * 
 * cdef TMMParamCpp TmmParamFromStr(str paramStr):             # <<<<<<<<<<<<<<
 *     return paramDict[paramStr]
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.TmmParamFromStr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = (enum TMM::TMMParam) 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":102
 *     cdef object _parent;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         # This object is always allocated by Material class
 *         _thisptr = NULL
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__cinit__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor___cinit__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_self) {
  CYTHON_UNUSED void *__pyx_v__thisptr;
  int __pyx_r;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":104
 *     def __cinit__(self):
 *         # This object is always allocated by Material class
 *         _thisptr = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  __pyx_v__thisptr = NULL;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":102
 *     cdef object _parent;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         # This object is always allocated by Material class
 *         _thisptr = NULL
*/

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":106
 *         _thisptr = NULL
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         # This object is always deallocated by Material class
 *         pass
*/

/* Python wrapper */
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_2__dealloc__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_2__dealloc__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_self) {

  /* function exit code */
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":110
 *         pass
 * 
 *     cdef _Init(self, Chi2TensorCpp* ptr, object parent):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         self._parent = parent # Avoid dealloc of parent
*/

static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_self, TMM::Chi2Tensor *__pyx_v_ptr, PyObject *__pyx_v_parent) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_Init", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":111
 * 
 *     cdef _Init(self, Chi2TensorCpp* ptr, object parent):
 *         self._thisptr = ptr             # <<<<<<<<<<<<<<
 *         self._parent = parent # Avoid dealloc of parent
 * 
*/
  __pyx_v_self->_thisptr = __pyx_v_ptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":112
 *     cdef _Init(self, Chi2TensorCpp* ptr, object parent):
 *         self._thisptr = ptr
 *         self._parent = parent # Avoid dealloc of parent             # <<<<<<<<<<<<<<
 * 
 *     def Update(self, **kwargs):
*/
  __Pyx_INCREF(__pyx_v_parent);
  __Pyx_GIVEREF(__pyx_v_parent);
  __Pyx_GOTREF(__pyx_v_self->_parent);
  __Pyx_DECREF(__pyx_v_self->_parent);
  __pyx_v_self->_parent = __pyx_v_parent;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":110
 *         pass
 * 
 *     cdef _Init(self, Chi2TensorCpp* ptr, object parent):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         self._parent = parent # Avoid dealloc of parent
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":114
 *         self._parent = parent # Avoid dealloc of parent
 * 
 *     def Update(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """Update(**kwargs)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_5Update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_4Update, "Update(**kwargs)\n\n        This method changes the values of second-order nonlinearity tensor. By\n        default the tensor is zero, not rotated and distinctFields set to True.\n\n        Parameters\n        ----------\n        chiXYZ : float, optional\n            Updates tensor value of second-order nonlinearity, where X, Y, Z\n            denote numbers from 1..3 corresponding to x-, y-, z-axis.\n        dXK : float, optional\n            Updates tensor values by contracted notation, where dXK = 0.5 * chiXYZ\n            and K is defined as: 1 = 11; 2 = 22; 3 = 33; 4 = 23, 32; 5 = 31, 13; 6 = 12, 21.\n        distinctFields : bool, optional\n            If set to True, then assumes that two input waves in nonlinear process\n            are the same (e.g second harmonic generation). If False, then the\n            input fields are not equal (as a result the result will be multiplied by 2).\n            Default is true.\n        phiX : float, optional\n            The rotation angle (radians) around x-axis of the second order nonlinear tensor.\n        phiY : float, optional\n            The rotation angle (radians) around y-axis of the second order nonlinear tensor.\n        phiZ : float, optional\n            The rotation angle (radians) around z-axis of the second order nonlinear tensor.\n\n        Examples\n        --------\n        Updates nonlinear tensor of material and then rotates it.\n\n        >>> mat = Material.Static(1.5)\n        >>> mat.chi2.Update(d11 = 1e-12, d22 = 2e-12, chi23 = 3e-12)\n        >>> mat.chi2.Update(phiX = 0.2, phiY = 0.5, phiZ = -0.1)\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_5Update = {"Update", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_5Update, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_4Update};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_5Update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Update (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("Update", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (__pyx_kwds_len > 0) {
    if (unlikely(__Pyx_CheckKeywordStrings("Update", __pyx_kwds) == -1)) return NULL;
    __pyx_v_kwargs = __Pyx_KwargsAsDict_FASTCALL(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_4Update(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_4Update(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_self, PyObject *__pyx_v_kwargs) {
  int __pyx_v_i1;
  int __pyx_v_i2;
  int __pyx_v_i3;
  double __pyx_v_value;
  PyObject *__pyx_v_k = NULL;
  bool __pyx_v_distinctFields;
  double __pyx_v_phiX;
  double __pyx_v_phiY;
  double __pyx_v_phiZ;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  double __pyx_t_11;
  bool __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Update", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":153
 * 
 *         # d-values
 *         for k in list(kwargs.keys()):             # <<<<<<<<<<<<<<
 *             if (not k.startswith("d")) or (len(k) != 3):
 *                 continue
*/
  __pyx_t_1 = __Pyx_PyDict_Keys(__pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 153, __pyx_L1_error)
      #endif
      if (__pyx_t_3 >= __pyx_temp) break;
    }
    __pyx_t_2 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
    ++__pyx_t_3;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":154
 *         # d-values
 *         for k in list(kwargs.keys()):
 *             if (not k.startswith("d")) or (len(k) != 3):             # <<<<<<<<<<<<<<
 *                 continue
 * 
*/
    __pyx_t_5 = __pyx_v_k;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_n_u_d};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_startswith, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_8 = (!__pyx_t_7);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_4 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_9 = PyObject_Length(__pyx_v_k); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 154, __pyx_L1_error)
    __pyx_t_8 = (__pyx_t_9 != 3);
    __pyx_t_4 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_4) {

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":155
 *         for k in list(kwargs.keys()):
 *             if (not k.startswith("d")) or (len(k) != 3):
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             i1 = int(k[1])
*/
      goto __pyx_L3_continue;

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":154
 *         # d-values
 *         for k in list(kwargs.keys()):
 *             if (not k.startswith("d")) or (len(k) != 3):             # <<<<<<<<<<<<<<
 *                 continue
 * 
*/
    }

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":157
 *                 continue
 * 
 *             i1 = int(k[1])             # <<<<<<<<<<<<<<
 *             i2 = int(k[2])
 *             value = kwargs.pop(k)
*/
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_k, 1, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_10 = __Pyx_PyLong_As_int(__pyx_t_5); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_i1 = __pyx_t_10;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":158
 * 
 *             i1 = int(k[1])
 *             i2 = int(k[2])             # <<<<<<<<<<<<<<
 *             value = kwargs.pop(k)
 *             self._thisptr.SetD(i1, i2, value)
*/
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_k, 2, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_10 = __Pyx_PyLong_As_int(__pyx_t_2); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_i2 = __pyx_t_10;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":159
 *             i1 = int(k[1])
 *             i2 = int(k[2])
 *             value = kwargs.pop(k)             # <<<<<<<<<<<<<<
 *             self._thisptr.SetD(i1, i2, value)
 * 
*/
    __pyx_t_2 = __Pyx_PyDict_Pop(__pyx_v_kwargs, __pyx_v_k, ((PyObject *)NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __Pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 159, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_value = __pyx_t_11;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":160
 *             i2 = int(k[2])
 *             value = kwargs.pop(k)
 *             self._thisptr.SetD(i1, i2, value)             # <<<<<<<<<<<<<<
 * 
 *         # chi2-values
*/
    try {
      __pyx_v_self->_thisptr->SetD(__pyx_v_i1, __pyx_v_i2, __pyx_v_value);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 160, __pyx_L1_error)
    }

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":153
 * 
 *         # d-values
 *         for k in list(kwargs.keys()):             # <<<<<<<<<<<<<<
 *             if (not k.startswith("d")) or (len(k) != 3):
 *                 continue
*/
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":163
 * 
 *         # chi2-values
 *         for k in list(kwargs.keys()):             # <<<<<<<<<<<<<<
 *             if (not k.startswith("chi")) or (len(k) != 6):
 *                 continue
*/
  __pyx_t_1 = __Pyx_PyDict_Keys(__pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 163, __pyx_L1_error)
      #endif
      if (__pyx_t_3 >= __pyx_temp) break;
    }
    __pyx_t_2 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
    ++__pyx_t_3;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":164
 *         # chi2-values
 *         for k in list(kwargs.keys()):
 *             if (not k.startswith("chi")) or (len(k) != 6):             # <<<<<<<<<<<<<<
 *                 continue
 *             i1 = int(k[3])
*/
    __pyx_t_5 = __pyx_v_k;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_mstate_global->__pyx_n_u_chi};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_startswith, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 164, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = (!__pyx_t_8);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_4 = __pyx_t_7;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_9 = PyObject_Length(__pyx_v_k); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 164, __pyx_L1_error)
    __pyx_t_7 = (__pyx_t_9 != 6);
    __pyx_t_4 = __pyx_t_7;
    __pyx_L12_bool_binop_done:;
    if (__pyx_t_4) {

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":165
 *         for k in list(kwargs.keys()):
 *             if (not k.startswith("chi")) or (len(k) != 6):
 *                 continue             # <<<<<<<<<<<<<<
 *             i1 = int(k[3])
 *             i2 = int(k[4])
*/
      goto __pyx_L9_continue;

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":164
 *         # chi2-values
 *         for k in list(kwargs.keys()):
 *             if (not k.startswith("chi")) or (len(k) != 6):             # <<<<<<<<<<<<<<
 *                 continue
 *             i1 = int(k[3])
*/
    }

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":166
 *             if (not k.startswith("chi")) or (len(k) != 6):
 *                 continue
 *             i1 = int(k[3])             # <<<<<<<<<<<<<<
 *             i2 = int(k[4])
 *             i3 = int(k[5])
*/
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_k, 3, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_10 = __Pyx_PyLong_As_int(__pyx_t_5); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_i1 = __pyx_t_10;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":167
 *                 continue
 *             i1 = int(k[3])
 *             i2 = int(k[4])             # <<<<<<<<<<<<<<
 *             i3 = int(k[5])
 *             value = kwargs.pop(k)
*/
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_k, 4, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_10 = __Pyx_PyLong_As_int(__pyx_t_2); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_i2 = __pyx_t_10;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":168
 *             i1 = int(k[3])
 *             i2 = int(k[4])
 *             i3 = int(k[5])             # <<<<<<<<<<<<<<
 *             value = kwargs.pop(k)
 *             self._thisptr.SetChi2(i1, i2, i3, value)
*/
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_k, 5, long, 1, __Pyx_PyLong_From_long, 0, 0, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 168, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_10 = __Pyx_PyLong_As_int(__pyx_t_5); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 168, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_i3 = __pyx_t_10;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":169
 *             i2 = int(k[4])
 *             i3 = int(k[5])
 *             value = kwargs.pop(k)             # <<<<<<<<<<<<<<
 *             self._thisptr.SetChi2(i1, i2, i3, value)
 * 
*/
    __pyx_t_5 = __Pyx_PyDict_Pop(__pyx_v_kwargs, __pyx_v_k, ((PyObject *)NULL)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = __Pyx_PyFloat_AsDouble(__pyx_t_5); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 169, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_value = __pyx_t_11;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":170
 *             i3 = int(k[5])
 *             value = kwargs.pop(k)
 *             self._thisptr.SetChi2(i1, i2, i3, value)             # <<<<<<<<<<<<<<
 * 
 *         # Distinct fields
*/
    try {
      __pyx_v_self->_thisptr->SetChi2(__pyx_v_i1, __pyx_v_i2, __pyx_v_i3, __pyx_v_value);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 170, __pyx_L1_error)
    }

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":163
 * 
 *         # chi2-values
 *         for k in list(kwargs.keys()):             # <<<<<<<<<<<<<<
 *             if (not k.startswith("chi")) or (len(k) != 6):
 *                 continue
*/
    __pyx_L9_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":174
 *         # Distinct fields
 *         cdef bool distinctFields
 *         if "distinctFields" in kwargs:             # <<<<<<<<<<<<<<
 *             distinctFields = kwargs.pop("distinctFields")
 *             self._thisptr.SetDistinctFields(distinctFields)
*/
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_mstate_global->__pyx_n_u_distinctFields, __pyx_v_kwargs, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 174, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":175
 *         cdef bool distinctFields
 *         if "distinctFields" in kwargs:
 *             distinctFields = kwargs.pop("distinctFields")             # <<<<<<<<<<<<<<
 *             self._thisptr.SetDistinctFields(distinctFields)
 * 
*/
    __pyx_t_1 = __Pyx_PyDict_Pop(__pyx_v_kwargs, __pyx_mstate_global->__pyx_n_u_distinctFields, ((PyObject *)NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_12 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_distinctFields = __pyx_t_12;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":176
 *         if "distinctFields" in kwargs:
 *             distinctFields = kwargs.pop("distinctFields")
 *             self._thisptr.SetDistinctFields(distinctFields)             # <<<<<<<<<<<<<<
 * 
 *         # Phis (may generate exception if no phiY-Z available)
*/
    try {
      __pyx_v_self->_thisptr->SetDistinctFields(__pyx_v_distinctFields);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 176, __pyx_L1_error)
    }

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":174
 *         # Distinct fields
 *         cdef bool distinctFields
 *         if "distinctFields" in kwargs:             # <<<<<<<<<<<<<<
 *             distinctFields = kwargs.pop("distinctFields")
 *             self._thisptr.SetDistinctFields(distinctFields)
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":180
 *         # Phis (may generate exception if no phiY-Z available)
 *         cdef double phiX, phiY, phiZ
 *         if "phiX" in kwargs:             # <<<<<<<<<<<<<<
 *             phiX = kwargs.pop("phiX")
 *             phiY = kwargs.pop("phiY")
*/
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_mstate_global->__pyx_n_u_phiX, __pyx_v_kwargs, Py_EQ)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 180, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":181
 *         cdef double phiX, phiY, phiZ
 *         if "phiX" in kwargs:
 *             phiX = kwargs.pop("phiX")             # <<<<<<<<<<<<<<
 *             phiY = kwargs.pop("phiY")
 *             phiZ = kwargs.pop("phiZ")
*/
    __pyx_t_1 = __Pyx_PyDict_Pop(__pyx_v_kwargs, __pyx_mstate_global->__pyx_n_u_phiX, ((PyObject *)NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = __Pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_phiX = __pyx_t_11;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":182
 *         if "phiX" in kwargs:
 *             phiX = kwargs.pop("phiX")
 *             phiY = kwargs.pop("phiY")             # <<<<<<<<<<<<<<
 *             phiZ = kwargs.pop("phiZ")
 *             self._thisptr.SetRotation(phiX, phiY, phiZ)
*/
    __pyx_t_1 = __Pyx_PyDict_Pop(__pyx_v_kwargs, __pyx_mstate_global->__pyx_n_u_phiY, ((PyObject *)NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = __Pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_phiY = __pyx_t_11;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":183
 *             phiX = kwargs.pop("phiX")
 *             phiY = kwargs.pop("phiY")
 *             phiZ = kwargs.pop("phiZ")             # <<<<<<<<<<<<<<
 *             self._thisptr.SetRotation(phiX, phiY, phiZ)
 * 
*/
    __pyx_t_1 = __Pyx_PyDict_Pop(__pyx_v_kwargs, __pyx_mstate_global->__pyx_n_u_phiZ, ((PyObject *)NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = __Pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_phiZ = __pyx_t_11;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":184
 *             phiY = kwargs.pop("phiY")
 *             phiZ = kwargs.pop("phiZ")
 *             self._thisptr.SetRotation(phiX, phiY, phiZ)             # <<<<<<<<<<<<<<
 * 
 *         # Errors
*/
    try {
      __pyx_v_self->_thisptr->SetRotation(__pyx_v_phiX, __pyx_v_phiY, __pyx_v_phiZ);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 184, __pyx_L1_error)
    }

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":180
 *         # Phis (may generate exception if no phiY-Z available)
 *         cdef double phiX, phiY, phiZ
 *         if "phiX" in kwargs:             # <<<<<<<<<<<<<<
 *             phiX = kwargs.pop("phiX")
 *             phiY = kwargs.pop("phiY")
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":187
 * 
 *         # Errors
 *         for k in list(kwargs.keys()):             # <<<<<<<<<<<<<<
 *             raise ValueError("Unknown kwarg %s" % (k))
 * 
*/
  __pyx_t_1 = __Pyx_PyDict_Keys(__pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PySequence_ListKeepNew(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_t_5; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 187, __pyx_L1_error)
      #endif
      if (__pyx_t_3 >= __pyx_temp) break;
    }
    __pyx_t_5 = __Pyx_PyList_GetItemRefFast(__pyx_t_1, __pyx_t_3, __Pyx_ReferenceSharing_OwnStrongReference);
    ++__pyx_t_3;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":188
 *         # Errors
 *         for k in list(kwargs.keys()):
 *             raise ValueError("Unknown kwarg %s" % (k))             # <<<<<<<<<<<<<<
 * 
 *     def GetChi2Tensor(self):
*/
    __pyx_t_2 = NULL;
    __pyx_t_13 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_Unknown_kwarg_s, __pyx_v_k); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_6 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_13};
      __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 188, __pyx_L1_error)

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":187
 * 
 *         # Errors
 *         for k in list(kwargs.keys()):             # <<<<<<<<<<<<<<
 *             raise ValueError("Unknown kwarg %s" % (k))
 * 
*/
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":114
 *         self._parent = parent # Avoid dealloc of parent
 * 
 *     def Update(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """Update(**kwargs)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Chi2Tensor.Update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":190
 *             raise ValueError("Unknown kwarg %s" % (k))
 * 
 *     def GetChi2Tensor(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns rotated 3x3x3 second-order nonlinearity tensor.
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_7GetChi2Tensor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_6GetChi2Tensor, "\n        Returns rotated 3x3x3 second-order nonlinearity tensor.\n\n        Returns\n        -------\n        ndarray(3, 3, 3) of floats\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_7GetChi2Tensor = {"GetChi2Tensor", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_7GetChi2Tensor, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_6GetChi2Tensor};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_7GetChi2Tensor(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("GetChi2Tensor (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("GetChi2Tensor", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("GetChi2Tensor", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_6GetChi2Tensor(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_6GetChi2Tensor(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_self) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  PyArrayObject *__pyx_v_res = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  double __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("GetChi2Tensor", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":199
 *         """
 *         cdef int i, j, k
 *         cdef np.ndarray res = np.zeros([3, 3, 3], dtype = float)             # <<<<<<<<<<<<<<
 * 
 *         # Eigency currently doesn't support Tensors
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyList_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_int_3);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_mstate_global->__pyx_int_3) != (0)) __PYX_ERR(0, 199, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_int_3);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_mstate_global->__pyx_int_3) != (0)) __PYX_ERR(0, 199, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_int_3);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 2, __pyx_mstate_global->__pyx_int_3) != (0)) __PYX_ERR(0, 199, __pyx_L1_error);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_2, __pyx_t_3};
    __pyx_t_6 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_dtype, ((PyObject *)(&PyFloat_Type)), __pyx_t_6, __pyx_callargs+2, 0) < (0)) __PYX_ERR(0, 199, __pyx_L1_error)
    __pyx_t_1 = __Pyx_Object_Vectorcall_CallFromBuilder((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 199, __pyx_L1_error)
  __pyx_v_res = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":202
 * 
 *         # Eigency currently doesn't support Tensors
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             for j in range(3):
 *                 for k in range(3):
*/
  for (__pyx_t_7 = 0; __pyx_t_7 < 3; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":203
 *         # Eigency currently doesn't support Tensors
 *         for i in range(3):
 *             for j in range(3):             # <<<<<<<<<<<<<<
 *                 for k in range(3):
 *                     res[i, j, k] = self._thisptr.GetChi2Element(i + 1, j + 1, k + 1)
*/
    for (__pyx_t_8 = 0; __pyx_t_8 < 3; __pyx_t_8+=1) {
      __pyx_v_j = __pyx_t_8;

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":204
 *         for i in range(3):
 *             for j in range(3):
 *                 for k in range(3):             # <<<<<<<<<<<<<<
 *                     res[i, j, k] = self._thisptr.GetChi2Element(i + 1, j + 1, k + 1)
 *         return res
*/
      for (__pyx_t_9 = 0; __pyx_t_9 < 3; __pyx_t_9+=1) {
        __pyx_v_k = __pyx_t_9;

        /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":205
 *             for j in range(3):
 *                 for k in range(3):
 *                     res[i, j, k] = self._thisptr.GetChi2Element(i + 1, j + 1, k + 1)             # <<<<<<<<<<<<<<
 *         return res
 * 
*/
        try {
          __pyx_t_10 = __pyx_v_self->_thisptr->GetChi2Element((__pyx_v_i + 1), (__pyx_v_j + 1), (__pyx_v_k + 1));
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 205, __pyx_L1_error)
        }
        __pyx_t_1 = PyFloat_FromDouble(__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 205, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyLong_From_int(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 205, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyLong_From_int(__pyx_v_j); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 205, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_3 = __Pyx_PyLong_From_int(__pyx_v_k); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 205, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 205, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_4);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4) != (0)) __PYX_ERR(0, 205, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_6);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_6) != (0)) __PYX_ERR(0, 205, __pyx_L1_error);
        __Pyx_GIVEREF(__pyx_t_3);
        if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_3) != (0)) __PYX_ERR(0, 205, __pyx_L1_error);
        __pyx_t_4 = 0;
        __pyx_t_6 = 0;
        __pyx_t_3 = 0;
        if (unlikely((PyObject_SetItem(((PyObject *)__pyx_v_res), __pyx_t_2, __pyx_t_1) < 0))) __PYX_ERR(0, 205, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    }
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":206
 *                 for k in range(3):
 *                     res[i, j, k] = self._thisptr.GetChi2Element(i + 1, j + 1, k + 1)
 *         return res             # <<<<<<<<<<<<<<
 * 
 * #===============================================================================
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_res);
  __pyx_r = ((PyObject *)__pyx_v_res);
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":190
 *             raise ValueError("Unknown kwarg %s" % (k))
 * 
 *     def GetChi2Tensor(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns rotated 3x3x3 second-order nonlinearity tensor.
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Chi2Tensor.GetChi2Tensor", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_9__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_8__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Chi2Tensor.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_11__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Chi2Tensor.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_10__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Chi2Tensor.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":236
 *     cdef readonly object chi2
 * 
 *     def __cinit__(self, np.ndarray[double, ndim = 1] wls, np.ndarray[double complex, ndim = 1] ns):             # <<<<<<<<<<<<<<
 *         # Copy is made in c++
 *         self._thisptr = new MaterialCpp(Map[ArrayXd](wls), Map[ArrayXcd](ns))
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_wls = 0;
  PyArrayObject *__pyx_v_ns = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_wls,&__pyx_mstate_global->__pyx_n_u_ns,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 236, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 236, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 236, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__cinit__", 0) < (0)) __PYX_ERR(0, 236, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, i); __PYX_ERR(0, 236, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 236, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 236, __pyx_L3_error)
    }
    __pyx_v_wls = ((PyArrayObject *)values[0]);
    __pyx_v_ns = ((PyArrayObject *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 236, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.Material.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_wls), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "wls", 0))) __PYX_ERR(0, 236, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_ns), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "ns", 0))) __PYX_ERR(0, 236, __pyx_L1_error)
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *)__pyx_v_self), __pyx_v_wls, __pyx_v_ns);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self, PyArrayObject *__pyx_v_wls, PyArrayObject *__pyx_v_ns) {
  TMM::Chi2Tensor *__pyx_v_chi2Ptr;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *__pyx_v_chi2 = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_ns;
  __Pyx_Buffer __pyx_pybuffer_ns;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_wls;
  __Pyx_Buffer __pyx_pybuffer_wls;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  eigency::Map<Eigen::ArrayXd>  __pyx_t_1;
  eigency::Map<Eigen::ArrayXcd>  __pyx_t_2;
  TMM::Material *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __pyx_pybuffer_wls.pybuffer.buf = NULL;
  __pyx_pybuffer_wls.refcount = 0;
  __pyx_pybuffernd_wls.data = NULL;
  __pyx_pybuffernd_wls.rcbuffer = &__pyx_pybuffer_wls;
  __pyx_pybuffer_ns.pybuffer.buf = NULL;
  __pyx_pybuffer_ns.refcount = 0;
  __pyx_pybuffernd_ns.data = NULL;
  __pyx_pybuffernd_ns.rcbuffer = &__pyx_pybuffer_ns;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_wls.rcbuffer->pybuffer, (PyObject*)__pyx_v_wls, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 236, __pyx_L1_error)
  }
  __pyx_pybuffernd_wls.diminfo[0].strides = __pyx_pybuffernd_wls.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_wls.diminfo[0].shape = __pyx_pybuffernd_wls.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_ns.rcbuffer->pybuffer, (PyObject*)__pyx_v_ns, &__Pyx_TypeInfo___pyx_t_double_complex, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 236, __pyx_L1_error)
  }
  __pyx_pybuffernd_ns.diminfo[0].strides = __pyx_pybuffernd_ns.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_ns.diminfo[0].shape = __pyx_pybuffernd_ns.rcbuffer->pybuffer.shape[0];

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":238
 *     def __cinit__(self, np.ndarray[double, ndim = 1] wls, np.ndarray[double complex, ndim = 1] ns):
 *         # Copy is made in c++
 *         self._thisptr = new MaterialCpp(Map[ArrayXd](wls), Map[ArrayXcd](ns))             # <<<<<<<<<<<<<<
 * 
 *         cdef Chi2TensorCpp *chi2Ptr = &self._thisptr.chi2
*/
  try {
    __pyx_t_1 = eigency::Map<Eigen::ArrayXd> (((PyArrayObject *)__pyx_v_wls));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 238, __pyx_L1_error)
  }
  try {
    __pyx_t_2 = eigency::Map<Eigen::ArrayXcd> (((PyArrayObject *)__pyx_v_ns));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 238, __pyx_L1_error)
  }
  try {
    __pyx_t_3 = new TMM::Material(__pyx_t_1, __pyx_t_2);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 238, __pyx_L1_error)
  }
  __pyx_v_self->_thisptr = __pyx_t_3;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":240
 *         self._thisptr = new MaterialCpp(Map[ArrayXd](wls), Map[ArrayXcd](ns))
 * 
 *         cdef Chi2TensorCpp *chi2Ptr = &self._thisptr.chi2             # <<<<<<<<<<<<<<
 *         chi2 = _Chi2Tensor()
 *         chi2._Init(chi2Ptr, self) # This class is passed as parent (avoids dealloc before chi2)
*/
  __pyx_v_chi2Ptr = (&__pyx_v_self->_thisptr->chi2);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":241
 * 
 *         cdef Chi2TensorCpp *chi2Ptr = &self._thisptr.chi2
 *         chi2 = _Chi2Tensor()             # <<<<<<<<<<<<<<
 *         chi2._Init(chi2Ptr, self) # This class is passed as parent (avoids dealloc before chi2)
 *         self.chi2 = chi2
*/
  __pyx_t_5 = NULL;
  __pyx_t_6 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_4);
  }
  __pyx_v_chi2 = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":242
 *         cdef Chi2TensorCpp *chi2Ptr = &self._thisptr.chi2
 *         chi2 = _Chi2Tensor()
 *         chi2._Init(chi2Ptr, self) # This class is passed as parent (avoids dealloc before chi2)             # <<<<<<<<<<<<<<
 *         self.chi2 = chi2
 * 
*/
  __pyx_t_4 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *)__pyx_v_chi2->__pyx_vtab)->_Init(__pyx_v_chi2, __pyx_v_chi2Ptr, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":243
 *         chi2 = _Chi2Tensor()
 *         chi2._Init(chi2Ptr, self) # This class is passed as parent (avoids dealloc before chi2)
 *         self.chi2 = chi2             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  __Pyx_INCREF((PyObject *)__pyx_v_chi2);
  __Pyx_GIVEREF((PyObject *)__pyx_v_chi2);
  __Pyx_GOTREF(__pyx_v_self->chi2);
  __Pyx_DECREF(__pyx_v_self->chi2);
  __pyx_v_self->chi2 = ((PyObject *)__pyx_v_chi2);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":236
 *     cdef readonly object chi2
 * 
 *     def __cinit__(self, np.ndarray[double, ndim = 1] wls, np.ndarray[double complex, ndim = 1] ns):             # <<<<<<<<<<<<<<
 *         # Copy is made in c++
 *         self._thisptr = new MaterialCpp(Map[ArrayXd](wls), Map[ArrayXcd](ns))
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_ns.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_wls.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.Material.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_ns.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_wls.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_chi2);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":245
 *         self.chi2 = chi2
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._thisptr
 * 
*/

/* Python wrapper */
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_2__dealloc__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self) {

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":246
 * 
 *     def __dealloc__(self):
 *         del self._thisptr             # <<<<<<<<<<<<<<
 * 
 *     def __call__(self, wl): # For compatibility
*/
  delete __pyx_v_self->_thisptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":245
 *         self.chi2 = chi2
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._thisptr
 * 
*/

  /* function exit code */
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":248
 *         del self._thisptr
 * 
 *     def __call__(self, wl): # For compatibility             # <<<<<<<<<<<<<<
 *         return self.GetN(wl)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_5__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_5__call__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_wl = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__call__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_wl,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 248, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 248, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__call__", 0) < (0)) __PYX_ERR(0, 248, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__call__", 1, 1, 1, i); __PYX_ERR(0, 248, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 248, __pyx_L3_error)
    }
    __pyx_v_wl = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__call__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 248, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.Material.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_4__call__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *)__pyx_v_self), __pyx_v_wl);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_4__call__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self, PyObject *__pyx_v_wl) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__call__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":249
 * 
 *     def __call__(self, wl): # For compatibility
 *         return self.GetN(wl)             # <<<<<<<<<<<<<<
 * 
 *     def GetN(self, wl):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = ((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_wl};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_GetN, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":248
 *         del self._thisptr
 * 
 *     def __call__(self, wl): # For compatibility             # <<<<<<<<<<<<<<
 *         return self.GetN(wl)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.Material.__call__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":251
 *         return self.GetN(wl)
 * 
 *     def GetN(self, wl):             # <<<<<<<<<<<<<<
 *         """GetN(wl)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_7GetN(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_6GetN, "GetN(wl)\n\n        Returns refractive index of material at specified wavelength.\n\n        Parameters\n        ----------\n        wl : float\n            Wavelength in meters.\n\n        Returns\n        -------\n        complex\n            Complex refractive index at wavelength wl.\n\n        Examples\n        --------\n        >>> wls = np.array([400e-9, 600e-9, 800e-9], dtype = float)\n        >>> ns = np.array([1.5 + 0.3j, 1.7 + 0.2j, 1.8 + 0.1j], dtype = complex)\n        >>> mat = Material(wls, ns)\n        >>> mat(600e-9)\n        1.7 + 0.2j\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_7GetN = {"GetN", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_7GetN, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_6GetN};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_7GetN(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_wl = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("GetN (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_wl,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 251, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 251, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "GetN", 0) < (0)) __PYX_ERR(0, 251, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("GetN", 1, 1, 1, i); __PYX_ERR(0, 251, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 251, __pyx_L3_error)
    }
    __pyx_v_wl = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("GetN", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 251, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.Material.GetN", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_6GetN(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *)__pyx_v_self), __pyx_v_wl);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_6GetN(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self, PyObject *__pyx_v_wl) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __pyx_t_double_complex __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("GetN", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":275
 * 
 *         """
 *         return self._thisptr.GetN(wl)             # <<<<<<<<<<<<<<
 * 
 *     def IsNonlinear(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyFloat_AsDouble(__pyx_v_wl); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 275, __pyx_L1_error)
  try {
    __pyx_t_2 = __pyx_v_self->_thisptr->GetN(__pyx_t_1);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 275, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_PyComplex_FromComplex(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":251
 *         return self.GetN(wl)
 * 
 *     def GetN(self, wl):             # <<<<<<<<<<<<<<
 *         """GetN(wl)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.Material.GetN", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":277
 *         return self._thisptr.GetN(wl)
 * 
 *     def IsNonlinear(self):             # <<<<<<<<<<<<<<
 *         """IsNonlinear()
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_9IsNonlinear(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_8IsNonlinear, "IsNonlinear()\n\n        Returns True if the material nonlinearity tensor ``chi2`` is nonzero,\n        False otherwise.\n\n        Returns\n        -------\n        bool\n            True if chi2 tensor is nonzero, False otherwise.\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_9IsNonlinear = {"IsNonlinear", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_9IsNonlinear, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_8IsNonlinear};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_9IsNonlinear(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("IsNonlinear (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("IsNonlinear", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("IsNonlinear", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_8IsNonlinear(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_8IsNonlinear(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("IsNonlinear", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":288
 *             True if chi2 tensor is nonzero, False otherwise.
 *         """
 *         return self._thisptr.IsNonlinear()             # <<<<<<<<<<<<<<
 * 
 * #===============================================================================
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->IsNonlinear();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 288, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":277
 *         return self._thisptr.GetN(wl)
 * 
 *     def IsNonlinear(self):             # <<<<<<<<<<<<<<
 *         """IsNonlinear()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.Material.IsNonlinear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":234
 *     """
 *     cdef MaterialCpp *_thisptr
 *     cdef readonly object chi2             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, np.ndarray[double, ndim = 1] wls, np.ndarray[double complex, ndim = 1] ns):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_4chi2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_4chi2_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_4chi2___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_4chi2___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->chi2);
  __pyx_r = __pyx_v_self->chi2;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_11__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_10__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.Material.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_13__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.Material.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_12__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.Material.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":377
 *     cdef bool _needsDealloc;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._thisptr = NULL
 *         self._needsDealloc = False;
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__cinit__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  int __pyx_r;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":378
 * 
 *     def __cinit__(self):
 *         self._thisptr = NULL             # <<<<<<<<<<<<<<
 *         self._needsDealloc = False;
 * 
*/
  __pyx_v_self->_thisptr = NULL;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":379
 *     def __cinit__(self):
 *         self._thisptr = NULL
 *         self._needsDealloc = False;             # <<<<<<<<<<<<<<
 * 
 *     cdef _Init(self, WaveCpp *ptr):
*/
  __pyx_v_self->_needsDealloc = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":377
 *     cdef bool _needsDealloc;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._thisptr = NULL
 *         self._needsDealloc = False;
*/

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":381
 *         self._needsDealloc = False;
 * 
 *     cdef _Init(self, WaveCpp *ptr):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         self._needsDealloc = False
*/

static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, TMM::Wave *__pyx_v_ptr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_Init", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":382
 * 
 *     cdef _Init(self, WaveCpp *ptr):
 *         self._thisptr = ptr             # <<<<<<<<<<<<<<
 *         self._needsDealloc = False
 * 
*/
  __pyx_v_self->_thisptr = __pyx_v_ptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":383
 *     cdef _Init(self, WaveCpp *ptr):
 *         self._thisptr = ptr
 *         self._needsDealloc = False             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  __pyx_v_self->_needsDealloc = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":381
 *         self._needsDealloc = False;
 * 
 *     cdef _Init(self, WaveCpp *ptr):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         self._needsDealloc = False
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":385
 *         self._needsDealloc = False
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._needsDealloc:
 *             del self._thisptr
*/

/* Python wrapper */
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2__dealloc__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  int __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":386
 * 
 *     def __dealloc__(self):
 *         if self._needsDealloc:             # <<<<<<<<<<<<<<
 *             del self._thisptr
 * 
*/
  __pyx_t_1 = (__pyx_v_self->_needsDealloc != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":387
 *     def __dealloc__(self):
 *         if self._needsDealloc:
 *             del self._thisptr             # <<<<<<<<<<<<<<
 * 
 *     def SetParams(self, **kwargs):
*/
    delete __pyx_v_self->_thisptr;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":386
 * 
 *     def __dealloc__(self):
 *         if self._needsDealloc:             # <<<<<<<<<<<<<<
 *             del self._thisptr
 * 
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":385
 *         self._needsDealloc = False
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._needsDealloc:
 *             del self._thisptr
*/

  /* function exit code */
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":389
 *             del self._thisptr
 * 
 *     def SetParams(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """SetParams(**kwargs)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_5SetParams(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4SetParams, "SetParams(**kwargs)\n\n        Helper method to set all the attributes of the class.\n\n        Returns\n        -------\n        None\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_5SetParams = {"SetParams", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_5SetParams, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4SetParams};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_5SetParams(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("SetParams (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("SetParams", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (__pyx_kwds_len > 0) {
    if (unlikely(__Pyx_CheckKeywordStrings("SetParams", __pyx_kwds) == -1)) return NULL;
    __pyx_v_kwargs = __Pyx_KwargsAsDict_FASTCALL(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4SetParams(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4SetParams(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("SetParams", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":400
 *         """
 * 
 *         for name, value in kwargs.items():             # <<<<<<<<<<<<<<
 *             if name not in waveParamsSet:
 *                 raise ValueError("Unknown kwarg %s" % (name))
*/
  __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_v_kwargs, 1, __pyx_mstate_global->__pyx_n_u_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":401
 * 
 *         for name, value in kwargs.items():
 *             if name not in waveParamsSet:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Unknown kwarg %s" % (name))
 *             setattr(self, name, value)
*/
    __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_v_name, __pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_waveParamsSet, Py_NE)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 401, __pyx_L1_error)
    if (unlikely(__pyx_t_8)) {

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":402
 *         for name, value in kwargs.items():
 *             if name not in waveParamsSet:
 *                 raise ValueError("Unknown kwarg %s" % (name))             # <<<<<<<<<<<<<<
 *             setattr(self, name, value)
 * 
*/
      __pyx_t_5 = NULL;
      __pyx_t_9 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_Unknown_kwarg_s, __pyx_v_name); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 402, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_9};
        __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 402, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 402, __pyx_L1_error)

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":401
 * 
 *         for name, value in kwargs.items():
 *             if name not in waveParamsSet:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Unknown kwarg %s" % (name))
 *             setattr(self, name, value)
*/
    }

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":403
 *             if name not in waveParamsSet:
 *                 raise ValueError("Unknown kwarg %s" % (name))
 *             setattr(self, name, value)             # <<<<<<<<<<<<<<
 * 
 *     # Getters
*/
    __pyx_t_11 = PyObject_SetAttr(((PyObject *)__pyx_v_self), __pyx_v_name, __pyx_v_value); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 403, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":389
 *             del self._thisptr
 * 
 *     def SetParams(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """SetParams(**kwargs)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.SetParams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":408
 *     #---------------------------------------------------------------------------
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def waveType(self):
 *         raise NotImplementedError()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8waveType_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8waveType_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8waveType___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8waveType___get__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":410
 *     @property
 *     def waveType(self):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_NotImplementedError)), __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 410, __pyx_L1_error)

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":408
 *     #---------------------------------------------------------------------------
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def waveType(self):
 *         raise NotImplementedError()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.waveType.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":499
 *     #---------------------------------------------------------------------------
 * 
 *     @waveType.setter             # <<<<<<<<<<<<<<
 *     def waveType(self, str waveTypeStr):  # @DuplicatedSignature
 *         self._thisptr.SetWaveType(WaveTypeFromStr(waveTypeStr))
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8waveType_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_waveTypeStr); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8waveType_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_waveTypeStr) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_waveTypeStr), (&PyUnicode_Type), 1, "waveTypeStr", 1))) __PYX_ERR(0, 500, __pyx_L1_error)
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8waveType_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), ((PyObject*)__pyx_v_waveTypeStr));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  goto __pyx_L5_cleaned_up;
  __pyx_L0:;
  __pyx_L5_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8waveType_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, PyObject *__pyx_v_waveTypeStr) {
  int __pyx_r;
  enum TMM::WaveType __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":501
 *     @waveType.setter
 *     def waveType(self, str waveTypeStr):  # @DuplicatedSignature
 *         self._thisptr.SetWaveType(WaveTypeFromStr(waveTypeStr))             # <<<<<<<<<<<<<<
 * 
 *     @pwr.setter
*/
  __pyx_t_1 = __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_WaveTypeFromStr(__pyx_v_waveTypeStr); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 501, __pyx_L1_error)
  try {
    __pyx_v_self->_thisptr->SetWaveType(__pyx_t_1);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 501, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":499
 *     #---------------------------------------------------------------------------
 * 
 *     @waveType.setter             # <<<<<<<<<<<<<<
 *     def waveType(self, str waveTypeStr):  # @DuplicatedSignature
 *         self._thisptr.SetWaveType(WaveTypeFromStr(waveTypeStr))
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.waveType.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":412
 *         raise NotImplementedError()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def pwr(self):
 *         return self._thisptr.GetPwr()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3pwr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3pwr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3pwr___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3pwr___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":414
 *     @property
 *     def pwr(self):
 *         return self._thisptr.GetPwr()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetPwr();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 414, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":412
 *         raise NotImplementedError()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def pwr(self):
 *         return self._thisptr.GetPwr()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.pwr.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":503
 *         self._thisptr.SetWaveType(WaveTypeFromStr(waveTypeStr))
 * 
 *     @pwr.setter             # <<<<<<<<<<<<<<
 *     def pwr(self, double value):  # @DuplicatedSignature
 *         self._thisptr.SetPwr(value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3pwr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3pwr_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 504, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.pwr.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3pwr_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3pwr_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":505
 *     @pwr.setter
 *     def pwr(self, double value):  # @DuplicatedSignature
 *         self._thisptr.SetPwr(value)             # <<<<<<<<<<<<<<
 * 
 *     @overrideE0.setter
*/
  try {
    __pyx_v_self->_thisptr->SetPwr(__pyx_v_value);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 505, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":503
 *         self._thisptr.SetWaveType(WaveTypeFromStr(waveTypeStr))
 * 
 *     @pwr.setter             # <<<<<<<<<<<<<<
 *     def pwr(self, double value):  # @DuplicatedSignature
 *         self._thisptr.SetPwr(value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.pwr.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":416
 *         return self._thisptr.GetPwr()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def overrideE0(self):
 *         return self._thisptr.GetOverrideE0()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_10overrideE0_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_10overrideE0_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_10overrideE0___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_10overrideE0___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":418
 *     @property
 *     def overrideE0(self):
 *         return self._thisptr.GetOverrideE0()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetOverrideE0();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 418, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":416
 *         return self._thisptr.GetPwr()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def overrideE0(self):
 *         return self._thisptr.GetOverrideE0()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.overrideE0.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":507
 *         self._thisptr.SetPwr(value)
 * 
 *     @overrideE0.setter             # <<<<<<<<<<<<<<
 *     def overrideE0(self, bool value): # @DuplicatedSignature
 *         self._thisptr.SetOverrideE0(value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_10overrideE0_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_10overrideE0_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  bool __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyObject_IsTrue(__pyx_arg_value); if (unlikely((__pyx_v_value == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 508, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.overrideE0.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_10overrideE0_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), ((bool)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_10overrideE0_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, bool __pyx_v_value) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":509
 *     @overrideE0.setter
 *     def overrideE0(self, bool value): # @DuplicatedSignature
 *         self._thisptr.SetOverrideE0(value)             # <<<<<<<<<<<<<<
 * 
 *     @E0.setter
*/
  try {
    __pyx_v_self->_thisptr->SetOverrideE0(__pyx_v_value);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 509, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":507
 *         self._thisptr.SetPwr(value)
 * 
 *     @overrideE0.setter             # <<<<<<<<<<<<<<
 *     def overrideE0(self, bool value): # @DuplicatedSignature
 *         self._thisptr.SetOverrideE0(value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.overrideE0.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":420
 *         return self._thisptr.GetOverrideE0()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def E0(self):
 *         return self._thisptr.GetE0()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2E0_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2E0_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2E0___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2E0___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":422
 *     @property
 *     def E0(self):
 *         return self._thisptr.GetE0()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetE0();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 422, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":420
 *         return self._thisptr.GetOverrideE0()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def E0(self):
 *         return self._thisptr.GetE0()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.E0.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":511
 *         self._thisptr.SetOverrideE0(value)
 * 
 *     @E0.setter             # <<<<<<<<<<<<<<
 *     def E0(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetE0(value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2E0_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2E0_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 512, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.E0.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2E0_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2E0_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":513
 *     @E0.setter
 *     def E0(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetE0(value)             # <<<<<<<<<<<<<<
 * 
 *     @w0.setter
*/
  try {
    __pyx_v_self->_thisptr->SetE0(__pyx_v_value);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 513, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":511
 *         self._thisptr.SetOverrideE0(value)
 * 
 *     @E0.setter             # <<<<<<<<<<<<<<
 *     def E0(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetE0(value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.E0.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":424
 *         return self._thisptr.GetE0()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def w0(self):
 *         return self._thisptr.GetW0()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2w0_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2w0_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2w0___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2w0___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":426
 *     @property
 *     def w0(self):
 *         return self._thisptr.GetW0()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetW0();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 426, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":424
 *         return self._thisptr.GetE0()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def w0(self):
 *         return self._thisptr.GetW0()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.w0.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":515
 *         self._thisptr.SetE0(value)
 * 
 *     @w0.setter             # <<<<<<<<<<<<<<
 *     def w0(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetW0(value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2w0_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2w0_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 516, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.w0.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2w0_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2w0_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":517
 *     @w0.setter
 *     def w0(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetW0(value)             # <<<<<<<<<<<<<<
 * 
 *     @Ly.setter
*/
  try {
    __pyx_v_self->_thisptr->SetW0(__pyx_v_value);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 517, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":515
 *         self._thisptr.SetE0(value)
 * 
 *     @w0.setter             # <<<<<<<<<<<<<<
 *     def w0(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetW0(value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.w0.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":428
 *         return self._thisptr.GetW0()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def Ly(self):
 *         return self._thisptr.GetLy()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2Ly_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2Ly_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2Ly___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2Ly___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":430
 *     @property
 *     def Ly(self):
 *         return self._thisptr.GetLy()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetLy();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 430, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":428
 *         return self._thisptr.GetW0()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def Ly(self):
 *         return self._thisptr.GetLy()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.Ly.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":519
 *         self._thisptr.SetW0(value)
 * 
 *     @Ly.setter             # <<<<<<<<<<<<<<
 *     def Ly(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetLy(value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2Ly_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2Ly_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 520, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.Ly.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2Ly_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2Ly_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":521
 *     @Ly.setter
 *     def Ly(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetLy(value)             # <<<<<<<<<<<<<<
 * 
 *     @a.setter
*/
  try {
    __pyx_v_self->_thisptr->SetLy(__pyx_v_value);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 521, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":519
 *         self._thisptr.SetW0(value)
 * 
 *     @Ly.setter             # <<<<<<<<<<<<<<
 *     def Ly(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetLy(value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.Ly.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":432
 *         return self._thisptr.GetLy()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def a(self):
 *         return self._thisptr.GetA()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1a_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1a_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1a___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1a___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":434
 *     @property
 *     def a(self):
 *         return self._thisptr.GetA()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetA();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 434, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":432
 *         return self._thisptr.GetLy()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def a(self):
 *         return self._thisptr.GetA()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.a.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":523
 *         self._thisptr.SetLy(value)
 * 
 *     @a.setter             # <<<<<<<<<<<<<<
 *     def a(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetA(value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1a_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1a_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 524, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.a.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1a_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1a_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":525
 *     @a.setter
 *     def a(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetA(value)             # <<<<<<<<<<<<<<
 * 
 *     @nPointsInteg.setter
*/
  try {
    __pyx_v_self->_thisptr->SetA(__pyx_v_value);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 525, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":523
 *         self._thisptr.SetLy(value)
 * 
 *     @a.setter             # <<<<<<<<<<<<<<
 *     def a(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetA(value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.a.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":436
 *         return self._thisptr.GetA()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def nPointsInteg(self):
 *         return self._thisptr.GetNPointsInteg()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12nPointsInteg_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12nPointsInteg_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12nPointsInteg___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12nPointsInteg___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":438
 *     @property
 *     def nPointsInteg(self):
 *         return self._thisptr.GetNPointsInteg()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetNPointsInteg();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 438, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":436
 *         return self._thisptr.GetA()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def nPointsInteg(self):
 *         return self._thisptr.GetNPointsInteg()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.nPointsInteg.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":527
 *         self._thisptr.SetA(value)
 * 
 *     @nPointsInteg.setter             # <<<<<<<<<<<<<<
 *     def nPointsInteg(self, int value): # @DuplicatedSignature
 *         self._thisptr.SetNPointsInteg(value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12nPointsInteg_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12nPointsInteg_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  int __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyLong_As_int(__pyx_arg_value); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 528, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.nPointsInteg.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12nPointsInteg_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), ((int)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12nPointsInteg_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, int __pyx_v_value) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":529
 *     @nPointsInteg.setter
 *     def nPointsInteg(self, int value): # @DuplicatedSignature
 *         self._thisptr.SetNPointsInteg(value)             # <<<<<<<<<<<<<<
 * 
 *     @maxX.setter
*/
  try {
    __pyx_v_self->_thisptr->SetNPointsInteg(__pyx_v_value);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 529, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":527
 *         self._thisptr.SetA(value)
 * 
 *     @nPointsInteg.setter             # <<<<<<<<<<<<<<
 *     def nPointsInteg(self, int value): # @DuplicatedSignature
 *         self._thisptr.SetNPointsInteg(value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.nPointsInteg.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":440
 *         return self._thisptr.GetNPointsInteg()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def maxX(self):
 *         return self._thisptr.GetMaxX()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4maxX_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4maxX_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4maxX___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4maxX___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":442
 *     @property
 *     def maxX(self):
 *         return self._thisptr.GetMaxX()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetMaxX();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 442, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":440
 *         return self._thisptr.GetNPointsInteg()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def maxX(self):
 *         return self._thisptr.GetMaxX()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.maxX.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":531
 *         self._thisptr.SetNPointsInteg(value)
 * 
 *     @maxX.setter             # <<<<<<<<<<<<<<
 *     def maxX(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetMaxX(value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4maxX_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4maxX_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 532, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.maxX.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4maxX_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4maxX_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":533
 *     @maxX.setter
 *     def maxX(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetMaxX(value)             # <<<<<<<<<<<<<<
 * 
 *     @dynamicMaxX.setter
*/
  try {
    __pyx_v_self->_thisptr->SetMaxX(__pyx_v_value);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 533, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":531
 *         self._thisptr.SetNPointsInteg(value)
 * 
 *     @maxX.setter             # <<<<<<<<<<<<<<
 *     def maxX(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetMaxX(value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.maxX.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":444
 *         return self._thisptr.GetMaxX()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamicMaxX(self):
 *         return self._thisptr.IsDynamicMaxXEnabled()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_11dynamicMaxX_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_11dynamicMaxX_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_11dynamicMaxX___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_11dynamicMaxX___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":446
 *     @property
 *     def dynamicMaxX(self):
 *         return self._thisptr.IsDynamicMaxXEnabled()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->IsDynamicMaxXEnabled();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 446, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":444
 *         return self._thisptr.GetMaxX()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamicMaxX(self):
 *         return self._thisptr.IsDynamicMaxXEnabled()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.dynamicMaxX.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":535
 *         self._thisptr.SetMaxX(value)
 * 
 *     @dynamicMaxX.setter             # <<<<<<<<<<<<<<
 *     def dynamicMaxX(self, bool value): # @DuplicatedSignature
 *         self._thisptr.EnableDynamicMaxX(value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_11dynamicMaxX_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_11dynamicMaxX_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  bool __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyObject_IsTrue(__pyx_arg_value); if (unlikely((__pyx_v_value == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 536, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.dynamicMaxX.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_11dynamicMaxX_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), ((bool)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_11dynamicMaxX_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, bool __pyx_v_value) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":537
 *     @dynamicMaxX.setter
 *     def dynamicMaxX(self, bool value): # @DuplicatedSignature
 *         self._thisptr.EnableDynamicMaxX(value)             # <<<<<<<<<<<<<<
 * 
 *     @dynamicMaxXCoef.setter
*/
  try {
    __pyx_v_self->_thisptr->EnableDynamicMaxX(__pyx_v_value);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 537, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":535
 *         self._thisptr.SetMaxX(value)
 * 
 *     @dynamicMaxX.setter             # <<<<<<<<<<<<<<
 *     def dynamicMaxX(self, bool value): # @DuplicatedSignature
 *         self._thisptr.EnableDynamicMaxX(value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.dynamicMaxX.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":448
 *         return self._thisptr.IsDynamicMaxXEnabled()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamicMaxXCoef(self):
 *         return self._thisptr.GetDynamicMaxXCoef()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_15dynamicMaxXCoef_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_15dynamicMaxXCoef_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_15dynamicMaxXCoef___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_15dynamicMaxXCoef___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":450
 *     @property
 *     def dynamicMaxXCoef(self):
 *         return self._thisptr.GetDynamicMaxXCoef()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetDynamicMaxXCoef();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 450, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":448
 *         return self._thisptr.IsDynamicMaxXEnabled()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamicMaxXCoef(self):
 *         return self._thisptr.GetDynamicMaxXCoef()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.dynamicMaxXCoef.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":539
 *         self._thisptr.EnableDynamicMaxX(value)
 * 
 *     @dynamicMaxXCoef.setter             # <<<<<<<<<<<<<<
 *     def dynamicMaxXCoef(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetDynamicMaxXCoef(value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_15dynamicMaxXCoef_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_15dynamicMaxXCoef_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 540, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.dynamicMaxXCoef.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_15dynamicMaxXCoef_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_15dynamicMaxXCoef_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":541
 *     @dynamicMaxXCoef.setter
 *     def dynamicMaxXCoef(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetDynamicMaxXCoef(value)             # <<<<<<<<<<<<<<
 * 
 *     @dynamicMaxXAddition.setter
*/
  try {
    __pyx_v_self->_thisptr->SetDynamicMaxXCoef(__pyx_v_value);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 541, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":539
 *         self._thisptr.EnableDynamicMaxX(value)
 * 
 *     @dynamicMaxXCoef.setter             # <<<<<<<<<<<<<<
 *     def dynamicMaxXCoef(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetDynamicMaxXCoef(value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.dynamicMaxXCoef.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":452
 *         return self._thisptr.GetDynamicMaxXCoef()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamicMaxXAddition(self):
 *         return self._thisptr.GetDynamicMaxXAddition()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_19dynamicMaxXAddition_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_19dynamicMaxXAddition_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_19dynamicMaxXAddition___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_19dynamicMaxXAddition___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":454
 *     @property
 *     def dynamicMaxXAddition(self):
 *         return self._thisptr.GetDynamicMaxXAddition()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetDynamicMaxXAddition();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 454, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":452
 *         return self._thisptr.GetDynamicMaxXCoef()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def dynamicMaxXAddition(self):
 *         return self._thisptr.GetDynamicMaxXAddition()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.dynamicMaxXAddition.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":543
 *         self._thisptr.SetDynamicMaxXCoef(value)
 * 
 *     @dynamicMaxXAddition.setter             # <<<<<<<<<<<<<<
 *     def dynamicMaxXAddition(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetDynamicMaxXAddition(value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_19dynamicMaxXAddition_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_19dynamicMaxXAddition_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 544, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.dynamicMaxXAddition.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_19dynamicMaxXAddition_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_19dynamicMaxXAddition_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":545
 *     @dynamicMaxXAddition.setter
 *     def dynamicMaxXAddition(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetDynamicMaxXAddition(value)             # <<<<<<<<<<<<<<
 * 
 *     @maxPhi.setter
*/
  try {
    __pyx_v_self->_thisptr->SetDynamicMaxXAddition(__pyx_v_value);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 545, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":543
 *         self._thisptr.SetDynamicMaxXCoef(value)
 * 
 *     @dynamicMaxXAddition.setter             # <<<<<<<<<<<<<<
 *     def dynamicMaxXAddition(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetDynamicMaxXAddition(value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.dynamicMaxXAddition.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":456
 *         return self._thisptr.GetDynamicMaxXAddition()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def maxPhi(self):
 *         return self._thisptr.GetMaxPhi()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6maxPhi_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6maxPhi_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6maxPhi___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6maxPhi___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":458
 *     @property
 *     def maxPhi(self):
 *         return self._thisptr.GetMaxPhi()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetMaxPhi();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 458, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":456
 *         return self._thisptr.GetDynamicMaxXAddition()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def maxPhi(self):
 *         return self._thisptr.GetMaxPhi()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.maxPhi.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":547
 *         self._thisptr.SetDynamicMaxXAddition(value)
 * 
 *     @maxPhi.setter             # <<<<<<<<<<<<<<
 *     def maxPhi(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetMaxPhi(value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6maxPhi_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6maxPhi_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 548, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.maxPhi.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6maxPhi_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6maxPhi_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":549
 *     @maxPhi.setter
 *     def maxPhi(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetMaxPhi(value)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  try {
    __pyx_v_self->_thisptr->SetMaxPhi(__pyx_v_value);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 549, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":547
 *         self._thisptr.SetDynamicMaxXAddition(value)
 * 
 *     @maxPhi.setter             # <<<<<<<<<<<<<<
 *     def maxPhi(self, double value): # @DuplicatedSignature
 *         self._thisptr.SetMaxPhi(value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.maxPhi.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":460
 *         return self._thisptr.GetMaxPhi()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def xRange(self):
 *         cdef pair[double, double] r = self._thisptr.GetXRange()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6xRange_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6xRange_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6xRange___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6xRange___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  std::pair<double,double>  __pyx_v_r;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  std::pair<double,double>  __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":462
 *     @property
 *     def xRange(self):
 *         cdef pair[double, double] r = self._thisptr.GetXRange()             # <<<<<<<<<<<<<<
 *         return (r.first, r.second)
 * 
*/
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetXRange();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 462, __pyx_L1_error)
  }
  __pyx_v_r = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":463
 *     def xRange(self):
 *         cdef pair[double, double] r = self._thisptr.GetXRange()
 *         return (r.first, r.second)             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_r.first); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_r.second); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2) != (0)) __PYX_ERR(0, 463, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3) != (0)) __PYX_ERR(0, 463, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":460
 *         return self._thisptr.GetMaxPhi()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def xRange(self):
 *         cdef pair[double, double] r = self._thisptr.GetXRange()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.xRange.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":465
 *         return (r.first, r.second)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def betas(self):
 *         return ndarray_copy(self._thisptr.GetBetas()).squeeze()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_5betas_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_5betas_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_5betas___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_5betas___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Eigen::ArrayXd __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":467
 *     @property
 *     def betas(self):
 *         return ndarray_copy(self._thisptr.GetBetas()).squeeze()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_3 = __pyx_v_self->_thisptr->GetBetas();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 467, __pyx_L1_error)
  }
  __pyx_t_4 = ((PyObject *)eigency::ndarray_copy(__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __pyx_t_4;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":465
 *         return (r.first, r.second)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def betas(self):
 *         return ndarray_copy(self._thisptr.GetBetas()).squeeze()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.betas.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":469
 *         return ndarray_copy(self._thisptr.GetBetas()).squeeze()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def phis(self):
 *         return ndarray_copy(self._thisptr.GetPhis()).squeeze()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4phis_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4phis_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4phis___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4phis___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Eigen::ArrayXd __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":471
 *     @property
 *     def phis(self):
 *         return ndarray_copy(self._thisptr.GetPhis()).squeeze()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_3 = __pyx_v_self->_thisptr->GetPhis();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 471, __pyx_L1_error)
  }
  __pyx_t_4 = ((PyObject *)eigency::ndarray_copy(__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __pyx_t_4;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":469
 *         return ndarray_copy(self._thisptr.GetBetas()).squeeze()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def phis(self):
 *         return ndarray_copy(self._thisptr.GetPhis()).squeeze()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.phis.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":473
 *         return ndarray_copy(self._thisptr.GetPhis()).squeeze()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def kxs(self):
 *         return ndarray_copy(self._thisptr.GetKxs()).squeeze()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3kxs_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3kxs_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3kxs___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3kxs___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Eigen::ArrayXd __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":475
 *     @property
 *     def kxs(self):
 *         return ndarray_copy(self._thisptr.GetKxs()).squeeze()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_3 = __pyx_v_self->_thisptr->GetKxs();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 475, __pyx_L1_error)
  }
  __pyx_t_4 = ((PyObject *)eigency::ndarray_copy(__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __pyx_t_4;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":473
 *         return ndarray_copy(self._thisptr.GetPhis()).squeeze()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def kxs(self):
 *         return ndarray_copy(self._thisptr.GetKxs()).squeeze()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.kxs.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":477
 *         return ndarray_copy(self._thisptr.GetKxs()).squeeze()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def kzs(self):
 *         return ndarray_copy(self._thisptr.GetKzs()).squeeze()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3kzs_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3kzs_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3kzs___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3kzs___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Eigen::ArrayXd __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":479
 *     @property
 *     def kzs(self):
 *         return ndarray_copy(self._thisptr.GetKzs()).squeeze()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_3 = __pyx_v_self->_thisptr->GetKzs();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 479, __pyx_L1_error)
  }
  __pyx_t_4 = ((PyObject *)eigency::ndarray_copy(__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __pyx_t_4;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":477
 *         return ndarray_copy(self._thisptr.GetKxs()).squeeze()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def kzs(self):
 *         return ndarray_copy(self._thisptr.GetKzs()).squeeze()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.kzs.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":481
 *         return ndarray_copy(self._thisptr.GetKzs()).squeeze()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def fieldProfile(self):
 *         xs = ndarray_copy(self._thisptr.GetFieldProfileXs()).squeeze()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12fieldProfile_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12fieldProfile_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12fieldProfile___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12fieldProfile___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_v_xs = NULL;
  PyObject *__pyx_v_fieldProfile = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Eigen::ArrayXd __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":483
 *     @property
 *     def fieldProfile(self):
 *         xs = ndarray_copy(self._thisptr.GetFieldProfileXs()).squeeze()             # <<<<<<<<<<<<<<
 *         fieldProfile = ndarray_copy(self._thisptr.GetFieldProfile()).squeeze()
 *         return xs, fieldProfile
*/
  try {
    __pyx_t_3 = __pyx_v_self->_thisptr->GetFieldProfileXs();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 483, __pyx_L1_error)
  }
  __pyx_t_4 = ((PyObject *)eigency::ndarray_copy(__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __pyx_t_4;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_xs = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":484
 *     def fieldProfile(self):
 *         xs = ndarray_copy(self._thisptr.GetFieldProfileXs()).squeeze()
 *         fieldProfile = ndarray_copy(self._thisptr.GetFieldProfile()).squeeze()             # <<<<<<<<<<<<<<
 *         return xs, fieldProfile
 * 
*/
  try {
    __pyx_t_3 = __pyx_v_self->_thisptr->GetFieldProfile();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 484, __pyx_L1_error)
  }
  __pyx_t_2 = ((PyObject *)eigency::ndarray_copy(__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_fieldProfile = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":485
 *         xs = ndarray_copy(self._thisptr.GetFieldProfileXs()).squeeze()
 *         fieldProfile = ndarray_copy(self._thisptr.GetFieldProfile()).squeeze()
 *         return xs, fieldProfile             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_xs);
  __Pyx_GIVEREF(__pyx_v_xs);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_xs) != (0)) __PYX_ERR(0, 485, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_fieldProfile);
  __Pyx_GIVEREF(__pyx_v_fieldProfile);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_fieldProfile) != (0)) __PYX_ERR(0, 485, __pyx_L1_error);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":481
 *         return ndarray_copy(self._thisptr.GetKzs()).squeeze()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def fieldProfile(self):
 *         xs = ndarray_copy(self._thisptr.GetFieldProfileXs()).squeeze()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.fieldProfile.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_xs);
  __Pyx_XDECREF(__pyx_v_fieldProfile);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":487
 *         return xs, fieldProfile
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def expansionCoefsKx(self):
 *         return ndarray_copy(self._thisptr.GetExpansionCoefsKx()).squeeze()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_16expansionCoefsKx_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_16expansionCoefsKx_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_16expansionCoefsKx___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_16expansionCoefsKx___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Eigen::ArrayXcd __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":489
 *     @property
 *     def expansionCoefsKx(self):
 *         return ndarray_copy(self._thisptr.GetExpansionCoefsKx()).squeeze()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_3 = __pyx_v_self->_thisptr->GetExpansionCoefsKx();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 489, __pyx_L1_error)
  }
  __pyx_t_4 = ((PyObject *)eigency::ndarray_copy(__pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __pyx_t_4;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":487
 *         return xs, fieldProfile
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def expansionCoefsKx(self):
 *         return ndarray_copy(self._thisptr.GetExpansionCoefsKx()).squeeze()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.expansionCoefsKx.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":491
 *         return ndarray_copy(self._thisptr.GetExpansionCoefsKx()).squeeze()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def beamArea(self):
 *         return self._thisptr.GetBeamArea()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8beamArea_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8beamArea_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8beamArea___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8beamArea___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":493
 *     @property
 *     def beamArea(self):
 *         return self._thisptr.GetBeamArea()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetBeamArea();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 493, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":491
 *         return ndarray_copy(self._thisptr.GetExpansionCoefsKx()).squeeze()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def beamArea(self):
 *         return self._thisptr.GetBeamArea()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.beamArea.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Wave.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":579
 *     cdef readonly double I, R, T
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__cinit__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities___cinit__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self) {
  int __pyx_r;

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":582
 *         pass
 * 
 *     cdef _Init(self, IntensitiesCpp *ptr):             # <<<<<<<<<<<<<<
 *         self.inc = ptr.inc
 *         self.r = ptr.r
*/

static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self, TMM::Intensities *__pyx_v_ptr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __pyx_t_double_complex __pyx_t_1;
  double __pyx_t_2;
  __Pyx_RefNannySetupContext("_Init", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":583
 * 
 *     cdef _Init(self, IntensitiesCpp *ptr):
 *         self.inc = ptr.inc             # <<<<<<<<<<<<<<
 *         self.r = ptr.r
 *         self.t = ptr.t
*/
  __pyx_t_1 = __pyx_v_ptr->inc;
  __pyx_v_self->inc = __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":584
 *     cdef _Init(self, IntensitiesCpp *ptr):
 *         self.inc = ptr.inc
 *         self.r = ptr.r             # <<<<<<<<<<<<<<
 *         self.t = ptr.t
 *         self.I = ptr.I
*/
  __pyx_t_1 = __pyx_v_ptr->r;
  __pyx_v_self->r = __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":585
 *         self.inc = ptr.inc
 *         self.r = ptr.r
 *         self.t = ptr.t             # <<<<<<<<<<<<<<
 *         self.I = ptr.I
 *         self.R = ptr.R
*/
  __pyx_t_1 = __pyx_v_ptr->t;
  __pyx_v_self->t = __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":586
 *         self.r = ptr.r
 *         self.t = ptr.t
 *         self.I = ptr.I             # <<<<<<<<<<<<<<
 *         self.R = ptr.R
 *         self.T = ptr.T
*/
  __pyx_t_2 = __pyx_v_ptr->I;
  __pyx_v_self->I = __pyx_t_2;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":587
 *         self.t = ptr.t
 *         self.I = ptr.I
 *         self.R = ptr.R             # <<<<<<<<<<<<<<
 *         self.T = ptr.T
 * 
*/
  __pyx_t_2 = __pyx_v_ptr->R;
  __pyx_v_self->R = __pyx_t_2;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":588
 *         self.I = ptr.I
 *         self.R = ptr.R
 *         self.T = ptr.T             # <<<<<<<<<<<<<<
 * 
 * #===============================================================================
*/
  __pyx_t_2 = __pyx_v_ptr->T;
  __pyx_v_self->T = __pyx_t_2;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":582
 *         pass
 * 
 *     cdef _Init(self, IntensitiesCpp *ptr):             # <<<<<<<<<<<<<<
 *         self.inc = ptr.inc
 *         self.r = ptr.r
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":576
 * 
 *     """
 *     cdef readonly double complex inc, r, t             # <<<<<<<<<<<<<<
 *     cdef readonly double I, R, T
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_3inc_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_3inc_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_3inc___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_3inc___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_PyComplex_FromComplex(__pyx_v_self->inc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Intensities.inc.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1r_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1r_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1r___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1r___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_PyComplex_FromComplex(__pyx_v_self->r); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Intensities.r.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1t_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1t_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1t___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1t___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_PyComplex_FromComplex(__pyx_v_self->t); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Intensities.t.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":577
 *     """
 *     cdef readonly double complex inc, r, t
 *     cdef readonly double I, R, T             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1I_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1I_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1I___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1I___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->I); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Intensities.I.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1R_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1R_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1R___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1R___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->R); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Intensities.R.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1T_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1T_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1T___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1T___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->T); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Intensities.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_2__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Intensities.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Intensities.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_4__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._Intensities.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":624
 *     cdef object _parent;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._needDealloc = False
 *         pass
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__cinit__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self) {
  int __pyx_r;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":625
 * 
 *     def __cinit__(self):
 *         self._needDealloc = False             # <<<<<<<<<<<<<<
 *         pass
 * 
*/
  __pyx_v_self->_needDealloc = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":624
 *     cdef object _parent;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._needDealloc = False
 *         pass
*/

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":628
 *         pass
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._needDealloc:
 *             del self._thisptr
*/

/* Python wrapper */
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2__dealloc__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self) {
  int __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":629
 * 
 *     def __dealloc__(self):
 *         if self._needDealloc:             # <<<<<<<<<<<<<<
 *             del self._thisptr
 * 
*/
  __pyx_t_1 = (__pyx_v_self->_needDealloc != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":630
 *     def __dealloc__(self):
 *         if self._needDealloc:
 *             del self._thisptr             # <<<<<<<<<<<<<<
 * 
 *     cdef _Init(self, SweepResultNonlinearTMMCpp *ptr, bool needDealloc = True, object parent = None):
*/
    delete __pyx_v_self->_thisptr;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":629
 * 
 *     def __dealloc__(self):
 *         if self._needDealloc:             # <<<<<<<<<<<<<<
 *             del self._thisptr
 * 
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":628
 *         pass
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._needDealloc:
 *             del self._thisptr
*/

  /* function exit code */
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":632
 *             del self._thisptr
 * 
 *     cdef _Init(self, SweepResultNonlinearTMMCpp *ptr, bool needDealloc = True, object parent = None):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         self._needDealloc = needDealloc
*/

static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self, TMM::SweepResultNonlinearTMM *__pyx_v_ptr, struct __pyx_opt_args_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM__Init *__pyx_optional_args) {
  bool __pyx_v_needDealloc = ((bool)1);
  PyObject *__pyx_v_parent = ((PyObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_Init", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_needDealloc = __pyx_optional_args->needDealloc;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_parent = __pyx_optional_args->parent;
      }
    }
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":633
 * 
 *     cdef _Init(self, SweepResultNonlinearTMMCpp *ptr, bool needDealloc = True, object parent = None):
 *         self._thisptr = ptr             # <<<<<<<<<<<<<<
 *         self._needDealloc = needDealloc
 *         self._parent = parent # To avoid dealloc of parent before this
*/
  __pyx_v_self->_thisptr = __pyx_v_ptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":634
 *     cdef _Init(self, SweepResultNonlinearTMMCpp *ptr, bool needDealloc = True, object parent = None):
 *         self._thisptr = ptr
 *         self._needDealloc = needDealloc             # <<<<<<<<<<<<<<
 *         self._parent = parent # To avoid dealloc of parent before this
 * 
*/
  __pyx_v_self->_needDealloc = __pyx_v_needDealloc;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":635
 *         self._thisptr = ptr
 *         self._needDealloc = needDealloc
 *         self._parent = parent # To avoid dealloc of parent before this             # <<<<<<<<<<<<<<
 * 
 *         self.inc = ndarray_view(self._thisptr.inc).squeeze()
*/
  __Pyx_INCREF(__pyx_v_parent);
  __Pyx_GIVEREF(__pyx_v_parent);
  __Pyx_GOTREF(__pyx_v_self->_parent);
  __Pyx_DECREF(__pyx_v_self->_parent);
  __pyx_v_self->_parent = __pyx_v_parent;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":637
 *         self._parent = parent # To avoid dealloc of parent before this
 * 
 *         self.inc = ndarray_view(self._thisptr.inc).squeeze()             # <<<<<<<<<<<<<<
 *         self.r = ndarray_view(self._thisptr.r).squeeze()
 *         self.t = ndarray_view(self._thisptr.t).squeeze()
*/
  __pyx_t_3 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->inc)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->inc);
  __Pyx_DECREF((PyObject *)__pyx_v_self->inc);
  __pyx_v_self->inc = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":638
 * 
 *         self.inc = ndarray_view(self._thisptr.inc).squeeze()
 *         self.r = ndarray_view(self._thisptr.r).squeeze()             # <<<<<<<<<<<<<<
 *         self.t = ndarray_view(self._thisptr.t).squeeze()
 *         self.Ii = ndarray_view(self._thisptr.Ii).squeeze()
*/
  __pyx_t_2 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->r)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->r);
  __Pyx_DECREF((PyObject *)__pyx_v_self->r);
  __pyx_v_self->r = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":639
 *         self.inc = ndarray_view(self._thisptr.inc).squeeze()
 *         self.r = ndarray_view(self._thisptr.r).squeeze()
 *         self.t = ndarray_view(self._thisptr.t).squeeze()             # <<<<<<<<<<<<<<
 *         self.Ii = ndarray_view(self._thisptr.Ii).squeeze()
 *         self.Ir = ndarray_view(self._thisptr.Ir).squeeze()
*/
  __pyx_t_3 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->t)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 639, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->t);
  __Pyx_DECREF((PyObject *)__pyx_v_self->t);
  __pyx_v_self->t = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":640
 *         self.r = ndarray_view(self._thisptr.r).squeeze()
 *         self.t = ndarray_view(self._thisptr.t).squeeze()
 *         self.Ii = ndarray_view(self._thisptr.Ii).squeeze()             # <<<<<<<<<<<<<<
 *         self.Ir = ndarray_view(self._thisptr.Ir).squeeze()
 *         self.It = ndarray_view(self._thisptr.It).squeeze()
*/
  __pyx_t_2 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->Ii)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->Ii);
  __Pyx_DECREF((PyObject *)__pyx_v_self->Ii);
  __pyx_v_self->Ii = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":641
 *         self.t = ndarray_view(self._thisptr.t).squeeze()
 *         self.Ii = ndarray_view(self._thisptr.Ii).squeeze()
 *         self.Ir = ndarray_view(self._thisptr.Ir).squeeze()             # <<<<<<<<<<<<<<
 *         self.It = ndarray_view(self._thisptr.It).squeeze()
 *         self.Ia = ndarray_view(self._thisptr.Ia).squeeze()
*/
  __pyx_t_3 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->Ir)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->Ir);
  __Pyx_DECREF((PyObject *)__pyx_v_self->Ir);
  __pyx_v_self->Ir = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":642
 *         self.Ii = ndarray_view(self._thisptr.Ii).squeeze()
 *         self.Ir = ndarray_view(self._thisptr.Ir).squeeze()
 *         self.It = ndarray_view(self._thisptr.It).squeeze()             # <<<<<<<<<<<<<<
 *         self.Ia = ndarray_view(self._thisptr.Ia).squeeze()
 *         self.enh = ndarray_view(self._thisptr.enh).squeeze()
*/
  __pyx_t_2 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->It)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->It);
  __Pyx_DECREF((PyObject *)__pyx_v_self->It);
  __pyx_v_self->It = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":643
 *         self.Ir = ndarray_view(self._thisptr.Ir).squeeze()
 *         self.It = ndarray_view(self._thisptr.It).squeeze()
 *         self.Ia = ndarray_view(self._thisptr.Ia).squeeze()             # <<<<<<<<<<<<<<
 *         self.enh = ndarray_view(self._thisptr.enh).squeeze()
 * 
*/
  __pyx_t_3 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->Ia)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->Ia);
  __Pyx_DECREF((PyObject *)__pyx_v_self->Ia);
  __pyx_v_self->Ia = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":644
 *         self.It = ndarray_view(self._thisptr.It).squeeze()
 *         self.Ia = ndarray_view(self._thisptr.Ia).squeeze()
 *         self.enh = ndarray_view(self._thisptr.enh).squeeze()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->enh)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->enh);
  __Pyx_DECREF((PyObject *)__pyx_v_self->enh);
  __pyx_v_self->enh = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":632
 *             del self._thisptr
 * 
 *     cdef _Init(self, SweepResultNonlinearTMMCpp *ptr, bool needDealloc = True, object parent = None):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         self._needDealloc = needDealloc
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._SweepResultNonlinearTMM._Init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":619
 *     """
 *     cdef SweepResultNonlinearTMMCpp *_thisptr
 *     cdef readonly np.ndarray inc, r, t             # <<<<<<<<<<<<<<
 *     cdef readonly np.ndarray Ii, Ir, It, Ia, enh
 *     cdef bool _needDealloc;
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3inc_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3inc_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3inc___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3inc___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->inc);
  __pyx_r = ((PyObject *)__pyx_v_self->inc);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1r_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1r_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1r___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1r___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->r);
  __pyx_r = ((PyObject *)__pyx_v_self->r);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1t_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1t_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1t___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1t___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->t);
  __pyx_r = ((PyObject *)__pyx_v_self->t);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":620
 *     cdef SweepResultNonlinearTMMCpp *_thisptr
 *     cdef readonly np.ndarray inc, r, t
 *     cdef readonly np.ndarray Ii, Ir, It, Ia, enh             # <<<<<<<<<<<<<<
 *     cdef bool _needDealloc;
 *     cdef object _parent;
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ii_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ii_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ii___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ii___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->Ii);
  __pyx_r = ((PyObject *)__pyx_v_self->Ii);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ir_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ir_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ir___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ir___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->Ir);
  __pyx_r = ((PyObject *)__pyx_v_self->Ir);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2It_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2It_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2It___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2It___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->It);
  __pyx_r = ((PyObject *)__pyx_v_self->It);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ia_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ia_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ia___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ia___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->Ia);
  __pyx_r = ((PyObject *)__pyx_v_self->Ia);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3enh_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3enh_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3enh___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3enh___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->enh);
  __pyx_r = ((PyObject *)__pyx_v_self->enh);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_4__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._SweepResultNonlinearTMM.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._SweepResultNonlinearTMM.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_6__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._SweepResultNonlinearTMM.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":674
 *     cdef object _parent;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._needDealloc = False
 *         pass
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__cinit__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self) {
  int __pyx_r;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":675
 * 
 *     def __cinit__(self):
 *         self._needDealloc = False             # <<<<<<<<<<<<<<
 *         pass
 * 
*/
  __pyx_v_self->_needDealloc = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":674
 *     cdef object _parent;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._needDealloc = False
 *         pass
*/

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":678
 *         pass
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._needDealloc:
 *             del self._thisptr
*/

/* Python wrapper */
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2__dealloc__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self) {
  int __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":679
 * 
 *     def __dealloc__(self):
 *         if self._needDealloc:             # <<<<<<<<<<<<<<
 *             del self._thisptr
 * 
*/
  __pyx_t_1 = (__pyx_v_self->_needDealloc != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":680
 *     def __dealloc__(self):
 *         if self._needDealloc:
 *             del self._thisptr             # <<<<<<<<<<<<<<
 * 
 *     cdef _Init(self, WaveSweepResultNonlinearTMMCpp *ptr, bool needDealloc = True, object parent = None):
*/
    delete __pyx_v_self->_thisptr;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":679
 * 
 *     def __dealloc__(self):
 *         if self._needDealloc:             # <<<<<<<<<<<<<<
 *             del self._thisptr
 * 
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":678
 *         pass
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._needDealloc:
 *             del self._thisptr
*/

  /* function exit code */
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":682
 *             del self._thisptr
 * 
 *     cdef _Init(self, WaveSweepResultNonlinearTMMCpp *ptr, bool needDealloc = True, object parent = None):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         self._needDealloc = needDealloc
*/

static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self, TMM::WaveSweepResultNonlinearTMM *__pyx_v_ptr, struct __pyx_opt_args_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM__Init *__pyx_optional_args) {
  bool __pyx_v_needDealloc = ((bool)1);
  PyObject *__pyx_v_parent = ((PyObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_Init", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_needDealloc = __pyx_optional_args->needDealloc;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_parent = __pyx_optional_args->parent;
      }
    }
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":683
 * 
 *     cdef _Init(self, WaveSweepResultNonlinearTMMCpp *ptr, bool needDealloc = True, object parent = None):
 *         self._thisptr = ptr             # <<<<<<<<<<<<<<
 *         self._needDealloc = needDealloc
 *         self._parent = parent # To avoid dealloc of parent before this
*/
  __pyx_v_self->_thisptr = __pyx_v_ptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":684
 *     cdef _Init(self, WaveSweepResultNonlinearTMMCpp *ptr, bool needDealloc = True, object parent = None):
 *         self._thisptr = ptr
 *         self._needDealloc = needDealloc             # <<<<<<<<<<<<<<
 *         self._parent = parent # To avoid dealloc of parent before this
 * 
*/
  __pyx_v_self->_needDealloc = __pyx_v_needDealloc;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":685
 *         self._thisptr = ptr
 *         self._needDealloc = needDealloc
 *         self._parent = parent # To avoid dealloc of parent before this             # <<<<<<<<<<<<<<
 * 
 *         self.Pi = ndarray_view(self._thisptr.Pi).squeeze()
*/
  __Pyx_INCREF(__pyx_v_parent);
  __Pyx_GIVEREF(__pyx_v_parent);
  __Pyx_GOTREF(__pyx_v_self->_parent);
  __Pyx_DECREF(__pyx_v_self->_parent);
  __pyx_v_self->_parent = __pyx_v_parent;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":687
 *         self._parent = parent # To avoid dealloc of parent before this
 * 
 *         self.Pi = ndarray_view(self._thisptr.Pi).squeeze()             # <<<<<<<<<<<<<<
 *         self.Pr = ndarray_view(self._thisptr.Pr).squeeze()
 *         self.Pt = ndarray_view(self._thisptr.Pt).squeeze()
*/
  __pyx_t_3 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->Pi)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 687, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->Pi);
  __Pyx_DECREF((PyObject *)__pyx_v_self->Pi);
  __pyx_v_self->Pi = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":688
 * 
 *         self.Pi = ndarray_view(self._thisptr.Pi).squeeze()
 *         self.Pr = ndarray_view(self._thisptr.Pr).squeeze()             # <<<<<<<<<<<<<<
 *         self.Pt = ndarray_view(self._thisptr.Pt).squeeze()
 *         self.enh = ndarray_view(self._thisptr.enh).squeeze()
*/
  __pyx_t_2 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->Pr)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 688, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->Pr);
  __Pyx_DECREF((PyObject *)__pyx_v_self->Pr);
  __pyx_v_self->Pr = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":689
 *         self.Pi = ndarray_view(self._thisptr.Pi).squeeze()
 *         self.Pr = ndarray_view(self._thisptr.Pr).squeeze()
 *         self.Pt = ndarray_view(self._thisptr.Pt).squeeze()             # <<<<<<<<<<<<<<
 *         self.enh = ndarray_view(self._thisptr.enh).squeeze()
 *         self.beamArea = ndarray_view(self._thisptr.beamArea).squeeze()
*/
  __pyx_t_3 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->Pt)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 689, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 689, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->Pt);
  __Pyx_DECREF((PyObject *)__pyx_v_self->Pt);
  __pyx_v_self->Pt = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":690
 *         self.Pr = ndarray_view(self._thisptr.Pr).squeeze()
 *         self.Pt = ndarray_view(self._thisptr.Pt).squeeze()
 *         self.enh = ndarray_view(self._thisptr.enh).squeeze()             # <<<<<<<<<<<<<<
 *         self.beamArea = ndarray_view(self._thisptr.beamArea).squeeze()
 * 
*/
  __pyx_t_2 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->enh)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->enh);
  __Pyx_DECREF((PyObject *)__pyx_v_self->enh);
  __pyx_v_self->enh = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":691
 *         self.Pt = ndarray_view(self._thisptr.Pt).squeeze()
 *         self.enh = ndarray_view(self._thisptr.enh).squeeze()
 *         self.beamArea = ndarray_view(self._thisptr.beamArea).squeeze()             # <<<<<<<<<<<<<<
 * 
 * #===============================================================================
*/
  __pyx_t_3 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->beamArea)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 691, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->beamArea);
  __Pyx_DECREF((PyObject *)__pyx_v_self->beamArea);
  __pyx_v_self->beamArea = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":682
 *             del self._thisptr
 * 
 *     cdef _Init(self, WaveSweepResultNonlinearTMMCpp *ptr, bool needDealloc = True, object parent = None):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         self._needDealloc = needDealloc
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._WaveSweepResultNonlinearTMM._Init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":670
 *     """
 *     cdef WaveSweepResultNonlinearTMMCpp *_thisptr
 *     cdef readonly np.ndarray Pi, Pr, Pt, enh, beamArea             # <<<<<<<<<<<<<<
 *     cdef bool _needDealloc;
 *     cdef object _parent;
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pi_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pi_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pi___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pi___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->Pi);
  __pyx_r = ((PyObject *)__pyx_v_self->Pi);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pr_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pr_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pr___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pr___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->Pr);
  __pyx_r = ((PyObject *)__pyx_v_self->Pr);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pt_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pt_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pt___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pt___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->Pt);
  __pyx_r = ((PyObject *)__pyx_v_self->Pt);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_3enh_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_3enh_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_3enh___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_3enh___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->enh);
  __pyx_r = ((PyObject *)__pyx_v_self->enh);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_8beamArea_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_8beamArea_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_8beamArea___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_8beamArea___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->beamArea);
  __pyx_r = ((PyObject *)__pyx_v_self->beamArea);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_4__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._WaveSweepResultNonlinearTMM.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._WaveSweepResultNonlinearTMM.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_6__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._WaveSweepResultNonlinearTMM.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":716
 *     cdef readonly np.ndarray E, H;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__cinit__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ___cinit__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_self) {
  int __pyx_r;

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":719
 *         pass
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._thisptr
 * 
*/

/* Python wrapper */
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_2__dealloc__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_self) {

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":720
 * 
 *     def __dealloc__(self):
 *         del self._thisptr             # <<<<<<<<<<<<<<
 * 
 *     cdef _Init(self, FieldsZCpp *ptr):
*/
  delete __pyx_v_self->_thisptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":719
 *         pass
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._thisptr
 * 
*/

  /* function exit code */
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":722
 *         del self._thisptr
 * 
 *     cdef _Init(self, FieldsZCpp *ptr):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 * 
*/

static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_self, TMM::FieldsZ *__pyx_v_ptr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_Init", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":723
 * 
 *     cdef _Init(self, FieldsZCpp *ptr):
 *         self._thisptr = ptr             # <<<<<<<<<<<<<<
 * 
 *         self.E = ndarray_view(self._thisptr.E)
*/
  __pyx_v_self->_thisptr = __pyx_v_ptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":725
 *         self._thisptr = ptr
 * 
 *         self.E = ndarray_view(self._thisptr.E)             # <<<<<<<<<<<<<<
 *         self.H = ndarray_view(self._thisptr.H)
 * 
*/
  __pyx_t_1 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->E)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->E);
  __Pyx_DECREF((PyObject *)__pyx_v_self->E);
  __pyx_v_self->E = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":726
 * 
 *         self.E = ndarray_view(self._thisptr.E)
 *         self.H = ndarray_view(self._thisptr.H)             # <<<<<<<<<<<<<<
 * 
 * #===============================================================================
*/
  __pyx_t_1 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->H)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->H);
  __Pyx_DECREF((PyObject *)__pyx_v_self->H);
  __pyx_v_self->H = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":722
 *         del self._thisptr
 * 
 *     cdef _Init(self, FieldsZCpp *ptr):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._FieldsZ._Init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":714
 *     """
 *     cdef FieldsZCpp *_thisptr;
 *     cdef readonly np.ndarray E, H;             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1E_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1E_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1E___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1E___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->E);
  __pyx_r = ((PyObject *)__pyx_v_self->E);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1H_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1H_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1H___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1H___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->H);
  __pyx_r = ((PyObject *)__pyx_v_self->H);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_4__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._FieldsZ.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._FieldsZ.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_6__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._FieldsZ.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":767
 *     cdef readonly np.ndarray Ex, Ey, Ez, Hx, Hy, Hz;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__cinit__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX___cinit__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self) {
  int __pyx_r;

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":770
 *         pass
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._thisptr
 * 
*/

/* Python wrapper */
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2__dealloc__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self) {

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":771
 * 
 *     def __dealloc__(self):
 *         del self._thisptr             # <<<<<<<<<<<<<<
 * 
 *     cdef _Init(self, FieldsZXCpp *ptr):
*/
  delete __pyx_v_self->_thisptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":770
 *         pass
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._thisptr
 * 
*/

  /* function exit code */
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":773
 *         del self._thisptr
 * 
 *     cdef _Init(self, FieldsZXCpp *ptr):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
*/

static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self, TMM::FieldsZX *__pyx_v_ptr) {
  enum TMM::Polarization __pyx_v_pol;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_Init", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":774
 * 
 *     cdef _Init(self, FieldsZXCpp *ptr):
 *         self._thisptr = ptr             # <<<<<<<<<<<<<<
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
 * 
*/
  __pyx_v_self->_thisptr = __pyx_v_ptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":775
 *     cdef _Init(self, FieldsZXCpp *ptr):
 *         self._thisptr = ptr
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()             # <<<<<<<<<<<<<<
 * 
 *         if pol == P_POL:
*/
  __pyx_v_pol = __pyx_v_self->_thisptr->GetPol();

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":777
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
 * 
 *         if pol == P_POL:             # <<<<<<<<<<<<<<
 *             self.Hy = ndarray_view(self._thisptr.Hy)
 *             self.Ex = ndarray_view(self._thisptr.Ex)
*/
  switch (__pyx_v_pol) {
    case TMM::Polarization::P_POL:

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":778
 * 
 *         if pol == P_POL:
 *             self.Hy = ndarray_view(self._thisptr.Hy)             # <<<<<<<<<<<<<<
 *             self.Ex = ndarray_view(self._thisptr.Ex)
 *             self.Ez = ndarray_view(self._thisptr.Ez)
*/
    __pyx_t_1 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->Hy)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->Hy);
    __Pyx_DECREF((PyObject *)__pyx_v_self->Hy);
    __pyx_v_self->Hy = ((PyArrayObject *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":779
 *         if pol == P_POL:
 *             self.Hy = ndarray_view(self._thisptr.Hy)
 *             self.Ex = ndarray_view(self._thisptr.Ex)             # <<<<<<<<<<<<<<
 *             self.Ez = ndarray_view(self._thisptr.Ez)
 *             self.Hx = None
*/
    __pyx_t_1 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->Ex)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 779, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->Ex);
    __Pyx_DECREF((PyObject *)__pyx_v_self->Ex);
    __pyx_v_self->Ex = ((PyArrayObject *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":780
 *             self.Hy = ndarray_view(self._thisptr.Hy)
 *             self.Ex = ndarray_view(self._thisptr.Ex)
 *             self.Ez = ndarray_view(self._thisptr.Ez)             # <<<<<<<<<<<<<<
 *             self.Hx = None
 *             self.Ey = None
*/
    __pyx_t_1 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->Ez)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->Ez);
    __Pyx_DECREF((PyObject *)__pyx_v_self->Ez);
    __pyx_v_self->Ez = ((PyArrayObject *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":781
 *             self.Ex = ndarray_view(self._thisptr.Ex)
 *             self.Ez = ndarray_view(self._thisptr.Ez)
 *             self.Hx = None             # <<<<<<<<<<<<<<
 *             self.Ey = None
 *             self.Hz = None
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->Hx);
    __Pyx_DECREF((PyObject *)__pyx_v_self->Hx);
    __pyx_v_self->Hx = ((PyArrayObject *)Py_None);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":782
 *             self.Ez = ndarray_view(self._thisptr.Ez)
 *             self.Hx = None
 *             self.Ey = None             # <<<<<<<<<<<<<<
 *             self.Hz = None
 *         elif pol == S_POL:
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->Ey);
    __Pyx_DECREF((PyObject *)__pyx_v_self->Ey);
    __pyx_v_self->Ey = ((PyArrayObject *)Py_None);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":783
 *             self.Hx = None
 *             self.Ey = None
 *             self.Hz = None             # <<<<<<<<<<<<<<
 *         elif pol == S_POL:
 *             self.Hx = ndarray_view(self._thisptr.Hx)
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->Hz);
    __Pyx_DECREF((PyObject *)__pyx_v_self->Hz);
    __pyx_v_self->Hz = ((PyArrayObject *)Py_None);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":777
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
 * 
 *         if pol == P_POL:             # <<<<<<<<<<<<<<
 *             self.Hy = ndarray_view(self._thisptr.Hy)
 *             self.Ex = ndarray_view(self._thisptr.Ex)
*/
    break;
    case TMM::Polarization::S_POL:

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":785
 *             self.Hz = None
 *         elif pol == S_POL:
 *             self.Hx = ndarray_view(self._thisptr.Hx)             # <<<<<<<<<<<<<<
 *             self.Ey = ndarray_view(self._thisptr.Ey)
 *             self.Hz = ndarray_view(self._thisptr.Hz)
*/
    __pyx_t_1 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->Hx)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 785, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->Hx);
    __Pyx_DECREF((PyObject *)__pyx_v_self->Hx);
    __pyx_v_self->Hx = ((PyArrayObject *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":786
 *         elif pol == S_POL:
 *             self.Hx = ndarray_view(self._thisptr.Hx)
 *             self.Ey = ndarray_view(self._thisptr.Ey)             # <<<<<<<<<<<<<<
 *             self.Hz = ndarray_view(self._thisptr.Hz)
 *             self.Hy = None
*/
    __pyx_t_1 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->Ey)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->Ey);
    __Pyx_DECREF((PyObject *)__pyx_v_self->Ey);
    __pyx_v_self->Ey = ((PyArrayObject *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":787
 *             self.Hx = ndarray_view(self._thisptr.Hx)
 *             self.Ey = ndarray_view(self._thisptr.Ey)
 *             self.Hz = ndarray_view(self._thisptr.Hz)             # <<<<<<<<<<<<<<
 *             self.Hy = None
 *             self.Ex = None
*/
    __pyx_t_1 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->Hz)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->Hz);
    __Pyx_DECREF((PyObject *)__pyx_v_self->Hz);
    __pyx_v_self->Hz = ((PyArrayObject *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":788
 *             self.Ey = ndarray_view(self._thisptr.Ey)
 *             self.Hz = ndarray_view(self._thisptr.Hz)
 *             self.Hy = None             # <<<<<<<<<<<<<<
 *             self.Ex = None
 *             self.Ez = None
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->Hy);
    __Pyx_DECREF((PyObject *)__pyx_v_self->Hy);
    __pyx_v_self->Hy = ((PyArrayObject *)Py_None);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":789
 *             self.Hz = ndarray_view(self._thisptr.Hz)
 *             self.Hy = None
 *             self.Ex = None             # <<<<<<<<<<<<<<
 *             self.Ez = None
 *         else:
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->Ex);
    __Pyx_DECREF((PyObject *)__pyx_v_self->Ex);
    __pyx_v_self->Ex = ((PyArrayObject *)Py_None);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":790
 *             self.Hy = None
 *             self.Ex = None
 *             self.Ez = None             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Unknown polarization.")
*/
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->Ez);
    __Pyx_DECREF((PyObject *)__pyx_v_self->Ez);
    __pyx_v_self->Ez = ((PyArrayObject *)Py_None);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":784
 *             self.Ey = None
 *             self.Hz = None
 *         elif pol == S_POL:             # <<<<<<<<<<<<<<
 *             self.Hx = ndarray_view(self._thisptr.Hx)
 *             self.Ey = ndarray_view(self._thisptr.Ey)
*/
    break;
    default:

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":792
 *             self.Ez = None
 *         else:
 *             raise ValueError("Unknown polarization.")             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
    __pyx_t_2 = NULL;
    __pyx_t_3 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Unknown_polarization};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 792, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 792, __pyx_L1_error)
    break;
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":773
 *         del self._thisptr
 * 
 *     cdef _Init(self, FieldsZXCpp *ptr):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._FieldsZX._Init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":794
 *             raise ValueError("Unknown polarization.")
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def EN(self):
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2EN_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2EN_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2EN___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2EN___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self) {
  enum TMM::Polarization __pyx_v_pol;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":796
 *     @property
 *     def EN(self):
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()             # <<<<<<<<<<<<<<
 *         if pol == P_POL:
 *             return np.sqrt(self.Ex.real ** 2 + self.Ex.imag ** 2 + \
*/
  __pyx_v_pol = __pyx_v_self->_thisptr->GetPol();

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":797
 *     def EN(self):
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
 *         if pol == P_POL:             # <<<<<<<<<<<<<<
 *             return np.sqrt(self.Ex.real ** 2 + self.Ex.imag ** 2 + \
 *                            self.Ez.real ** 2 + self.Ez.imag ** 2)
*/
  switch (__pyx_v_pol) {
    case TMM::Polarization::P_POL:

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":798
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
 *         if pol == P_POL:
 *             return np.sqrt(self.Ex.real ** 2 + self.Ex.imag ** 2 + \             # <<<<<<<<<<<<<<
 *                            self.Ez.real ** 2 + self.Ez.imag ** 2)
 *         elif pol == S_POL:
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_sqrt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->Ex), __pyx_mstate_global->__pyx_n_u_real); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyNumber_Power(__pyx_t_3, __pyx_mstate_global->__pyx_int_2, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->Ex), __pyx_mstate_global->__pyx_n_u_imag); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = PyNumber_Power(__pyx_t_3, __pyx_mstate_global->__pyx_int_2, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":799
 *         if pol == P_POL:
 *             return np.sqrt(self.Ex.real ** 2 + self.Ex.imag ** 2 + \
 *                            self.Ez.real ** 2 + self.Ez.imag ** 2)             # <<<<<<<<<<<<<<
 *         elif pol == S_POL:
 *             return np.sqrt(self.Ey.real ** 2 + self.Ey.imag ** 2)
*/
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->Ez), __pyx_mstate_global->__pyx_n_u_real); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = PyNumber_Power(__pyx_t_6, __pyx_mstate_global->__pyx_int_2, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":798
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
 *         if pol == P_POL:
 *             return np.sqrt(self.Ex.real ** 2 + self.Ex.imag ** 2 + \             # <<<<<<<<<<<<<<
 *                            self.Ez.real ** 2 + self.Ez.imag ** 2)
 *         elif pol == S_POL:
*/
    __pyx_t_6 = PyNumber_Add(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":799
 *         if pol == P_POL:
 *             return np.sqrt(self.Ex.real ** 2 + self.Ex.imag ** 2 + \
 *                            self.Ez.real ** 2 + self.Ez.imag ** 2)             # <<<<<<<<<<<<<<
 *         elif pol == S_POL:
 *             return np.sqrt(self.Ey.real ** 2 + self.Ey.imag ** 2)
*/
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->Ez), __pyx_mstate_global->__pyx_n_u_imag); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PyNumber_Power(__pyx_t_5, __pyx_mstate_global->__pyx_int_2, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Add(__pyx_t_6, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 799, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":797
 *     def EN(self):
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
 *         if pol == P_POL:             # <<<<<<<<<<<<<<
 *             return np.sqrt(self.Ex.real ** 2 + self.Ex.imag ** 2 + \
 *                            self.Ez.real ** 2 + self.Ez.imag ** 2)
*/
    break;
    case TMM::Polarization::S_POL:

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":801
 *                            self.Ez.real ** 2 + self.Ez.imag ** 2)
 *         elif pol == S_POL:
 *             return np.sqrt(self.Ey.real ** 2 + self.Ey.imag ** 2)             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Unknown polarization.")
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 801, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_sqrt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 801, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->Ey), __pyx_mstate_global->__pyx_n_u_real); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 801, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = PyNumber_Power(__pyx_t_5, __pyx_mstate_global->__pyx_int_2, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 801, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->Ey), __pyx_mstate_global->__pyx_n_u_imag); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 801, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyNumber_Power(__pyx_t_5, __pyx_mstate_global->__pyx_int_2, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 801, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Add(__pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 801, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":800
 *             return np.sqrt(self.Ex.real ** 2 + self.Ex.imag ** 2 + \
 *                            self.Ez.real ** 2 + self.Ez.imag ** 2)
 *         elif pol == S_POL:             # <<<<<<<<<<<<<<
 *             return np.sqrt(self.Ey.real ** 2 + self.Ey.imag ** 2)
 *         else:
*/
    break;
    default:

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":803
 *             return np.sqrt(self.Ey.real ** 2 + self.Ey.imag ** 2)
 *         else:
 *             raise ValueError("Unknown polarization.")             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
    __pyx_t_2 = NULL;
    __pyx_t_7 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Unknown_polarization};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 803, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 803, __pyx_L1_error)
    break;
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":794
 *             raise ValueError("Unknown polarization.")
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def EN(self):
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._FieldsZX.EN.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":805
 *             raise ValueError("Unknown polarization.")
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def HN(self):
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2HN_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2HN_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2HN___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2HN___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self) {
  enum TMM::Polarization __pyx_v_pol;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":807
 *     @property
 *     def HN(self):
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()             # <<<<<<<<<<<<<<
 *         if pol == P_POL:
 *             return np.sqrt(self.Hy.real ** 2 + self.Hy.imag ** 2)
*/
  __pyx_v_pol = __pyx_v_self->_thisptr->GetPol();

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":808
 *     def HN(self):
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
 *         if pol == P_POL:             # <<<<<<<<<<<<<<
 *             return np.sqrt(self.Hy.real ** 2 + self.Hy.imag ** 2)
 *         elif pol == S_POL:
*/
  switch (__pyx_v_pol) {
    case TMM::Polarization::P_POL:

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":809
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
 *         if pol == P_POL:
 *             return np.sqrt(self.Hy.real ** 2 + self.Hy.imag ** 2)             # <<<<<<<<<<<<<<
 *         elif pol == S_POL:
 *             return np.sqrt(self.Hx.real ** 2 + self.Hx.imag ** 2 + \
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_sqrt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->Hy), __pyx_mstate_global->__pyx_n_u_real); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyNumber_Power(__pyx_t_3, __pyx_mstate_global->__pyx_int_2, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->Hy), __pyx_mstate_global->__pyx_n_u_imag); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = PyNumber_Power(__pyx_t_3, __pyx_mstate_global->__pyx_int_2, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_4, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 809, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":808
 *     def HN(self):
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
 *         if pol == P_POL:             # <<<<<<<<<<<<<<
 *             return np.sqrt(self.Hy.real ** 2 + self.Hy.imag ** 2)
 *         elif pol == S_POL:
*/
    break;
    case TMM::Polarization::S_POL:

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":811
 *             return np.sqrt(self.Hy.real ** 2 + self.Hy.imag ** 2)
 *         elif pol == S_POL:
 *             return np.sqrt(self.Hx.real ** 2 + self.Hx.imag ** 2 + \             # <<<<<<<<<<<<<<
 *                            self.Hz.real ** 2 + self.Hz.imag ** 2)
 *         else:
*/
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_sqrt); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->Hx), __pyx_mstate_global->__pyx_n_u_real); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = PyNumber_Power(__pyx_t_3, __pyx_mstate_global->__pyx_int_2, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->Hx), __pyx_mstate_global->__pyx_n_u_imag); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyNumber_Power(__pyx_t_3, __pyx_mstate_global->__pyx_int_2, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Add(__pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":812
 *         elif pol == S_POL:
 *             return np.sqrt(self.Hx.real ** 2 + self.Hx.imag ** 2 + \
 *                            self.Hz.real ** 2 + self.Hz.imag ** 2)             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Unknown polarization.")
*/
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->Hz), __pyx_mstate_global->__pyx_n_u_real); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 812, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyNumber_Power(__pyx_t_5, __pyx_mstate_global->__pyx_int_2, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 812, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":811
 *             return np.sqrt(self.Hy.real ** 2 + self.Hy.imag ** 2)
 *         elif pol == S_POL:
 *             return np.sqrt(self.Hx.real ** 2 + self.Hx.imag ** 2 + \             # <<<<<<<<<<<<<<
 *                            self.Hz.real ** 2 + self.Hz.imag ** 2)
 *         else:
*/
    __pyx_t_5 = PyNumber_Add(__pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":812
 *         elif pol == S_POL:
 *             return np.sqrt(self.Hx.real ** 2 + self.Hx.imag ** 2 + \
 *                            self.Hz.real ** 2 + self.Hz.imag ** 2)             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Unknown polarization.")
*/
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->Hz), __pyx_mstate_global->__pyx_n_u_imag); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 812, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = PyNumber_Power(__pyx_t_6, __pyx_mstate_global->__pyx_int_2, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 812, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 812, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_t_2, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 811, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":810
 *         if pol == P_POL:
 *             return np.sqrt(self.Hy.real ** 2 + self.Hy.imag ** 2)
 *         elif pol == S_POL:             # <<<<<<<<<<<<<<
 *             return np.sqrt(self.Hx.real ** 2 + self.Hx.imag ** 2 + \
 *                            self.Hz.real ** 2 + self.Hz.imag ** 2)
*/
    break;
    default:

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":814
 *                            self.Hz.real ** 2 + self.Hz.imag ** 2)
 *         else:
 *             raise ValueError("Unknown polarization.")             # <<<<<<<<<<<<<<
 * 
 * #===============================================================================
*/
    __pyx_t_2 = NULL;
    __pyx_t_7 = 1;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_kp_u_Unknown_polarization};
      __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 814, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 814, __pyx_L1_error)
    break;
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":805
 *             raise ValueError("Unknown polarization.")
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def HN(self):
 *         cdef PolarizationCpp pol = self._thisptr.GetPol()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._FieldsZX.HN.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":765
 *     """
 *     cdef FieldsZXCpp *_thisptr;
 *     cdef readonly np.ndarray Ex, Ey, Ez, Hx, Hy, Hz;             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ex_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ex_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ex___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ex___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->Ex);
  __pyx_r = ((PyObject *)__pyx_v_self->Ex);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ey_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ey_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ey___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ey___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->Ey);
  __pyx_r = ((PyObject *)__pyx_v_self->Ey);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ez_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ez_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ez___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ez___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->Ez);
  __pyx_r = ((PyObject *)__pyx_v_self->Ez);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hx_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hx_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hx___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hx___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->Hx);
  __pyx_r = ((PyObject *)__pyx_v_self->Hx);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hy_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hy_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hy___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hy___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->Hy);
  __pyx_r = ((PyObject *)__pyx_v_self->Hy);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hz_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hz_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hz___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hz___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->Hz);
  __pyx_r = ((PyObject *)__pyx_v_self->Hz);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_4__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._FieldsZX.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._FieldsZX.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_6__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._FieldsZX.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":834
 *     cdef HomogeneousWaveCpp *_thisptr
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._thisptr = NULL
 * 
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__cinit__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_self) {
  int __pyx_r;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":835
 * 
 *     def __cinit__(self):
 *         self._thisptr = NULL             # <<<<<<<<<<<<<<
 * 
 *     cdef _Init(self, HomogeneousWaveCpp *ptr):
*/
  __pyx_v_self->_thisptr = NULL;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":834
 *     cdef HomogeneousWaveCpp *_thisptr
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._thisptr = NULL
 * 
*/

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":837
 *         self._thisptr = NULL
 * 
 *     cdef _Init(self, HomogeneousWaveCpp *ptr):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 * 
*/

static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_self, TMM::HomogeneousWave *__pyx_v_ptr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_Init", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":838
 * 
 *     cdef _Init(self, HomogeneousWaveCpp *ptr):
 *         self._thisptr = ptr             # <<<<<<<<<<<<<<
 * 
 *     def GetMainFields(self, double z):
*/
  __pyx_v_self->_thisptr = __pyx_v_ptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":837
 *         self._thisptr = NULL
 * 
 *     cdef _Init(self, HomogeneousWaveCpp *ptr):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":840
 *         self._thisptr = ptr
 * 
 *     def GetMainFields(self, double z):             # <<<<<<<<<<<<<<
 *         """GetMainFields(z)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_3GetMainFields(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_2GetMainFields, "GetMainFields(z)\n\n        Returns forward and backward component of the main fields at the\n        distance z.  In case of s-pol the main field is Ey and in case of p-pol\n        the main field is Hy.\n\n        Parameters\n        ----------\n        z : float\n            z-coordinate for the main fields calculation. z=0 is the beginning of\n            the layer.\n\n        Returns\n        -------\n        ndarray(2,) of complex\n            The forward and backward component of the main fields.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_3GetMainFields = {"GetMainFields", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_3GetMainFields, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_2GetMainFields};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_3GetMainFields(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  double __pyx_v_z;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("GetMainFields (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_z,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 840, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 840, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "GetMainFields", 0) < (0)) __PYX_ERR(0, 840, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("GetMainFields", 1, 1, 1, i); __PYX_ERR(0, 840, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 840, __pyx_L3_error)
    }
    __pyx_v_z = __Pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_z == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 840, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("GetMainFields", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 840, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._HomogeneousWave.GetMainFields", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_2GetMainFields(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *)__pyx_v_self), __pyx_v_z);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_2GetMainFields(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_self, double __pyx_v_z) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Eigen::Array2cd __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("GetMainFields", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":860
 *         """
 *         # Have to copy, because result is in stack
 *         return ndarray_copy(self._thisptr.GetMainFields(z))             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetMainFields(__pyx_v_z);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 860, __pyx_L1_error)
  }
  __pyx_t_2 = ((PyObject *)eigency::ndarray_copy(__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":840
 *         self._thisptr = ptr
 * 
 *     def GetMainFields(self, double z):             # <<<<<<<<<<<<<<
 *         """GetMainFields(z)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._HomogeneousWave.GetMainFields", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":862
 *         return ndarray_copy(self._thisptr.GetMainFields(z))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def kzF(self):
 *         return self._thisptr.GetKzF()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_3kzF_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_3kzF_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_3kzF___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_3kzF___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __pyx_t_double_complex __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":864
 *     @property
 *     def kzF(self):
 *         return self._thisptr.GetKzF()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetKzF();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 864, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_PyComplex_FromComplex(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":862
 *         return ndarray_copy(self._thisptr.GetMainFields(z))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def kzF(self):
 *         return self._thisptr.GetKzF()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._HomogeneousWave.kzF.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":866
 *         return self._thisptr.GetKzF()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def kx(self):
 *         return self._thisptr.GetKx()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_2kx_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_2kx_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_2kx___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_2kx___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":868
 *     @property
 *     def kx(self):
 *         return self._thisptr.GetKx()             # <<<<<<<<<<<<<<
 * 
 * #===============================================================================
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetKx();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 868, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 868, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":866
 *         return self._thisptr.GetKzF()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def kx(self):
 *         return self._thisptr.GetKx()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._HomogeneousWave.kx.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_4__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._HomogeneousWave.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._HomogeneousWave.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_6__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._HomogeneousWave.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":889
 *     cdef int _layerNr;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._thisptr = NULL
 *         self.hw = None
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__cinit__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":890
 * 
 *     def __cinit__(self):
 *         self._thisptr = NULL             # <<<<<<<<<<<<<<
 *         self.hw = None
 * 
*/
  __pyx_v_self->_thisptr = NULL;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":891
 *     def __cinit__(self):
 *         self._thisptr = NULL
 *         self.hw = None             # <<<<<<<<<<<<<<
 * 
 *     cdef _Init(self, NonlinearLayerCpp *ptr, int layerNr, object parent):
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->hw);
  __Pyx_DECREF(__pyx_v_self->hw);
  __pyx_v_self->hw = Py_None;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":889
 *     cdef int _layerNr;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._thisptr = NULL
 *         self.hw = None
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":893
 *         self.hw = None
 * 
 *     cdef _Init(self, NonlinearLayerCpp *ptr, int layerNr, object parent):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         self._parent = parent
*/

static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self, TMM::NonlinearLayer *__pyx_v_ptr, int __pyx_v_layerNr, PyObject *__pyx_v_parent) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *__pyx_v_hw = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  TMM::HomogeneousWave *__pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_Init", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":894
 * 
 *     cdef _Init(self, NonlinearLayerCpp *ptr, int layerNr, object parent):
 *         self._thisptr = ptr             # <<<<<<<<<<<<<<
 *         self._parent = parent
 *         self._layerNr = layerNr
*/
  __pyx_v_self->_thisptr = __pyx_v_ptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":895
 *     cdef _Init(self, NonlinearLayerCpp *ptr, int layerNr, object parent):
 *         self._thisptr = ptr
 *         self._parent = parent             # <<<<<<<<<<<<<<
 *         self._layerNr = layerNr
 * 
*/
  __Pyx_INCREF(__pyx_v_parent);
  __Pyx_GIVEREF(__pyx_v_parent);
  __Pyx_GOTREF(__pyx_v_self->_parent);
  __Pyx_DECREF(__pyx_v_self->_parent);
  __pyx_v_self->_parent = __pyx_v_parent;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":896
 *         self._thisptr = ptr
 *         self._parent = parent
 *         self._layerNr = layerNr             # <<<<<<<<<<<<<<
 * 
 *         hw = _HomogeneousWave()
*/
  __pyx_v_self->_layerNr = __pyx_v_layerNr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":898
 *         self._layerNr = layerNr
 * 
 *         hw = _HomogeneousWave()             # <<<<<<<<<<<<<<
 *         hw._Init(self._thisptr.GetHw())
 *         self.hw = hw
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 898, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_hw = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":899
 * 
 *         hw = _HomogeneousWave()
 *         hw._Init(self._thisptr.GetHw())             # <<<<<<<<<<<<<<
 *         self.hw = hw
 * 
*/
  try {
    __pyx_t_4 = __pyx_v_self->_thisptr->GetHw();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 899, __pyx_L1_error)
  }
  __pyx_t_1 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *)__pyx_v_hw->__pyx_vtab)->_Init(__pyx_v_hw, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 899, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":900
 *         hw = _HomogeneousWave()
 *         hw._Init(self._thisptr.GetHw())
 *         self.hw = hw             # <<<<<<<<<<<<<<
 * 
 *     def GetIntensity(self, double z):
*/
  __Pyx_INCREF((PyObject *)__pyx_v_hw);
  __Pyx_GIVEREF((PyObject *)__pyx_v_hw);
  __Pyx_GOTREF(__pyx_v_self->hw);
  __Pyx_DECREF(__pyx_v_self->hw);
  __pyx_v_self->hw = ((PyObject *)__pyx_v_hw);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":893
 *         self.hw = None
 * 
 *     cdef _Init(self, NonlinearLayerCpp *ptr, int layerNr, object parent):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         self._parent = parent
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._NonlinearLayer._Init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_hw);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":902
 *         self.hw = hw
 * 
 *     def GetIntensity(self, double z):             # <<<<<<<<<<<<<<
 *         """GetIntensity(z)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_3GetIntensity(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_2GetIntensity, "GetIntensity(z)\n\n        Returns intensity of the plane wave at coordinate ``z``.\n\n        Parameters\n        ----------\n        z : float\n            z-coordinate\n\n        Returns\n        -------\n        float\n            The intensity of the plane waves in layer.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_3GetIntensity = {"GetIntensity", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_3GetIntensity, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_2GetIntensity};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_3GetIntensity(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  double __pyx_v_z;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("GetIntensity (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_z,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 902, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 902, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "GetIntensity", 0) < (0)) __PYX_ERR(0, 902, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("GetIntensity", 1, 1, 1, i); __PYX_ERR(0, 902, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 902, __pyx_L3_error)
    }
    __pyx_v_z = __Pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_z == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 902, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("GetIntensity", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 902, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._NonlinearLayer.GetIntensity", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_2GetIntensity(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)__pyx_v_self), __pyx_v_z);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_2GetIntensity(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self, double __pyx_v_z) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("GetIntensity", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":918
 * 
 *         """
 *         return self._thisptr.GetIntensity(z);             # <<<<<<<<<<<<<<
 * 
 *     def GetAbsorbedIntensity(self):
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetIntensity(__pyx_v_z);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 918, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":902
 *         self.hw = hw
 * 
 *     def GetIntensity(self, double z):             # <<<<<<<<<<<<<<
 *         """GetIntensity(z)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._NonlinearLayer.GetIntensity", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":920
 *         return self._thisptr.GetIntensity(z);
 * 
 *     def GetAbsorbedIntensity(self):             # <<<<<<<<<<<<<<
 *         """GetAbsorbedIntensity()
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_5GetAbsorbedIntensity(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_4GetAbsorbedIntensity, "GetAbsorbedIntensity()\n\n        Returns absorbed intensity in the layer.\n\n        Returns\n        -------\n        float\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_5GetAbsorbedIntensity = {"GetAbsorbedIntensity", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_5GetAbsorbedIntensity, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_4GetAbsorbedIntensity};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_5GetAbsorbedIntensity(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("GetAbsorbedIntensity (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("GetAbsorbedIntensity", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("GetAbsorbedIntensity", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_4GetAbsorbedIntensity(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_4GetAbsorbedIntensity(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("GetAbsorbedIntensity", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":930
 * 
 *         """
 *         return self._thisptr.GetAbsorbedIntensity();             # <<<<<<<<<<<<<<
 * 
 *     def GetSrcIntensity(self):
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetAbsorbedIntensity();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 930, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":920
 *         return self._thisptr.GetIntensity(z);
 * 
 *     def GetAbsorbedIntensity(self):             # <<<<<<<<<<<<<<
 *         """GetAbsorbedIntensity()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._NonlinearLayer.GetAbsorbedIntensity", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":932
 *         return self._thisptr.GetAbsorbedIntensity();
 * 
 *     def GetSrcIntensity(self):             # <<<<<<<<<<<<<<
 *         """GetSrcIntensity()
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_7GetSrcIntensity(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_6GetSrcIntensity, "GetSrcIntensity()\n\n        Calculates source power of the layer. Only nonzero if the material is\n        nonlinear.\n\n        Returns\n        -------\n        float\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_7GetSrcIntensity = {"GetSrcIntensity", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_7GetSrcIntensity, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_6GetSrcIntensity};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_7GetSrcIntensity(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("GetSrcIntensity (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("GetSrcIntensity", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("GetSrcIntensity", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_6GetSrcIntensity(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_6GetSrcIntensity(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("GetSrcIntensity", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":943
 * 
 *         """
 *         return self._thisptr.GetSrcIntensity();             # <<<<<<<<<<<<<<
 * 
 *     # Getter
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetSrcIntensity();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 943, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":932
 *         return self._thisptr.GetAbsorbedIntensity();
 * 
 *     def GetSrcIntensity(self):             # <<<<<<<<<<<<<<
 *         """GetSrcIntensity()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._NonlinearLayer.GetSrcIntensity", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":948
 *     #---------------------------------------------------------------------------
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def d(self):
 *         return self._thisptr.GetThickness()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1d_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1d_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1d___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1d___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":950
 *     @property
 *     def d(self):
 *         return self._thisptr.GetThickness()             # <<<<<<<<<<<<<<
 * 
 *     # Setter
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetThickness();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 950, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":948
 *     #---------------------------------------------------------------------------
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def d(self):
 *         return self._thisptr.GetThickness()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._NonlinearLayer.d.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":955
 *     #---------------------------------------------------------------------------
 * 
 *     @d.setter             # <<<<<<<<<<<<<<
 *     def d(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetThickness(value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1d_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1d_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1d_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1d_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":957
 *     @d.setter
 *     def d(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetThickness(value)             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __Pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 957, __pyx_L1_error)
  try {
    __pyx_v_self->_thisptr->SetThickness(__pyx_t_1);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 957, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":955
 *     #---------------------------------------------------------------------------
 * 
 *     @d.setter             # <<<<<<<<<<<<<<
 *     def d(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetThickness(value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._NonlinearLayer.d.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":885
 *     """
 *     cdef NonlinearLayerCpp *_thisptr
 *     cdef readonly object hw             # <<<<<<<<<<<<<<
 *     cdef object _parent;
 *     cdef int _layerNr;
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_2hw_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_2hw_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_2hw___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_2hw___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->hw);
  __pyx_r = __pyx_v_self->hw;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_9__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_8__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._NonlinearLayer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_11__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._NonlinearLayer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_10__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._NonlinearLayer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1016
 *     cdef readonly _Wave wave;
 * 
 *     def __cinit__(self, bool initStruct = True, object parent = None, **kwargs):             # <<<<<<<<<<<<<<
 *         if initStruct:
 *             self._thisptr = new NonlinearTMMCpp()
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  bool __pyx_v_initStruct;
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_initStruct,&__pyx_mstate_global->__pyx_n_u_parent,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1016, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1016, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1016, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values, kwd_pos_args, __pyx_kwds_len, "__cinit__", 1) < (0)) __PYX_ERR(0, 1016, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1016, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1016, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject *)Py_None));
    }
    if (values[0]) {
      __pyx_v_initStruct = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_initStruct == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1016, __pyx_L3_error)
    } else {
      __pyx_v_initStruct = ((bool)1);
    }
    __pyx_v_parent = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 2, __pyx_nargs); __PYX_ERR(0, 1016, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), __pyx_v_initStruct, __pyx_v_parent, __pyx_v_kwargs);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, bool __pyx_v_initStruct, PyObject *__pyx_v_parent, PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  TMM::NonlinearTMM *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1017
 * 
 *     def __cinit__(self, bool initStruct = True, object parent = None, **kwargs):
 *         if initStruct:             # <<<<<<<<<<<<<<
 *             self._thisptr = new NonlinearTMMCpp()
 *             self._Init(self._thisptr)
*/
  __pyx_t_1 = (__pyx_v_initStruct != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1018
 *     def __cinit__(self, bool initStruct = True, object parent = None, **kwargs):
 *         if initStruct:
 *             self._thisptr = new NonlinearTMMCpp()             # <<<<<<<<<<<<<<
 *             self._Init(self._thisptr)
 *             self._needDealloc = True
*/
    try {
      __pyx_t_2 = new TMM::NonlinearTMM();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 1018, __pyx_L1_error)
    }
    __pyx_v_self->_thisptr = __pyx_t_2;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1019
 *         if initStruct:
 *             self._thisptr = new NonlinearTMMCpp()
 *             self._Init(self._thisptr)             # <<<<<<<<<<<<<<
 *             self._needDealloc = True
 *         else:
*/
    __pyx_t_3 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self->__pyx_vtab)->_Init(__pyx_v_self, __pyx_v_self->_thisptr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1019, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1020
 *             self._thisptr = new NonlinearTMMCpp()
 *             self._Init(self._thisptr)
 *             self._needDealloc = True             # <<<<<<<<<<<<<<
 *         else:
 *             self._thisptr = NULL
*/
    __pyx_v_self->_needDealloc = 1;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1017
 * 
 *     def __cinit__(self, bool initStruct = True, object parent = None, **kwargs):
 *         if initStruct:             # <<<<<<<<<<<<<<
 *             self._thisptr = new NonlinearTMMCpp()
 *             self._Init(self._thisptr)
*/
    goto __pyx_L3;
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1022
 *             self._needDealloc = True
 *         else:
 *             self._thisptr = NULL             # <<<<<<<<<<<<<<
 *             self._needDealloc = False
 *         self._parent = parent
*/
  /*else*/ {
    __pyx_v_self->_thisptr = NULL;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1023
 *         else:
 *             self._thisptr = NULL
 *             self._needDealloc = False             # <<<<<<<<<<<<<<
 *         self._parent = parent
 * 
*/
    __pyx_v_self->_needDealloc = 0;
  }
  __pyx_L3:;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1024
 *             self._thisptr = NULL
 *             self._needDealloc = False
 *         self._parent = parent             # <<<<<<<<<<<<<<
 * 
 *         self.materialsCache = []
*/
  __Pyx_INCREF(__pyx_v_parent);
  __Pyx_GIVEREF(__pyx_v_parent);
  __Pyx_GOTREF(__pyx_v_self->_parent);
  __Pyx_DECREF(__pyx_v_self->_parent);
  __pyx_v_self->_parent = __pyx_v_parent;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1026
 *         self._parent = parent
 * 
 *         self.materialsCache = []             # <<<<<<<<<<<<<<
 *         self.layers = []
 *         self.SetParams(**kwargs)
*/
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1026, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->materialsCache);
  __Pyx_DECREF(__pyx_v_self->materialsCache);
  __pyx_v_self->materialsCache = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1027
 * 
 *         self.materialsCache = []
 *         self.layers = []             # <<<<<<<<<<<<<<
 *         self.SetParams(**kwargs)
 * 
*/
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1027, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->layers);
  __Pyx_DECREF(__pyx_v_self->layers);
  __pyx_v_self->layers = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1028
 *         self.materialsCache = []
 *         self.layers = []
 *         self.SetParams(**kwargs)             # <<<<<<<<<<<<<<
 * 
 *     cdef _Init(self, NonlinearTMMCpp* ptr):
*/
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_SetParams); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_mstate_global->__pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1016
 *     cdef readonly _Wave wave;
 * 
 *     def __cinit__(self, bool initStruct = True, object parent = None, **kwargs):             # <<<<<<<<<<<<<<
 *         if initStruct:
 *             self._thisptr = new NonlinearTMMCpp()
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1030
 *         self.SetParams(**kwargs)
 * 
 *     cdef _Init(self, NonlinearTMMCpp* ptr):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         cdef WaveCpp *wavePtr = self._thisptr.GetWave()
*/

static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, TMM::NonlinearTMM *__pyx_v_ptr) {
  TMM::Wave *__pyx_v_wavePtr;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *__pyx_v_wave = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  TMM::Wave *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_Init", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1031
 * 
 *     cdef _Init(self, NonlinearTMMCpp* ptr):
 *         self._thisptr = ptr             # <<<<<<<<<<<<<<
 *         cdef WaveCpp *wavePtr = self._thisptr.GetWave()
 *         wave = _Wave()
*/
  __pyx_v_self->_thisptr = __pyx_v_ptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1032
 *     cdef _Init(self, NonlinearTMMCpp* ptr):
 *         self._thisptr = ptr
 *         cdef WaveCpp *wavePtr = self._thisptr.GetWave()             # <<<<<<<<<<<<<<
 *         wave = _Wave()
 *         wave._Init(wavePtr)
*/
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetWave();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1032, __pyx_L1_error)
  }
  __pyx_v_wavePtr = __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1033
 *         self._thisptr = ptr
 *         cdef WaveCpp *wavePtr = self._thisptr.GetWave()
 *         wave = _Wave()             # <<<<<<<<<<<<<<
 *         wave._Init(wavePtr)
 *         self.wave = wave
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1033, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_wave = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1034
 *         cdef WaveCpp *wavePtr = self._thisptr.GetWave()
 *         wave = _Wave()
 *         wave._Init(wavePtr)             # <<<<<<<<<<<<<<
 *         self.wave = wave
 * 
*/
  __pyx_t_2 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)__pyx_v_wave->__pyx_vtab)->_Init(__pyx_v_wave, __pyx_v_wavePtr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1035
 *         wave = _Wave()
 *         wave._Init(wavePtr)
 *         self.wave = wave             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  __Pyx_INCREF((PyObject *)__pyx_v_wave);
  __Pyx_GIVEREF((PyObject *)__pyx_v_wave);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->wave);
  __Pyx_DECREF((PyObject *)__pyx_v_self->wave);
  __pyx_v_self->wave = __pyx_v_wave;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1030
 *         self.SetParams(**kwargs)
 * 
 *     cdef _Init(self, NonlinearTMMCpp* ptr):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 *         cdef WaveCpp *wavePtr = self._thisptr.GetWave()
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM._Init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wave);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1037
 *         self.wave = wave
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._needDealloc:
 *             del self._thisptr
*/

/* Python wrapper */
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2__dealloc__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self) {
  int __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1038
 * 
 *     def __dealloc__(self):
 *         if self._needDealloc:             # <<<<<<<<<<<<<<
 *             del self._thisptr
 * 
*/
  __pyx_t_1 = (__pyx_v_self->_needDealloc != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1039
 *     def __dealloc__(self):
 *         if self._needDealloc:
 *             del self._thisptr             # <<<<<<<<<<<<<<
 * 
 *     # Methods
*/
    delete __pyx_v_self->_thisptr;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1038
 * 
 *     def __dealloc__(self):
 *         if self._needDealloc:             # <<<<<<<<<<<<<<
 *             del self._thisptr
 * 
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1037
 *         self.wave = wave
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._needDealloc:
 *             del self._thisptr
*/

  /* function exit code */
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1044
 *     #---------------------------------------------------------------------------
 * 
 *     def AddLayer(self, double d, Material material):             # <<<<<<<<<<<<<<
 *         """AddLayer(d, material)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_5AddLayer(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4AddLayer, "AddLayer(d, material)\n\n        Adds layer to the TMM.\n\n        Parameters\n        ----------\n        d : float\n            Thickness of the layer in meters. First and the last medium have\n            infinite thickness.\n        material : :class:`Material`\n            Instance of the helper class :class:`Material`\n\n        Examples\n        --------\n        >>> tmm = TMM()\n        >>> tmm.AddLayer(float(\"inf\"), Material.Static(1.5))\n        >>> tmm.AddLayer(50e-9, Material.Static(0.3 + 3j))\n        >>> tmm.AddLayer(float(\"inf\"), Material.Static(1.0))\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_5AddLayer = {"AddLayer", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_5AddLayer, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4AddLayer};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_5AddLayer(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  double __pyx_v_d;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_material = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("AddLayer (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_d,&__pyx_mstate_global->__pyx_n_u_material,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1044, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1044, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1044, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "AddLayer", 0) < (0)) __PYX_ERR(0, 1044, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("AddLayer", 1, 2, 2, i); __PYX_ERR(0, 1044, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1044, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1044, __pyx_L3_error)
    }
    __pyx_v_d = __Pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_d == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1044, __pyx_L3_error)
    __pyx_v_material = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("AddLayer", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1044, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.AddLayer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_material), __pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material, 1, "material", 0))) __PYX_ERR(0, 1044, __pyx_L1_error)
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4AddLayer(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), __pyx_v_d, __pyx_v_material);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4AddLayer(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, double __pyx_v_d, struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_material) {
  int __pyx_v_lastLayer;
  TMM::NonlinearLayer *__pyx_v_layerptr;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *__pyx_v_layer = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  TMM::NonlinearLayer *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("AddLayer", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1065
 *         """
 *         # No copy of material is made
 *         self._thisptr.AddLayer(d, material._thisptr)             # <<<<<<<<<<<<<<
 * 
 *         # Cache material classes, avoids dealloc
*/
  try {
    __pyx_v_self->_thisptr->AddLayer(__pyx_v_d, __pyx_v_material->_thisptr);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1065, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1068
 * 
 *         # Cache material classes, avoids dealloc
 *         self.materialsCache.append(material)             # <<<<<<<<<<<<<<
 * 
 *         # Layers list
*/
  if (unlikely(__pyx_v_self->materialsCache == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1068, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_Append(__pyx_v_self->materialsCache, ((PyObject *)__pyx_v_material)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1068, __pyx_L1_error)

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1071
 * 
 *         # Layers list
 *         cdef int lastLayer = self._thisptr.LayersCount() - 1             # <<<<<<<<<<<<<<
 *         cdef NonlinearLayerCpp *layerptr = self._thisptr.GetLayer(lastLayer)
 *         layer = _NonlinearLayer()
*/
  try {
    __pyx_t_2 = __pyx_v_self->_thisptr->LayersCount();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1071, __pyx_L1_error)
  }
  __pyx_v_lastLayer = (__pyx_t_2 - 1);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1072
 *         # Layers list
 *         cdef int lastLayer = self._thisptr.LayersCount() - 1
 *         cdef NonlinearLayerCpp *layerptr = self._thisptr.GetLayer(lastLayer)             # <<<<<<<<<<<<<<
 *         layer = _NonlinearLayer()
 *         layer._Init(layerptr, lastLayer, self)
*/
  try {
    __pyx_t_3 = __pyx_v_self->_thisptr->GetLayer(__pyx_v_lastLayer);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1072, __pyx_L1_error)
  }
  __pyx_v_layerptr = __pyx_t_3;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1073
 *         cdef int lastLayer = self._thisptr.LayersCount() - 1
 *         cdef NonlinearLayerCpp *layerptr = self._thisptr.GetLayer(lastLayer)
 *         layer = _NonlinearLayer()             # <<<<<<<<<<<<<<
 *         layer._Init(layerptr, lastLayer, self)
 *         self.layers.append(layer)
*/
  __pyx_t_5 = NULL;
  __pyx_t_6 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1073, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_4);
  }
  __pyx_v_layer = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1074
 *         cdef NonlinearLayerCpp *layerptr = self._thisptr.GetLayer(lastLayer)
 *         layer = _NonlinearLayer()
 *         layer._Init(layerptr, lastLayer, self)             # <<<<<<<<<<<<<<
 *         self.layers.append(layer)
 * 
*/
  __pyx_t_4 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)__pyx_v_layer->__pyx_vtab)->_Init(__pyx_v_layer, __pyx_v_layerptr, __pyx_v_lastLayer, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1075
 *         layer = _NonlinearLayer()
 *         layer._Init(layerptr, lastLayer, self)
 *         self.layers.append(layer)             # <<<<<<<<<<<<<<
 * 
 *     def SetParams(self, **kwargs):
*/
  if (unlikely(__pyx_v_self->layers == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1075, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_Append(__pyx_v_self->layers, ((PyObject *)__pyx_v_layer)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1075, __pyx_L1_error)

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1044
 *     #---------------------------------------------------------------------------
 * 
 *     def AddLayer(self, double d, Material material):             # <<<<<<<<<<<<<<
 *         """AddLayer(d, material)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.AddLayer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_layer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1077
 *         self.layers.append(layer)
 * 
 *     def SetParams(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """SetParams(**kwargs)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_7SetParams(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_6SetParams, "SetParams(**kwargs)\n\n        Helper method to set the values of all the attributes. See the docstring\n        of :class:`NonlinearTMM`.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_7SetParams = {"SetParams", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_7SetParams, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_6SetParams};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_7SetParams(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("SetParams (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("SetParams", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (__pyx_kwds_len > 0) {
    if (unlikely(__Pyx_CheckKeywordStrings("SetParams", __pyx_kwds) == -1)) return NULL;
    __pyx_v_kwargs = __Pyx_KwargsAsDict_FASTCALL(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_6SetParams(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_6SetParams(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("SetParams", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1084
 * 
 *         """
 *         for name, value in kwargs.items():             # <<<<<<<<<<<<<<
 *             if name not in paramDict:
 *                 raise ValueError("Unknown kwarg %s" % (name))
*/
  __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_v_kwargs, 1, __pyx_mstate_global->__pyx_n_u_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 1084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1085
 *         """
 *         for name, value in kwargs.items():
 *             if name not in paramDict:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Unknown kwarg %s" % (name))
 *             setattr(self, name, value)
*/
    __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_v_name, __pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_paramDict, Py_NE)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1085, __pyx_L1_error)
    if (unlikely(__pyx_t_8)) {

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1086
 *         for name, value in kwargs.items():
 *             if name not in paramDict:
 *                 raise ValueError("Unknown kwarg %s" % (name))             # <<<<<<<<<<<<<<
 *             setattr(self, name, value)
 * 
*/
      __pyx_t_5 = NULL;
      __pyx_t_9 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_Unknown_kwarg_s, __pyx_v_name); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1086, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_9};
        __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1086, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 1086, __pyx_L1_error)

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1085
 *         """
 *         for name, value in kwargs.items():
 *             if name not in paramDict:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Unknown kwarg %s" % (name))
 *             setattr(self, name, value)
*/
    }

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1087
 *             if name not in paramDict:
 *                 raise ValueError("Unknown kwarg %s" % (name))
 *             setattr(self, name, value)             # <<<<<<<<<<<<<<
 * 
 *     def Solve(self, **kwargs):
*/
    __pyx_t_11 = PyObject_SetAttr(((PyObject *)__pyx_v_self), __pyx_v_name, __pyx_v_value); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 1087, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1077
 *         self.layers.append(layer)
 * 
 *     def SetParams(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """SetParams(**kwargs)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.SetParams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1089
 *             setattr(self, name, value)
 * 
 *     def Solve(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """Solve(**kwargs)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_9Solve(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_8Solve, "Solve(**kwargs)\n\n        Solves the structure.\n\n        Parameters\n        ----------\n        wl : float, optional\n             Wavelength in meters.\n        pol : str, optional\n            \"p\" or \"s\" denoting the polarization of the waves.\n        beta : float, optional\n            Normalized tangential wave vector.\n        E0 : complex, optional\n            Input plane wave amplitude in vacuum. Only used if ``overrideE0`` is\n            True.\n        I0 : float, optional\n            Input plane wave intensity in the first layer. Only used if\n            ``overrideE0`` is False.\n        overrideE0 : bool, optional\n            Selects between ``E0`` and ``I0``.\n\n        Examples\n        --------\n        >>> tmm.Solve(wl = 532e-9, beta = 0.0)\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_9Solve = {"Solve", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_9Solve, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_8Solve};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_9Solve(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Solve (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("Solve", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (__pyx_kwds_len > 0) {
    if (unlikely(__Pyx_CheckKeywordStrings("Solve", __pyx_kwds) == -1)) return NULL;
    __pyx_v_kwargs = __Pyx_KwargsAsDict_FASTCALL(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_8Solve(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_8Solve(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Solve", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1116
 * 
 *         """
 *         self.SetParams(**kwargs)             # <<<<<<<<<<<<<<
 *         self._thisptr.Solve()
 * 
*/
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_SetParams); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_mstate_global->__pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1117
 *         """
 *         self.SetParams(**kwargs)
 *         self._thisptr.Solve()             # <<<<<<<<<<<<<<
 * 
 *     def GetIntensities(self):
*/
  try {
    __pyx_v_self->_thisptr->Solve();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1117, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1089
 *             setattr(self, name, value)
 * 
 *     def Solve(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """Solve(**kwargs)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.Solve", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1119
 *         self._thisptr.Solve()
 * 
 *     def GetIntensities(self):             # <<<<<<<<<<<<<<
 *         """GetIntensities()
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_11GetIntensities(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10GetIntensities, "GetIntensities()\n\n        Returns the intensities and amplitudes of incident, reflected and\n        transmitted wave. The structure must be solved first.\n\n        Returns\n        -------\n        :class:`_Intensities`\n            Helper class to hold intensity data.\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_11GetIntensities = {"GetIntensities", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_11GetIntensities, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10GetIntensities};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_11GetIntensities(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("GetIntensities (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("GetIntensities", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("GetIntensities", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10GetIntensities(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10GetIntensities(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self) {
  TMM::Intensities __pyx_v_resCpp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_res = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  TMM::Intensities __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("GetIntensities", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1130
 *             Helper class to hold intensity data.
 *         """
 *         cdef IntensitiesCpp resCpp = self._thisptr.GetIntensities()             # <<<<<<<<<<<<<<
 *         res = _Intensities()
 *         res._Init(&resCpp)
*/
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetIntensities();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1130, __pyx_L1_error)
  }
  __pyx_v_resCpp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1131
 *         """
 *         cdef IntensitiesCpp resCpp = self._thisptr.GetIntensities()
 *         res = _Intensities()             # <<<<<<<<<<<<<<
 *         res._Init(&resCpp)
 *         return res
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1131, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_res = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1132
 *         cdef IntensitiesCpp resCpp = self._thisptr.GetIntensities()
 *         res = _Intensities()
 *         res._Init(&resCpp)             # <<<<<<<<<<<<<<
 *         return res
 * 
*/
  __pyx_t_2 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_v_res->__pyx_vtab)->_Init(__pyx_v_res, (&__pyx_v_resCpp)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1133
 *         res = _Intensities()
 *         res._Init(&resCpp)
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     def Sweep(self, str paramStr, np.ndarray[double, ndim = 1] values, int layerNr = 0, double layerZ = 0.0, bool outPwr = True, bool outAbs = False, outEnh = False):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_res);
  __pyx_r = ((PyObject *)__pyx_v_res);
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1119
 *         self._thisptr.Solve()
 * 
 *     def GetIntensities(self):             # <<<<<<<<<<<<<<
 *         """GetIntensities()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.GetIntensities", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1135
 *         return res
 * 
 *     def Sweep(self, str paramStr, np.ndarray[double, ndim = 1] values, int layerNr = 0, double layerZ = 0.0, bool outPwr = True, bool outAbs = False, outEnh = False):             # <<<<<<<<<<<<<<
 *         """Sweep(paramStr, values, layerNr = 0, layerZ = 0, outPwr = True, outAbs = False, outEnh = False)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_13Sweep(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_12Sweep, "Sweep(paramStr, values, layerNr = 0, layerZ = 0, outPwr = True, outAbs = False, outEnh = False)\n\n        Solves the structure for a series of ``values`` of param ``paramStr``. Using\n        this function is more convenient and faster than just changing params\n        and solving the structure.\n\n        Parameters\n        ----------\n        paramStr : str\n            'wl':\n                Wavelength in meters\n            'beta':\n                normalized tangential wavevector\n            'I0':\n                intensity of the wave\n            'd_i':\n                thickness of layer i (0..N-1)\n        values : ndarray of floats\n            Corresponding values of param ``paramStr``.\n        layerNr : int\n            Specifies layer, where electrical field enhancement is calculated.\n        layerZ : double\n            Specifies z-coordinate of enhancement calculation inside ``layerNr``.\n        outPwr : bool\n            Turns calculation of intensities on/off.\n        outAbs : bool\n            Turns calculation of absorption in the entire structure on/off.\n        outEnh : bool\n            Turns calculation of enhancement in layer ``layerNr`` at distance\n            ``layerZ`` on/off.\n\n        Returns\n        -------\n        :class:`_SweepResultNonlinearTMM`\n            Helper class to store the result.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_13Sweep = {"Sweep", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_13Sweep, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_12Sweep};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_13Sweep(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_paramStr = 0;
  PyArrayObject *__pyx_v_values = 0;
  int __pyx_v_layerNr;
  double __pyx_v_layerZ;
  bool __pyx_v_outPwr;
  bool __pyx_v_outAbs;
  PyObject *__pyx_v_outEnh = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[7] = {0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Sweep (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_paramStr,&__pyx_mstate_global->__pyx_n_u_values,&__pyx_mstate_global->__pyx_n_u_layerNr,&__pyx_mstate_global->__pyx_n_u_layerZ,&__pyx_mstate_global->__pyx_n_u_outPwr,&__pyx_mstate_global->__pyx_n_u_outAbs,&__pyx_mstate_global->__pyx_n_u_outEnh,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1135, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 1135, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1135, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1135, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1135, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1135, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1135, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1135, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "Sweep", 0) < (0)) __PYX_ERR(0, 1135, __pyx_L3_error)
      if (!values[6]) values[6] = __Pyx_NewRef(((PyObject *)Py_False));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("Sweep", 0, 2, 7, i); __PYX_ERR(0, 1135, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 1135, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1135, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1135, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1135, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1135, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1135, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1135, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[6]) values[6] = __Pyx_NewRef(((PyObject *)Py_False));
    }
    __pyx_v_paramStr = ((PyObject*)values[0]);
    __pyx_v_values = ((PyArrayObject *)values[1]);
    if (values[2]) {
      __pyx_v_layerNr = __Pyx_PyLong_As_int(values[2]); if (unlikely((__pyx_v_layerNr == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1135, __pyx_L3_error)
    } else {
      __pyx_v_layerNr = ((int)0);
    }
    if (values[3]) {
      __pyx_v_layerZ = __Pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_layerZ == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1135, __pyx_L3_error)
    } else {
      __pyx_v_layerZ = ((double)0.0);
    }
    if (values[4]) {
      __pyx_v_outPwr = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_outPwr == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1135, __pyx_L3_error)
    } else {
      __pyx_v_outPwr = ((bool)1);
    }
    if (values[5]) {
      __pyx_v_outAbs = __Pyx_PyObject_IsTrue(values[5]); if (unlikely((__pyx_v_outAbs == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1135, __pyx_L3_error)
    } else {
      __pyx_v_outAbs = ((bool)0);
    }
    __pyx_v_outEnh = values[6];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("Sweep", 0, 2, 7, __pyx_nargs); __PYX_ERR(0, 1135, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.Sweep", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_paramStr), (&PyUnicode_Type), 1, "paramStr", 1))) __PYX_ERR(0, 1135, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_values), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "values", 0))) __PYX_ERR(0, 1135, __pyx_L1_error)
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_12Sweep(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), __pyx_v_paramStr, __pyx_v_values, __pyx_v_layerNr, __pyx_v_layerZ, __pyx_v_outPwr, __pyx_v_outAbs, __pyx_v_outEnh);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_12Sweep(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_paramStr, PyArrayObject *__pyx_v_values, int __pyx_v_layerNr, double __pyx_v_layerZ, bool __pyx_v_outPwr, bool __pyx_v_outAbs, PyObject *__pyx_v_outEnh) {
  TMM::SweepResultNonlinearTMM *__pyx_v_resCpp;
  int __pyx_v_outmask;
  enum TMM::TMMParam __pyx_v_param;
  int __pyx_v_paramLayer;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_res = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_values;
  __Pyx_Buffer __pyx_pybuffer_values;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  enum TMM::TMMParam __pyx_t_5;
  eigency::Map<Eigen::ArrayXd>  __pyx_t_6;
  std::unique_ptr<TMM::SweepResultNonlinearTMM>  __pyx_t_7;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Sweep", 0);
  __pyx_pybuffer_values.pybuffer.buf = NULL;
  __pyx_pybuffer_values.refcount = 0;
  __pyx_pybuffernd_values.data = NULL;
  __pyx_pybuffernd_values.rcbuffer = &__pyx_pybuffer_values;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_values.rcbuffer->pybuffer, (PyObject*)__pyx_v_values, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1135, __pyx_L1_error)
  }
  __pyx_pybuffernd_values.diminfo[0].strides = __pyx_pybuffernd_values.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_values.diminfo[0].shape = __pyx_pybuffernd_values.rcbuffer->pybuffer.shape[0];

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1174
 *         """
 *         cdef SweepResultNonlinearTMMCpp *resCpp;
 *         cdef int outmask = 0             # <<<<<<<<<<<<<<
 * 
 *         if outPwr:
*/
  __pyx_v_outmask = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1176
 *         cdef int outmask = 0
 * 
 *         if outPwr:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_PWRFLOWS
 *         if outAbs:
*/
  __pyx_t_1 = (__pyx_v_outPwr != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1177
 * 
 *         if outPwr:
 *             outmask |= SWEEP_PWRFLOWS             # <<<<<<<<<<<<<<
 *         if outAbs:
 *             outmask |= SWEEP_ABS
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_PWRFLOWS);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1176
 *         cdef int outmask = 0
 * 
 *         if outPwr:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_PWRFLOWS
 *         if outAbs:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1178
 *         if outPwr:
 *             outmask |= SWEEP_PWRFLOWS
 *         if outAbs:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_ABS
 *         if outEnh:
*/
  __pyx_t_1 = (__pyx_v_outAbs != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1179
 *             outmask |= SWEEP_PWRFLOWS
 *         if outAbs:
 *             outmask |= SWEEP_ABS             # <<<<<<<<<<<<<<
 *         if outEnh:
 *             outmask |= SWEEP_ENH
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_ABS);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1178
 *         if outPwr:
 *             outmask |= SWEEP_PWRFLOWS
 *         if outAbs:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_ABS
 *         if outEnh:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1180
 *         if outAbs:
 *             outmask |= SWEEP_ABS
 *         if outEnh:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_ENH
 * 
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_outEnh); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1180, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1181
 *             outmask |= SWEEP_ABS
 *         if outEnh:
 *             outmask |= SWEEP_ENH             # <<<<<<<<<<<<<<
 * 
 *         cdef TMMParamCpp param;
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_ENH);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1180
 *         if outAbs:
 *             outmask |= SWEEP_ABS
 *         if outEnh:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_ENH
 * 
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1184
 * 
 *         cdef TMMParamCpp param;
 *         cdef int paramLayer = -1;             # <<<<<<<<<<<<<<
 *         if (paramStr.startswith("d_")):
 *             param = PARAM_LAYER_D
*/
  __pyx_v_paramLayer = -1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1185
 *         cdef TMMParamCpp param;
 *         cdef int paramLayer = -1;
 *         if (paramStr.startswith("d_")):             # <<<<<<<<<<<<<<
 *             param = PARAM_LAYER_D
 *             paramLayer = int(paramStr[2:])
*/
  if (unlikely(__pyx_v_paramStr == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "startswith");
    __PYX_ERR(0, 1185, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyUnicode_Tailmatch(__pyx_v_paramStr, __pyx_mstate_global->__pyx_n_u_d_2, 0, PY_SSIZE_T_MAX, -1); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1185, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1186
 *         cdef int paramLayer = -1;
 *         if (paramStr.startswith("d_")):
 *             param = PARAM_LAYER_D             # <<<<<<<<<<<<<<
 *             paramLayer = int(paramStr[2:])
 *         else:
*/
    __pyx_v_param = TMM::TMMParam::PARAM_LAYER_D;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1187
 *         if (paramStr.startswith("d_")):
 *             param = PARAM_LAYER_D
 *             paramLayer = int(paramStr[2:])             # <<<<<<<<<<<<<<
 *         else:
 *             param = TmmParamFromStr(paramStr)
*/
    if (unlikely(__pyx_v_paramStr == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 1187, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyUnicode_Substring(__pyx_v_paramStr, 2, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1187, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_paramLayer = __pyx_t_4;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1185
 *         cdef TMMParamCpp param;
 *         cdef int paramLayer = -1;
 *         if (paramStr.startswith("d_")):             # <<<<<<<<<<<<<<
 *             param = PARAM_LAYER_D
 *             paramLayer = int(paramStr[2:])
*/
    goto __pyx_L6;
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1189
 *             paramLayer = int(paramStr[2:])
 *         else:
 *             param = TmmParamFromStr(paramStr)             # <<<<<<<<<<<<<<
 * 
 *         resCpp = self._thisptr.Sweep(param, Map[ArrayXd](values), outmask, paramLayer, layerNr, layerZ).release()
*/
  /*else*/ {
    __pyx_t_5 = __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_TmmParamFromStr(__pyx_v_paramStr); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1189, __pyx_L1_error)
    __pyx_v_param = __pyx_t_5;
  }
  __pyx_L6:;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1191
 *             param = TmmParamFromStr(paramStr)
 * 
 *         resCpp = self._thisptr.Sweep(param, Map[ArrayXd](values), outmask, paramLayer, layerNr, layerZ).release()             # <<<<<<<<<<<<<<
 *         res = _SweepResultNonlinearTMM()
 *         res._Init(resCpp);
*/
  try {
    __pyx_t_6 = eigency::Map<Eigen::ArrayXd> (((PyArrayObject *)__pyx_v_values));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1191, __pyx_L1_error)
  }
  try {
    __pyx_t_7 = __pyx_v_self->_thisptr->Sweep(__pyx_v_param, __pyx_t_6, __pyx_v_outmask, __pyx_v_paramLayer, __pyx_v_layerNr, __pyx_v_layerZ);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1191, __pyx_L1_error)
  }
  __pyx_v_resCpp = __pyx_t_7.release();

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1192
 * 
 *         resCpp = self._thisptr.Sweep(param, Map[ArrayXd](values), outmask, paramLayer, layerNr, layerZ).release()
 *         res = _SweepResultNonlinearTMM()             # <<<<<<<<<<<<<<
 *         res._Init(resCpp);
 *         return res
*/
  __pyx_t_2 = NULL;
  __pyx_t_8 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1192, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
  }
  __pyx_v_res = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1193
 *         resCpp = self._thisptr.Sweep(param, Map[ArrayXd](values), outmask, paramLayer, layerNr, layerZ).release()
 *         res = _SweepResultNonlinearTMM()
 *         res._Init(resCpp);             # <<<<<<<<<<<<<<
 *         return res
 * 
*/
  __pyx_t_3 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_res->__pyx_vtab)->_Init(__pyx_v_res, __pyx_v_resCpp, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1194
 *         res = _SweepResultNonlinearTMM()
 *         res._Init(resCpp);
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     def GetFields(self, np.ndarray[double, ndim = 1] zs, str dir = "total"):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_res);
  __pyx_r = ((PyObject *)__pyx_v_res);
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1135
 *         return res
 * 
 *     def Sweep(self, str paramStr, np.ndarray[double, ndim = 1] values, int layerNr = 0, double layerZ = 0.0, bool outPwr = True, bool outAbs = False, outEnh = False):             # <<<<<<<<<<<<<<
 *         """Sweep(paramStr, values, layerNr = 0, layerZ = 0, outPwr = True, outAbs = False, outEnh = False)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_values.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.Sweep", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_values.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1196
 *         return res
 * 
 *     def GetFields(self, np.ndarray[double, ndim = 1] zs, str dir = "total"):             # <<<<<<<<<<<<<<
 *         """GetFields(zs, dir = "total")
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_15GetFields(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_14GetFields, "GetFields(zs, dir = \"total\")\n\n        Calculates electrical and magnetic fields along z-axis. The structure\n        must be solved first.\n\n        Parameters\n        ----------\n        zs : ndarray of floats\n            Points on z-axis where to calculate the fields. The beginning of the\n            first layer is at z = 0.\n        dir : {'total', 'forward', 'backward'}\n            Specifies the components of the output fields.\n\n        Returns\n        -------\n        :class:`_FieldsZ`\n            Helper class to store electric and magnetic fields.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_15GetFields = {"GetFields", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_15GetFields, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_14GetFields};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_15GetFields(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyArrayObject *__pyx_v_zs = 0;
  PyObject *__pyx_v_dir = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("GetFields (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_zs,&__pyx_mstate_global->__pyx_n_u_dir,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1196, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1196, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1196, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "GetFields", 0) < (0)) __PYX_ERR(0, 1196, __pyx_L3_error)
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject*)__pyx_mstate_global->__pyx_n_u_total));
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("GetFields", 0, 1, 2, i); __PYX_ERR(0, 1196, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1196, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1196, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[1]) values[1] = __Pyx_NewRef(((PyObject*)__pyx_mstate_global->__pyx_n_u_total));
    }
    __pyx_v_zs = ((PyArrayObject *)values[0]);
    __pyx_v_dir = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("GetFields", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1196, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.GetFields", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_zs), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "zs", 0))) __PYX_ERR(0, 1196, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dir), (&PyUnicode_Type), 1, "dir", 1))) __PYX_ERR(0, 1196, __pyx_L1_error)
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_14GetFields(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), __pyx_v_zs, __pyx_v_dir);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_14GetFields(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyArrayObject *__pyx_v_zs, PyObject *__pyx_v_dir) {
  TMM::FieldsZ *__pyx_v_resCpp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *__pyx_v_res = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_zs;
  __Pyx_Buffer __pyx_pybuffer_zs;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  eigency::Map<Eigen::ArrayXd>  __pyx_t_1;
  enum TMM::WaveDirection __pyx_t_2;
  std::unique_ptr<TMM::FieldsZ>  __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("GetFields", 0);
  __pyx_pybuffer_zs.pybuffer.buf = NULL;
  __pyx_pybuffer_zs.refcount = 0;
  __pyx_pybuffernd_zs.data = NULL;
  __pyx_pybuffernd_zs.rcbuffer = &__pyx_pybuffer_zs;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_zs.rcbuffer->pybuffer, (PyObject*)__pyx_v_zs, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1196, __pyx_L1_error)
  }
  __pyx_pybuffernd_zs.diminfo[0].strides = __pyx_pybuffernd_zs.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_zs.diminfo[0].shape = __pyx_pybuffernd_zs.rcbuffer->pybuffer.shape[0];

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1217
 *         """
 *         cdef FieldsZCpp *resCpp;
 *         resCpp = self._thisptr.GetFields(Map[ArrayXd](zs), WaveDirectionFromStr(dir)).release()             # <<<<<<<<<<<<<<
 *         res = _FieldsZ()
 *         res._Init(resCpp)
*/
  try {
    __pyx_t_1 = eigency::Map<Eigen::ArrayXd> (((PyArrayObject *)__pyx_v_zs));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1217, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_WaveDirectionFromStr(__pyx_v_dir); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1217, __pyx_L1_error)
  try {
    __pyx_t_3 = __pyx_v_self->_thisptr->GetFields(__pyx_t_1, __pyx_t_2);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1217, __pyx_L1_error)
  }
  __pyx_v_resCpp = __pyx_t_3.release();

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1218
 *         cdef FieldsZCpp *resCpp;
 *         resCpp = self._thisptr.GetFields(Map[ArrayXd](zs), WaveDirectionFromStr(dir)).release()
 *         res = _FieldsZ()             # <<<<<<<<<<<<<<
 *         res._Init(resCpp)
 *         return res
*/
  __pyx_t_5 = NULL;
  __pyx_t_6 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1218, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_4);
  }
  __pyx_v_res = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1219
 *         resCpp = self._thisptr.GetFields(Map[ArrayXd](zs), WaveDirectionFromStr(dir)).release()
 *         res = _FieldsZ()
 *         res._Init(resCpp)             # <<<<<<<<<<<<<<
 *         return res
 * 
*/
  __pyx_t_4 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *)__pyx_v_res->__pyx_vtab)->_Init(__pyx_v_res, __pyx_v_resCpp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1220
 *         res = _FieldsZ()
 *         res._Init(resCpp)
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     def GetFields2D(self, np.ndarray[double, ndim = 1] zs, np.ndarray[double, ndim = 1] xs, str dir = "total"):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_res);
  __pyx_r = ((PyObject *)__pyx_v_res);
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1196
 *         return res
 * 
 *     def GetFields(self, np.ndarray[double, ndim = 1] zs, str dir = "total"):             # <<<<<<<<<<<<<<
 *         """GetFields(zs, dir = "total")
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_zs.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.GetFields", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_zs.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1222
 *         return res
 * 
 *     def GetFields2D(self, np.ndarray[double, ndim = 1] zs, np.ndarray[double, ndim = 1] xs, str dir = "total"):             # <<<<<<<<<<<<<<
 *         """GetFields2D(zs, xs, dir = "total")
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_17GetFields2D(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_16GetFields2D, "GetFields2D(zs, xs, dir = \"total\")\n\n        Calculates electrical and magnetic fields in xz-plane. Made as fast as\n        possible. The structure must be solved first.\n\n        Parameters\n        ----------\n        zs : ndarray of floats\n            Points on z-axis where to calculate the fields. The beginning of the\n            first layer is at z = 0.\n        xs : ndarray of floats\n            Points on x-axis where to calculate the fields.\n        dir : {'total', 'forward', 'backward'}\n            Specifies the components of the output fields.\n\n        Returns\n        -------\n        :class:`_FieldsZX`\n            Helper class to store electric and magnetic fields in regular grid.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_17GetFields2D = {"GetFields2D", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_17GetFields2D, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_16GetFields2D};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_17GetFields2D(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyArrayObject *__pyx_v_zs = 0;
  PyArrayObject *__pyx_v_xs = 0;
  PyObject *__pyx_v_dir = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("GetFields2D (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_zs,&__pyx_mstate_global->__pyx_n_u_xs,&__pyx_mstate_global->__pyx_n_u_dir,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1222, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "GetFields2D", 0) < (0)) __PYX_ERR(0, 1222, __pyx_L3_error)
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject*)__pyx_mstate_global->__pyx_n_u_total));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("GetFields2D", 0, 2, 3, i); __PYX_ERR(0, 1222, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1222, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1222, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1222, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject*)__pyx_mstate_global->__pyx_n_u_total));
    }
    __pyx_v_zs = ((PyArrayObject *)values[0]);
    __pyx_v_xs = ((PyArrayObject *)values[1]);
    __pyx_v_dir = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("GetFields2D", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 1222, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.GetFields2D", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_zs), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "zs", 0))) __PYX_ERR(0, 1222, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_xs), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "xs", 0))) __PYX_ERR(0, 1222, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dir), (&PyUnicode_Type), 1, "dir", 1))) __PYX_ERR(0, 1222, __pyx_L1_error)
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_16GetFields2D(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), __pyx_v_zs, __pyx_v_xs, __pyx_v_dir);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_16GetFields2D(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyArrayObject *__pyx_v_zs, PyArrayObject *__pyx_v_xs, PyObject *__pyx_v_dir) {
  TMM::FieldsZX *__pyx_v_resCpp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_res = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_xs;
  __Pyx_Buffer __pyx_pybuffer_xs;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_zs;
  __Pyx_Buffer __pyx_pybuffer_zs;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  eigency::Map<Eigen::ArrayXd>  __pyx_t_1;
  eigency::Map<Eigen::ArrayXd>  __pyx_t_2;
  enum TMM::WaveDirection __pyx_t_3;
  std::unique_ptr<TMM::FieldsZX>  __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("GetFields2D", 0);
  __pyx_pybuffer_zs.pybuffer.buf = NULL;
  __pyx_pybuffer_zs.refcount = 0;
  __pyx_pybuffernd_zs.data = NULL;
  __pyx_pybuffernd_zs.rcbuffer = &__pyx_pybuffer_zs;
  __pyx_pybuffer_xs.pybuffer.buf = NULL;
  __pyx_pybuffer_xs.refcount = 0;
  __pyx_pybuffernd_xs.data = NULL;
  __pyx_pybuffernd_xs.rcbuffer = &__pyx_pybuffer_xs;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_zs.rcbuffer->pybuffer, (PyObject*)__pyx_v_zs, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1222, __pyx_L1_error)
  }
  __pyx_pybuffernd_zs.diminfo[0].strides = __pyx_pybuffernd_zs.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_zs.diminfo[0].shape = __pyx_pybuffernd_zs.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_xs.rcbuffer->pybuffer, (PyObject*)__pyx_v_xs, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1222, __pyx_L1_error)
  }
  __pyx_pybuffernd_xs.diminfo[0].strides = __pyx_pybuffernd_xs.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_xs.diminfo[0].shape = __pyx_pybuffernd_xs.rcbuffer->pybuffer.shape[0];

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1245
 *         """
 *         cdef FieldsZXCpp *resCpp;
 *         resCpp = self._thisptr.GetFields2D(Map[ArrayXd](zs), Map[ArrayXd](xs), WaveDirectionFromStr(dir)).release()             # <<<<<<<<<<<<<<
 *         res = _FieldsZX()
 *         res._Init(resCpp)
*/
  try {
    __pyx_t_1 = eigency::Map<Eigen::ArrayXd> (((PyArrayObject *)__pyx_v_zs));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1245, __pyx_L1_error)
  }
  try {
    __pyx_t_2 = eigency::Map<Eigen::ArrayXd> (((PyArrayObject *)__pyx_v_xs));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1245, __pyx_L1_error)
  }
  __pyx_t_3 = __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_WaveDirectionFromStr(__pyx_v_dir); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1245, __pyx_L1_error)
  try {
    __pyx_t_4 = __pyx_v_self->_thisptr->GetFields2D(__pyx_t_1, __pyx_t_2, __pyx_t_3);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1245, __pyx_L1_error)
  }
  __pyx_v_resCpp = __pyx_t_4.release();

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1246
 *         cdef FieldsZXCpp *resCpp;
 *         resCpp = self._thisptr.GetFields2D(Map[ArrayXd](zs), Map[ArrayXd](xs), WaveDirectionFromStr(dir)).release()
 *         res = _FieldsZX()             # <<<<<<<<<<<<<<
 *         res._Init(resCpp)
 *         return res
*/
  __pyx_t_6 = NULL;
  __pyx_t_7 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
    __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1246, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_5);
  }
  __pyx_v_res = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1247
 *         resCpp = self._thisptr.GetFields2D(Map[ArrayXd](zs), Map[ArrayXd](xs), WaveDirectionFromStr(dir)).release()
 *         res = _FieldsZX()
 *         res._Init(resCpp)             # <<<<<<<<<<<<<<
 *         return res
 * 
*/
  __pyx_t_5 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_res->__pyx_vtab)->_Init(__pyx_v_res, __pyx_v_resCpp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1248
 *         res = _FieldsZX()
 *         res._Init(resCpp)
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     def GetAbsorbedIntensity(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_res);
  __pyx_r = ((PyObject *)__pyx_v_res);
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1222
 *         return res
 * 
 *     def GetFields2D(self, np.ndarray[double, ndim = 1] zs, np.ndarray[double, ndim = 1] xs, str dir = "total"):             # <<<<<<<<<<<<<<
 *         """GetFields2D(zs, xs, dir = "total")
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_xs.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_zs.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.GetFields2D", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_xs.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_zs.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1250
 *         return res
 * 
 *     def GetAbsorbedIntensity(self):             # <<<<<<<<<<<<<<
 *         """GetAbsorbedIntensity()
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_19GetAbsorbedIntensity(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_18GetAbsorbedIntensity, "GetAbsorbedIntensity()\n\n        Calculates intensity of absorption. The structure must be solved first.\n\n        Returns\n        -------\n        float\n            Absorption intensity\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_19GetAbsorbedIntensity = {"GetAbsorbedIntensity", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_19GetAbsorbedIntensity, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_18GetAbsorbedIntensity};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_19GetAbsorbedIntensity(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("GetAbsorbedIntensity (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("GetAbsorbedIntensity", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("GetAbsorbedIntensity", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_18GetAbsorbedIntensity(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_18GetAbsorbedIntensity(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("GetAbsorbedIntensity", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1261
 * 
 *         """
 *         return self._thisptr.GetAbsorbedIntensity();             # <<<<<<<<<<<<<<
 * 
 *     def GetEnhancement(self, int layerNr, double z = 0.0):
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetAbsorbedIntensity();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1261, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1250
 *         return res
 * 
 *     def GetAbsorbedIntensity(self):             # <<<<<<<<<<<<<<
 *         """GetAbsorbedIntensity()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.GetAbsorbedIntensity", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1263
 *         return self._thisptr.GetAbsorbedIntensity();
 * 
 *     def GetEnhancement(self, int layerNr, double z = 0.0):             # <<<<<<<<<<<<<<
 *         """GetEnhancement(layerNr, z = 0.0)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_21GetEnhancement(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_20GetEnhancement, "GetEnhancement(layerNr, z = 0.0)\n\n        Calculates the enhancement of electrical field norm at specified layer at\n        fixed z-coordinate.\n\n        Parameters\n        ----------\n        layerNr : int\n            Number of layer where to calculate the enhancement.\n        z : float\n            Z-distance from the beginning of the layer.\n\n        Returns\n        -------\n        float\n            The enhancement of the electrical field norm in comparison to the\n            input wave in vacuum.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_21GetEnhancement = {"GetEnhancement", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_21GetEnhancement, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_20GetEnhancement};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_21GetEnhancement(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_layerNr;
  double __pyx_v_z;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("GetEnhancement (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_layerNr,&__pyx_mstate_global->__pyx_n_u_z,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1263, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1263, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1263, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "GetEnhancement", 0) < (0)) __PYX_ERR(0, 1263, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("GetEnhancement", 0, 1, 2, i); __PYX_ERR(0, 1263, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1263, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1263, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_layerNr = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_layerNr == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1263, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_z = __Pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_z == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1263, __pyx_L3_error)
    } else {
      __pyx_v_z = ((double)0.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("GetEnhancement", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1263, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.GetEnhancement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_20GetEnhancement(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), __pyx_v_layerNr, __pyx_v_z);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_20GetEnhancement(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, int __pyx_v_layerNr, double __pyx_v_z) {
  double __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("GetEnhancement", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1284
 *         """
 *         cdef double res
 *         res = self._thisptr.GetEnhancement(layerNr, z)             # <<<<<<<<<<<<<<
 *         return res
 * 
*/
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetEnhancement(__pyx_v_layerNr, __pyx_v_z);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1284, __pyx_L1_error)
  }
  __pyx_v_res = __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1285
 *         cdef double res
 *         res = self._thisptr.GetEnhancement(layerNr, z)
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     # Waves
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_res); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1263
 *         return self._thisptr.GetAbsorbedIntensity();
 * 
 *     def GetEnhancement(self, int layerNr, double z = 0.0):             # <<<<<<<<<<<<<<
 *         """GetEnhancement(layerNr, z = 0.0)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.GetEnhancement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1289
 *     # Waves
 * 
 *     def WaveGetPowerFlows(self, int layerNr, double x0 = float("nan"), double x1 = float("nan"), double z = 0.0):             # <<<<<<<<<<<<<<
 *         """WaveGetPowerFlows(layerNr, x0 = float("nan"), x1 = float("nan"), z = 0.0)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_23WaveGetPowerFlows(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_22WaveGetPowerFlows, "WaveGetPowerFlows(layerNr, x0 = float(\"nan\"), x1 = float(\"nan\"), z = 0.0)\n\n        Analogous to ``GetIntensities``, but calculates the powers of the\n        beams instead of the intensities of the plane waves.\n\n        Parameters\n        ----------\n        layerNr : int\n            Specifies layer number where to calculate the power of the beam.\n        x0 : float\n            Specifies the starting point of the integration of the power in the\n            x-direction. By default this parameter is selected by the ``_Wave``\n            class.\n        x1 : float\n            Specifies the end point of the integration of the power in the\n            x-direction. By default this parameter is selected by the ``_Wave``\n            class.\n        z : float\n            Specifies the z-position of the line through which the integration\n            of the power of the beam is done.\n\n        Returns\n        -------\n        tuple of floats (PB, PF)\n            PB is the power propagated into negative infinity and PF denotes\n            the power propagating in the positive direction of z-axis.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_23WaveGetPowerFlows = {"WaveGetPowerFlows", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_23WaveGetPowerFlows, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_22WaveGetPowerFlows};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_23WaveGetPowerFlows(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_layerNr;
  double __pyx_v_x0;
  double __pyx_v_x1;
  double __pyx_v_z;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("WaveGetPowerFlows (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_layerNr,&__pyx_mstate_global->__pyx_n_u_x0,&__pyx_mstate_global->__pyx_n_u_x1,&__pyx_mstate_global->__pyx_n_u_z,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1289, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1289, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1289, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1289, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1289, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "WaveGetPowerFlows", 0) < (0)) __PYX_ERR(0, 1289, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("WaveGetPowerFlows", 0, 1, 4, i); __PYX_ERR(0, 1289, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1289, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1289, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1289, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1289, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_layerNr = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_layerNr == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_x0 = __Pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_x0 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L3_error)
    } else {
      __pyx_v_x0 = __pyx_mstate_global->__pyx_k_;
    }
    if (values[2]) {
      __pyx_v_x1 = __Pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_x1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L3_error)
    } else {
      __pyx_v_x1 = __pyx_mstate_global->__pyx_k__2;
    }
    if (values[3]) {
      __pyx_v_z = __Pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_z == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L3_error)
    } else {
      __pyx_v_z = ((double)0.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("WaveGetPowerFlows", 0, 1, 4, __pyx_nargs); __PYX_ERR(0, 1289, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.WaveGetPowerFlows", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_22WaveGetPowerFlows(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), __pyx_v_layerNr, __pyx_v_x0, __pyx_v_x1, __pyx_v_z);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_22WaveGetPowerFlows(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, int __pyx_v_layerNr, double __pyx_v_x0, double __pyx_v_x1, double __pyx_v_z) {
  std::pair<double,double>  __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  std::pair<double,double>  __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("WaveGetPowerFlows", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1320
 *         # NonlinearLayer has its own specific method
 *         cdef pair[double, double] res;
 *         res = self._thisptr.WaveGetPowerFlows(layerNr, x0, x1, z)             # <<<<<<<<<<<<<<
 *         return (res.first, res.second)
 * 
*/
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->WaveGetPowerFlows(__pyx_v_layerNr, __pyx_v_x0, __pyx_v_x1, __pyx_v_z);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1320, __pyx_L1_error)
  }
  __pyx_v_res = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1321
 *         cdef pair[double, double] res;
 *         res = self._thisptr.WaveGetPowerFlows(layerNr, x0, x1, z)
 *         return (res.first, res.second)             # <<<<<<<<<<<<<<
 * 
 *     def WaveGetEnhancement(self, int layerNr, double z = 0.0):
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_res.first); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_res.second); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2) != (0)) __PYX_ERR(0, 1321, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3) != (0)) __PYX_ERR(0, 1321, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1289
 *     # Waves
 * 
 *     def WaveGetPowerFlows(self, int layerNr, double x0 = float("nan"), double x1 = float("nan"), double z = 0.0):             # <<<<<<<<<<<<<<
 *         """WaveGetPowerFlows(layerNr, x0 = float("nan"), x1 = float("nan"), z = 0.0)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.WaveGetPowerFlows", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1323
 *         return (res.first, res.second)
 * 
 *     def WaveGetEnhancement(self, int layerNr, double z = 0.0):             # <<<<<<<<<<<<<<
 *         """WaveGetEnhancement(layerNr, z = 0.0)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_25WaveGetEnhancement(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_24WaveGetEnhancement, "WaveGetEnhancement(layerNr, z = 0.0)\n\n        Calculates enhancement of electrical field norm in comparison to the input\n        beam in vacuum. Analogous to ``GetEnhancement``.\n\n        Parameters\n        ----------\n        layerNr : int\n            Number of layer where to calculate the enhancement.\n        z : float\n            Z-distance from the beginning of the layer.\n\n        Returns\n        -------\n        float\n            The enhancement of the electrical field norm in comparison to the\n            input wave in vacuum.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_25WaveGetEnhancement = {"WaveGetEnhancement", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_25WaveGetEnhancement, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_24WaveGetEnhancement};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_25WaveGetEnhancement(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_layerNr;
  double __pyx_v_z;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("WaveGetEnhancement (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_layerNr,&__pyx_mstate_global->__pyx_n_u_z,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1323, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1323, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1323, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "WaveGetEnhancement", 0) < (0)) __PYX_ERR(0, 1323, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("WaveGetEnhancement", 0, 1, 2, i); __PYX_ERR(0, 1323, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1323, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1323, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_layerNr = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_layerNr == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1323, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_z = __Pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_z == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1323, __pyx_L3_error)
    } else {
      __pyx_v_z = ((double)0.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("WaveGetEnhancement", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1323, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.WaveGetEnhancement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_24WaveGetEnhancement(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), __pyx_v_layerNr, __pyx_v_z);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_24WaveGetEnhancement(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, int __pyx_v_layerNr, double __pyx_v_z) {
  double __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("WaveGetEnhancement", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1344
 *         """
 *         cdef double res
 *         res = self._thisptr.WaveGetEnhancement(layerNr, z)             # <<<<<<<<<<<<<<
 *         return res
 * 
*/
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->WaveGetEnhancement(__pyx_v_layerNr, __pyx_v_z);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1344, __pyx_L1_error)
  }
  __pyx_v_res = __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1345
 *         cdef double res
 *         res = self._thisptr.WaveGetEnhancement(layerNr, z)
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] values, \
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_res); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1323
 *         return (res.first, res.second)
 * 
 *     def WaveGetEnhancement(self, int layerNr, double z = 0.0):             # <<<<<<<<<<<<<<
 *         """WaveGetEnhancement(layerNr, z = 0.0)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.WaveGetEnhancement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1347
 *         return res
 * 
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] values, \             # <<<<<<<<<<<<<<
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \
 *             bool outR = False, bool outT = False, bool outEnh = False):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_27WaveSweep(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_26WaveSweep, "WaveSweep(paramStr, values, layerNr = 0, layerZ = 0.0, outPwr = True, outR = False, outT = False, outEnh = False)\n\n        Solves the structure for waves for a series of ``values`` of param\n        ``paramStr``. Using this function is more convenient and faster than just\n        changing params and solving the structure. Analogous to ``Sweep``.\n\n        Parameters\n        ----------\n        paramStr : str\n            'wl':\n                Wavelength in meters\n            'beta':\n                normalized tangential wavevector\n            'I0':\n                intensity of the wave\n            'd_i':\n                thickness of layer i (0..N-1)\n            'w0':\n                waist size of the input beam\n        values : ndarray of floats\n            Corresponding values of param ``paramStr``.\n        layerNr : int\n            Specifies layer, where electrical field enhancement is calculated.\n        layerZ : double\n            Specifies z-coordinate of enhancement calculation inside ``layerNr``.\n        outPwr : bool\n            Turns calculation of all powers on/off.\n        outR : bool\n            Turns calculation of reflected power on/off.\n        outT : bool\n            Turns calculation of transmitted power on/off.\n        outEnh : bool\n            Turns calculation of enhancement in layer ``layerNr`` at distance\n            ``layerZ`` on/off.\n\n        Returns\n        -------\n        :class:`_SweepResultNonlinearTMM`\n            Helper class to store the result.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_27WaveSweep = {"WaveSweep", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_27WaveSweep, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_26WaveSweep};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_27WaveSweep(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_paramStr = 0;
  PyArrayObject *__pyx_v_values = 0;
  int __pyx_v_layerNr;
  double __pyx_v_layerZ;
  bool __pyx_v_outPwr;
  bool __pyx_v_outR;
  bool __pyx_v_outT;
  bool __pyx_v_outEnh;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[8] = {0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("WaveSweep (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_paramStr,&__pyx_mstate_global->__pyx_n_u_values,&__pyx_mstate_global->__pyx_n_u_layerNr,&__pyx_mstate_global->__pyx_n_u_layerZ,&__pyx_mstate_global->__pyx_n_u_outPwr,&__pyx_mstate_global->__pyx_n_u_outR,&__pyx_mstate_global->__pyx_n_u_outT,&__pyx_mstate_global->__pyx_n_u_outEnh,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1347, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  8:
        values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 1347, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 1347, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1347, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1347, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1347, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1347, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1347, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1347, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "WaveSweep", 0) < (0)) __PYX_ERR(0, 1347, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("WaveSweep", 0, 2, 8, i); __PYX_ERR(0, 1347, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  8:
        values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 1347, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 1347, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1347, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1347, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1347, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1347, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1347, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1347, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_paramStr = ((PyObject*)values[0]);
    __pyx_v_values = ((PyArrayObject *)values[1]);
    if (values[2]) {
      __pyx_v_layerNr = __Pyx_PyLong_As_int(values[2]); if (unlikely((__pyx_v_layerNr == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1348, __pyx_L3_error)
    } else {
      __pyx_v_layerNr = ((int)0);
    }
    if (values[3]) {
      __pyx_v_layerZ = __Pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_layerZ == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1348, __pyx_L3_error)
    } else {
      __pyx_v_layerZ = ((double)0.0);
    }
    if (values[4]) {
      __pyx_v_outPwr = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_outPwr == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1348, __pyx_L3_error)
    } else {

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1348
 * 
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] values, \
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \             # <<<<<<<<<<<<<<
 *             bool outR = False, bool outT = False, bool outEnh = False):
 *         """WaveSweep(paramStr, values, layerNr = 0, layerZ = 0.0, outPwr = True, outR = False, outT = False, outEnh = False)
*/
      __pyx_v_outPwr = ((bool)1);
    }
    if (values[5]) {
      __pyx_v_outR = __Pyx_PyObject_IsTrue(values[5]); if (unlikely((__pyx_v_outR == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1349, __pyx_L3_error)
    } else {

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1349
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] values, \
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \
 *             bool outR = False, bool outT = False, bool outEnh = False):             # <<<<<<<<<<<<<<
 *         """WaveSweep(paramStr, values, layerNr = 0, layerZ = 0.0, outPwr = True, outR = False, outT = False, outEnh = False)
 * 
*/
      __pyx_v_outR = ((bool)0);
    }
    if (values[6]) {
      __pyx_v_outT = __Pyx_PyObject_IsTrue(values[6]); if (unlikely((__pyx_v_outT == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1349, __pyx_L3_error)
    } else {
      __pyx_v_outT = ((bool)0);
    }
    if (values[7]) {
      __pyx_v_outEnh = __Pyx_PyObject_IsTrue(values[7]); if (unlikely((__pyx_v_outEnh == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1349, __pyx_L3_error)
    } else {
      __pyx_v_outEnh = ((bool)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("WaveSweep", 0, 2, 8, __pyx_nargs); __PYX_ERR(0, 1347, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.WaveSweep", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_paramStr), (&PyUnicode_Type), 1, "paramStr", 1))) __PYX_ERR(0, 1347, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_values), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "values", 0))) __PYX_ERR(0, 1347, __pyx_L1_error)
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_26WaveSweep(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), __pyx_v_paramStr, __pyx_v_values, __pyx_v_layerNr, __pyx_v_layerZ, __pyx_v_outPwr, __pyx_v_outR, __pyx_v_outT, __pyx_v_outEnh);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1347
 *         return res
 * 
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] values, \             # <<<<<<<<<<<<<<
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \
 *             bool outR = False, bool outT = False, bool outEnh = False):
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_26WaveSweep(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_paramStr, PyArrayObject *__pyx_v_values, int __pyx_v_layerNr, double __pyx_v_layerZ, bool __pyx_v_outPwr, bool __pyx_v_outR, bool __pyx_v_outT, bool __pyx_v_outEnh) {
  TMM::WaveSweepResultNonlinearTMM *__pyx_v_resCpp;
  int __pyx_v_outmask;
  enum TMM::TMMParam __pyx_v_param;
  int __pyx_v_paramLayer;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_res = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_values;
  __Pyx_Buffer __pyx_pybuffer_values;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  enum TMM::TMMParam __pyx_t_5;
  eigency::Map<Eigen::ArrayXd>  __pyx_t_6;
  std::unique_ptr<TMM::WaveSweepResultNonlinearTMM>  __pyx_t_7;
  size_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("WaveSweep", 0);
  __pyx_pybuffer_values.pybuffer.buf = NULL;
  __pyx_pybuffer_values.refcount = 0;
  __pyx_pybuffernd_values.data = NULL;
  __pyx_pybuffernd_values.rcbuffer = &__pyx_pybuffer_values;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_values.rcbuffer->pybuffer, (PyObject*)__pyx_v_values, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1347, __pyx_L1_error)
  }
  __pyx_pybuffernd_values.diminfo[0].strides = __pyx_pybuffernd_values.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_values.diminfo[0].shape = __pyx_pybuffernd_values.rcbuffer->pybuffer.shape[0];

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1392
 *         """
 *         cdef WaveSweepResultNonlinearTMMCpp *resCpp;
 *         cdef int outmask = 0             # <<<<<<<<<<<<<<
 * 
 *         if outPwr:
*/
  __pyx_v_outmask = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1394
 *         cdef int outmask = 0
 * 
 *         if outPwr:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_PWRFLOWS
 *         if outEnh:
*/
  __pyx_t_1 = (__pyx_v_outPwr != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1395
 * 
 *         if outPwr:
 *             outmask |= SWEEP_PWRFLOWS             # <<<<<<<<<<<<<<
 *         if outEnh:
 *             outmask |= SWEEP_ENH
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_PWRFLOWS);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1394
 *         cdef int outmask = 0
 * 
 *         if outPwr:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_PWRFLOWS
 *         if outEnh:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1396
 *         if outPwr:
 *             outmask |= SWEEP_PWRFLOWS
 *         if outEnh:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_ENH
 *         if outR:
*/
  __pyx_t_1 = (__pyx_v_outEnh != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1397
 *             outmask |= SWEEP_PWRFLOWS
 *         if outEnh:
 *             outmask |= SWEEP_ENH             # <<<<<<<<<<<<<<
 *         if outR:
 *             outmask |= SWEEP_R
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_ENH);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1396
 *         if outPwr:
 *             outmask |= SWEEP_PWRFLOWS
 *         if outEnh:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_ENH
 *         if outR:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1398
 *         if outEnh:
 *             outmask |= SWEEP_ENH
 *         if outR:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_R
 *         if outT:
*/
  __pyx_t_1 = (__pyx_v_outR != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1399
 *             outmask |= SWEEP_ENH
 *         if outR:
 *             outmask |= SWEEP_R             # <<<<<<<<<<<<<<
 *         if outT:
 *             outmask |= SWEEP_T
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_R);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1398
 *         if outEnh:
 *             outmask |= SWEEP_ENH
 *         if outR:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_R
 *         if outT:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1400
 *         if outR:
 *             outmask |= SWEEP_R
 *         if outT:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_T
 * 
*/
  __pyx_t_1 = (__pyx_v_outT != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1401
 *             outmask |= SWEEP_R
 *         if outT:
 *             outmask |= SWEEP_T             # <<<<<<<<<<<<<<
 * 
 *         cdef TMMParamCpp param;
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_T);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1400
 *         if outR:
 *             outmask |= SWEEP_R
 *         if outT:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_T
 * 
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1404
 * 
 *         cdef TMMParamCpp param;
 *         cdef int paramLayer = -1;             # <<<<<<<<<<<<<<
 *         if (paramStr.startswith("d_")):
 *             param = PARAM_LAYER_D
*/
  __pyx_v_paramLayer = -1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1405
 *         cdef TMMParamCpp param;
 *         cdef int paramLayer = -1;
 *         if (paramStr.startswith("d_")):             # <<<<<<<<<<<<<<
 *             param = PARAM_LAYER_D
 *             paramLayer = int(paramStr[2:])
*/
  if (unlikely(__pyx_v_paramStr == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "startswith");
    __PYX_ERR(0, 1405, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyUnicode_Tailmatch(__pyx_v_paramStr, __pyx_mstate_global->__pyx_n_u_d_2, 0, PY_SSIZE_T_MAX, -1); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1405, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1406
 *         cdef int paramLayer = -1;
 *         if (paramStr.startswith("d_")):
 *             param = PARAM_LAYER_D             # <<<<<<<<<<<<<<
 *             paramLayer = int(paramStr[2:])
 *         else:
*/
    __pyx_v_param = TMM::TMMParam::PARAM_LAYER_D;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1407
 *         if (paramStr.startswith("d_")):
 *             param = PARAM_LAYER_D
 *             paramLayer = int(paramStr[2:])             # <<<<<<<<<<<<<<
 *         else:
 *             param = TmmParamFromStr(paramStr)
*/
    if (unlikely(__pyx_v_paramStr == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 1407, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyUnicode_Substring(__pyx_v_paramStr, 2, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1407, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_paramLayer = __pyx_t_4;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1405
 *         cdef TMMParamCpp param;
 *         cdef int paramLayer = -1;
 *         if (paramStr.startswith("d_")):             # <<<<<<<<<<<<<<
 *             param = PARAM_LAYER_D
 *             paramLayer = int(paramStr[2:])
*/
    goto __pyx_L7;
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1409
 *             paramLayer = int(paramStr[2:])
 *         else:
 *             param = TmmParamFromStr(paramStr)             # <<<<<<<<<<<<<<
 * 
 *         resCpp = self._thisptr.WaveSweep(param, Map[ArrayXd](values), outmask, paramLayer, layerNr, layerZ).release()
*/
  /*else*/ {
    __pyx_t_5 = __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_TmmParamFromStr(__pyx_v_paramStr); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1409, __pyx_L1_error)
    __pyx_v_param = __pyx_t_5;
  }
  __pyx_L7:;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1411
 *             param = TmmParamFromStr(paramStr)
 * 
 *         resCpp = self._thisptr.WaveSweep(param, Map[ArrayXd](values), outmask, paramLayer, layerNr, layerZ).release()             # <<<<<<<<<<<<<<
 *         res = _WaveSweepResultNonlinearTMM()
 *         res._Init(resCpp);
*/
  try {
    __pyx_t_6 = eigency::Map<Eigen::ArrayXd> (((PyArrayObject *)__pyx_v_values));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1411, __pyx_L1_error)
  }
  try {
    __pyx_t_7 = __pyx_v_self->_thisptr->WaveSweep(__pyx_v_param, __pyx_t_6, __pyx_v_outmask, __pyx_v_paramLayer, __pyx_v_layerNr, __pyx_v_layerZ);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1411, __pyx_L1_error)
  }
  __pyx_v_resCpp = __pyx_t_7.release();

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1412
 * 
 *         resCpp = self._thisptr.WaveSweep(param, Map[ArrayXd](values), outmask, paramLayer, layerNr, layerZ).release()
 *         res = _WaveSweepResultNonlinearTMM()             # <<<<<<<<<<<<<<
 *         res._Init(resCpp);
 *         return res
*/
  __pyx_t_2 = NULL;
  __pyx_t_8 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM, __pyx_callargs+__pyx_t_8, (1-__pyx_t_8) | (__pyx_t_8*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1412, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
  }
  __pyx_v_res = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1413
 *         resCpp = self._thisptr.WaveSweep(param, Map[ArrayXd](values), outmask, paramLayer, layerNr, layerZ).release()
 *         res = _WaveSweepResultNonlinearTMM()
 *         res._Init(resCpp);             # <<<<<<<<<<<<<<
 *         return res
 * 
*/
  __pyx_t_3 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_v_res->__pyx_vtab)->_Init(__pyx_v_res, __pyx_v_resCpp, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1414
 *         res = _WaveSweepResultNonlinearTMM()
 *         res._Init(resCpp);
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     def WaveGetFields2D(self, np.ndarray[double, ndim = 1] zs, \
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_res);
  __pyx_r = ((PyObject *)__pyx_v_res);
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1347
 *         return res
 * 
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] values, \             # <<<<<<<<<<<<<<
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \
 *             bool outR = False, bool outT = False, bool outEnh = False):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_values.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.WaveSweep", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_values.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1416
 *         return res
 * 
 *     def WaveGetFields2D(self, np.ndarray[double, ndim = 1] zs, \             # <<<<<<<<<<<<<<
 *             np.ndarray[double, ndim = 1] xs, str dirStr = "total"):
 *         """WaveGetFields2D(zs, xs, dirStr = "total")
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_29WaveGetFields2D(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_28WaveGetFields2D, "WaveGetFields2D(zs, xs, dirStr = \"total\")\n\n        Calculates 2D electric and magnetic fields of beam propagating in the\n        structure. Analogous to ``GetFields2D`` of plane waves.\n\n        Parameters\n        ----------\n        zs : ndarray of floats\n            Points on z-axis where to calculate the fields. The beginning of the\n            first layer is at z = 0.\n        xs : ndarray of floats\n            Points on x-axis where to calculate the fields.\n        dirStr : {'total', 'forward', 'backward'}\n            Specifies the components of the output fields.\n\n        Returns\n        -------\n        :class:`_FieldsZX`\n            Helper class to store electric and magnetic fields in regular grid.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_29WaveGetFields2D = {"WaveGetFields2D", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_29WaveGetFields2D, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_28WaveGetFields2D};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_29WaveGetFields2D(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyArrayObject *__pyx_v_zs = 0;
  PyArrayObject *__pyx_v_xs = 0;
  PyObject *__pyx_v_dirStr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("WaveGetFields2D (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_zs,&__pyx_mstate_global->__pyx_n_u_xs,&__pyx_mstate_global->__pyx_n_u_dirStr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1416, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1416, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1416, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1416, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "WaveGetFields2D", 0) < (0)) __PYX_ERR(0, 1416, __pyx_L3_error)
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject*)__pyx_mstate_global->__pyx_n_u_total));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("WaveGetFields2D", 0, 2, 3, i); __PYX_ERR(0, 1416, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1416, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1416, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1416, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject*)__pyx_mstate_global->__pyx_n_u_total));
    }
    __pyx_v_zs = ((PyArrayObject *)values[0]);
    __pyx_v_xs = ((PyArrayObject *)values[1]);
    __pyx_v_dirStr = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("WaveGetFields2D", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 1416, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.WaveGetFields2D", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_zs), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "zs", 0))) __PYX_ERR(0, 1416, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_xs), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "xs", 0))) __PYX_ERR(0, 1417, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dirStr), (&PyUnicode_Type), 1, "dirStr", 1))) __PYX_ERR(0, 1417, __pyx_L1_error)
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_28WaveGetFields2D(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), __pyx_v_zs, __pyx_v_xs, __pyx_v_dirStr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_28WaveGetFields2D(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyArrayObject *__pyx_v_zs, PyArrayObject *__pyx_v_xs, PyObject *__pyx_v_dirStr) {
  TMM::FieldsZX *__pyx_v_resCpp;
  enum TMM::WaveDirection __pyx_v_direction;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_res = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_xs;
  __Pyx_Buffer __pyx_pybuffer_xs;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_zs;
  __Pyx_Buffer __pyx_pybuffer_zs;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum TMM::WaveDirection __pyx_t_1;
  eigency::Map<Eigen::ArrayXd>  __pyx_t_2;
  eigency::Map<Eigen::ArrayXd>  __pyx_t_3;
  std::unique_ptr<TMM::FieldsZX>  __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("WaveGetFields2D", 0);
  __pyx_pybuffer_zs.pybuffer.buf = NULL;
  __pyx_pybuffer_zs.refcount = 0;
  __pyx_pybuffernd_zs.data = NULL;
  __pyx_pybuffernd_zs.rcbuffer = &__pyx_pybuffer_zs;
  __pyx_pybuffer_xs.pybuffer.buf = NULL;
  __pyx_pybuffer_xs.refcount = 0;
  __pyx_pybuffernd_xs.data = NULL;
  __pyx_pybuffernd_xs.rcbuffer = &__pyx_pybuffer_xs;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_zs.rcbuffer->pybuffer, (PyObject*)__pyx_v_zs, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1416, __pyx_L1_error)
  }
  __pyx_pybuffernd_zs.diminfo[0].strides = __pyx_pybuffernd_zs.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_zs.diminfo[0].shape = __pyx_pybuffernd_zs.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_xs.rcbuffer->pybuffer, (PyObject*)__pyx_v_xs, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1416, __pyx_L1_error)
  }
  __pyx_pybuffernd_xs.diminfo[0].strides = __pyx_pybuffernd_xs.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_xs.diminfo[0].shape = __pyx_pybuffernd_xs.rcbuffer->pybuffer.shape[0];

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1440
 *         """
 *         cdef FieldsZXCpp *resCpp;
 *         cdef WaveDirectionCpp direction = WaveDirectionFromStr(dirStr)             # <<<<<<<<<<<<<<
 *         resCpp = self._thisptr.WaveGetFields2D(Map[ArrayXd](zs), Map[ArrayXd](xs), direction).release()
 *         res = _FieldsZX()
*/
  __pyx_t_1 = __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_WaveDirectionFromStr(__pyx_v_dirStr); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1440, __pyx_L1_error)
  __pyx_v_direction = __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1441
 *         cdef FieldsZXCpp *resCpp;
 *         cdef WaveDirectionCpp direction = WaveDirectionFromStr(dirStr)
 *         resCpp = self._thisptr.WaveGetFields2D(Map[ArrayXd](zs), Map[ArrayXd](xs), direction).release()             # <<<<<<<<<<<<<<
 *         res = _FieldsZX()
 *         res._Init(resCpp)
*/
  try {
    __pyx_t_2 = eigency::Map<Eigen::ArrayXd> (((PyArrayObject *)__pyx_v_zs));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1441, __pyx_L1_error)
  }
  try {
    __pyx_t_3 = eigency::Map<Eigen::ArrayXd> (((PyArrayObject *)__pyx_v_xs));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1441, __pyx_L1_error)
  }
  try {
    __pyx_t_4 = __pyx_v_self->_thisptr->WaveGetFields2D(__pyx_t_2, __pyx_t_3, __pyx_v_direction);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1441, __pyx_L1_error)
  }
  __pyx_v_resCpp = __pyx_t_4.release();

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1442
 *         cdef WaveDirectionCpp direction = WaveDirectionFromStr(dirStr)
 *         resCpp = self._thisptr.WaveGetFields2D(Map[ArrayXd](zs), Map[ArrayXd](xs), direction).release()
 *         res = _FieldsZX()             # <<<<<<<<<<<<<<
 *         res._Init(resCpp)
 *         return res
*/
  __pyx_t_6 = NULL;
  __pyx_t_7 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
    __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1442, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_5);
  }
  __pyx_v_res = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1443
 *         resCpp = self._thisptr.WaveGetFields2D(Map[ArrayXd](zs), Map[ArrayXd](xs), direction).release()
 *         res = _FieldsZX()
 *         res._Init(resCpp)             # <<<<<<<<<<<<<<
 *         return res
 * 
*/
  __pyx_t_5 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_res->__pyx_vtab)->_Init(__pyx_v_res, __pyx_v_resCpp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1444
 *         res = _FieldsZX()
 *         res._Init(resCpp)
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     # Getters
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_res);
  __pyx_r = ((PyObject *)__pyx_v_res);
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1416
 *         return res
 * 
 *     def WaveGetFields2D(self, np.ndarray[double, ndim = 1] zs, \             # <<<<<<<<<<<<<<
 *             np.ndarray[double, ndim = 1] xs, str dirStr = "total"):
 *         """WaveGetFields2D(zs, xs, dirStr = "total")
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_xs.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_zs.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.WaveGetFields2D", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_xs.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_zs.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1449
 *     #---------------------------------------------------------------------------
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def wl(self):
 *         return self._thisptr.GetWl()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2wl_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2wl_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2wl___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2wl___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1451
 *     @property
 *     def wl(self):
 *         return self._thisptr.GetWl()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetWl();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1451, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1449
 *     #---------------------------------------------------------------------------
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def wl(self):
 *         return self._thisptr.GetWl()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.wl.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1480
 *     #---------------------------------------------------------------------------
 * 
 *     @wl.setter             # <<<<<<<<<<<<<<
 *     def wl(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetWl(<double>value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2wl_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2wl_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2wl_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2wl_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1482
 *     @wl.setter
 *     def wl(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetWl(<double>value)             # <<<<<<<<<<<<<<
 * 
 *     @beta.setter
*/
  __pyx_t_1 = __Pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1482, __pyx_L1_error)
  try {
    __pyx_v_self->_thisptr->SetWl(((double)__pyx_t_1));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1482, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1480
 *     #---------------------------------------------------------------------------
 * 
 *     @wl.setter             # <<<<<<<<<<<<<<
 *     def wl(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetWl(<double>value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.wl.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1453
 *         return self._thisptr.GetWl()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def beta(self):
 *         return self._thisptr.GetBeta()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4beta_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4beta_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4beta___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4beta___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1455
 *     @property
 *     def beta(self):
 *         return self._thisptr.GetBeta()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetBeta();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1455, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1453
 *         return self._thisptr.GetWl()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def beta(self):
 *         return self._thisptr.GetBeta()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.beta.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1484
 *         self._thisptr.SetWl(<double>value)
 * 
 *     @beta.setter             # <<<<<<<<<<<<<<
 *     def beta(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetBeta(<double>value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4beta_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4beta_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4beta_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4beta_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1486
 *     @beta.setter
 *     def beta(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetBeta(<double>value)             # <<<<<<<<<<<<<<
 * 
 *     @pol.setter
*/
  __pyx_t_1 = __Pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1486, __pyx_L1_error)
  try {
    __pyx_v_self->_thisptr->SetBeta(((double)__pyx_t_1));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1486, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1484
 *         self._thisptr.SetWl(<double>value)
 * 
 *     @beta.setter             # <<<<<<<<<<<<<<
 *     def beta(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetBeta(<double>value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.beta.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1457
 *         return self._thisptr.GetBeta()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def pol(self):
 *         return "ps"[<int>self._thisptr.GetPolarization()]
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3pol_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3pol_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3pol___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3pol___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum TMM::Polarization __pyx_t_1;
  int __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1459
 *     @property
 *     def pol(self):
 *         return "ps"[<int>self._thisptr.GetPolarization()]             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetPolarization();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1459, __pyx_L1_error)
  }
  __pyx_t_2 = ((int)__pyx_t_1);
  __pyx_t_3 = __Pyx_GetItemInt_Unicode(__pyx_mstate_global->__pyx_n_u_ps, __pyx_t_2, int, 1, __Pyx_PyLong_From_int, 0, 1, 1, 1, __Pyx_ReferenceSharing_SharedReference); if (unlikely(__pyx_t_3 == (Py_UCS4)-1)) __PYX_ERR(0, 1459, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyUnicode_FromOrdinal(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1457
 *         return self._thisptr.GetBeta()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def pol(self):
 *         return "ps"[<int>self._thisptr.GetPolarization()]
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.pol.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1488
 *         self._thisptr.SetBeta(<double>value)
 * 
 *     @pol.setter             # <<<<<<<<<<<<<<
 *     def pol(self, polStr):  # @DuplicatedSignature
 *         self._thisptr.SetPolarization(PolarizationFromStr(polStr))
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3pol_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_polStr); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3pol_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_polStr) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3pol_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), ((PyObject *)__pyx_v_polStr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3pol_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_polStr) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  enum TMM::Polarization __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1490
 *     @pol.setter
 *     def pol(self, polStr):  # @DuplicatedSignature
 *         self._thisptr.SetPolarization(PolarizationFromStr(polStr))             # <<<<<<<<<<<<<<
 * 
 *     @I0.setter
*/
  __pyx_t_1 = __pyx_v_polStr;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 1490, __pyx_L1_error)
  __pyx_t_2 = __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_PolarizationFromStr(((PyObject*)__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  try {
    __pyx_v_self->_thisptr->SetPolarization(__pyx_t_2);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1490, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1488
 *         self._thisptr.SetBeta(<double>value)
 * 
 *     @pol.setter             # <<<<<<<<<<<<<<
 *     def pol(self, polStr):  # @DuplicatedSignature
 *         self._thisptr.SetPolarization(PolarizationFromStr(polStr))
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.pol.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1461
 *         return "ps"[<int>self._thisptr.GetPolarization()]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def I0(self):
 *         return self._thisptr.GetI0()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2I0_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2I0_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2I0___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2I0___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1463
 *     @property
 *     def I0(self):
 *         return self._thisptr.GetI0()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetI0();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1463, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1461
 *         return "ps"[<int>self._thisptr.GetPolarization()]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def I0(self):
 *         return self._thisptr.GetI0()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.I0.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1492
 *         self._thisptr.SetPolarization(PolarizationFromStr(polStr))
 * 
 *     @I0.setter             # <<<<<<<<<<<<<<
 *     def I0(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetI0(<double>value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2I0_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2I0_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2I0_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2I0_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1494
 *     @I0.setter
 *     def I0(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetI0(<double>value)             # <<<<<<<<<<<<<<
 * 
 *     @overrideE0.setter
*/
  __pyx_t_1 = __Pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1494, __pyx_L1_error)
  try {
    __pyx_v_self->_thisptr->SetI0(((double)__pyx_t_1));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1494, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1492
 *         self._thisptr.SetPolarization(PolarizationFromStr(polStr))
 * 
 *     @I0.setter             # <<<<<<<<<<<<<<
 *     def I0(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetI0(<double>value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.I0.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1465
 *         return self._thisptr.GetI0()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def overrideE0(self):
 *         return self._thisptr.GetOverrideE0()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10overrideE0_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10overrideE0_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10overrideE0___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10overrideE0___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  bool __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1467
 *     @property
 *     def overrideE0(self):
 *         return self._thisptr.GetOverrideE0()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetOverrideE0();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1467, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1465
 *         return self._thisptr.GetI0()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def overrideE0(self):
 *         return self._thisptr.GetOverrideE0()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.overrideE0.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1496
 *         self._thisptr.SetI0(<double>value)
 * 
 *     @overrideE0.setter             # <<<<<<<<<<<<<<
 *     def overrideE0(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetOverrideE0(<bool>value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10overrideE0_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10overrideE0_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10overrideE0_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10overrideE0_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  bool __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1498
 *     @overrideE0.setter
 *     def overrideE0(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetOverrideE0(<bool>value)             # <<<<<<<<<<<<<<
 * 
 *     @E0.setter
*/
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1498, __pyx_L1_error)
  try {
    __pyx_v_self->_thisptr->SetOverrideE0(((bool)__pyx_t_1));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1498, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1496
 *         self._thisptr.SetI0(<double>value)
 * 
 *     @overrideE0.setter             # <<<<<<<<<<<<<<
 *     def overrideE0(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetOverrideE0(<bool>value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.overrideE0.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1469
 *         return self._thisptr.GetOverrideE0()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def E0(self):
 *         return self._thisptr.GetE0()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2E0_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2E0_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2E0___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2E0___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __pyx_t_double_complex __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1471
 *     @property
 *     def E0(self):
 *         return self._thisptr.GetE0()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetE0();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1471, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_PyComplex_FromComplex(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1469
 *         return self._thisptr.GetOverrideE0()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def E0(self):
 *         return self._thisptr.GetE0()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.E0.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1500
 *         self._thisptr.SetOverrideE0(<bool>value)
 * 
 *     @E0.setter             # <<<<<<<<<<<<<<
 *     def E0(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetE0(<double complex>value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2E0_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2E0_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2E0_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2E0_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __pyx_t_double_complex __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1502
 *     @E0.setter
 *     def E0(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetE0(<double complex>value)             # <<<<<<<<<<<<<<
 * 
 *     @mode.setter
*/
  __pyx_t_1 = __Pyx_PyComplex_As___pyx_t_double_complex(__pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1502, __pyx_L1_error)
  try {
    __pyx_v_self->_thisptr->SetE0(__pyx_t_double_complex_from_parts(((double)__Pyx_CREAL(__pyx_t_1)), ((double)__Pyx_CIMAG(__pyx_t_1))));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1502, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1500
 *         self._thisptr.SetOverrideE0(<bool>value)
 * 
 *     @E0.setter             # <<<<<<<<<<<<<<
 *     def E0(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetE0(<double complex>value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.E0.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1473
 *         return self._thisptr.GetE0()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def mode(self):
 *         return ["incident", "nonlinear"][<int>self._thisptr.GetMode()]
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4mode_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4mode_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4mode___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4mode___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  enum TMM::NonlinearTmmMode __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1475
 *     @property
 *     def mode(self):
 *         return ["incident", "nonlinear"][<int>self._thisptr.GetMode()]             # <<<<<<<<<<<<<<
 * 
 *     # Setter
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_incident);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_incident);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_mstate_global->__pyx_n_u_incident) != (0)) __PYX_ERR(0, 1475, __pyx_L1_error);
  __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_nonlinear);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_nonlinear);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_n_u_nonlinear) != (0)) __PYX_ERR(0, 1475, __pyx_L1_error);
  try {
    __pyx_t_2 = __pyx_v_self->_thisptr->GetMode();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1475, __pyx_L1_error)
  }
  __pyx_t_3 = ((int)__pyx_t_2);
  __pyx_t_4 = __Pyx_GetItemInt_List(__pyx_t_1, __pyx_t_3, int, 1, __Pyx_PyLong_From_int, 1, 1, 1, 1, __Pyx_ReferenceSharing_OwnStrongReference); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1473
 *         return self._thisptr.GetE0()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def mode(self):
 *         return ["incident", "nonlinear"][<int>self._thisptr.GetMode()]
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.mode.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1504
 *         self._thisptr.SetE0(<double complex>value)
 * 
 *     @mode.setter             # <<<<<<<<<<<<<<
 *     def mode(self, modeStr):  # @DuplicatedSignature
 *         self._thisptr.SetMode(NonlinearTmmModeFromStr(modeStr))
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4mode_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_modeStr); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4mode_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_modeStr) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4mode_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), ((PyObject *)__pyx_v_modeStr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4mode_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, PyObject *__pyx_v_modeStr) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  enum TMM::NonlinearTmmMode __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1506
 *     @mode.setter
 *     def mode(self, modeStr):  # @DuplicatedSignature
 *         self._thisptr.SetMode(NonlinearTmmModeFromStr(modeStr))             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_1 = __pyx_v_modeStr;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_1))) __PYX_ERR(0, 1506, __pyx_L1_error)
  __pyx_t_2 = __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTmmModeFromStr(((PyObject*)__pyx_t_1)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1506, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  try {
    __pyx_v_self->_thisptr->SetMode(__pyx_t_2);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1506, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1504
 *         self._thisptr.SetE0(<double complex>value)
 * 
 *     @mode.setter             # <<<<<<<<<<<<<<
 *     def mode(self, modeStr):  # @DuplicatedSignature
 *         self._thisptr.SetMode(NonlinearTmmModeFromStr(modeStr))
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.mode.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1010
 * 
 *     cdef NonlinearTMMCpp *_thisptr
 *     cdef readonly list materialsCache             # <<<<<<<<<<<<<<
 *     cdef readonly list layers
 *     cdef bool _needDealloc
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_14materialsCache_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_14materialsCache_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_14materialsCache___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_14materialsCache___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->materialsCache);
  __pyx_r = __pyx_v_self->materialsCache;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1011
 *     cdef NonlinearTMMCpp *_thisptr
 *     cdef readonly list materialsCache
 *     cdef readonly list layers             # <<<<<<<<<<<<<<
 *     cdef bool _needDealloc
 *     cdef object _parent;
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_6layers_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_6layers_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_6layers___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_6layers___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->layers);
  __pyx_r = __pyx_v_self->layers;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1014
 *     cdef bool _needDealloc
 *     cdef object _parent;
 *     cdef readonly _Wave wave;             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, bool initStruct = True, object parent = None, **kwargs):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4wave_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4wave_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4wave___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4wave___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->wave);
  __pyx_r = ((PyObject *)__pyx_v_self->wave);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_31__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_31__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_31__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_31__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_30__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_30__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_33__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_33__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_33__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_33__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_32__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_32__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1533
 *     cdef readonly object P1, P2, Gen
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__cinit__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities___cinit__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_self) {
  int __pyx_r;

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1536
 *         pass
 * 
 *     cdef _Init(self, SecondOrderNLIntensitiesCpp* ptr):             # <<<<<<<<<<<<<<
 *         P1 = _Intensities()
 *         P1._Init(&ptr.P1)
*/

static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_self, TMM::SecondOrderNLIntensities *__pyx_v_ptr) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_P1 = NULL;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_P2 = NULL;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *__pyx_v_Gen = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_Init", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1537
 * 
 *     cdef _Init(self, SecondOrderNLIntensitiesCpp* ptr):
 *         P1 = _Intensities()             # <<<<<<<<<<<<<<
 *         P1._Init(&ptr.P1)
 * 
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1537, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_P1 = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1538
 *     cdef _Init(self, SecondOrderNLIntensitiesCpp* ptr):
 *         P1 = _Intensities()
 *         P1._Init(&ptr.P1)             # <<<<<<<<<<<<<<
 * 
 *         P2 = _Intensities()
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_v_P1->__pyx_vtab)->_Init(__pyx_v_P1, (&__pyx_v_ptr->P1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1540
 *         P1._Init(&ptr.P1)
 * 
 *         P2 = _Intensities()             # <<<<<<<<<<<<<<
 *         P2._Init(&ptr.P2)
 * 
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_P2 = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1541
 * 
 *         P2 = _Intensities()
 *         P2._Init(&ptr.P2)             # <<<<<<<<<<<<<<
 * 
 *         Gen = _Intensities()
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_v_P2->__pyx_vtab)->_Init(__pyx_v_P2, (&__pyx_v_ptr->P2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1543
 *         P2._Init(&ptr.P2)
 * 
 *         Gen = _Intensities()             # <<<<<<<<<<<<<<
 *         Gen._Init(&ptr.Gen)
 * 
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1543, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_Gen = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1544
 * 
 *         Gen = _Intensities()
 *         Gen._Init(&ptr.Gen)             # <<<<<<<<<<<<<<
 * 
 *         self.P1 = P1
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)__pyx_v_Gen->__pyx_vtab)->_Init(__pyx_v_Gen, (&__pyx_v_ptr->Gen)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1546
 *         Gen._Init(&ptr.Gen)
 * 
 *         self.P1 = P1             # <<<<<<<<<<<<<<
 *         self.P2 = P2
 *         self.Gen = Gen
*/
  __Pyx_INCREF((PyObject *)__pyx_v_P1);
  __Pyx_GIVEREF((PyObject *)__pyx_v_P1);
  __Pyx_GOTREF(__pyx_v_self->P1);
  __Pyx_DECREF(__pyx_v_self->P1);
  __pyx_v_self->P1 = ((PyObject *)__pyx_v_P1);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1547
 * 
 *         self.P1 = P1
 *         self.P2 = P2             # <<<<<<<<<<<<<<
 *         self.Gen = Gen
 * 
*/
  __Pyx_INCREF((PyObject *)__pyx_v_P2);
  __Pyx_GIVEREF((PyObject *)__pyx_v_P2);
  __Pyx_GOTREF(__pyx_v_self->P2);
  __Pyx_DECREF(__pyx_v_self->P2);
  __pyx_v_self->P2 = ((PyObject *)__pyx_v_P2);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1548
 *         self.P1 = P1
 *         self.P2 = P2
 *         self.Gen = Gen             # <<<<<<<<<<<<<<
 * 
 * #===============================================================================
*/
  __Pyx_INCREF((PyObject *)__pyx_v_Gen);
  __Pyx_GIVEREF((PyObject *)__pyx_v_Gen);
  __Pyx_GOTREF(__pyx_v_self->Gen);
  __Pyx_DECREF(__pyx_v_self->Gen);
  __pyx_v_self->Gen = ((PyObject *)__pyx_v_Gen);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1536
 *         pass
 * 
 *     cdef _Init(self, SecondOrderNLIntensitiesCpp* ptr):             # <<<<<<<<<<<<<<
 *         P1 = _Intensities()
 *         P1._Init(&ptr.P1)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._SecondOrderNLIntensities._Init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_P1);
  __Pyx_XDECREF((PyObject *)__pyx_v_P2);
  __Pyx_XDECREF((PyObject *)__pyx_v_Gen);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1531
 * 
 *     """
 *     cdef readonly object P1, P2, Gen             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2P1_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2P1_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2P1___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2P1___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->P1);
  __pyx_r = __pyx_v_self->P1;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2P2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2P2_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2P2___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2P2___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->P2);
  __pyx_r = __pyx_v_self->P2;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_3Gen_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_3Gen_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_3Gen___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_3Gen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->Gen);
  __pyx_r = __pyx_v_self->Gen;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._SecondOrderNLIntensities.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._SecondOrderNLIntensities.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_4__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._SecondOrderNLIntensities.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1576
 *     cdef readonly np.ndarray wlsGen, betasGen;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._thisptr = NULL
 *         pass
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__cinit__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self) {
  int __pyx_r;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1577
 * 
 *     def __cinit__(self):
 *         self._thisptr = NULL             # <<<<<<<<<<<<<<
 *         pass
 * 
*/
  __pyx_v_self->_thisptr = NULL;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1576
 *     cdef readonly np.ndarray wlsGen, betasGen;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._thisptr = NULL
 *         pass
*/

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1580
 *         pass
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._thisptr:
 *             del self._thisptr
*/

/* Python wrapper */
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2__dealloc__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self) {
  int __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1581
 * 
 *     def __dealloc__(self):
 *         if self._thisptr:             # <<<<<<<<<<<<<<
 *             del self._thisptr
 * 
*/
  __pyx_t_1 = (__pyx_v_self->_thisptr != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1582
 *     def __dealloc__(self):
 *         if self._thisptr:
 *             del self._thisptr             # <<<<<<<<<<<<<<
 * 
 *     cdef _Init(self, SweepResultSecondOrderNLTMMCpp *ptr):
*/
    delete __pyx_v_self->_thisptr;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1581
 * 
 *     def __dealloc__(self):
 *         if self._thisptr:             # <<<<<<<<<<<<<<
 *             del self._thisptr
 * 
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1580
 *         pass
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._thisptr:
 *             del self._thisptr
*/

  /* function exit code */
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1584
 *             del self._thisptr
 * 
 *     cdef _Init(self, SweepResultSecondOrderNLTMMCpp *ptr):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 * 
*/

static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self, TMM::SweepResultSecondOrderNLTMM *__pyx_v_ptr) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_P1 = NULL;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_P2 = NULL;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *__pyx_v_Gen = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  struct __pyx_opt_args_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM__Init __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_Init", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1585
 * 
 *     cdef _Init(self, SweepResultSecondOrderNLTMMCpp *ptr):
 *         self._thisptr = ptr             # <<<<<<<<<<<<<<
 * 
 *         P1 = _SweepResultNonlinearTMM()
*/
  __pyx_v_self->_thisptr = __pyx_v_ptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1587
 *         self._thisptr = ptr
 * 
 *         P1 = _SweepResultNonlinearTMM()             # <<<<<<<<<<<<<<
 *         P2 = _SweepResultNonlinearTMM()
 *         Gen = _SweepResultNonlinearTMM()
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1587, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_P1 = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1588
 * 
 *         P1 = _SweepResultNonlinearTMM()
 *         P2 = _SweepResultNonlinearTMM()             # <<<<<<<<<<<<<<
 *         Gen = _SweepResultNonlinearTMM()
 * 
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1588, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_P2 = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1589
 *         P1 = _SweepResultNonlinearTMM()
 *         P2 = _SweepResultNonlinearTMM()
 *         Gen = _SweepResultNonlinearTMM()             # <<<<<<<<<<<<<<
 * 
 *         P1._Init(&ptr.P1, False, self)
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1589, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_Gen = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1591
 *         Gen = _SweepResultNonlinearTMM()
 * 
 *         P1._Init(&ptr.P1, False, self)             # <<<<<<<<<<<<<<
 *         P2._Init(&ptr.P2, False, self)
 *         Gen._Init(&ptr.Gen, False, self)
*/
  __pyx_t_4.__pyx_n = 2;
  __pyx_t_4.needDealloc = 0;
  __pyx_t_4.parent = ((PyObject *)__pyx_v_self);
  __pyx_t_1 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_P1->__pyx_vtab)->_Init(__pyx_v_P1, (&__pyx_v_ptr->P1), &__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1592
 * 
 *         P1._Init(&ptr.P1, False, self)
 *         P2._Init(&ptr.P2, False, self)             # <<<<<<<<<<<<<<
 *         Gen._Init(&ptr.Gen, False, self)
 * 
*/
  __pyx_t_4.__pyx_n = 2;
  __pyx_t_4.needDealloc = 0;
  __pyx_t_4.parent = ((PyObject *)__pyx_v_self);
  __pyx_t_1 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_P2->__pyx_vtab)->_Init(__pyx_v_P2, (&__pyx_v_ptr->P2), &__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1593
 *         P1._Init(&ptr.P1, False, self)
 *         P2._Init(&ptr.P2, False, self)
 *         Gen._Init(&ptr.Gen, False, self)             # <<<<<<<<<<<<<<
 * 
 *         self.P1 = P1
*/
  __pyx_t_4.__pyx_n = 2;
  __pyx_t_4.needDealloc = 0;
  __pyx_t_4.parent = ((PyObject *)__pyx_v_self);
  __pyx_t_1 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)__pyx_v_Gen->__pyx_vtab)->_Init(__pyx_v_Gen, (&__pyx_v_ptr->Gen), &__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1595
 *         Gen._Init(&ptr.Gen, False, self)
 * 
 *         self.P1 = P1             # <<<<<<<<<<<<<<
 *         self.P2 = P2
 *         self.Gen = Gen
*/
  __Pyx_INCREF((PyObject *)__pyx_v_P1);
  __Pyx_GIVEREF((PyObject *)__pyx_v_P1);
  __Pyx_GOTREF(__pyx_v_self->P1);
  __Pyx_DECREF(__pyx_v_self->P1);
  __pyx_v_self->P1 = ((PyObject *)__pyx_v_P1);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1596
 * 
 *         self.P1 = P1
 *         self.P2 = P2             # <<<<<<<<<<<<<<
 *         self.Gen = Gen
 *         self.wlsGen = ndarray_view(self._thisptr.wlsGen).squeeze()
*/
  __Pyx_INCREF((PyObject *)__pyx_v_P2);
  __Pyx_GIVEREF((PyObject *)__pyx_v_P2);
  __Pyx_GOTREF(__pyx_v_self->P2);
  __Pyx_DECREF(__pyx_v_self->P2);
  __pyx_v_self->P2 = ((PyObject *)__pyx_v_P2);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1597
 *         self.P1 = P1
 *         self.P2 = P2
 *         self.Gen = Gen             # <<<<<<<<<<<<<<
 *         self.wlsGen = ndarray_view(self._thisptr.wlsGen).squeeze()
 *         self.betasGen = ndarray_view(self._thisptr.betasGen).squeeze()
*/
  __Pyx_INCREF((PyObject *)__pyx_v_Gen);
  __Pyx_GIVEREF((PyObject *)__pyx_v_Gen);
  __Pyx_GOTREF(__pyx_v_self->Gen);
  __Pyx_DECREF(__pyx_v_self->Gen);
  __pyx_v_self->Gen = ((PyObject *)__pyx_v_Gen);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1598
 *         self.P2 = P2
 *         self.Gen = Gen
 *         self.wlsGen = ndarray_view(self._thisptr.wlsGen).squeeze()             # <<<<<<<<<<<<<<
 *         self.betasGen = ndarray_view(self._thisptr.betasGen).squeeze()
 * 
*/
  __pyx_t_5 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->wlsGen)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __pyx_t_5;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1598, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1598, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->wlsGen);
  __Pyx_DECREF((PyObject *)__pyx_v_self->wlsGen);
  __pyx_v_self->wlsGen = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1599
 *         self.Gen = Gen
 *         self.wlsGen = ndarray_view(self._thisptr.wlsGen).squeeze()
 *         self.betasGen = ndarray_view(self._thisptr.betasGen).squeeze()             # <<<<<<<<<<<<<<
 * 
 * #===============================================================================
*/
  __pyx_t_2 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->betasGen)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1599, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1599, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->betasGen);
  __Pyx_DECREF((PyObject *)__pyx_v_self->betasGen);
  __pyx_v_self->betasGen = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1584
 *             del self._thisptr
 * 
 *     cdef _Init(self, SweepResultSecondOrderNLTMMCpp *ptr):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._SweepResultSecondOrderNLTMM._Init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_P1);
  __Pyx_XDECREF((PyObject *)__pyx_v_P2);
  __Pyx_XDECREF((PyObject *)__pyx_v_Gen);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1573
 *     """
 *     cdef SweepResultSecondOrderNLTMMCpp *_thisptr
 *     cdef readonly object P1, P2, Gen             # <<<<<<<<<<<<<<
 *     cdef readonly np.ndarray wlsGen, betasGen;
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2P1_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2P1_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2P1___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2P1___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->P1);
  __pyx_r = __pyx_v_self->P1;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2P2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2P2_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2P2___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2P2___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->P2);
  __pyx_r = __pyx_v_self->P2;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_3Gen_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_3Gen_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_3Gen___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_3Gen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->Gen);
  __pyx_r = __pyx_v_self->Gen;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1574
 *     cdef SweepResultSecondOrderNLTMMCpp *_thisptr
 *     cdef readonly object P1, P2, Gen
 *     cdef readonly np.ndarray wlsGen, betasGen;             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_6wlsGen_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_6wlsGen_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_6wlsGen___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_6wlsGen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->wlsGen);
  __pyx_r = ((PyObject *)__pyx_v_self->wlsGen);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_8betasGen_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_8betasGen_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_8betasGen___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_8betasGen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->betasGen);
  __pyx_r = ((PyObject *)__pyx_v_self->betasGen);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_4__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._SweepResultSecondOrderNLTMM.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._SweepResultSecondOrderNLTMM.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_6__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._SweepResultSecondOrderNLTMM.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1628
 *     cdef readonly np.ndarray wlsGen, betasGen;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._thisptr = NULL
 *         pass
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return -1;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__cinit__", __pyx_kwds); return -1;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self) {
  int __pyx_r;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1629
 * 
 *     def __cinit__(self):
 *         self._thisptr = NULL             # <<<<<<<<<<<<<<
 *         pass
 * 
*/
  __pyx_v_self->_thisptr = NULL;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1628
 *     cdef readonly np.ndarray wlsGen, betasGen;
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._thisptr = NULL
 *         pass
*/

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1632
 *         pass
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._thisptr:
 *             del self._thisptr
*/

/* Python wrapper */
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2__dealloc__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self) {
  int __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1633
 * 
 *     def __dealloc__(self):
 *         if self._thisptr:             # <<<<<<<<<<<<<<
 *             del self._thisptr
 * 
*/
  __pyx_t_1 = (__pyx_v_self->_thisptr != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1634
 *     def __dealloc__(self):
 *         if self._thisptr:
 *             del self._thisptr             # <<<<<<<<<<<<<<
 * 
 *     cdef _Init(self, WaveSweepResultSecondOrderNLTMMCpp *ptr):
*/
    delete __pyx_v_self->_thisptr;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1633
 * 
 *     def __dealloc__(self):
 *         if self._thisptr:             # <<<<<<<<<<<<<<
 *             del self._thisptr
 * 
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1632
 *         pass
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._thisptr:
 *             del self._thisptr
*/

  /* function exit code */
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1636
 *             del self._thisptr
 * 
 *     cdef _Init(self, WaveSweepResultSecondOrderNLTMMCpp *ptr):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 * 
*/

static PyObject *__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM__Init(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self, TMM::WaveSweepResultSecondOrderNLTMM *__pyx_v_ptr) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_P1 = NULL;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_P2 = NULL;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *__pyx_v_Gen = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  struct __pyx_opt_args_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM__Init __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_Init", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1637
 * 
 *     cdef _Init(self, WaveSweepResultSecondOrderNLTMMCpp *ptr):
 *         self._thisptr = ptr             # <<<<<<<<<<<<<<
 * 
 *         P1 = _WaveSweepResultNonlinearTMM()
*/
  __pyx_v_self->_thisptr = __pyx_v_ptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1639
 *         self._thisptr = ptr
 * 
 *         P1 = _WaveSweepResultNonlinearTMM()             # <<<<<<<<<<<<<<
 *         P2 = _WaveSweepResultNonlinearTMM()
 *         Gen = _WaveSweepResultNonlinearTMM()
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1639, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_P1 = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1640
 * 
 *         P1 = _WaveSweepResultNonlinearTMM()
 *         P2 = _WaveSweepResultNonlinearTMM()             # <<<<<<<<<<<<<<
 *         Gen = _WaveSweepResultNonlinearTMM()
 * 
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1640, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_P2 = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1641
 *         P1 = _WaveSweepResultNonlinearTMM()
 *         P2 = _WaveSweepResultNonlinearTMM()
 *         Gen = _WaveSweepResultNonlinearTMM()             # <<<<<<<<<<<<<<
 * 
 *         P1._Init(&ptr.P1, False, self)
*/
  __pyx_t_2 = NULL;
  __pyx_t_3 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (__pyx_t_3*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1641, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_v_Gen = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1643
 *         Gen = _WaveSweepResultNonlinearTMM()
 * 
 *         P1._Init(&ptr.P1, False, self)             # <<<<<<<<<<<<<<
 *         P2._Init(&ptr.P2, False, self)
 *         Gen._Init(&ptr.Gen, False, self)
*/
  __pyx_t_4.__pyx_n = 2;
  __pyx_t_4.needDealloc = 0;
  __pyx_t_4.parent = ((PyObject *)__pyx_v_self);
  __pyx_t_1 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_v_P1->__pyx_vtab)->_Init(__pyx_v_P1, (&__pyx_v_ptr->P1), &__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1644
 * 
 *         P1._Init(&ptr.P1, False, self)
 *         P2._Init(&ptr.P2, False, self)             # <<<<<<<<<<<<<<
 *         Gen._Init(&ptr.Gen, False, self)
 * 
*/
  __pyx_t_4.__pyx_n = 2;
  __pyx_t_4.needDealloc = 0;
  __pyx_t_4.parent = ((PyObject *)__pyx_v_self);
  __pyx_t_1 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_v_P2->__pyx_vtab)->_Init(__pyx_v_P2, (&__pyx_v_ptr->P2), &__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1645
 *         P1._Init(&ptr.P1, False, self)
 *         P2._Init(&ptr.P2, False, self)
 *         Gen._Init(&ptr.Gen, False, self)             # <<<<<<<<<<<<<<
 * 
 *         self.P1 = P1
*/
  __pyx_t_4.__pyx_n = 2;
  __pyx_t_4.needDealloc = 0;
  __pyx_t_4.parent = ((PyObject *)__pyx_v_self);
  __pyx_t_1 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)__pyx_v_Gen->__pyx_vtab)->_Init(__pyx_v_Gen, (&__pyx_v_ptr->Gen), &__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1647
 *         Gen._Init(&ptr.Gen, False, self)
 * 
 *         self.P1 = P1             # <<<<<<<<<<<<<<
 *         self.P2 = P2
 *         self.Gen = Gen
*/
  __Pyx_INCREF((PyObject *)__pyx_v_P1);
  __Pyx_GIVEREF((PyObject *)__pyx_v_P1);
  __Pyx_GOTREF(__pyx_v_self->P1);
  __Pyx_DECREF(__pyx_v_self->P1);
  __pyx_v_self->P1 = ((PyObject *)__pyx_v_P1);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1648
 * 
 *         self.P1 = P1
 *         self.P2 = P2             # <<<<<<<<<<<<<<
 *         self.Gen = Gen
 *         self.wlsGen = ndarray_view(self._thisptr.wlsGen).squeeze()
*/
  __Pyx_INCREF((PyObject *)__pyx_v_P2);
  __Pyx_GIVEREF((PyObject *)__pyx_v_P2);
  __Pyx_GOTREF(__pyx_v_self->P2);
  __Pyx_DECREF(__pyx_v_self->P2);
  __pyx_v_self->P2 = ((PyObject *)__pyx_v_P2);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1649
 *         self.P1 = P1
 *         self.P2 = P2
 *         self.Gen = Gen             # <<<<<<<<<<<<<<
 *         self.wlsGen = ndarray_view(self._thisptr.wlsGen).squeeze()
 *         self.betasGen = ndarray_view(self._thisptr.betasGen).squeeze()
*/
  __Pyx_INCREF((PyObject *)__pyx_v_Gen);
  __Pyx_GIVEREF((PyObject *)__pyx_v_Gen);
  __Pyx_GOTREF(__pyx_v_self->Gen);
  __Pyx_DECREF(__pyx_v_self->Gen);
  __pyx_v_self->Gen = ((PyObject *)__pyx_v_Gen);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1650
 *         self.P2 = P2
 *         self.Gen = Gen
 *         self.wlsGen = ndarray_view(self._thisptr.wlsGen).squeeze()             # <<<<<<<<<<<<<<
 *         self.betasGen = ndarray_view(self._thisptr.betasGen).squeeze()
 * 
*/
  __pyx_t_5 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->wlsGen)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __pyx_t_5;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1650, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1650, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->wlsGen);
  __Pyx_DECREF((PyObject *)__pyx_v_self->wlsGen);
  __pyx_v_self->wlsGen = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1651
 *         self.Gen = Gen
 *         self.wlsGen = ndarray_view(self._thisptr.wlsGen).squeeze()
 *         self.betasGen = ndarray_view(self._thisptr.betasGen).squeeze()             # <<<<<<<<<<<<<<
 * 
 * 
*/
  __pyx_t_2 = ((PyObject *)eigency::ndarray_view(__pyx_v_self->_thisptr->betasGen)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_squeeze, __pyx_callargs+__pyx_t_3, (1-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1651, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->betasGen);
  __Pyx_DECREF((PyObject *)__pyx_v_self->betasGen);
  __pyx_v_self->betasGen = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1636
 *             del self._thisptr
 * 
 *     cdef _Init(self, WaveSweepResultSecondOrderNLTMMCpp *ptr):             # <<<<<<<<<<<<<<
 *         self._thisptr = ptr
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._WaveSweepResultSecondOrderNLTMM._Init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_P1);
  __Pyx_XDECREF((PyObject *)__pyx_v_P2);
  __Pyx_XDECREF((PyObject *)__pyx_v_Gen);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1625
 *     """
 *     cdef WaveSweepResultSecondOrderNLTMMCpp *_thisptr
 *     cdef readonly object P1, P2, Gen             # <<<<<<<<<<<<<<
 *     cdef readonly np.ndarray wlsGen, betasGen;
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2P1_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2P1_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2P1___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2P1___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->P1);
  __pyx_r = __pyx_v_self->P1;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2P2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2P2_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2P2___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2P2___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->P2);
  __pyx_r = __pyx_v_self->P2;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_3Gen_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_3Gen_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_3Gen___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_3Gen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->Gen);
  __pyx_r = __pyx_v_self->Gen;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1626
 *     cdef WaveSweepResultSecondOrderNLTMMCpp *_thisptr
 *     cdef readonly object P1, P2, Gen
 *     cdef readonly np.ndarray wlsGen, betasGen;             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_6wlsGen_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_6wlsGen_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_6wlsGen___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_6wlsGen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->wlsGen);
  __pyx_r = ((PyObject *)__pyx_v_self->wlsGen);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_8betasGen_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_8betasGen_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_8betasGen___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_8betasGen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->betasGen);
  __pyx_r = ((PyObject *)__pyx_v_self->betasGen);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_4__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._WaveSweepResultSecondOrderNLTMM.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._WaveSweepResultSecondOrderNLTMM.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_6__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython._WaveSweepResultSecondOrderNLTMM.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1696
 *     cdef readonly object P1, P2, Gen
 * 
 *     def __cinit__(self, str mode = "sfg", **kwargs):             # <<<<<<<<<<<<<<
 *         self._thisptr = new SecondOrderNLTMMCpp()
 *         self._thisptr.SetProcess(NonlinearProcessFromStr(mode))
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_mode = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_mode,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1696, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1696, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values, kwd_pos_args, __pyx_kwds_len, "__cinit__", 1) < (0)) __PYX_ERR(0, 1696, __pyx_L3_error)
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject*)__pyx_mstate_global->__pyx_n_u_sfg));
    } else {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1696, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[0]) values[0] = __Pyx_NewRef(((PyObject*)__pyx_mstate_global->__pyx_n_u_sfg));
    }
    __pyx_v_mode = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 1696, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mode), (&PyUnicode_Type), 1, "mode", 1))) __PYX_ERR(0, 1696, __pyx_L1_error)
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM___cinit__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self), __pyx_v_mode, __pyx_v_kwargs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM___cinit__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyObject *__pyx_v_mode, PyObject *__pyx_v_kwargs) {
  TMM::NonlinearTMM *__pyx_v_tmmP1Ptr;
  TMM::NonlinearTMM *__pyx_v_tmmP2Ptr;
  TMM::NonlinearTMM *__pyx_v_tmmGenPtr;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_tmmP1Py = 0;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_tmmP2Py = 0;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *__pyx_v_tmmGenPy = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  TMM::SecondOrderNLTMM *__pyx_t_1;
  enum TMM::NonlinearProcess __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  TMM::NonlinearTMM *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1697
 * 
 *     def __cinit__(self, str mode = "sfg", **kwargs):
 *         self._thisptr = new SecondOrderNLTMMCpp()             # <<<<<<<<<<<<<<
 *         self._thisptr.SetProcess(NonlinearProcessFromStr(mode))
 *         self.materialsCache = []
*/
  try {
    __pyx_t_1 = new TMM::SecondOrderNLTMM();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1697, __pyx_L1_error)
  }
  __pyx_v_self->_thisptr = __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1698
 *     def __cinit__(self, str mode = "sfg", **kwargs):
 *         self._thisptr = new SecondOrderNLTMMCpp()
 *         self._thisptr.SetProcess(NonlinearProcessFromStr(mode))             # <<<<<<<<<<<<<<
 *         self.materialsCache = []
 * 
*/
  __pyx_t_2 = __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearProcessFromStr(__pyx_v_mode); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1698, __pyx_L1_error)
  try {
    __pyx_v_self->_thisptr->SetProcess(__pyx_t_2);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1698, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1699
 *         self._thisptr = new SecondOrderNLTMMCpp()
 *         self._thisptr.SetProcess(NonlinearProcessFromStr(mode))
 *         self.materialsCache = []             # <<<<<<<<<<<<<<
 * 
 *         # Pointers to TMM-s
*/
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->materialsCache);
  __Pyx_DECREF(__pyx_v_self->materialsCache);
  __pyx_v_self->materialsCache = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1702
 * 
 *         # Pointers to TMM-s
 *         cdef NonlinearTMMCpp* tmmP1Ptr = self._thisptr.GetP1()             # <<<<<<<<<<<<<<
 *         cdef NonlinearTMMCpp* tmmP2Ptr = self._thisptr.GetP2()
 *         cdef NonlinearTMMCpp* tmmGenPtr = self._thisptr.GetGen()
*/
  try {
    __pyx_t_4 = __pyx_v_self->_thisptr->GetP1();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1702, __pyx_L1_error)
  }
  __pyx_v_tmmP1Ptr = __pyx_t_4;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1703
 *         # Pointers to TMM-s
 *         cdef NonlinearTMMCpp* tmmP1Ptr = self._thisptr.GetP1()
 *         cdef NonlinearTMMCpp* tmmP2Ptr = self._thisptr.GetP2()             # <<<<<<<<<<<<<<
 *         cdef NonlinearTMMCpp* tmmGenPtr = self._thisptr.GetGen()
 * 
*/
  try {
    __pyx_t_4 = __pyx_v_self->_thisptr->GetP2();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1703, __pyx_L1_error)
  }
  __pyx_v_tmmP2Ptr = __pyx_t_4;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1704
 *         cdef NonlinearTMMCpp* tmmP1Ptr = self._thisptr.GetP1()
 *         cdef NonlinearTMMCpp* tmmP2Ptr = self._thisptr.GetP2()
 *         cdef NonlinearTMMCpp* tmmGenPtr = self._thisptr.GetGen()             # <<<<<<<<<<<<<<
 * 
 *         # Wrapper classes
*/
  try {
    __pyx_t_4 = __pyx_v_self->_thisptr->GetGen();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1704, __pyx_L1_error)
  }
  __pyx_v_tmmGenPtr = __pyx_t_4;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1707
 * 
 *         # Wrapper classes
 *         cdef NonlinearTMM tmmP1Py = NonlinearTMM(False)             # <<<<<<<<<<<<<<
 *         cdef NonlinearTMM tmmP2Py = NonlinearTMM(False)
 *         cdef NonlinearTMM tmmGenPy = NonlinearTMM(False)
*/
  __pyx_t_5 = NULL;
  __pyx_t_6 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, Py_False};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1707, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
  }
  __pyx_v_tmmP1Py = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1708
 *         # Wrapper classes
 *         cdef NonlinearTMM tmmP1Py = NonlinearTMM(False)
 *         cdef NonlinearTMM tmmP2Py = NonlinearTMM(False)             # <<<<<<<<<<<<<<
 *         cdef NonlinearTMM tmmGenPy = NonlinearTMM(False)
 * 
*/
  __pyx_t_5 = NULL;
  __pyx_t_6 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, Py_False};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1708, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
  }
  __pyx_v_tmmP2Py = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1709
 *         cdef NonlinearTMM tmmP1Py = NonlinearTMM(False)
 *         cdef NonlinearTMM tmmP2Py = NonlinearTMM(False)
 *         cdef NonlinearTMM tmmGenPy = NonlinearTMM(False)             # <<<<<<<<<<<<<<
 * 
 *         # Init wrappers
*/
  __pyx_t_5 = NULL;
  __pyx_t_6 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, Py_False};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1709, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
  }
  __pyx_v_tmmGenPy = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1712
 * 
 *         # Init wrappers
 *         tmmP1Py._Init(tmmP1Ptr)             # <<<<<<<<<<<<<<
 *         tmmP2Py._Init(tmmP2Ptr)
 *         tmmGenPy._Init(tmmGenPtr)
*/
  __pyx_t_3 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_tmmP1Py->__pyx_vtab)->_Init(__pyx_v_tmmP1Py, __pyx_v_tmmP1Ptr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1713
 *         # Init wrappers
 *         tmmP1Py._Init(tmmP1Ptr)
 *         tmmP2Py._Init(tmmP2Ptr)             # <<<<<<<<<<<<<<
 *         tmmGenPy._Init(tmmGenPtr)
 * 
*/
  __pyx_t_3 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_tmmP2Py->__pyx_vtab)->_Init(__pyx_v_tmmP2Py, __pyx_v_tmmP2Ptr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1713, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1714
 *         tmmP1Py._Init(tmmP1Ptr)
 *         tmmP2Py._Init(tmmP2Ptr)
 *         tmmGenPy._Init(tmmGenPtr)             # <<<<<<<<<<<<<<
 * 
 *         # Save to members
*/
  __pyx_t_3 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)__pyx_v_tmmGenPy->__pyx_vtab)->_Init(__pyx_v_tmmGenPy, __pyx_v_tmmGenPtr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1717
 * 
 *         # Save to members
 *         self.P1 = tmmP1Py             # <<<<<<<<<<<<<<
 *         self.P2 = tmmP2Py
 *         self.Gen = tmmGenPy
*/
  __Pyx_INCREF((PyObject *)__pyx_v_tmmP1Py);
  __Pyx_GIVEREF((PyObject *)__pyx_v_tmmP1Py);
  __Pyx_GOTREF(__pyx_v_self->P1);
  __Pyx_DECREF(__pyx_v_self->P1);
  __pyx_v_self->P1 = ((PyObject *)__pyx_v_tmmP1Py);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1718
 *         # Save to members
 *         self.P1 = tmmP1Py
 *         self.P2 = tmmP2Py             # <<<<<<<<<<<<<<
 *         self.Gen = tmmGenPy
 * 
*/
  __Pyx_INCREF((PyObject *)__pyx_v_tmmP2Py);
  __Pyx_GIVEREF((PyObject *)__pyx_v_tmmP2Py);
  __Pyx_GOTREF(__pyx_v_self->P2);
  __Pyx_DECREF(__pyx_v_self->P2);
  __pyx_v_self->P2 = ((PyObject *)__pyx_v_tmmP2Py);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1719
 *         self.P1 = tmmP1Py
 *         self.P2 = tmmP2Py
 *         self.Gen = tmmGenPy             # <<<<<<<<<<<<<<
 * 
 *         # Set params
*/
  __Pyx_INCREF((PyObject *)__pyx_v_tmmGenPy);
  __Pyx_GIVEREF((PyObject *)__pyx_v_tmmGenPy);
  __Pyx_GOTREF(__pyx_v_self->Gen);
  __Pyx_DECREF(__pyx_v_self->Gen);
  __pyx_v_self->Gen = ((PyObject *)__pyx_v_tmmGenPy);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1722
 * 
 *         # Set params
 *         self.SetParams(**kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_SetParams); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_mstate_global->__pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1696
 *     cdef readonly object P1, P2, Gen
 * 
 *     def __cinit__(self, str mode = "sfg", **kwargs):             # <<<<<<<<<<<<<<
 *         self._thisptr = new SecondOrderNLTMMCpp()
 *         self._thisptr.SetProcess(NonlinearProcessFromStr(mode))
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_tmmP1Py);
  __Pyx_XDECREF((PyObject *)__pyx_v_tmmP2Py);
  __Pyx_XDECREF((PyObject *)__pyx_v_tmmGenPy);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1724
 *         self.SetParams(**kwargs)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._thisptr
 * 
*/

/* Python wrapper */
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2__dealloc__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2__dealloc__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self) {

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1725
 * 
 *     def __dealloc__(self):
 *         del self._thisptr             # <<<<<<<<<<<<<<
 * 
 *     # Methods
*/
  delete __pyx_v_self->_thisptr;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1724
 *         self.SetParams(**kwargs)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         del self._thisptr
 * 
*/

  /* function exit code */
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1730
 *     #---------------------------------------------------------------------------
 * 
 *     def SetParams(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """SetParams(**kwargs)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_5SetParams(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_4SetParams, "SetParams(**kwargs)\n\n        Helper method to set the values of all the attributes. See the docstring\n        of :class:`SecondOrderNLTMM`.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_5SetParams = {"SetParams", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_5SetParams, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_4SetParams};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_5SetParams(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("SetParams (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("SetParams", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (__pyx_kwds_len > 0) {
    if (unlikely(__Pyx_CheckKeywordStrings("SetParams", __pyx_kwds) == -1)) return NULL;
    __pyx_v_kwargs = __Pyx_KwargsAsDict_FASTCALL(__pyx_kwds, __pyx_kwvalues);
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = PyDict_New();
    if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  }
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_4SetParams(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_4SetParams(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("SetParams", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1737
 * 
 *         """
 *         for name, value in kwargs.items():             # <<<<<<<<<<<<<<
 *             if name not in paramDictSecondOrderNLTMM:
 *                 raise ValueError("Unknown kwarg %s" % (name))
*/
  __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_v_kwargs, 1, __pyx_mstate_global->__pyx_n_u_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 1737, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1738
 *         """
 *         for name, value in kwargs.items():
 *             if name not in paramDictSecondOrderNLTMM:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Unknown kwarg %s" % (name))
 *             setattr(self, name, value)
*/
    __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_v_name, __pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_paramDictSecondOrderNLTMM, Py_NE)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1738, __pyx_L1_error)
    if (unlikely(__pyx_t_8)) {

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1739
 *         for name, value in kwargs.items():
 *             if name not in paramDictSecondOrderNLTMM:
 *                 raise ValueError("Unknown kwarg %s" % (name))             # <<<<<<<<<<<<<<
 *             setattr(self, name, value)
 * 
*/
      __pyx_t_5 = NULL;
      __pyx_t_9 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_Unknown_kwarg_s, __pyx_v_name); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1739, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = 1;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_9};
        __pyx_t_6 = __Pyx_PyObject_FastCall((PyObject*)(((PyTypeObject*)PyExc_ValueError)), __pyx_callargs+__pyx_t_10, (2-__pyx_t_10) | (__pyx_t_10*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1739, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 1739, __pyx_L1_error)

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1738
 *         """
 *         for name, value in kwargs.items():
 *             if name not in paramDictSecondOrderNLTMM:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Unknown kwarg %s" % (name))
 *             setattr(self, name, value)
*/
    }

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1740
 *             if name not in paramDictSecondOrderNLTMM:
 *                 raise ValueError("Unknown kwarg %s" % (name))
 *             setattr(self, name, value)             # <<<<<<<<<<<<<<
 * 
 *     def AddLayer(self, double d, Material material):
*/
    __pyx_t_11 = PyObject_SetAttr(((PyObject *)__pyx_v_self), __pyx_v_name, __pyx_v_value); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 1740, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1730
 *     #---------------------------------------------------------------------------
 * 
 *     def SetParams(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """SetParams(**kwargs)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.SetParams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1742
 *             setattr(self, name, value)
 * 
 *     def AddLayer(self, double d, Material material):             # <<<<<<<<<<<<<<
 *         """AddLayer(d, material)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_7AddLayer(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_6AddLayer, "AddLayer(d, material)\n\n        Adds layer to SecondOrderNLTMM. Equivalent of adding layer to P1, P2\n        and Gen.\n\n        Parameters\n        ----------\n        d : float\n            layer thickness (m)\n        material : :class:`Material`\n            The class containing the material parameters.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_7AddLayer = {"AddLayer", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_7AddLayer, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_6AddLayer};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_7AddLayer(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  double __pyx_v_d;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_material = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("AddLayer (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_d,&__pyx_mstate_global->__pyx_n_u_material,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1742, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1742, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1742, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "AddLayer", 0) < (0)) __PYX_ERR(0, 1742, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("AddLayer", 1, 2, 2, i); __PYX_ERR(0, 1742, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1742, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1742, __pyx_L3_error)
    }
    __pyx_v_d = __Pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_d == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1742, __pyx_L3_error)
    __pyx_v_material = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("AddLayer", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1742, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.AddLayer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_material), __pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material, 1, "material", 0))) __PYX_ERR(0, 1742, __pyx_L1_error)
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_6AddLayer(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self), __pyx_v_d, __pyx_v_material);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_6AddLayer(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, double __pyx_v_d, struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *__pyx_v_material) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("AddLayer", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1757
 *         """
 *         # No copy of material is made
 *         self.P1.AddLayer(d, material)             # <<<<<<<<<<<<<<
 *         self.P2.AddLayer(d, material)
 *         self.Gen.AddLayer(d, material)
*/
  __pyx_t_2 = __pyx_v_self->P1;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_d); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_3, ((PyObject *)__pyx_v_material)};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_AddLayer, __pyx_callargs+__pyx_t_4, (3-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1758
 *         # No copy of material is made
 *         self.P1.AddLayer(d, material)
 *         self.P2.AddLayer(d, material)             # <<<<<<<<<<<<<<
 *         self.Gen.AddLayer(d, material)
 * 
*/
  __pyx_t_3 = __pyx_v_self->P2;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_d); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_t_2, ((PyObject *)__pyx_v_material)};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_AddLayer, __pyx_callargs+__pyx_t_4, (3-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1759
 *         self.P1.AddLayer(d, material)
 *         self.P2.AddLayer(d, material)
 *         self.Gen.AddLayer(d, material)             # <<<<<<<<<<<<<<
 * 
 *         # Cache material classes, avoids material dealloc
*/
  __pyx_t_2 = __pyx_v_self->Gen;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_d); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_3, ((PyObject *)__pyx_v_material)};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod((PyObject*)__pyx_mstate_global->__pyx_n_u_AddLayer, __pyx_callargs+__pyx_t_4, (3-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1759, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1762
 * 
 *         # Cache material classes, avoids material dealloc
 *         self.materialsCache.append(material)             # <<<<<<<<<<<<<<
 * 
 *     def Solve(self):
*/
  if (unlikely(__pyx_v_self->materialsCache == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
    __PYX_ERR(0, 1762, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_self->materialsCache, ((PyObject *)__pyx_v_material)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 1762, __pyx_L1_error)

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1742
 *             setattr(self, name, value)
 * 
 *     def AddLayer(self, double d, Material material):             # <<<<<<<<<<<<<<
 *         """AddLayer(d, material)
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.AddLayer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1764
 *         self.materialsCache.append(material)
 * 
 *     def Solve(self):             # <<<<<<<<<<<<<<
 *         """Solve()
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_9Solve(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_8Solve, "Solve()\n\n        Solves the structure.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_9Solve = {"Solve", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_9Solve, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_8Solve};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_9Solve(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Solve (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("Solve", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("Solve", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_8Solve(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_8Solve(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Solve", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1770
 * 
 *         """
 *         self._thisptr.Solve()             # <<<<<<<<<<<<<<
 * 
 *     def UpdateGenParams(self):
*/
  try {
    __pyx_v_self->_thisptr->Solve();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1770, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1764
 *         self.materialsCache.append(material)
 * 
 *     def Solve(self):             # <<<<<<<<<<<<<<
 *         """Solve()
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.Solve", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1772
 *         self._thisptr.Solve()
 * 
 *     def UpdateGenParams(self):             # <<<<<<<<<<<<<<
 *         """UpdateGenParams()
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11UpdateGenParams(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_10UpdateGenParams, "UpdateGenParams()\n\n        Forces the update of the wavelength and beta of the generated beam ``Gen``.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11UpdateGenParams = {"UpdateGenParams", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11UpdateGenParams, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_10UpdateGenParams};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11UpdateGenParams(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("UpdateGenParams (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("UpdateGenParams", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("UpdateGenParams", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_10UpdateGenParams(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_10UpdateGenParams(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("UpdateGenParams", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1778
 * 
 *         """
 *         self._thisptr.UpdateGenParams()             # <<<<<<<<<<<<<<
 * 
 *     def GetIntensities(self):
*/
  try {
    __pyx_v_self->_thisptr->UpdateGenParams();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1778, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1772
 *         self._thisptr.Solve()
 * 
 *     def UpdateGenParams(self):             # <<<<<<<<<<<<<<
 *         """UpdateGenParams()
 * 
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.UpdateGenParams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1780
 *         self._thisptr.UpdateGenParams()
 * 
 *     def GetIntensities(self):             # <<<<<<<<<<<<<<
 *         """GetIntensities()
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_13GetIntensities(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_12GetIntensities, "GetIntensities()\n\n        Returns the intensities and amplitudes of incident, reflected and\n        transmitted wave for ``P1``, ``P2`` and ``Gen``.\n        The structure must be solved first.\n\n        Returns\n        -------\n        :class:`_SecondOrderNLIntensities`\n            Helper class to hold intensity data.\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_13GetIntensities = {"GetIntensities", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_13GetIntensities, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_12GetIntensities};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_13GetIntensities(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("GetIntensities (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("GetIntensities", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("GetIntensities", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_12GetIntensities(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_12GetIntensities(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self) {
  TMM::SecondOrderNLIntensities __pyx_v_resCpp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *__pyx_v_res = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  TMM::SecondOrderNLIntensities __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("GetIntensities", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1793
 *         """
 *         cdef SecondOrderNLIntensitiesCpp resCpp;
 *         resCpp = self._thisptr.GetIntensities();             # <<<<<<<<<<<<<<
 * 
 *         res = _SecondOrderNLIntensities()
*/
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetIntensities();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1793, __pyx_L1_error)
  }
  __pyx_v_resCpp = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1795
 *         resCpp = self._thisptr.GetIntensities();
 * 
 *         res = _SecondOrderNLIntensities()             # <<<<<<<<<<<<<<
 *         res._Init(&resCpp)
 *         return res
*/
  __pyx_t_3 = NULL;
  __pyx_t_4 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities, __pyx_callargs+__pyx_t_4, (1-__pyx_t_4) | (__pyx_t_4*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1795, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_2);
  }
  __pyx_v_res = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1796
 * 
 *         res = _SecondOrderNLIntensities()
 *         res._Init(&resCpp)             # <<<<<<<<<<<<<<
 *         return res
 * 
*/
  __pyx_t_2 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *)__pyx_v_res->__pyx_vtab)->_Init(__pyx_v_res, (&__pyx_v_resCpp)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1797
 *         res = _SecondOrderNLIntensities()
 *         res._Init(&resCpp)
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     def Sweep(self, str paramStr, np.ndarray[double, ndim = 1] valuesP1, np.ndarray[double, ndim = 1] valuesP2, int layerNr = 0, double layerZ = 0.0, bool outPwr = True, bool outAbs = False, bool outEnh = False, bool outP1 = True, bool outP2 = True, bool outGen = True):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_res);
  __pyx_r = ((PyObject *)__pyx_v_res);
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1780
 *         self._thisptr.UpdateGenParams()
 * 
 *     def GetIntensities(self):             # <<<<<<<<<<<<<<
 *         """GetIntensities()
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.GetIntensities", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1799
 *         return res
 * 
 *     def Sweep(self, str paramStr, np.ndarray[double, ndim = 1] valuesP1, np.ndarray[double, ndim = 1] valuesP2, int layerNr = 0, double layerZ = 0.0, bool outPwr = True, bool outAbs = False, bool outEnh = False, bool outP1 = True, bool outP2 = True, bool outGen = True):             # <<<<<<<<<<<<<<
 *         """Sweep(paramStr, valuesP1, valuesP2, layerNr = 0, layerZ = 0.0, outPwr = True, outAbs = False, outEnh = False, outP1 = True, outP2 = True, outGen = True)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_15Sweep(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14Sweep, "Sweep(paramStr, valuesP1, valuesP2, layerNr = 0, layerZ = 0.0, outPwr = True, outAbs = False, outEnh = False, outP1 = True, outP2 = True, outGen = True)\n\n        Solves the structure for a series of ``valuesP1`` and ``valuesP2``\n        of param ``paramStr``. Using this function is more convenient and\n        faster than just changing params and solving the structure (parallelized\n        with OpenMP).\n\n        Parameters\n        ----------\n        paramStr : str\n            'wl':\n                Wavelength in meters\n            'beta':\n                normalized tangential wavevector\n            'I0':\n                intensity of the wave\n            'd_i':\n                thickness of layer i (0..N-1)\n        valuesP1 : ndarray of floats\n            Corresponding values of param ``paramStr`` for ``P1``.\n        valuesP2 : ndarray of floats\n            Corresponding values of param ``paramStr`` for ``P2``.\n        layerNr : int\n            Specifies layer, where electrical field enhancement is calculated.\n        layerZ : double\n            Specifies z-coordinate of enhancement calculation inside ``layerNr``.\n        outPwr : bool\n            Turns calculation of intensities on/off.\n        outAbs : bool\n            Turns calculation of absorption in the entire structure on/off.\n        outEnh : bool\n            Turns calculation of enhancement in layer ``layerNr`` at distance\n            ``layerZ`` on/off.\n        outP1 : bool\n            Turns calculation of ``P1`` on/off.\n        outP2 : bool\n            Turns calculation of ``P2`` on/off.\n        outGen : bool\n            Turns calculation of ``Gen`` on/off.\n\n        Returns\n        -------\n        :class:`_SweepResultSecondOrderNLTMM`\n            Helper class to store the result.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_15Sweep = {"Sweep", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_15Sweep, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14Sweep};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_15Sweep(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_paramStr = 0;
  PyArrayObject *__pyx_v_valuesP1 = 0;
  PyArrayObject *__pyx_v_valuesP2 = 0;
  int __pyx_v_layerNr;
  double __pyx_v_layerZ;
  bool __pyx_v_outPwr;
  bool __pyx_v_outAbs;
  bool __pyx_v_outEnh;
  bool __pyx_v_outP1;
  bool __pyx_v_outP2;
  bool __pyx_v_outGen;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("Sweep (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_paramStr,&__pyx_mstate_global->__pyx_n_u_valuesP1,&__pyx_mstate_global->__pyx_n_u_valuesP2,&__pyx_mstate_global->__pyx_n_u_layerNr,&__pyx_mstate_global->__pyx_n_u_layerZ,&__pyx_mstate_global->__pyx_n_u_outPwr,&__pyx_mstate_global->__pyx_n_u_outAbs,&__pyx_mstate_global->__pyx_n_u_outEnh,&__pyx_mstate_global->__pyx_n_u_outP1,&__pyx_mstate_global->__pyx_n_u_outP2,&__pyx_mstate_global->__pyx_n_u_outGen,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1799, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case 11:
        values[10] = __Pyx_ArgRef_FASTCALL(__pyx_args, 10);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 10:
        values[9] = __Pyx_ArgRef_FASTCALL(__pyx_args, 9);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  9:
        values[8] = __Pyx_ArgRef_FASTCALL(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "Sweep", 0) < (0)) __PYX_ERR(0, 1799, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("Sweep", 0, 3, 11, i); __PYX_ERR(0, 1799, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case 11:
        values[10] = __Pyx_ArgRef_FASTCALL(__pyx_args, 10);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 10:
        values[9] = __Pyx_ArgRef_FASTCALL(__pyx_args, 9);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  9:
        values[8] = __Pyx_ArgRef_FASTCALL(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1799, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1799, __pyx_L3_error)
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1799, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1799, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_paramStr = ((PyObject*)values[0]);
    __pyx_v_valuesP1 = ((PyArrayObject *)values[1]);
    __pyx_v_valuesP2 = ((PyArrayObject *)values[2]);
    if (values[3]) {
      __pyx_v_layerNr = __Pyx_PyLong_As_int(values[3]); if (unlikely((__pyx_v_layerNr == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1799, __pyx_L3_error)
    } else {
      __pyx_v_layerNr = ((int)0);
    }
    if (values[4]) {
      __pyx_v_layerZ = __Pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_layerZ == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1799, __pyx_L3_error)
    } else {
      __pyx_v_layerZ = ((double)0.0);
    }
    if (values[5]) {
      __pyx_v_outPwr = __Pyx_PyObject_IsTrue(values[5]); if (unlikely((__pyx_v_outPwr == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1799, __pyx_L3_error)
    } else {
      __pyx_v_outPwr = ((bool)1);
    }
    if (values[6]) {
      __pyx_v_outAbs = __Pyx_PyObject_IsTrue(values[6]); if (unlikely((__pyx_v_outAbs == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1799, __pyx_L3_error)
    } else {
      __pyx_v_outAbs = ((bool)0);
    }
    if (values[7]) {
      __pyx_v_outEnh = __Pyx_PyObject_IsTrue(values[7]); if (unlikely((__pyx_v_outEnh == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1799, __pyx_L3_error)
    } else {
      __pyx_v_outEnh = ((bool)0);
    }
    if (values[8]) {
      __pyx_v_outP1 = __Pyx_PyObject_IsTrue(values[8]); if (unlikely((__pyx_v_outP1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1799, __pyx_L3_error)
    } else {
      __pyx_v_outP1 = ((bool)1);
    }
    if (values[9]) {
      __pyx_v_outP2 = __Pyx_PyObject_IsTrue(values[9]); if (unlikely((__pyx_v_outP2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1799, __pyx_L3_error)
    } else {
      __pyx_v_outP2 = ((bool)1);
    }
    if (values[10]) {
      __pyx_v_outGen = __Pyx_PyObject_IsTrue(values[10]); if (unlikely((__pyx_v_outGen == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1799, __pyx_L3_error)
    } else {
      __pyx_v_outGen = ((bool)1);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("Sweep", 0, 3, 11, __pyx_nargs); __PYX_ERR(0, 1799, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.Sweep", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_paramStr), (&PyUnicode_Type), 1, "paramStr", 1))) __PYX_ERR(0, 1799, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_valuesP1), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "valuesP1", 0))) __PYX_ERR(0, 1799, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_valuesP2), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "valuesP2", 0))) __PYX_ERR(0, 1799, __pyx_L1_error)
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14Sweep(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self), __pyx_v_paramStr, __pyx_v_valuesP1, __pyx_v_valuesP2, __pyx_v_layerNr, __pyx_v_layerZ, __pyx_v_outPwr, __pyx_v_outAbs, __pyx_v_outEnh, __pyx_v_outP1, __pyx_v_outP2, __pyx_v_outGen);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14Sweep(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyObject *__pyx_v_paramStr, PyArrayObject *__pyx_v_valuesP1, PyArrayObject *__pyx_v_valuesP2, int __pyx_v_layerNr, double __pyx_v_layerZ, bool __pyx_v_outPwr, bool __pyx_v_outAbs, bool __pyx_v_outEnh, bool __pyx_v_outP1, bool __pyx_v_outP2, bool __pyx_v_outGen) {
  TMM::SweepResultSecondOrderNLTMM *__pyx_v_resCpp;
  int __pyx_v_outmask;
  enum TMM::TMMParam __pyx_v_param;
  int __pyx_v_paramLayer;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *__pyx_v_res = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_valuesP1;
  __Pyx_Buffer __pyx_pybuffer_valuesP1;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_valuesP2;
  __Pyx_Buffer __pyx_pybuffer_valuesP2;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  enum TMM::TMMParam __pyx_t_5;
  eigency::Map<Eigen::ArrayXd>  __pyx_t_6;
  eigency::Map<Eigen::ArrayXd>  __pyx_t_7;
  std::unique_ptr<TMM::SweepResultSecondOrderNLTMM>  __pyx_t_8;
  size_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("Sweep", 0);
  __pyx_pybuffer_valuesP1.pybuffer.buf = NULL;
  __pyx_pybuffer_valuesP1.refcount = 0;
  __pyx_pybuffernd_valuesP1.data = NULL;
  __pyx_pybuffernd_valuesP1.rcbuffer = &__pyx_pybuffer_valuesP1;
  __pyx_pybuffer_valuesP2.pybuffer.buf = NULL;
  __pyx_pybuffer_valuesP2.refcount = 0;
  __pyx_pybuffernd_valuesP2.data = NULL;
  __pyx_pybuffernd_valuesP2.rcbuffer = &__pyx_pybuffer_valuesP2;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_valuesP1.rcbuffer->pybuffer, (PyObject*)__pyx_v_valuesP1, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1799, __pyx_L1_error)
  }
  __pyx_pybuffernd_valuesP1.diminfo[0].strides = __pyx_pybuffernd_valuesP1.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_valuesP1.diminfo[0].shape = __pyx_pybuffernd_valuesP1.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_valuesP2.rcbuffer->pybuffer, (PyObject*)__pyx_v_valuesP2, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1799, __pyx_L1_error)
  }
  __pyx_pybuffernd_valuesP2.diminfo[0].strides = __pyx_pybuffernd_valuesP2.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_valuesP2.diminfo[0].shape = __pyx_pybuffernd_valuesP2.rcbuffer->pybuffer.shape[0];

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1847
 *         """
 *         cdef SweepResultSecondOrderNLTMMCpp *resCpp;
 *         cdef int outmask = 0             # <<<<<<<<<<<<<<
 *         if outP1:
 *             outmask |= SWEEP_P1
*/
  __pyx_v_outmask = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1848
 *         cdef SweepResultSecondOrderNLTMMCpp *resCpp;
 *         cdef int outmask = 0
 *         if outP1:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_P1
 *         if outP2:
*/
  __pyx_t_1 = (__pyx_v_outP1 != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1849
 *         cdef int outmask = 0
 *         if outP1:
 *             outmask |= SWEEP_P1             # <<<<<<<<<<<<<<
 *         if outP2:
 *             outmask |= SWEEP_P2
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_P1);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1848
 *         cdef SweepResultSecondOrderNLTMMCpp *resCpp;
 *         cdef int outmask = 0
 *         if outP1:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_P1
 *         if outP2:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1850
 *         if outP1:
 *             outmask |= SWEEP_P1
 *         if outP2:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_P2
 *         if outGen:
*/
  __pyx_t_1 = (__pyx_v_outP2 != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1851
 *             outmask |= SWEEP_P1
 *         if outP2:
 *             outmask |= SWEEP_P2             # <<<<<<<<<<<<<<
 *         if outGen:
 *             outmask |= SWEEP_GEN
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_P2);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1850
 *         if outP1:
 *             outmask |= SWEEP_P1
 *         if outP2:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_P2
 *         if outGen:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1852
 *         if outP2:
 *             outmask |= SWEEP_P2
 *         if outGen:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_GEN
 *         if outPwr:
*/
  __pyx_t_1 = (__pyx_v_outGen != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1853
 *             outmask |= SWEEP_P2
 *         if outGen:
 *             outmask |= SWEEP_GEN             # <<<<<<<<<<<<<<
 *         if outPwr:
 *             outmask |= SWEEP_PWRFLOWS
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_GEN);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1852
 *         if outP2:
 *             outmask |= SWEEP_P2
 *         if outGen:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_GEN
 *         if outPwr:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1854
 *         if outGen:
 *             outmask |= SWEEP_GEN
 *         if outPwr:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_PWRFLOWS
 *         if outAbs:
*/
  __pyx_t_1 = (__pyx_v_outPwr != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1855
 *             outmask |= SWEEP_GEN
 *         if outPwr:
 *             outmask |= SWEEP_PWRFLOWS             # <<<<<<<<<<<<<<
 *         if outAbs:
 *             outmask |= SWEEP_ABS
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_PWRFLOWS);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1854
 *         if outGen:
 *             outmask |= SWEEP_GEN
 *         if outPwr:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_PWRFLOWS
 *         if outAbs:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1856
 *         if outPwr:
 *             outmask |= SWEEP_PWRFLOWS
 *         if outAbs:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_ABS
 *         if outEnh:
*/
  __pyx_t_1 = (__pyx_v_outAbs != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1857
 *             outmask |= SWEEP_PWRFLOWS
 *         if outAbs:
 *             outmask |= SWEEP_ABS             # <<<<<<<<<<<<<<
 *         if outEnh:
 *             outmask |= SWEEP_ENH
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_ABS);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1856
 *         if outPwr:
 *             outmask |= SWEEP_PWRFLOWS
 *         if outAbs:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_ABS
 *         if outEnh:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1858
 *         if outAbs:
 *             outmask |= SWEEP_ABS
 *         if outEnh:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_ENH
 * 
*/
  __pyx_t_1 = (__pyx_v_outEnh != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1859
 *             outmask |= SWEEP_ABS
 *         if outEnh:
 *             outmask |= SWEEP_ENH             # <<<<<<<<<<<<<<
 * 
 *         cdef TMMParamCpp param;
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_ENH);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1858
 *         if outAbs:
 *             outmask |= SWEEP_ABS
 *         if outEnh:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_ENH
 * 
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1862
 * 
 *         cdef TMMParamCpp param;
 *         cdef int paramLayer = -1;             # <<<<<<<<<<<<<<
 *         if (paramStr.startswith("d_")):
 *             param = PARAM_LAYER_D
*/
  __pyx_v_paramLayer = -1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1863
 *         cdef TMMParamCpp param;
 *         cdef int paramLayer = -1;
 *         if (paramStr.startswith("d_")):             # <<<<<<<<<<<<<<
 *             param = PARAM_LAYER_D
 *             paramLayer = int(paramStr[2:])
*/
  if (unlikely(__pyx_v_paramStr == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "startswith");
    __PYX_ERR(0, 1863, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyUnicode_Tailmatch(__pyx_v_paramStr, __pyx_mstate_global->__pyx_n_u_d_2, 0, PY_SSIZE_T_MAX, -1); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1863, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1864
 *         cdef int paramLayer = -1;
 *         if (paramStr.startswith("d_")):
 *             param = PARAM_LAYER_D             # <<<<<<<<<<<<<<
 *             paramLayer = int(paramStr[2:])
 *         else:
*/
    __pyx_v_param = TMM::TMMParam::PARAM_LAYER_D;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1865
 *         if (paramStr.startswith("d_")):
 *             param = PARAM_LAYER_D
 *             paramLayer = int(paramStr[2:])             # <<<<<<<<<<<<<<
 *         else:
 *             param = TmmParamFromStr(paramStr)
*/
    if (unlikely(__pyx_v_paramStr == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 1865, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyUnicode_Substring(__pyx_v_paramStr, 2, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1865, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_paramLayer = __pyx_t_4;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1863
 *         cdef TMMParamCpp param;
 *         cdef int paramLayer = -1;
 *         if (paramStr.startswith("d_")):             # <<<<<<<<<<<<<<
 *             param = PARAM_LAYER_D
 *             paramLayer = int(paramStr[2:])
*/
    goto __pyx_L9;
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1867
 *             paramLayer = int(paramStr[2:])
 *         else:
 *             param = TmmParamFromStr(paramStr)             # <<<<<<<<<<<<<<
 * 
 *         resCpp = self._thisptr.Sweep(param, Map[ArrayXd](valuesP1), Map[ArrayXd](valuesP2), outmask, paramLayer, layerNr, layerZ).release()
*/
  /*else*/ {
    __pyx_t_5 = __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_TmmParamFromStr(__pyx_v_paramStr); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1867, __pyx_L1_error)
    __pyx_v_param = __pyx_t_5;
  }
  __pyx_L9:;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1869
 *             param = TmmParamFromStr(paramStr)
 * 
 *         resCpp = self._thisptr.Sweep(param, Map[ArrayXd](valuesP1), Map[ArrayXd](valuesP2), outmask, paramLayer, layerNr, layerZ).release()             # <<<<<<<<<<<<<<
 *         res = _SweepResultSecondOrderNLTMM()
 *         res._Init(resCpp);
*/
  try {
    __pyx_t_6 = eigency::Map<Eigen::ArrayXd> (((PyArrayObject *)__pyx_v_valuesP1));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1869, __pyx_L1_error)
  }
  try {
    __pyx_t_7 = eigency::Map<Eigen::ArrayXd> (((PyArrayObject *)__pyx_v_valuesP2));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1869, __pyx_L1_error)
  }
  try {
    __pyx_t_8 = __pyx_v_self->_thisptr->Sweep(__pyx_v_param, __pyx_t_6, __pyx_t_7, __pyx_v_outmask, __pyx_v_paramLayer, __pyx_v_layerNr, __pyx_v_layerZ);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1869, __pyx_L1_error)
  }
  __pyx_v_resCpp = __pyx_t_8.release();

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1870
 * 
 *         resCpp = self._thisptr.Sweep(param, Map[ArrayXd](valuesP1), Map[ArrayXd](valuesP2), outmask, paramLayer, layerNr, layerZ).release()
 *         res = _SweepResultSecondOrderNLTMM()             # <<<<<<<<<<<<<<
 *         res._Init(resCpp);
 *         return res
*/
  __pyx_t_2 = NULL;
  __pyx_t_9 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM, __pyx_callargs+__pyx_t_9, (1-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1870, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
  }
  __pyx_v_res = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1871
 *         resCpp = self._thisptr.Sweep(param, Map[ArrayXd](valuesP1), Map[ArrayXd](valuesP2), outmask, paramLayer, layerNr, layerZ).release()
 *         res = _SweepResultSecondOrderNLTMM()
 *         res._Init(resCpp);             # <<<<<<<<<<<<<<
 *         return res
 * 
*/
  __pyx_t_3 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)__pyx_v_res->__pyx_vtab)->_Init(__pyx_v_res, __pyx_v_resCpp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1872
 *         res = _SweepResultSecondOrderNLTMM()
 *         res._Init(resCpp);
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     def WaveGetPowerFlows(self, int layerNr, double x0 = float("nan"), double x1 = float("nan"), double z = 0.0):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_res);
  __pyx_r = ((PyObject *)__pyx_v_res);
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1799
 *         return res
 * 
 *     def Sweep(self, str paramStr, np.ndarray[double, ndim = 1] valuesP1, np.ndarray[double, ndim = 1] valuesP2, int layerNr = 0, double layerZ = 0.0, bool outPwr = True, bool outAbs = False, bool outEnh = False, bool outP1 = True, bool outP2 = True, bool outGen = True):             # <<<<<<<<<<<<<<
 *         """Sweep(paramStr, valuesP1, valuesP2, layerNr = 0, layerZ = 0.0, outPwr = True, outAbs = False, outEnh = False, outP1 = True, outP2 = True, outGen = True)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_valuesP1.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_valuesP2.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.Sweep", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_valuesP1.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_valuesP2.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1874
 *         return res
 * 
 *     def WaveGetPowerFlows(self, int layerNr, double x0 = float("nan"), double x1 = float("nan"), double z = 0.0):             # <<<<<<<<<<<<<<
 *         """WaveGetPowerFlows(layerNr, x0 = float("nan"), x1 = float("nan"), z = 0.0)
 * 
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_17WaveGetPowerFlows(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_16WaveGetPowerFlows, "WaveGetPowerFlows(layerNr, x0 = float(\"nan\"), x1 = float(\"nan\"), z = 0.0)\n\n        Analogous to ``GetIntensities``, but calculates the powers of the\n        beams instead of the intensities of the plane waves. Only for the\n        calculation of the power of the generated beam. For pump beams use the\n        same method of ``NonlinearTMM``.\n\n        Parameters\n        ----------\n        layerNr : int\n            Specifies layer number where to calculate the power of the beam.\n        x0 : float\n            Specifies the starting point of the integration of the power in the\n            x-direction. By default this parameter is selected by the ``_Wave``\n            class.\n        x1 : float\n            Specifies the end point of the integration of the power in the\n            x-direction. By default this parameter is selected by the ``_Wave``\n            class.\n        z : float\n            Specifies the z-position of the line through which the integration\n            of the power of the beam is done.\n\n        Returns\n        -------\n        tuple of floats (PB, PF)\n            PB is the power propagated into negative infinity and PF denotes\n            the power propagating in the positive direction of z-axis.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_17WaveGetPowerFlows = {"WaveGetPowerFlows", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_17WaveGetPowerFlows, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_16WaveGetPowerFlows};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_17WaveGetPowerFlows(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int __pyx_v_layerNr;
  double __pyx_v_x0;
  double __pyx_v_x1;
  double __pyx_v_z;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("WaveGetPowerFlows (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_layerNr,&__pyx_mstate_global->__pyx_n_u_x0,&__pyx_mstate_global->__pyx_n_u_x1,&__pyx_mstate_global->__pyx_n_u_z,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1874, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1874, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1874, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1874, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1874, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "WaveGetPowerFlows", 0) < (0)) __PYX_ERR(0, 1874, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("WaveGetPowerFlows", 0, 1, 4, i); __PYX_ERR(0, 1874, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1874, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1874, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1874, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1874, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_layerNr = __Pyx_PyLong_As_int(values[0]); if (unlikely((__pyx_v_layerNr == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1874, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_x0 = __Pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_x0 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1874, __pyx_L3_error)
    } else {
      __pyx_v_x0 = __pyx_mstate_global->__pyx_k__3;
    }
    if (values[2]) {
      __pyx_v_x1 = __Pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_x1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1874, __pyx_L3_error)
    } else {
      __pyx_v_x1 = __pyx_mstate_global->__pyx_k__4;
    }
    if (values[3]) {
      __pyx_v_z = __Pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_z == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1874, __pyx_L3_error)
    } else {
      __pyx_v_z = ((double)0.0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("WaveGetPowerFlows", 0, 1, 4, __pyx_nargs); __PYX_ERR(0, 1874, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.WaveGetPowerFlows", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_16WaveGetPowerFlows(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self), __pyx_v_layerNr, __pyx_v_x0, __pyx_v_x1, __pyx_v_z);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_16WaveGetPowerFlows(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, int __pyx_v_layerNr, double __pyx_v_x0, double __pyx_v_x1, double __pyx_v_z) {
  std::pair<double,double>  __pyx_v_res;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  std::pair<double,double>  __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("WaveGetPowerFlows", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1906
 *         """
 *         cdef pair[double, double] res;
 *         res = self._thisptr.WaveGetPowerFlows(layerNr, x0, x1, z)             # <<<<<<<<<<<<<<
 *         return (res.first, res.second)
 * 
*/
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->WaveGetPowerFlows(__pyx_v_layerNr, __pyx_v_x0, __pyx_v_x1, __pyx_v_z);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1906, __pyx_L1_error)
  }
  __pyx_v_res = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_1);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1907
 *         cdef pair[double, double] res;
 *         res = self._thisptr.WaveGetPowerFlows(layerNr, x0, x1, z)
 *         return (res.first, res.second)             # <<<<<<<<<<<<<<
 * 
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] valuesP1, \
*/
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_res.first); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_res.second); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2) != (0)) __PYX_ERR(0, 1907, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3) != (0)) __PYX_ERR(0, 1907, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1874
 *         return res
 * 
 *     def WaveGetPowerFlows(self, int layerNr, double x0 = float("nan"), double x1 = float("nan"), double z = 0.0):             # <<<<<<<<<<<<<<
 *         """WaveGetPowerFlows(layerNr, x0 = float("nan"), x1 = float("nan"), z = 0.0)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.WaveGetPowerFlows", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1909
 *         return (res.first, res.second)
 * 
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] valuesP1, \             # <<<<<<<<<<<<<<
 *             np.ndarray[double, ndim = 1] valuesP2,
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_19WaveSweep(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_18WaveSweep, "WaveSweep(paramStr, valuesP1, valuesP2, layerNr = 0, layerZ = 0.0, outPwr = True, outR = False, outT = False, outEnh = False, outP1 = True, outP2 = True, outGen = True)\n\n        Solves the structure for waves for a series of ``valuesP1`` and ``valuesP2`` of param\n        ``paramStr``. Using this function is more convenient and faster than just\n        changing the params and solving the structure. Analogous to ``Sweep``.\n\n        Parameters\n        ----------\n        paramStr : str\n            'wl':\n                Wavelength in meters\n            'beta':\n                normalized tangential wavevector\n            'I0':\n                intensity of the wave\n            'd_i':\n                thickness of layer i (0..N-1)\n            'w0':\n                waist size of the input beam\n        valuesP1 : ndarray of floats\n            Corresponding values of param ``paramStr`` for ``P1``.\n        valuesP2 : ndarray of floats\n            Corresponding values of param ``paramStr`` for ``P2``.\n        layerNr : int\n            Specifies layer, where electrical field enhancement is calculated.\n        layerZ : double\n            Specifies z-coordinate of enhancement calculation inside ``layerNr``.\n        outPwr : bool\n            Turns calculation of all powers on/off.\n        outR : bool\n            Turns calculation of reflected power on/off.\n        outT : bool\n            Turns calculation of transmitted power on/off.\n        outEnh : bool\n            Turns calculation of enhancement in layer ``layerNr`` at distance\n            ``layerZ`` on/off.\n        outP1 : bool\n            Turns calculation of ``P1`` on/off.\n        outP2 : bool\n            Turns calculation of ``P2`` on/off.\n        outGen : bool\n            Turns calculation of ``Gen`` on/off.\n\n        Returns\n        -------\n        :class:`_WaveSweepResultSecondOrderNLTMM`\n            Helper class to store the result.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_19WaveSweep = {"WaveSweep", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_19WaveSweep, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_18WaveSweep};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_19WaveSweep(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_paramStr = 0;
  PyArrayObject *__pyx_v_valuesP1 = 0;
  PyArrayObject *__pyx_v_valuesP2 = 0;
  int __pyx_v_layerNr;
  double __pyx_v_layerZ;
  bool __pyx_v_outPwr;
  bool __pyx_v_outR;
  bool __pyx_v_outT;
  bool __pyx_v_outEnh;
  bool __pyx_v_outP1;
  bool __pyx_v_outP2;
  bool __pyx_v_outGen;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("WaveSweep (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_paramStr,&__pyx_mstate_global->__pyx_n_u_valuesP1,&__pyx_mstate_global->__pyx_n_u_valuesP2,&__pyx_mstate_global->__pyx_n_u_layerNr,&__pyx_mstate_global->__pyx_n_u_layerZ,&__pyx_mstate_global->__pyx_n_u_outPwr,&__pyx_mstate_global->__pyx_n_u_outR,&__pyx_mstate_global->__pyx_n_u_outT,&__pyx_mstate_global->__pyx_n_u_outEnh,&__pyx_mstate_global->__pyx_n_u_outP1,&__pyx_mstate_global->__pyx_n_u_outP2,&__pyx_mstate_global->__pyx_n_u_outGen,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1909, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case 12:
        values[11] = __Pyx_ArgRef_FASTCALL(__pyx_args, 11);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[11])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 11:
        values[10] = __Pyx_ArgRef_FASTCALL(__pyx_args, 10);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 10:
        values[9] = __Pyx_ArgRef_FASTCALL(__pyx_args, 9);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  9:
        values[8] = __Pyx_ArgRef_FASTCALL(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "WaveSweep", 0) < (0)) __PYX_ERR(0, 1909, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("WaveSweep", 0, 3, 12, i); __PYX_ERR(0, 1909, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case 12:
        values[11] = __Pyx_ArgRef_FASTCALL(__pyx_args, 11);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[11])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 11:
        values[10] = __Pyx_ArgRef_FASTCALL(__pyx_args, 10);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[10])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case 10:
        values[9] = __Pyx_ArgRef_FASTCALL(__pyx_args, 9);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[9])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  9:
        values[8] = __Pyx_ArgRef_FASTCALL(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_FASTCALL(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_FASTCALL(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_FASTCALL(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_FASTCALL(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_FASTCALL(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 1909, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1909, __pyx_L3_error)
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1909, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1909, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_paramStr = ((PyObject*)values[0]);
    __pyx_v_valuesP1 = ((PyArrayObject *)values[1]);
    __pyx_v_valuesP2 = ((PyArrayObject *)values[2]);
    if (values[3]) {
      __pyx_v_layerNr = __Pyx_PyLong_As_int(values[3]); if (unlikely((__pyx_v_layerNr == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1911, __pyx_L3_error)
    } else {
      __pyx_v_layerNr = ((int)0);
    }
    if (values[4]) {
      __pyx_v_layerZ = __Pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_layerZ == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1911, __pyx_L3_error)
    } else {
      __pyx_v_layerZ = ((double)0.0);
    }
    if (values[5]) {
      __pyx_v_outPwr = __Pyx_PyObject_IsTrue(values[5]); if (unlikely((__pyx_v_outPwr == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1911, __pyx_L3_error)
    } else {

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1911
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] valuesP1, \
 *             np.ndarray[double, ndim = 1] valuesP2,
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \             # <<<<<<<<<<<<<<
 *             bool outR = False, bool outT = False, bool outEnh = False, \
 *             bool outP1 = True, bool outP2 = True, bool outGen = True):
*/
      __pyx_v_outPwr = ((bool)1);
    }
    if (values[6]) {
      __pyx_v_outR = __Pyx_PyObject_IsTrue(values[6]); if (unlikely((__pyx_v_outR == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1912, __pyx_L3_error)
    } else {

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1912
 *             np.ndarray[double, ndim = 1] valuesP2,
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \
 *             bool outR = False, bool outT = False, bool outEnh = False, \             # <<<<<<<<<<<<<<
 *             bool outP1 = True, bool outP2 = True, bool outGen = True):
 *         """WaveSweep(paramStr, valuesP1, valuesP2, layerNr = 0, layerZ = 0.0, outPwr = True, outR = False, outT = False, outEnh = False, outP1 = True, outP2 = True, outGen = True)
*/
      __pyx_v_outR = ((bool)0);
    }
    if (values[7]) {
      __pyx_v_outT = __Pyx_PyObject_IsTrue(values[7]); if (unlikely((__pyx_v_outT == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1912, __pyx_L3_error)
    } else {
      __pyx_v_outT = ((bool)0);
    }
    if (values[8]) {
      __pyx_v_outEnh = __Pyx_PyObject_IsTrue(values[8]); if (unlikely((__pyx_v_outEnh == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1912, __pyx_L3_error)
    } else {
      __pyx_v_outEnh = ((bool)0);
    }
    if (values[9]) {
      __pyx_v_outP1 = __Pyx_PyObject_IsTrue(values[9]); if (unlikely((__pyx_v_outP1 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1913, __pyx_L3_error)
    } else {

      /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1913
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \
 *             bool outR = False, bool outT = False, bool outEnh = False, \
 *             bool outP1 = True, bool outP2 = True, bool outGen = True):             # <<<<<<<<<<<<<<
 *         """WaveSweep(paramStr, valuesP1, valuesP2, layerNr = 0, layerZ = 0.0, outPwr = True, outR = False, outT = False, outEnh = False, outP1 = True, outP2 = True, outGen = True)
 * 
*/
      __pyx_v_outP1 = ((bool)1);
    }
    if (values[10]) {
      __pyx_v_outP2 = __Pyx_PyObject_IsTrue(values[10]); if (unlikely((__pyx_v_outP2 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1913, __pyx_L3_error)
    } else {
      __pyx_v_outP2 = ((bool)1);
    }
    if (values[11]) {
      __pyx_v_outGen = __Pyx_PyObject_IsTrue(values[11]); if (unlikely((__pyx_v_outGen == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1913, __pyx_L3_error)
    } else {
      __pyx_v_outGen = ((bool)1);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("WaveSweep", 0, 3, 12, __pyx_nargs); __PYX_ERR(0, 1909, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.WaveSweep", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_paramStr), (&PyUnicode_Type), 1, "paramStr", 1))) __PYX_ERR(0, 1909, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_valuesP1), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "valuesP1", 0))) __PYX_ERR(0, 1909, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_valuesP2), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "valuesP2", 0))) __PYX_ERR(0, 1910, __pyx_L1_error)
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_18WaveSweep(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self), __pyx_v_paramStr, __pyx_v_valuesP1, __pyx_v_valuesP2, __pyx_v_layerNr, __pyx_v_layerZ, __pyx_v_outPwr, __pyx_v_outR, __pyx_v_outT, __pyx_v_outEnh, __pyx_v_outP1, __pyx_v_outP2, __pyx_v_outGen);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1909
 *         return (res.first, res.second)
 * 
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] valuesP1, \             # <<<<<<<<<<<<<<
 *             np.ndarray[double, ndim = 1] valuesP2,
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \
*/

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_18WaveSweep(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyObject *__pyx_v_paramStr, PyArrayObject *__pyx_v_valuesP1, PyArrayObject *__pyx_v_valuesP2, int __pyx_v_layerNr, double __pyx_v_layerZ, bool __pyx_v_outPwr, bool __pyx_v_outR, bool __pyx_v_outT, bool __pyx_v_outEnh, bool __pyx_v_outP1, bool __pyx_v_outP2, bool __pyx_v_outGen) {
  TMM::WaveSweepResultSecondOrderNLTMM *__pyx_v_resCpp;
  int __pyx_v_outmask;
  enum TMM::TMMParam __pyx_v_param;
  int __pyx_v_paramLayer;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *__pyx_v_res = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_valuesP1;
  __Pyx_Buffer __pyx_pybuffer_valuesP1;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_valuesP2;
  __Pyx_Buffer __pyx_pybuffer_valuesP2;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  enum TMM::TMMParam __pyx_t_5;
  eigency::Map<Eigen::ArrayXd>  __pyx_t_6;
  eigency::Map<Eigen::ArrayXd>  __pyx_t_7;
  std::unique_ptr<TMM::WaveSweepResultSecondOrderNLTMM>  __pyx_t_8;
  size_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("WaveSweep", 0);
  __pyx_pybuffer_valuesP1.pybuffer.buf = NULL;
  __pyx_pybuffer_valuesP1.refcount = 0;
  __pyx_pybuffernd_valuesP1.data = NULL;
  __pyx_pybuffernd_valuesP1.rcbuffer = &__pyx_pybuffer_valuesP1;
  __pyx_pybuffer_valuesP2.pybuffer.buf = NULL;
  __pyx_pybuffer_valuesP2.refcount = 0;
  __pyx_pybuffernd_valuesP2.data = NULL;
  __pyx_pybuffernd_valuesP2.rcbuffer = &__pyx_pybuffer_valuesP2;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_valuesP1.rcbuffer->pybuffer, (PyObject*)__pyx_v_valuesP1, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1909, __pyx_L1_error)
  }
  __pyx_pybuffernd_valuesP1.diminfo[0].strides = __pyx_pybuffernd_valuesP1.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_valuesP1.diminfo[0].shape = __pyx_pybuffernd_valuesP1.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_valuesP2.rcbuffer->pybuffer, (PyObject*)__pyx_v_valuesP2, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1909, __pyx_L1_error)
  }
  __pyx_pybuffernd_valuesP2.diminfo[0].strides = __pyx_pybuffernd_valuesP2.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_valuesP2.diminfo[0].shape = __pyx_pybuffernd_valuesP2.rcbuffer->pybuffer.shape[0];

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1965
 * 
 *         cdef WaveSweepResultSecondOrderNLTMMCpp *resCpp;
 *         cdef int outmask = 0             # <<<<<<<<<<<<<<
 *         if outP1:
 *             outmask |= SWEEP_P1
*/
  __pyx_v_outmask = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1966
 *         cdef WaveSweepResultSecondOrderNLTMMCpp *resCpp;
 *         cdef int outmask = 0
 *         if outP1:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_P1
 *         if outP2:
*/
  __pyx_t_1 = (__pyx_v_outP1 != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1967
 *         cdef int outmask = 0
 *         if outP1:
 *             outmask |= SWEEP_P1             # <<<<<<<<<<<<<<
 *         if outP2:
 *             outmask |= SWEEP_P2
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_P1);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1966
 *         cdef WaveSweepResultSecondOrderNLTMMCpp *resCpp;
 *         cdef int outmask = 0
 *         if outP1:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_P1
 *         if outP2:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1968
 *         if outP1:
 *             outmask |= SWEEP_P1
 *         if outP2:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_P2
 *         if outGen:
*/
  __pyx_t_1 = (__pyx_v_outP2 != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1969
 *             outmask |= SWEEP_P1
 *         if outP2:
 *             outmask |= SWEEP_P2             # <<<<<<<<<<<<<<
 *         if outGen:
 *             outmask |= SWEEP_GEN
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_P2);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1968
 *         if outP1:
 *             outmask |= SWEEP_P1
 *         if outP2:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_P2
 *         if outGen:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1970
 *         if outP2:
 *             outmask |= SWEEP_P2
 *         if outGen:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_GEN
 *         if outPwr:
*/
  __pyx_t_1 = (__pyx_v_outGen != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1971
 *             outmask |= SWEEP_P2
 *         if outGen:
 *             outmask |= SWEEP_GEN             # <<<<<<<<<<<<<<
 *         if outPwr:
 *             outmask |= SWEEP_PWRFLOWS
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_GEN);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1970
 *         if outP2:
 *             outmask |= SWEEP_P2
 *         if outGen:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_GEN
 *         if outPwr:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1972
 *         if outGen:
 *             outmask |= SWEEP_GEN
 *         if outPwr:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_PWRFLOWS
 *         if outEnh:
*/
  __pyx_t_1 = (__pyx_v_outPwr != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1973
 *             outmask |= SWEEP_GEN
 *         if outPwr:
 *             outmask |= SWEEP_PWRFLOWS             # <<<<<<<<<<<<<<
 *         if outEnh:
 *             outmask |= SWEEP_ENH
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_PWRFLOWS);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1972
 *         if outGen:
 *             outmask |= SWEEP_GEN
 *         if outPwr:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_PWRFLOWS
 *         if outEnh:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1974
 *         if outPwr:
 *             outmask |= SWEEP_PWRFLOWS
 *         if outEnh:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_ENH
 *         if outR:
*/
  __pyx_t_1 = (__pyx_v_outEnh != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1975
 *             outmask |= SWEEP_PWRFLOWS
 *         if outEnh:
 *             outmask |= SWEEP_ENH             # <<<<<<<<<<<<<<
 *         if outR:
 *             outmask |= SWEEP_R
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_ENH);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1974
 *         if outPwr:
 *             outmask |= SWEEP_PWRFLOWS
 *         if outEnh:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_ENH
 *         if outR:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1976
 *         if outEnh:
 *             outmask |= SWEEP_ENH
 *         if outR:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_R
 *         if outT:
*/
  __pyx_t_1 = (__pyx_v_outR != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1977
 *             outmask |= SWEEP_ENH
 *         if outR:
 *             outmask |= SWEEP_R             # <<<<<<<<<<<<<<
 *         if outT:
 *             outmask |= SWEEP_T
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_R);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1976
 *         if outEnh:
 *             outmask |= SWEEP_ENH
 *         if outR:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_R
 *         if outT:
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1978
 *         if outR:
 *             outmask |= SWEEP_R
 *         if outT:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_T
 * 
*/
  __pyx_t_1 = (__pyx_v_outT != 0);
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1979
 *             outmask |= SWEEP_R
 *         if outT:
 *             outmask |= SWEEP_T             # <<<<<<<<<<<<<<
 * 
 *         cdef TMMParamCpp param;
*/
    __pyx_v_outmask = (__pyx_v_outmask | TMM::SWEEP_T);

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1978
 *         if outR:
 *             outmask |= SWEEP_R
 *         if outT:             # <<<<<<<<<<<<<<
 *             outmask |= SWEEP_T
 * 
*/
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1982
 * 
 *         cdef TMMParamCpp param;
 *         cdef int paramLayer = -1;             # <<<<<<<<<<<<<<
 *         if (paramStr.startswith("d_")):
 *             param = PARAM_LAYER_D
*/
  __pyx_v_paramLayer = -1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1983
 *         cdef TMMParamCpp param;
 *         cdef int paramLayer = -1;
 *         if (paramStr.startswith("d_")):             # <<<<<<<<<<<<<<
 *             param = PARAM_LAYER_D
 *             paramLayer = int(paramStr[2:])
*/
  if (unlikely(__pyx_v_paramStr == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "startswith");
    __PYX_ERR(0, 1983, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyUnicode_Tailmatch(__pyx_v_paramStr, __pyx_mstate_global->__pyx_n_u_d_2, 0, PY_SSIZE_T_MAX, -1); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1983, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1984
 *         cdef int paramLayer = -1;
 *         if (paramStr.startswith("d_")):
 *             param = PARAM_LAYER_D             # <<<<<<<<<<<<<<
 *             paramLayer = int(paramStr[2:])
 *         else:
*/
    __pyx_v_param = TMM::TMMParam::PARAM_LAYER_D;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1985
 *         if (paramStr.startswith("d_")):
 *             param = PARAM_LAYER_D
 *             paramLayer = int(paramStr[2:])             # <<<<<<<<<<<<<<
 *         else:
 *             param = TmmParamFromStr(paramStr)
*/
    if (unlikely(__pyx_v_paramStr == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 1985, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyUnicode_Substring(__pyx_v_paramStr, 2, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1985, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1985, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyLong_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1985, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_paramLayer = __pyx_t_4;

    /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1983
 *         cdef TMMParamCpp param;
 *         cdef int paramLayer = -1;
 *         if (paramStr.startswith("d_")):             # <<<<<<<<<<<<<<
 *             param = PARAM_LAYER_D
 *             paramLayer = int(paramStr[2:])
*/
    goto __pyx_L10;
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1987
 *             paramLayer = int(paramStr[2:])
 *         else:
 *             param = TmmParamFromStr(paramStr)             # <<<<<<<<<<<<<<
 * 
 *         resCpp = self._thisptr.WaveSweep(param, Map[ArrayXd](valuesP1), Map[ArrayXd](valuesP2), outmask, paramLayer, layerNr, layerZ).release()
*/
  /*else*/ {
    __pyx_t_5 = __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_TmmParamFromStr(__pyx_v_paramStr); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1987, __pyx_L1_error)
    __pyx_v_param = __pyx_t_5;
  }
  __pyx_L10:;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1989
 *             param = TmmParamFromStr(paramStr)
 * 
 *         resCpp = self._thisptr.WaveSweep(param, Map[ArrayXd](valuesP1), Map[ArrayXd](valuesP2), outmask, paramLayer, layerNr, layerZ).release()             # <<<<<<<<<<<<<<
 *         res = _WaveSweepResultSecondOrderNLTMM()
 *         res._Init(resCpp);
*/
  try {
    __pyx_t_6 = eigency::Map<Eigen::ArrayXd> (((PyArrayObject *)__pyx_v_valuesP1));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1989, __pyx_L1_error)
  }
  try {
    __pyx_t_7 = eigency::Map<Eigen::ArrayXd> (((PyArrayObject *)__pyx_v_valuesP2));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1989, __pyx_L1_error)
  }
  try {
    __pyx_t_8 = __pyx_v_self->_thisptr->WaveSweep(__pyx_v_param, __pyx_t_6, __pyx_t_7, __pyx_v_outmask, __pyx_v_paramLayer, __pyx_v_layerNr, __pyx_v_layerZ);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 1989, __pyx_L1_error)
  }
  __pyx_v_resCpp = __pyx_t_8.release();

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1990
 * 
 *         resCpp = self._thisptr.WaveSweep(param, Map[ArrayXd](valuesP1), Map[ArrayXd](valuesP2), outmask, paramLayer, layerNr, layerZ).release()
 *         res = _WaveSweepResultSecondOrderNLTMM()             # <<<<<<<<<<<<<<
 *         res._Init(resCpp);
 *         return res
*/
  __pyx_t_2 = NULL;
  __pyx_t_9 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM, __pyx_callargs+__pyx_t_9, (1-__pyx_t_9) | (__pyx_t_9*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1990, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_3);
  }
  __pyx_v_res = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1991
 *         resCpp = self._thisptr.WaveSweep(param, Map[ArrayXd](valuesP1), Map[ArrayXd](valuesP2), outmask, paramLayer, layerNr, layerZ).release()
 *         res = _WaveSweepResultSecondOrderNLTMM()
 *         res._Init(resCpp);             # <<<<<<<<<<<<<<
 *         return res
 * 
*/
  __pyx_t_3 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)__pyx_v_res->__pyx_vtab)->_Init(__pyx_v_res, __pyx_v_resCpp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1991, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1992
 *         res = _WaveSweepResultSecondOrderNLTMM()
 *         res._Init(resCpp);
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     def WaveGetFields2D(self, np.ndarray[double, ndim = 1] zs, \
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_res);
  __pyx_r = ((PyObject *)__pyx_v_res);
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1909
 *         return (res.first, res.second)
 * 
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] valuesP1, \             # <<<<<<<<<<<<<<
 *             np.ndarray[double, ndim = 1] valuesP2,
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_valuesP1.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_valuesP2.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.WaveSweep", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_valuesP1.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_valuesP2.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1994
 *         return res
 * 
 *     def WaveGetFields2D(self, np.ndarray[double, ndim = 1] zs, \             # <<<<<<<<<<<<<<
 *                         np.ndarray[double, ndim = 1] xs, str dirStr = "total"):
 *         """WaveGetFields2D(zs, xs, dirStr = "total")
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_21WaveGetFields2D(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_20WaveGetFields2D, "WaveGetFields2D(zs, xs, dirStr = \"total\")\n\n        Calculates 2D electric and magnetic fields of the beam propagating in the\n        structure. Analogous to ``GetFields2D`` of plane waves. Only for the\n        calculation of the generated beam fields. For pump beams use the\n        same method of ``NonlinearTMM``.\n\n        Parameters\n        ----------\n        zs : ndarray of floats\n            Points on z-axis where to calculate the fields. The beginning of the\n            first layer is at z = 0.\n        xs : ndarray of floats\n            Points on x-axis where to calculate the fields.\n        dirStr : {'total', 'forward', 'backward'}\n            Specifies the components of the output fields.\n\n        Returns\n        -------\n        :class:`_FieldsZX`\n            Helper class to store electric and magnetic fields in regular grid.\n\n        ");
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_21WaveGetFields2D = {"WaveGetFields2D", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_21WaveGetFields2D, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_20WaveGetFields2D};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_21WaveGetFields2D(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyArrayObject *__pyx_v_zs = 0;
  PyArrayObject *__pyx_v_xs = 0;
  PyObject *__pyx_v_dirStr = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("WaveGetFields2D (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_zs,&__pyx_mstate_global->__pyx_n_u_xs,&__pyx_mstate_global->__pyx_n_u_dirStr,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 1994, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1994, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1994, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1994, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "WaveGetFields2D", 0) < (0)) __PYX_ERR(0, 1994, __pyx_L3_error)
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject*)__pyx_mstate_global->__pyx_n_u_total));
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("WaveGetFields2D", 0, 2, 3, i); __PYX_ERR(0, 1994, __pyx_L3_error) }
      }
    } else {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 1994, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 1994, __pyx_L3_error)
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 1994, __pyx_L3_error)
        break;
        default: goto __pyx_L5_argtuple_error;
      }
      if (!values[2]) values[2] = __Pyx_NewRef(((PyObject*)__pyx_mstate_global->__pyx_n_u_total));
    }
    __pyx_v_zs = ((PyArrayObject *)values[0]);
    __pyx_v_xs = ((PyArrayObject *)values[1]);
    __pyx_v_dirStr = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("WaveGetFields2D", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 1994, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.WaveGetFields2D", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_zs), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "zs", 0))) __PYX_ERR(0, 1994, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_xs), __pyx_mstate_global->__pyx_ptype_5numpy_ndarray, 1, "xs", 0))) __PYX_ERR(0, 1995, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dirStr), (&PyUnicode_Type), 1, "dirStr", 1))) __PYX_ERR(0, 1995, __pyx_L1_error)
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_20WaveGetFields2D(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self), __pyx_v_zs, __pyx_v_xs, __pyx_v_dirStr);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_20WaveGetFields2D(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyArrayObject *__pyx_v_zs, PyArrayObject *__pyx_v_xs, PyObject *__pyx_v_dirStr) {
  TMM::FieldsZX *__pyx_v_resCpp;
  enum TMM::WaveDirection __pyx_v_direction;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *__pyx_v_res = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_xs;
  __Pyx_Buffer __pyx_pybuffer_xs;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_zs;
  __Pyx_Buffer __pyx_pybuffer_zs;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  enum TMM::WaveDirection __pyx_t_1;
  eigency::Map<Eigen::ArrayXd>  __pyx_t_2;
  eigency::Map<Eigen::ArrayXd>  __pyx_t_3;
  std::unique_ptr<TMM::FieldsZX>  __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("WaveGetFields2D", 0);
  __pyx_pybuffer_zs.pybuffer.buf = NULL;
  __pyx_pybuffer_zs.refcount = 0;
  __pyx_pybuffernd_zs.data = NULL;
  __pyx_pybuffernd_zs.rcbuffer = &__pyx_pybuffer_zs;
  __pyx_pybuffer_xs.pybuffer.buf = NULL;
  __pyx_pybuffer_xs.refcount = 0;
  __pyx_pybuffernd_xs.data = NULL;
  __pyx_pybuffernd_xs.rcbuffer = &__pyx_pybuffer_xs;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_zs.rcbuffer->pybuffer, (PyObject*)__pyx_v_zs, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1994, __pyx_L1_error)
  }
  __pyx_pybuffernd_zs.diminfo[0].strides = __pyx_pybuffernd_zs.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_zs.diminfo[0].shape = __pyx_pybuffernd_zs.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_xs.rcbuffer->pybuffer, (PyObject*)__pyx_v_xs, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1994, __pyx_L1_error)
  }
  __pyx_pybuffernd_xs.diminfo[0].strides = __pyx_pybuffernd_xs.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_xs.diminfo[0].shape = __pyx_pybuffernd_xs.rcbuffer->pybuffer.shape[0];

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2021
 * 
 *         cdef FieldsZXCpp *resCpp;
 *         cdef WaveDirectionCpp direction = WaveDirectionFromStr(dirStr)             # <<<<<<<<<<<<<<
 *         resCpp = self._thisptr.WaveGetFields2D(Map[ArrayXd](zs), Map[ArrayXd](xs), direction).release()
 *         res = _FieldsZX()
*/
  __pyx_t_1 = __pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_WaveDirectionFromStr(__pyx_v_dirStr); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2021, __pyx_L1_error)
  __pyx_v_direction = __pyx_t_1;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2022
 *         cdef FieldsZXCpp *resCpp;
 *         cdef WaveDirectionCpp direction = WaveDirectionFromStr(dirStr)
 *         resCpp = self._thisptr.WaveGetFields2D(Map[ArrayXd](zs), Map[ArrayXd](xs), direction).release()             # <<<<<<<<<<<<<<
 *         res = _FieldsZX()
 *         res._Init(resCpp)
*/
  try {
    __pyx_t_2 = eigency::Map<Eigen::ArrayXd> (((PyArrayObject *)__pyx_v_zs));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2022, __pyx_L1_error)
  }
  try {
    __pyx_t_3 = eigency::Map<Eigen::ArrayXd> (((PyArrayObject *)__pyx_v_xs));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2022, __pyx_L1_error)
  }
  try {
    __pyx_t_4 = __pyx_v_self->_thisptr->WaveGetFields2D(__pyx_t_2, __pyx_t_3, __pyx_v_direction);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2022, __pyx_L1_error)
  }
  __pyx_v_resCpp = __pyx_t_4.release();

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2023
 *         cdef WaveDirectionCpp direction = WaveDirectionFromStr(dirStr)
 *         resCpp = self._thisptr.WaveGetFields2D(Map[ArrayXd](zs), Map[ArrayXd](xs), direction).release()
 *         res = _FieldsZX()             # <<<<<<<<<<<<<<
 *         res._Init(resCpp)
 *         return res
*/
  __pyx_t_6 = NULL;
  __pyx_t_7 = 1;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
    __pyx_t_5 = __Pyx_PyObject_FastCall((PyObject*)__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2023, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_5);
  }
  __pyx_v_res = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2024
 *         resCpp = self._thisptr.WaveGetFields2D(Map[ArrayXd](zs), Map[ArrayXd](xs), direction).release()
 *         res = _FieldsZX()
 *         res._Init(resCpp)             # <<<<<<<<<<<<<<
 *         return res
 * 
*/
  __pyx_t_5 = ((struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)__pyx_v_res->__pyx_vtab)->_Init(__pyx_v_res, __pyx_v_resCpp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2024, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2025
 *         res = _FieldsZX()
 *         res._Init(resCpp)
 *         return res             # <<<<<<<<<<<<<<
 * 
 *     # Getter
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_res);
  __pyx_r = ((PyObject *)__pyx_v_res);
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1994
 *         return res
 * 
 *     def WaveGetFields2D(self, np.ndarray[double, ndim = 1] zs, \             # <<<<<<<<<<<<<<
 *                         np.ndarray[double, ndim = 1] xs, str dirStr = "total"):
 *         """WaveGetFields2D(zs, xs, dirStr = "total")
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_xs.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_zs.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.WaveGetFields2D", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_xs.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_zs.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_res);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2030
 *     #---------------------------------------------------------------------------
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def deltaWlSpdc(self):
 *         return self._thisptr.GetDeltaWlSpdc()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11deltaWlSpdc_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11deltaWlSpdc_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11deltaWlSpdc___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11deltaWlSpdc___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2032
 *     @property
 *     def deltaWlSpdc(self):
 *         return self._thisptr.GetDeltaWlSpdc()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetDeltaWlSpdc();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2032, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2030
 *     #---------------------------------------------------------------------------
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def deltaWlSpdc(self):
 *         return self._thisptr.GetDeltaWlSpdc()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.deltaWlSpdc.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2045
 *     #---------------------------------------------------------------------------
 * 
 *     @deltaWlSpdc.setter             # <<<<<<<<<<<<<<
 *     def deltaWlSpdc(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetDeltaWlSpdc(<double>value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11deltaWlSpdc_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11deltaWlSpdc_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11deltaWlSpdc_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11deltaWlSpdc_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2047
 *     @deltaWlSpdc.setter
 *     def deltaWlSpdc(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetDeltaWlSpdc(<double>value)             # <<<<<<<<<<<<<<
 * 
 *     @solidAngleSpdc.setter
*/
  __pyx_t_1 = __Pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2047, __pyx_L1_error)
  try {
    __pyx_v_self->_thisptr->SetDeltaWlSpdc(((double)__pyx_t_1));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2047, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2045
 *     #---------------------------------------------------------------------------
 * 
 *     @deltaWlSpdc.setter             # <<<<<<<<<<<<<<
 *     def deltaWlSpdc(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetDeltaWlSpdc(<double>value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.deltaWlSpdc.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2034
 *         return self._thisptr.GetDeltaWlSpdc()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def solidAngleSpdc(self):
 *         return self._thisptr.GetSolidAngleSpdc()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14solidAngleSpdc_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14solidAngleSpdc_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14solidAngleSpdc___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14solidAngleSpdc___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2036
 *     @property
 *     def solidAngleSpdc(self):
 *         return self._thisptr.GetSolidAngleSpdc()             # <<<<<<<<<<<<<<
 * 
 *     @property
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetSolidAngleSpdc();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2036, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2034
 *         return self._thisptr.GetDeltaWlSpdc()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def solidAngleSpdc(self):
 *         return self._thisptr.GetSolidAngleSpdc()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.solidAngleSpdc.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2049
 *         self._thisptr.SetDeltaWlSpdc(<double>value)
 * 
 *     @solidAngleSpdc.setter             # <<<<<<<<<<<<<<
 *     def solidAngleSpdc(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetSolidAngleSpdc(<double>value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14solidAngleSpdc_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14solidAngleSpdc_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14solidAngleSpdc_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14solidAngleSpdc_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2051
 *     @solidAngleSpdc.setter
 *     def solidAngleSpdc(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetSolidAngleSpdc(<double>value)             # <<<<<<<<<<<<<<
 * 
 *     @deltaThetaSpdc.setter
*/
  __pyx_t_1 = __Pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2051, __pyx_L1_error)
  try {
    __pyx_v_self->_thisptr->SetSolidAngleSpdc(((double)__pyx_t_1));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2051, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2049
 *         self._thisptr.SetDeltaWlSpdc(<double>value)
 * 
 *     @solidAngleSpdc.setter             # <<<<<<<<<<<<<<
 *     def solidAngleSpdc(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetSolidAngleSpdc(<double>value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.solidAngleSpdc.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2038
 *         return self._thisptr.GetSolidAngleSpdc()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def deltaThetaSpdc(self):
 *         return self._thisptr.GetDeltaThetaSpdc()
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14deltaThetaSpdc_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14deltaThetaSpdc_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14deltaThetaSpdc___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14deltaThetaSpdc___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2040
 *     @property
 *     def deltaThetaSpdc(self):
 *         return self._thisptr.GetDeltaThetaSpdc()             # <<<<<<<<<<<<<<
 * 
 *     # Setter
*/
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->_thisptr->GetDeltaThetaSpdc();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2040, __pyx_L1_error)
  }
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2038
 *         return self._thisptr.GetSolidAngleSpdc()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def deltaThetaSpdc(self):
 *         return self._thisptr.GetDeltaThetaSpdc()
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.deltaThetaSpdc.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2053
 *         self._thisptr.SetSolidAngleSpdc(<double>value)
 * 
 *     @deltaThetaSpdc.setter             # <<<<<<<<<<<<<<
 *     def deltaThetaSpdc(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetDeltaThetaSpdc(<double>value)
*/

/* Python wrapper */
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14deltaThetaSpdc_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14deltaThetaSpdc_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14deltaThetaSpdc_2__set__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14deltaThetaSpdc_2__set__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2055
 *     @deltaThetaSpdc.setter
 *     def deltaThetaSpdc(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetDeltaThetaSpdc(<double>value)             # <<<<<<<<<<<<<<
*/
  __pyx_t_1 = __Pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2055, __pyx_L1_error)
  try {
    __pyx_v_self->_thisptr->SetDeltaThetaSpdc(((double)__pyx_t_1));
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 2055, __pyx_L1_error)
  }

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":2053
 *         self._thisptr.SetSolidAngleSpdc(<double>value)
 * 
 *     @deltaThetaSpdc.setter             # <<<<<<<<<<<<<<
 *     def deltaThetaSpdc(self, value):  # @DuplicatedSignature
 *         self._thisptr.SetDeltaThetaSpdc(<double>value)
*/

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.deltaThetaSpdc.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1694
 *     cdef SecondOrderNLTMMCpp *_thisptr
 *     cdef list materialsCache
 *     cdef readonly object P1, P2, Gen             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, str mode = "sfg", **kwargs):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2P1_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2P1_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2P1___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2P1___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->P1);
  __pyx_r = __pyx_v_self->P1;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2P2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2P2_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2P2___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2P2___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->P2);
  __pyx_r = __pyx_v_self->P2;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_3Gen_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_3Gen_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_3Gen___get__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_3Gen___get__(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->Gen);
  __pyx_r = __pyx_v_self->Gen;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_23__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_23__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_23__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_23__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_22__reduce_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

/* Python wrapper */
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_25__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_25__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_25__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_25__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(2, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < (0)) __PYX_ERR(2, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(2, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(2, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(2, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_24__setstate_cython__(((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(((PyObject *)(((PyTypeObject*)PyExc_TypeError))), __pyx_mstate_global->__pyx_kp_u_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(2, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor;

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *)o);
  p->__pyx_vtab = __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor;
  p->_parent = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_1__cinit__(o, __pyx_mstate_global->__pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor(PyObject *o) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_parent);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->_parent) {
    e = (*v)(p->_parent, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *)o;
  tmp = ((PyObject*)p->_parent);
  p->_parent = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor[] = {
  {"Update", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_5Update, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_4Update},
  {"GetChi2Tensor", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_7GetChi2Tensor, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_6GetChi2Tensor},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor},
  {Py_tp_doc, (void *)PyDoc_STR("_Chi2Tensor()\n\n    This is a helper class for Material to accommodate the second-order\n    susceptibility tensor. Allows setting nonlinearities by chi2- and by d-values.\n    The initial tensor can be rotated around all three axes.\n\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor},
  {Py_tp_clear, (void *)__pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython._Chi2Tensor",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""_Chi2Tensor", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("_Chi2Tensor()\n\n    This is a helper class for Material to accommodate the second-order\n    susceptibility tensor. Allows setting nonlinearities by chi2- and by d-values.\n    The initial tensor can be rotated around all three axes.\n\n    "), /*tp_doc*/
  __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor, /*tp_traverse*/
  __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython_Material(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *)o);
  p->chi2 = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython_Material(PyObject *o) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython_Material) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->chi2);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython_Material(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->chi2) {
    e = (*v)(p->chi2, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython_Material(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material *)o;
  tmp = ((PyObject*)p->chi2);
  p->chi2 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_chi2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_4chi2_1__get__(o);
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython_Material[] = {
  {"GetN", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_7GetN, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_6GetN},
  {"IsNonlinear", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_9IsNonlinear, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_8IsNonlinear},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython_Material[] = {
  {"chi2", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_chi2, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_Material_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython_Material},
  {Py_tp_call, (void *)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_5__call__},
  {Py_tp_doc, (void *)PyDoc_STR("Material(wls, ns)\n\n    This class describes the optical parameters of a nonlinear medium. The default\n    constructor takes arrays of wavelengths and complex refractive indices and\n    performs linear interpolation. As a shortcut, ``__call__`` is defined as an\n    alias for ``GetN``.\n\n    Parameters\n    ----------\n    wls : ndarray of floats\n        Array of wavelengths (m)\n    ns : ndarray of complex floats\n        Corresponding complex refractive indices to the wls array.\n\n    Attributes\n    ----------\n    chi2 : :class:`_Chi2Tensor`\n        Instance of the :class:`_Chi2Tensor` helper class to store second-order nonlinearity tensor.\n\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython_Material},
  {Py_tp_clear, (void *)__pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython_Material},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython_Material},
  {Py_tp_getset, (void *)__pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython_Material},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython_Material},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_Material_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython.Material",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_Material_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_Material = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""Material", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_Material), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython_Material, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_5__call__, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("Material(wls, ns)\n\n    This class describes the optical parameters of a nonlinear medium. The default\n    constructor takes arrays of wavelengths and complex refractive indices and\n    performs linear interpolation. As a shortcut, ``__call__`` is defined as an\n    alias for ``GetN``.\n\n    Parameters\n    ----------\n    wls : ndarray of floats\n        Array of wavelengths (m)\n    ns : ndarray of complex floats\n        Corresponding complex refractive indices to the wls array.\n\n    Attributes\n    ----------\n    chi2 : :class:`_Chi2Tensor`\n        Instance of the :class:`_Chi2Tensor` helper class to store second-order nonlinearity tensor.\n\n    "), /*tp_doc*/
  __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython_Material, /*tp_traverse*/
  __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython_Material, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython_Material, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython_Material, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython_Material, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave;

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)o);
  p->__pyx_vtab = __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave;
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1__cinit__(o, __pyx_mstate_global->__pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_waveType(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8waveType_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_waveType(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8waveType_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_pwr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3pwr_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_pwr(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3pwr_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_overrideE0(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_10overrideE0_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_overrideE0(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_10overrideE0_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_E0(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2E0_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_E0(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2E0_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_w0(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2w0_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_w0(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2w0_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_Ly(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2Ly_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_Ly(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_2Ly_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_a(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1a_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_a(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_1a_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_nPointsInteg(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12nPointsInteg_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_nPointsInteg(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12nPointsInteg_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_maxX(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4maxX_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_maxX(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4maxX_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_dynamicMaxX(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_11dynamicMaxX_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_dynamicMaxX(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_11dynamicMaxX_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_dynamicMaxXCoef(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_15dynamicMaxXCoef_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_dynamicMaxXCoef(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_15dynamicMaxXCoef_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_dynamicMaxXAddition(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_19dynamicMaxXAddition_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_dynamicMaxXAddition(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_19dynamicMaxXAddition_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_maxPhi(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6maxPhi_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_maxPhi(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6maxPhi_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_xRange(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_6xRange_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_betas(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_5betas_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_phis(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4phis_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_kxs(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3kxs_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_kzs(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_3kzs_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_fieldProfile(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_12fieldProfile_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_expansionCoefsKx(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_16expansionCoefsKx_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_beamArea(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_8beamArea_1__get__(o);
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave[] = {
  {"SetParams", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_5SetParams, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_4SetParams},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave[] = {
  {"waveType", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_waveType, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_waveType, 0, 0},
  {"pwr", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_pwr, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_pwr, 0, 0},
  {"overrideE0", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_overrideE0, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_overrideE0, 0, 0},
  {"E0", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_E0, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_E0, 0, 0},
  {"w0", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_w0, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_w0, 0, 0},
  {"Ly", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_Ly, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_Ly, 0, 0},
  {"a", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_a, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_a, 0, 0},
  {"nPointsInteg", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_nPointsInteg, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_nPointsInteg, 0, 0},
  {"maxX", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_maxX, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_maxX, 0, 0},
  {"dynamicMaxX", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_dynamicMaxX, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_dynamicMaxX, 0, 0},
  {"dynamicMaxXCoef", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_dynamicMaxXCoef, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_dynamicMaxXCoef, 0, 0},
  {"dynamicMaxXAddition", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_dynamicMaxXAddition, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_dynamicMaxXAddition, 0, 0},
  {"maxPhi", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_maxPhi, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_maxPhi, 0, 0},
  {"xRange", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_xRange, 0, 0, 0},
  {"betas", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_betas, 0, 0, 0},
  {"phis", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_phis, 0, 0, 0},
  {"kxs", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_kxs, 0, 0, 0},
  {"kzs", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_kzs, 0, 0, 0},
  {"fieldProfile", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_fieldProfile, 0, 0, 0},
  {"expansionCoefsKx", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_expansionCoefsKx, 0, 0, 0},
  {"beamArea", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_beamArea, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave},
  {Py_tp_doc, (void *)PyDoc_STR("_Wave()\n\n    This is a helper class for using non plane waves as an input.\n\n    Attributes\n    ----------\n    waveType : str\n        'planewave':\n            Usual plane wave\n        'gaussian':\n            Standard Gaussian beam with waist size w0\n        'tukey':\n            Rectangular wave with cosine tapered edges (Tukey window). The width\n            is given by ``w0`` and the tapered region is determined by ``a``.\n        'spdc':\n            Wave used to represent vacuum fluctuations in case of SPDC calculation.\n\n        Default is 'planewave'.\n    pwr : float\n        Power of the wave (in watts). Default 1.0. The area of the beam is\n        calculated as w0 * Ly.\n    overrideE0 : bool\n        if True, then ``pwr`` is ignored and ``E0`` is used. Default False.\n    E0 : float\n        Maximum electrical field of the beam in vacuum.\n    w0 : float\n        Waist size of the beam (in meters). Default is 0.1 mm.\n    Ly : float\n        Beam size in y-direction (in meters).\n        Default is 1 mm.\n    a : float\n        Parameter used for Tukey wave. If a = 1, then the wave is a perfect\n        rectangle; if a = 0.5, then half of the profile is tapered by cosines.\n        Default is 0.7.\n    nPointsInteg : int\n        Number of points used to represent the profile of the wave.\n        Default is 100.\n    maxX : float\n        The wave profile is given in x-range (-maxX..maxX). Only used if\n        dynamicMaxX is False. Default is 1 mm.\n    dynamicMaxX : bool\n        Selects between dynamic determination of x-span of the wave (region\n        where ``nPointsInteg`` samples are taken) or fixed span by ``maxX``.\n        Default is True.\n    dynamicMaxXCoef : float\n        Dynamic maxX equals to the beam width (corrected to the angle of\n        incidence) times ``dynamicMaxXCoef``. Default is 2.0.\n    dynamicMaxXAddition : float\n        Constant factor that is added to maxX if ``dynamicMaxX`` is true.\n        Default is 0.0.""\n    maxPhi : float\n        Determines the maximum angle of deviation from the direction of\n        propagation of the participating fields. Increases the range of x-span\n        (i.e ``maxX``) if it is necessary to limit the angular distribution of the\n        participating plane waves. Default is 0.17 rad.\n    xRange : tuple of floats (xMin, xMax)\n        Current x-span calculated from ``maxX``, ``dynamicMaxXCoef`` and ``maxPhi``.\n        The wave must be solved to access this quantity.\n    betas : ndarray of floats\n        Normalized tangential wave vectors of plane wave expansion. The wave\n        must be solved to access this quantity.\n    phis : ndarray of floats\n        Same information as in betas, but converted to angles in respect to the\n        direction of the propagation. The wave must be solved to access this\n        quantity.\n    kxs : ndarray of floats\n        Same information as in betas, but converted to the x-component of the\n        wave vector. The wave must be solved to access this quantity.\n    kzs : ndarray of floats\n        Same information as in betas, but converted to the z-component of the\n        wave vector. The wave must be solved to access this quantity.\n    fieldProfile : tuple(2,) of ndarray of floats (xs, fieldProfile)\n        The field profile of the input wave sampled by ``nPointsInteg`` points. The\n        wave must be solved to access this quantity.\n    expansionCoefsKx : ndarray of complex\n        Array of plane wave expansion coefs.\n\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave},
  {Py_tp_getset, (void *)__pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython._Wave",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""_Wave", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("_Wave()\n\n    This is a helper class for using non plane waves as an input.\n\n    Attributes\n    ----------\n    waveType : str\n        'planewave':\n            Usual plane wave\n        'gaussian':\n            Standard Gaussian beam with waist size w0\n        'tukey':\n            Rectangular wave with cosine tapered edges (Tukey window). The width\n            is given by ``w0`` and the tapered region is determined by ``a``.\n        'spdc':\n            Wave used to represent vacuum fluctuations in case of SPDC calculation.\n\n        Default is 'planewave'.\n    pwr : float\n        Power of the wave (in watts). Default 1.0. The area of the beam is\n        calculated as w0 * Ly.\n    overrideE0 : bool\n        if True, then ``pwr`` is ignored and ``E0`` is used. Default False.\n    E0 : float\n        Maximum electrical field of the beam in vacuum.\n    w0 : float\n        Waist size of the beam (in meters). Default is 0.1 mm.\n    Ly : float\n        Beam size in y-direction (in meters).\n        Default is 1 mm.\n    a : float\n        Parameter used for Tukey wave. If a = 1, then the wave is a perfect\n        rectangle; if a = 0.5, then half of the profile is tapered by cosines.\n        Default is 0.7.\n    nPointsInteg : int\n        Number of points used to represent the profile of the wave.\n        Default is 100.\n    maxX : float\n        The wave profile is given in x-range (-maxX..maxX). Only used if\n        dynamicMaxX is False. Default is 1 mm.\n    dynamicMaxX : bool\n        Selects between dynamic determination of x-span of the wave (region\n        where ``nPointsInteg`` samples are taken) or fixed span by ``maxX``.\n        Default is True.\n    dynamicMaxXCoef : float\n        Dynamic maxX equals to the beam width (corrected to the angle of\n        incidence) times ``dynamicMaxXCoef``. Default is 2.0.\n    dynamicMaxXAddition : float\n        Constant factor that is added to maxX if ``dynamicMaxX`` is true.\n        Default is 0.0.""\n    maxPhi : float\n        Determines the maximum angle of deviation from the direction of\n        propagation of the participating fields. Increases the range of x-span\n        (i.e ``maxX``) if it is necessary to limit the angular distribution of the\n        participating plane waves. Default is 0.17 rad.\n    xRange : tuple of floats (xMin, xMax)\n        Current x-span calculated from ``maxX``, ``dynamicMaxXCoef`` and ``maxPhi``.\n        The wave must be solved to access this quantity.\n    betas : ndarray of floats\n        Normalized tangential wave vectors of plane wave expansion. The wave\n        must be solved to access this quantity.\n    phis : ndarray of floats\n        Same information as in betas, but converted to angles in respect to the\n        direction of the propagation. The wave must be solved to access this\n        quantity.\n    kxs : ndarray of floats\n        Same information as in betas, but converted to the x-component of the\n        wave vector. The wave must be solved to access this quantity.\n    kzs : ndarray of floats\n        Same information as in betas, but converted to the z-component of the\n        wave vector. The wave must be solved to access this quantity.\n    fieldProfile : tuple(2,) of ndarray of floats (xs, fieldProfile)\n        The field profile of the input wave sampled by ``nPointsInteg`` points. The\n        wave must be solved to access this quantity.\n    expansionCoefsKx : ndarray of complex\n        Array of plane wave expansion coefs.\n\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities;

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *)o);
  p->__pyx_vtab = __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities;
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1__cinit__(o, __pyx_mstate_global->__pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_inc(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_3inc_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_r(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1r_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_t(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1t_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_I(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1I_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_R(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1R_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_T(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_1T_1__get__(o);
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities[] = {
  {"inc", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_inc, 0, 0, 0},
  {"r", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_r, 0, 0, 0},
  {"t", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_t, 0, 0, 0},
  {"I", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_I, 0, 0, 0},
  {"R", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_R, 0, 0, 0},
  {"T", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_T, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Helper class to store intensities for :class:`NonlinearTMM`.\n\n    Attributes\n    ----------\n    inc : complex\n        Electrical field amplitude of the incident plane wave.\n    r : complex\n        Electrical field amplitude of the reflected plane wave.\n    t : complex\n        Electrical field amplitude of the transmitted plane wave.\n    I : float\n        Intensity of the incident plane wave.\n    R : float\n        Intensity of the reflected plane wave.\n    T : float\n        Intensity of the transmitted plane wave.\n\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities},
  {Py_tp_getset, (void *)__pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython._Intensities",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""_Intensities", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    Helper class to store intensities for :class:`NonlinearTMM`.\n\n    Attributes\n    ----------\n    inc : complex\n        Electrical field amplitude of the incident plane wave.\n    r : complex\n        Electrical field amplitude of the reflected plane wave.\n    t : complex\n        Electrical field amplitude of the transmitted plane wave.\n    I : float\n        Intensity of the incident plane wave.\n    R : float\n        Intensity of the reflected plane wave.\n    T : float\n        Intensity of the transmitted plane wave.\n\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM;

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)o);
  p->__pyx_vtab = __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM;
  p->inc = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->r = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->t = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->Ii = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->Ir = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->It = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->Ia = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->enh = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->_parent = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1__cinit__(o, __pyx_mstate_global->__pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM(PyObject *o) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->inc);
  Py_CLEAR(p->r);
  Py_CLEAR(p->t);
  Py_CLEAR(p->Ii);
  Py_CLEAR(p->Ir);
  Py_CLEAR(p->It);
  Py_CLEAR(p->Ia);
  Py_CLEAR(p->enh);
  Py_CLEAR(p->_parent);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->inc) {
    e = (*v)(((PyObject *)p->inc), a); if (e) return e;
  }
  if (p->r) {
    e = (*v)(((PyObject *)p->r), a); if (e) return e;
  }
  if (p->t) {
    e = (*v)(((PyObject *)p->t), a); if (e) return e;
  }
  if (p->Ii) {
    e = (*v)(((PyObject *)p->Ii), a); if (e) return e;
  }
  if (p->Ir) {
    e = (*v)(((PyObject *)p->Ir), a); if (e) return e;
  }
  if (p->It) {
    e = (*v)(((PyObject *)p->It), a); if (e) return e;
  }
  if (p->Ia) {
    e = (*v)(((PyObject *)p->Ia), a); if (e) return e;
  }
  if (p->enh) {
    e = (*v)(((PyObject *)p->enh), a); if (e) return e;
  }
  if (p->_parent) {
    e = (*v)(p->_parent, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *)o;
  tmp = ((PyObject*)p->inc);
  p->inc = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->r);
  p->r = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->t);
  p->t = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Ii);
  p->Ii = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Ir);
  p->Ir = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->It);
  p->It = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Ia);
  p->Ia = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->enh);
  p->enh = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_parent);
  p->_parent = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_inc(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3inc_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_r(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1r_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_t(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_1t_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_Ii(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ii_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_Ir(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ir_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_It(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2It_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_Ia(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_2Ia_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_enh(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_3enh_1__get__(o);
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM[] = {
  {"inc", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_inc, 0, 0, 0},
  {"r", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_r, 0, 0, 0},
  {"t", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_t, 0, 0, 0},
  {"Ii", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_Ii, 0, 0, 0},
  {"Ir", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_Ir, 0, 0, 0},
  {"It", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_It, 0, 0, 0},
  {"Ia", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_Ia, 0, 0, 0},
  {"enh", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_enh, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Helper class to store the result of the ``NonlinearTMM.Sweep`` method.\n\n    Attributes\n    ----------\n    inc : ndarray of complex\n        The complex amplitudes of the incident plane waves.\n    r : ndarray of complex\n        The complex amplitudes of the reflected plane waves.\n    t : ndarray of complex\n        The complex amplitudes of the transmitted plane waves.\n    Ii : ndarray of double\n        The intensities of the incident plane waves.\n    Ir : ndarray of double\n        The intensities of the reflected plane waves.\n    It : ndarray of double\n        The intensities of the transmitted plane waves.\n    Ia : ndarray of double\n        The intensities of absorption in the structure.\n    enh : ndarray of double\n        The enhancement values of the electrical field norm.\n\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM},
  {Py_tp_clear, (void *)__pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM},
  {Py_tp_getset, (void *)__pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython._SweepResultNonlinearTMM",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""_SweepResultNonlinearTMM", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Helper class to store the result of the ``NonlinearTMM.Sweep`` method.\n\n    Attributes\n    ----------\n    inc : ndarray of complex\n        The complex amplitudes of the incident plane waves.\n    r : ndarray of complex\n        The complex amplitudes of the reflected plane waves.\n    t : ndarray of complex\n        The complex amplitudes of the transmitted plane waves.\n    Ii : ndarray of double\n        The intensities of the incident plane waves.\n    Ir : ndarray of double\n        The intensities of the reflected plane waves.\n    It : ndarray of double\n        The intensities of the transmitted plane waves.\n    Ia : ndarray of double\n        The intensities of absorption in the structure.\n    enh : ndarray of double\n        The enhancement values of the electrical field norm.\n\n    "), /*tp_doc*/
  __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM, /*tp_traverse*/
  __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM;

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)o);
  p->__pyx_vtab = __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM;
  p->Pi = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->Pr = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->Pt = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->enh = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->beamArea = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->_parent = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_1__cinit__(o, __pyx_mstate_global->__pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM(PyObject *o) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->Pi);
  Py_CLEAR(p->Pr);
  Py_CLEAR(p->Pt);
  Py_CLEAR(p->enh);
  Py_CLEAR(p->beamArea);
  Py_CLEAR(p->_parent);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->Pi) {
    e = (*v)(((PyObject *)p->Pi), a); if (e) return e;
  }
  if (p->Pr) {
    e = (*v)(((PyObject *)p->Pr), a); if (e) return e;
  }
  if (p->Pt) {
    e = (*v)(((PyObject *)p->Pt), a); if (e) return e;
  }
  if (p->enh) {
    e = (*v)(((PyObject *)p->enh), a); if (e) return e;
  }
  if (p->beamArea) {
    e = (*v)(((PyObject *)p->beamArea), a); if (e) return e;
  }
  if (p->_parent) {
    e = (*v)(p->_parent, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *)o;
  tmp = ((PyObject*)p->Pi);
  p->Pi = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Pr);
  p->Pr = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Pt);
  p->Pt = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->enh);
  p->enh = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->beamArea);
  p->beamArea = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_parent);
  p->_parent = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_Pi(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pi_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_Pr(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pr_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_Pt(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_2Pt_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_enh(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_3enh_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_beamArea(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_8beamArea_1__get__(o);
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM[] = {
  {"Pi", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_Pi, 0, 0, 0},
  {"Pr", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_Pr, 0, 0, 0},
  {"Pt", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_Pt, 0, 0, 0},
  {"enh", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_enh, 0, 0, 0},
  {"beamArea", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_beamArea, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM},
  {Py_tp_doc, (void *)PyDoc_STR("\n    The helper class to store the results of `WaveSweep`.\n\n    Attributes\n    ----------\n    Pi : ndarray of floats\n        The power of the incident beam.\n    Pr : ndarray of floats\n        The power of the reflected beam.\n    Pt : ndarray of floats\n        The power of the transmitted beam.\n    enh : ndarray of floats\n        The enhancement of the beam.\n    beamArea : ndarray of floats\n        The area of the beam with correction of angle of incidence.\n\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM},
  {Py_tp_clear, (void *)__pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM},
  {Py_tp_getset, (void *)__pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython._WaveSweepResultNonlinearTMM",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""_WaveSweepResultNonlinearTMM", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    The helper class to store the results of `WaveSweep`.\n\n    Attributes\n    ----------\n    Pi : ndarray of floats\n        The power of the incident beam.\n    Pr : ndarray of floats\n        The power of the reflected beam.\n    Pt : ndarray of floats\n        The power of the transmitted beam.\n    enh : ndarray of floats\n        The enhancement of the beam.\n    beamArea : ndarray of floats\n        The area of the beam with correction of angle of incidence.\n\n    "), /*tp_doc*/
  __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM, /*tp_traverse*/
  __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ;

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *)o);
  p->__pyx_vtab = __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ;
  p->E = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->H = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1__cinit__(o, __pyx_mstate_global->__pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ(PyObject *o) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->E);
  Py_CLEAR(p->H);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->E) {
    e = (*v)(((PyObject *)p->E), a); if (e) return e;
  }
  if (p->H) {
    e = (*v)(((PyObject *)p->H), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *)o;
  tmp = ((PyObject*)p->E);
  p->E = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->H);
  p->H = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_E(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1E_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_H(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_1H_1__get__(o);
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ[] = {
  {"E", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_E, 0, 0, 0},
  {"H", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_H, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Helper class to store the results of `GetFields`.\n\n    Attributes\n    ----------\n    E : ndarray(N, 3) of complex\n        Electrical fields along the z-coordinate. The first index corresponds to\n        the z-coordinate and the second corresponds to the x-, y-, z-component\n        (0..2).\n    H : ndarray(N, 3) of complex\n        Magnetic fields along the z-coordinate. The first index corresponds to\n        the z-coordinate and the second corresponds to the x-, y-, z-component\n        (0..2).\n\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ},
  {Py_tp_clear, (void *)__pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ},
  {Py_tp_getset, (void *)__pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython._FieldsZ",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""_FieldsZ", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Helper class to store the results of `GetFields`.\n\n    Attributes\n    ----------\n    E : ndarray(N, 3) of complex\n        Electrical fields along the z-coordinate. The first index corresponds to\n        the z-coordinate and the second corresponds to the x-, y-, z-component\n        (0..2).\n    H : ndarray(N, 3) of complex\n        Magnetic fields along the z-coordinate. The first index corresponds to\n        the z-coordinate and the second corresponds to the x-, y-, z-component\n        (0..2).\n\n    "), /*tp_doc*/
  __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ, /*tp_traverse*/
  __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX;

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)o);
  p->__pyx_vtab = __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX;
  p->Ex = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->Ey = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->Ez = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->Hx = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->Hy = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->Hz = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_1__cinit__(o, __pyx_mstate_global->__pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX(PyObject *o) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->Ex);
  Py_CLEAR(p->Ey);
  Py_CLEAR(p->Ez);
  Py_CLEAR(p->Hx);
  Py_CLEAR(p->Hy);
  Py_CLEAR(p->Hz);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->Ex) {
    e = (*v)(((PyObject *)p->Ex), a); if (e) return e;
  }
  if (p->Ey) {
    e = (*v)(((PyObject *)p->Ey), a); if (e) return e;
  }
  if (p->Ez) {
    e = (*v)(((PyObject *)p->Ez), a); if (e) return e;
  }
  if (p->Hx) {
    e = (*v)(((PyObject *)p->Hx), a); if (e) return e;
  }
  if (p->Hy) {
    e = (*v)(((PyObject *)p->Hy), a); if (e) return e;
  }
  if (p->Hz) {
    e = (*v)(((PyObject *)p->Hz), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *)o;
  tmp = ((PyObject*)p->Ex);
  p->Ex = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Ey);
  p->Ey = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Ez);
  p->Ez = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Hx);
  p->Hx = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Hy);
  p->Hy = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Hz);
  p->Hz = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_EN(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2EN_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_HN(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2HN_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_Ex(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ex_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_Ey(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ey_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_Ez(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Ez_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_Hx(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hx_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_Hy(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hy_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_Hz(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_2Hz_1__get__(o);
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX[] = {
  {"EN", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_EN, 0, 0, 0},
  {"HN", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_HN, 0, 0, 0},
  {"Ex", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_Ex, 0, 0, 0},
  {"Ey", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_Ey, 0, 0, 0},
  {"Ez", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_Ez, 0, 0, 0},
  {"Hx", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_Hx, 0, 0, 0},
  {"Hy", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_Hy, 0, 0, 0},
  {"Hz", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_Hz, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Helper class to store the result of `GetFields2D` and `WaveGetFields2D`.\n\n    Attributes\n    ----------\n    Ex : ndarray(N, M) of complex\n        Electrical field x-component in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n    Ey : ndarray(N, M) of complex\n        Electrical field y-component in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n    Ez : ndarray(N, M) of complex\n        Electrical field z-component in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n    EN : ndarray(N, M) of float\n        Electrical field norm in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n    Hx : ndarray(N, M) of complex\n        Magnetic field x-component in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n    Hy : ndarray(N, M) of complex\n        Magnetic field y-component in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n    Hz : ndarray(N, M) of complex\n        Magnetic field z-component in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n    HN : ndarray(N, M) of float\n        Magnetic field norm in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX},
  {Py_tp_clear, (void *)__pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX},
  {Py_tp_getset, (void *)__pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython._FieldsZX",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""_FieldsZX", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n    Helper class to store the result of `GetFields2D` and `WaveGetFields2D`.\n\n    Attributes\n    ----------\n    Ex : ndarray(N, M) of complex\n        Electrical field x-component in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n    Ey : ndarray(N, M) of complex\n        Electrical field y-component in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n    Ez : ndarray(N, M) of complex\n        Electrical field z-component in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n    EN : ndarray(N, M) of float\n        Electrical field norm in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n    Hx : ndarray(N, M) of complex\n        Magnetic field x-component in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n    Hy : ndarray(N, M) of complex\n        Magnetic field y-component in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n    Hz : ndarray(N, M) of complex\n        Magnetic field z-component in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n    HN : ndarray(N, M) of float\n        Magnetic field norm in the zx-plane. First index corresponds to\n        the z-coordinate and the second to the x-coordinate.\n\n    "), /*tp_doc*/
  __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX, /*tp_traverse*/
  __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave;

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *)o);
  p->__pyx_vtab = __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave;
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_1__cinit__(o, __pyx_mstate_global->__pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_kzF(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_3kzF_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_kx(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_2kx_1__get__(o);
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave[] = {
  {"GetMainFields", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_3GetMainFields, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_2GetMainFields},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave[] = {
  {"kzF", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_kzF, 0, 0, 0},
  {"kx", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_kx, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave},
  {Py_tp_doc, (void *)PyDoc_STR("\n    Helper class for `NonlinearLayer` to describe the parameters of the\n    homogeneous waves in the layer.\n\n    Attributes\n    ----------\n    kzF : complex\n        Z-component of the wave vector.\n    kx : float\n        X-component of the wave vector.\n    ")},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave},
  {Py_tp_getset, (void *)__pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython._HomogeneousWave",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""_HomogeneousWave", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    Helper class for `NonlinearLayer` to describe the parameters of the\n    homogeneous waves in the layer.\n\n    Attributes\n    ----------\n    kzF : complex\n        Z-component of the wave vector.\n    kx : float\n        X-component of the wave vector.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer;

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)o);
  p->__pyx_vtab = __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer;
  p->hw = Py_None; Py_INCREF(Py_None);
  p->_parent = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1__cinit__(o, __pyx_mstate_global->__pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer(PyObject *o) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->hw);
  Py_CLEAR(p->_parent);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->hw) {
    e = (*v)(p->hw, a); if (e) return e;
  }
  if (p->_parent) {
    e = (*v)(p->_parent, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *)o;
  tmp = ((PyObject*)p->hw);
  p->hw = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_parent);
  p->_parent = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_d(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1d_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_d(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_1d_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_hw(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_2hw_1__get__(o);
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer[] = {
  {"GetIntensity", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_3GetIntensity, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_2GetIntensity},
  {"GetAbsorbedIntensity", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_5GetAbsorbedIntensity, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_4GetAbsorbedIntensity},
  {"GetSrcIntensity", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_7GetSrcIntensity, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_6GetSrcIntensity},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer[] = {
  {"d", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_d, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_d, 0, 0},
  {"hw", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_hw, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer},
  {Py_tp_doc, (void *)PyDoc_STR("\n        Helper class for layer-specific data and methods.\n\n        Attributes\n        ----------\n        d : float\n            The thickness of the layer.\n\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer},
  {Py_tp_clear, (void *)__pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer},
  {Py_tp_getset, (void *)__pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython._NonlinearLayer",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""_NonlinearLayer", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("\n        Helper class for layer-specific data and methods.\n\n        Attributes\n        ----------\n        d : float\n            The thickness of the layer.\n\n    "), /*tp_doc*/
  __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer, /*tp_traverse*/
  __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM;

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)o);
  p->__pyx_vtab = __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM;
  p->materialsCache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->layers = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_parent = Py_None; Py_INCREF(Py_None);
  p->wave = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM(PyObject *o) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->materialsCache);
  Py_CLEAR(p->layers);
  Py_CLEAR(p->_parent);
  Py_CLEAR(p->wave);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->materialsCache) {
    e = (*v)(p->materialsCache, a); if (e) return e;
  }
  if (p->layers) {
    e = (*v)(p->layers, a); if (e) return e;
  }
  if (p->_parent) {
    e = (*v)(p->_parent, a); if (e) return e;
  }
  if (p->wave) {
    e = (*v)(((PyObject *)p->wave), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *)o;
  tmp = ((PyObject*)p->materialsCache);
  p->materialsCache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->layers);
  p->layers = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_parent);
  p->_parent = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->wave);
  p->wave = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_wl(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2wl_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_wl(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2wl_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_beta(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4beta_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_beta(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4beta_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_pol(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3pol_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_pol(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_3pol_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_I0(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2I0_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_I0(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2I0_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_overrideE0(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10overrideE0_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_overrideE0(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10overrideE0_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_E0(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2E0_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_E0(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_2E0_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_mode(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4mode_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_mode(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4mode_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_materialsCache(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_14materialsCache_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_layers(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_6layers_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_wave(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4wave_1__get__(o);
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM[] = {
  {"AddLayer", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_5AddLayer, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_4AddLayer},
  {"SetParams", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_7SetParams, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_6SetParams},
  {"Solve", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_9Solve, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_8Solve},
  {"GetIntensities", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_11GetIntensities, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_10GetIntensities},
  {"Sweep", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_13Sweep, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_12Sweep},
  {"GetFields", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_15GetFields, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_14GetFields},
  {"GetFields2D", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_17GetFields2D, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_16GetFields2D},
  {"GetAbsorbedIntensity", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_19GetAbsorbedIntensity, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_18GetAbsorbedIntensity},
  {"GetEnhancement", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_21GetEnhancement, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_20GetEnhancement},
  {"WaveGetPowerFlows", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_23WaveGetPowerFlows, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_22WaveGetPowerFlows},
  {"WaveGetEnhancement", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_25WaveGetEnhancement, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_24WaveGetEnhancement},
  {"WaveSweep", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_27WaveSweep, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_26WaveSweep},
  {"WaveGetFields2D", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_29WaveGetFields2D, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_28WaveGetFields2D},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_31__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_33__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM[] = {
  {"wl", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_wl, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_wl, 0, 0},
  {"beta", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_beta, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_beta, 0, 0},
  {"pol", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_pol, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_pol, 0, 0},
  {"I0", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_I0, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_I0, 0, 0},
  {"overrideE0", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_overrideE0, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_overrideE0, 0, 0},
  {"E0", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_E0, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_E0, 0, 0},
  {"mode", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_mode, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_mode, 0, 0},
  {"materialsCache", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_materialsCache, 0, 0, 0},
  {"layers", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_layers, 0, 0, 0},
  {"wave", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_wave, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM},
  {Py_tp_doc, (void *)PyDoc_STR("NonlinearTMM()\n\n    This class is mainly used to calculate linear propagation of plane waves\n    in stratified medium. It can work like ordinary TMM and calculate the propagation\n    of the input waves. It is also capable of nonlinear calculations, but\n    for this purpose use specialized class :class:`SecondOrderNLTMM`.\n\n    Default constructor takes no arguments.\n\n    Attributes\n    ----------\n    E0 : complex\n        The amplitude of the input plane wave (given in vacuum, not in the first layer).\n        Only used if ``overrideE0`` is set to True (default is False); otherwise intensity\n        ``I0`` is used instead.\n        Default: 1.0.\n    I0 : float\n        The intensity of the input beam in the first medium. Only used if\n        ``overrideE0`` is set to False (default).\n        Default: 1.0.\n    beta : float\n        Normalized tangential wave vector. Determines the angle of incidence through\n        relation beta = sin(theta) * n_p, where theta is angle of incidence and\n        n_p is the refractive index of the prism.\n        Default: not defined.\n    layers : list of :class:`_NonlinearLayer`\n        Allows accessing layers by index.\n    mode : str\n        Default is \"incident\", which corresponds to ordinary TMM. The other mode is\n        \"nonlinear\", but it is automatically set by :class:`SecondOrderNLTMM` (i.e, this\n        parameter has to be modified only in special cases).\n    overrideE0 : bool\n        If True, then parameter ``E0`` is used; otherwise intensity ``I0`` is used.\n        Default: False.\n    pol : str\n        \"p\" or \"s\" corresponding to p- and s-polarization, respectively.\n        Default: Not defined.\n    wave : :class:`_Wave`\n        Instance of helper class :class:`_Wave` for non-plane-wave calculations.\n    wl : float\n        Wavelength of calculations in meters.\n\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM},
  {Py_tp_clear, (void *)__pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM},
  {Py_tp_getset, (void *)__pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython.NonlinearTMM",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""NonlinearTMM", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("NonlinearTMM()\n\n    This class is mainly used to calculate linear propagation of plane waves\n    in stratified medium. It can work like ordinary TMM and calculate the propagation\n    of the input waves. It is also capable of nonlinear calculations, but\n    for this purpose use specialized class :class:`SecondOrderNLTMM`.\n\n    Default constructor takes no arguments.\n\n    Attributes\n    ----------\n    E0 : complex\n        The amplitude of the input plane wave (given in vacuum, not in the first layer).\n        Only used if ``overrideE0`` is set to True (default is False); otherwise intensity\n        ``I0`` is used instead.\n        Default: 1.0.\n    I0 : float\n        The intensity of the input beam in the first medium. Only used if\n        ``overrideE0`` is set to False (default).\n        Default: 1.0.\n    beta : float\n        Normalized tangential wave vector. Determines the angle of incidence through\n        relation beta = sin(theta) * n_p, where theta is angle of incidence and\n        n_p is the refractive index of the prism.\n        Default: not defined.\n    layers : list of :class:`_NonlinearLayer`\n        Allows accessing layers by index.\n    mode : str\n        Default is \"incident\", which corresponds to ordinary TMM. The other mode is\n        \"nonlinear\", but it is automatically set by :class:`SecondOrderNLTMM` (i.e, this\n        parameter has to be modified only in special cases).\n    overrideE0 : bool\n        If True, then parameter ``E0`` is used; otherwise intensity ``I0`` is used.\n        Default: False.\n    pol : str\n        \"p\" or \"s\" corresponding to p- and s-polarization, respectively.\n        Default: Not defined.\n    wave : :class:`_Wave`\n        Instance of helper class :class:`_Wave` for non-plane-wave calculations.\n    wl : float\n        Wavelength of calculations in meters.\n\n    "), /*tp_doc*/
  __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, /*tp_traverse*/
  __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities;

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *)o);
  p->__pyx_vtab = __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities;
  p->P1 = Py_None; Py_INCREF(Py_None);
  p->P2 = Py_None; Py_INCREF(Py_None);
  p->Gen = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_1__cinit__(o, __pyx_mstate_global->__pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities(PyObject *o) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->P1);
  Py_CLEAR(p->P2);
  Py_CLEAR(p->Gen);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->P1) {
    e = (*v)(p->P1, a); if (e) return e;
  }
  if (p->P2) {
    e = (*v)(p->P2, a); if (e) return e;
  }
  if (p->Gen) {
    e = (*v)(p->Gen, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *)o;
  tmp = ((PyObject*)p->P1);
  p->P1 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->P2);
  p->P2 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Gen);
  p->Gen = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_P1(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2P1_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_P2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_2P2_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_Gen(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_3Gen_1__get__(o);
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities[] = {
  {"P1", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_P1, 0, 0, 0},
  {"P2", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_P2, 0, 0, 0},
  {"Gen", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_Gen, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities},
  {Py_tp_doc, (void *)PyDoc_STR("_SecondOrderNLIntensities()\n    This is a helper class for `SecondOrderNLTMM`. It stores the results of\n    `GetIntensities`.\n\n    Attributes\n    ----------\n    P1 : :class:`_Intensities`\n        Instance of ``_Intensities`` class and holds the intensities of the first\n        input wave.\n    P2 : :class:`_Intensities`\n        Instance of ``_Intensities`` class and holds the intensities of the second\n        input wave.\n    Gen : :class:`_Intensities`\n        Instance of ``_Intensities`` class and holds the intensities of the generated\n        wave.\n\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities},
  {Py_tp_clear, (void *)__pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities},
  {Py_tp_getset, (void *)__pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython._SecondOrderNLIntensities",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""_SecondOrderNLIntensities", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("_SecondOrderNLIntensities()\n    This is a helper class for `SecondOrderNLTMM`. It stores the results of\n    `GetIntensities`.\n\n    Attributes\n    ----------\n    P1 : :class:`_Intensities`\n        Instance of ``_Intensities`` class and holds the intensities of the first\n        input wave.\n    P2 : :class:`_Intensities`\n        Instance of ``_Intensities`` class and holds the intensities of the second\n        input wave.\n    Gen : :class:`_Intensities`\n        Instance of ``_Intensities`` class and holds the intensities of the generated\n        wave.\n\n    "), /*tp_doc*/
  __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities, /*tp_traverse*/
  __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM;

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)o);
  p->__pyx_vtab = __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM;
  p->P1 = Py_None; Py_INCREF(Py_None);
  p->P2 = Py_None; Py_INCREF(Py_None);
  p->Gen = Py_None; Py_INCREF(Py_None);
  p->wlsGen = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->betasGen = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_1__cinit__(o, __pyx_mstate_global->__pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM(PyObject *o) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->P1);
  Py_CLEAR(p->P2);
  Py_CLEAR(p->Gen);
  Py_CLEAR(p->wlsGen);
  Py_CLEAR(p->betasGen);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->P1) {
    e = (*v)(p->P1, a); if (e) return e;
  }
  if (p->P2) {
    e = (*v)(p->P2, a); if (e) return e;
  }
  if (p->Gen) {
    e = (*v)(p->Gen, a); if (e) return e;
  }
  if (p->wlsGen) {
    e = (*v)(((PyObject *)p->wlsGen), a); if (e) return e;
  }
  if (p->betasGen) {
    e = (*v)(((PyObject *)p->betasGen), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *)o;
  tmp = ((PyObject*)p->P1);
  p->P1 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->P2);
  p->P2 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Gen);
  p->Gen = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->wlsGen);
  p->wlsGen = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->betasGen);
  p->betasGen = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_P1(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2P1_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_P2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_2P2_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_Gen(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_3Gen_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_wlsGen(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_6wlsGen_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_betasGen(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_8betasGen_1__get__(o);
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM[] = {
  {"P1", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_P1, 0, 0, 0},
  {"P2", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_P2, 0, 0, 0},
  {"Gen", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_Gen, 0, 0, 0},
  {"wlsGen", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_wlsGen, 0, 0, 0},
  {"betasGen", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_betasGen, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM},
  {Py_tp_doc, (void *)PyDoc_STR("_SweepResultSecondOrderNLTMM()\n    Helper class for `SecondOrderNLTMM`. Stores the results of `Sweep` method.\n\n    Attributes\n    ----------\n    P1 : :class:`_SweepResultNonlinearTMM`\n        Sweep results for the first input wave.\n    P2 : :class:`_SweepResultNonlinearTMM`\n        Sweep results for the second input wave.\n    Gen : :class:`_SweepResultNonlinearTMM`\n        Sweep results for the generated wave.\n    wlsGen : ndarray of floats\n        Stores the wavelength of the generated wave.\n    betasGen : ndarray of floats\n        Stores the normalized tangential wave vector of the generated wave.\n\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM},
  {Py_tp_clear, (void *)__pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM},
  {Py_tp_getset, (void *)__pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython._SweepResultSecondOrderNLTMM",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""_SweepResultSecondOrderNLTMM", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("_SweepResultSecondOrderNLTMM()\n    Helper class for `SecondOrderNLTMM`. Stores the results of `Sweep` method.\n\n    Attributes\n    ----------\n    P1 : :class:`_SweepResultNonlinearTMM`\n        Sweep results for the first input wave.\n    P2 : :class:`_SweepResultNonlinearTMM`\n        Sweep results for the second input wave.\n    Gen : :class:`_SweepResultNonlinearTMM`\n        Sweep results for the generated wave.\n    wlsGen : ndarray of floats\n        Stores the wavelength of the generated wave.\n    betasGen : ndarray of floats\n        Stores the normalized tangential wave vector of the generated wave.\n\n    "), /*tp_doc*/
  __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM, /*tp_traverse*/
  __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM;

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)o);
  p->__pyx_vtab = __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM;
  p->P1 = Py_None; Py_INCREF(Py_None);
  p->P2 = Py_None; Py_INCREF(Py_None);
  p->Gen = Py_None; Py_INCREF(Py_None);
  p->wlsGen = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  p->betasGen = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_1__cinit__(o, __pyx_mstate_global->__pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM(PyObject *o) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->P1);
  Py_CLEAR(p->P2);
  Py_CLEAR(p->Gen);
  Py_CLEAR(p->wlsGen);
  Py_CLEAR(p->betasGen);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->P1) {
    e = (*v)(p->P1, a); if (e) return e;
  }
  if (p->P2) {
    e = (*v)(p->P2, a); if (e) return e;
  }
  if (p->Gen) {
    e = (*v)(p->Gen, a); if (e) return e;
  }
  if (p->wlsGen) {
    e = (*v)(((PyObject *)p->wlsGen), a); if (e) return e;
  }
  if (p->betasGen) {
    e = (*v)(((PyObject *)p->betasGen), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *)o;
  tmp = ((PyObject*)p->P1);
  p->P1 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->P2);
  p->P2 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Gen);
  p->Gen = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->wlsGen);
  p->wlsGen = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->betasGen);
  p->betasGen = ((PyArrayObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_P1(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2P1_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_P2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_2P2_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_Gen(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_3Gen_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_wlsGen(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_6wlsGen_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_betasGen(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_8betasGen_1__get__(o);
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM[] = {
  {"P1", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_P1, 0, 0, 0},
  {"P2", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_P2, 0, 0, 0},
  {"Gen", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_Gen, 0, 0, 0},
  {"wlsGen", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_wlsGen, 0, 0, 0},
  {"betasGen", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_betasGen, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM},
  {Py_tp_doc, (void *)PyDoc_STR("_WaveSweepResultSecondOrderNLTMM()\n    Helper class for `SecondOrderNLTMM`. Stores the results of `WaveSweep`\n    method.\n\n    Attributes\n    ----------\n    P1 : :class:`_WaveSweepResultNonlinearTMM`\n        The sweep result of the first input beam.\n    P2 : :class:`_WaveSweepResultNonlinearTMM`\n        The sweep result of the second input beam.\n    Gen : :class:`_WaveSweepResultNonlinearTMM`\n        The sweep result of the generated beam.\n    wlsGen : ndarray of floats\n        Stores the wavelength of the generated wave.\n    betasGen : ndarray of floats\n        Stores the normalized tangential wave vector of the generated wave.\n\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM},
  {Py_tp_clear, (void *)__pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM},
  {Py_tp_getset, (void *)__pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython._WaveSweepResultSecondOrderNLTMM",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""_WaveSweepResultSecondOrderNLTMM", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("_WaveSweepResultSecondOrderNLTMM()\n    Helper class for `SecondOrderNLTMM`. Stores the results of `WaveSweep`\n    method.\n\n    Attributes\n    ----------\n    P1 : :class:`_WaveSweepResultNonlinearTMM`\n        The sweep result of the first input beam.\n    P2 : :class:`_WaveSweepResultNonlinearTMM`\n        The sweep result of the second input beam.\n    Gen : :class:`_WaveSweepResultNonlinearTMM`\n        The sweep result of the generated beam.\n    wlsGen : ndarray of floats\n        Stores the wavelength of the generated wave.\n    betasGen : ndarray of floats\n        Stores the normalized tangential wave vector of the generated wave.\n\n    "), /*tp_doc*/
  __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM, /*tp_traverse*/
  __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *p;
  PyObject *o;
  o = __Pyx_AllocateExtensionType(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)o);
  p->materialsCache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->P1 = Py_None; Py_INCREF(Py_None);
  p->P2 = Py_None; Py_INCREF(Py_None);
  p->Gen = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM(PyObject *o) {
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(__Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->materialsCache);
  Py_CLEAR(p->P1);
  Py_CLEAR(p->P2);
  Py_CLEAR(p->Gen);
  PyTypeObject *tp = Py_TYPE(o);
  #if CYTHON_USE_TYPE_SLOTS
  (*tp->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(tp, Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  Py_DECREF(tp);
  #endif
}

static int __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->materialsCache) {
    e = (*v)(p->materialsCache, a); if (e) return e;
  }
  if (p->P1) {
    e = (*v)(p->P1, a); if (e) return e;
  }
  if (p->P2) {
    e = (*v)(p->P2, a); if (e) return e;
  }
  if (p->Gen) {
    e = (*v)(p->Gen, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *p = (struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM *)o;
  tmp = ((PyObject*)p->materialsCache);
  p->materialsCache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->P1);
  p->P1 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->P2);
  p->P2 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->Gen);
  p->Gen = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_deltaWlSpdc(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11deltaWlSpdc_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_deltaWlSpdc(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11deltaWlSpdc_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_solidAngleSpdc(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14solidAngleSpdc_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_solidAngleSpdc(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14solidAngleSpdc_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_deltaThetaSpdc(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14deltaThetaSpdc_1__get__(o);
}

static int __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_deltaThetaSpdc(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14deltaThetaSpdc_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_P1(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2P1_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_P2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_2P2_1__get__(o);
}

static PyObject *__pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_Gen(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_3Gen_1__get__(o);
}

static PyMethodDef __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM[] = {
  {"SetParams", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_5SetParams, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_4SetParams},
  {"AddLayer", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_7AddLayer, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_6AddLayer},
  {"Solve", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_9Solve, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_8Solve},
  {"UpdateGenParams", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11UpdateGenParams, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_10UpdateGenParams},
  {"GetIntensities", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_13GetIntensities, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_12GetIntensities},
  {"Sweep", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_15Sweep, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_14Sweep},
  {"WaveGetPowerFlows", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_17WaveGetPowerFlows, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_16WaveGetPowerFlows},
  {"WaveSweep", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_19WaveSweep, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_18WaveSweep},
  {"WaveGetFields2D", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_21WaveGetFields2D, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_20WaveGetFields2D},
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_23__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_25__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM[] = {
  {"deltaWlSpdc", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_deltaWlSpdc, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_deltaWlSpdc, 0, 0},
  {"solidAngleSpdc", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_solidAngleSpdc, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_solidAngleSpdc, 0, 0},
  {"deltaThetaSpdc", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_deltaThetaSpdc, __pyx_setprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_deltaThetaSpdc, 0, 0},
  {"P1", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_P1, 0, 0, 0},
  {"P2", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_P2, 0, 0, 0},
  {"Gen", __pyx_getprop_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_Gen, 0, 0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM},
  {Py_tp_doc, (void *)PyDoc_STR("SecondOrderNLTMM(mode)\n\n    This class calculates second-order nonlinear processes (e.g. sum-frequency,\n    difference-frequency generation and SPDC) in layered structures. Relies on the\n    functionality of the ``NonlinearTMM`` class.\n\n    Parameters\n    ----------\n    mode : str {'sfg', 'dfg', 'spdc'}\n\n    Attributes\n    ----------\n    P1 : :class:`NonlinearTMM`\n        The instance of the first pump wave TMM.\n    P2 : :class:`NonlinearTMM`\n        The instance of the second pump wave TMM.\n    Gen : :class:`NonlinearTMM`\n        The instance of the generated wave TMM.\n    deltaWlSpdc : float\n        The spectral collection window (in nanometers) of the SPDC signal.\n        Only used if ``mode`` is set to ``\"spdc\"``.\n        Default: NaN.\n    solidAngleSpdc : float\n        The collection solid angle (srad) of the detector (given in vacuum).\n        Only used if ``mode`` is set to ``\"spdc\"``.\n        Default: NaN.\n    deltaThetaSpdc : float\n        The horizontal collection window (rad) of the detector (given in vacuum).\n        It is used to calculate the vertical span of the detector.\n        Only used if ``mode`` is set to ``\"spdc\"``.\n        Default: NaN.\n\n    ")},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM},
  {Py_tp_clear, (void *)__pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM},
  {Py_tp_methods, (void *)__pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM},
  {Py_tp_getset, (void *)__pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM},
  {Py_tp_new, (void *)__pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM},
  {0, 0},
};
static PyType_Spec __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM_spec = {
  "NonlinearTMM._SecondOrderNLTMMCython.SecondOrderNLTMM",
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM_slots,
};
#else

static PyTypeObject __pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM = {
  PyVarObject_HEAD_INIT(0, 0)
  "NonlinearTMM._SecondOrderNLTMMCython.""SecondOrderNLTMM", /*tp_name*/
  sizeof(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, /*tp_dealloc*/
  0, /*tp_vectorcall_offset*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  PyDoc_STR("SecondOrderNLTMM(mode)\n\n    This class calculates second-order nonlinear processes (e.g. sum-frequency,\n    difference-frequency generation and SPDC) in layered structures. Relies on the\n    functionality of the ``NonlinearTMM`` class.\n\n    Parameters\n    ----------\n    mode : str {'sfg', 'dfg', 'spdc'}\n\n    Attributes\n    ----------\n    P1 : :class:`NonlinearTMM`\n        The instance of the first pump wave TMM.\n    P2 : :class:`NonlinearTMM`\n        The instance of the second pump wave TMM.\n    Gen : :class:`NonlinearTMM`\n        The instance of the generated wave TMM.\n    deltaWlSpdc : float\n        The spectral collection window (in nanometers) of the SPDC signal.\n        Only used if ``mode`` is set to ``\"spdc\"``.\n        Default: NaN.\n    solidAngleSpdc : float\n        The collection solid angle (srad) of the detector (given in vacuum).\n        Only used if ``mode`` is set to ``\"spdc\"``.\n        Default: NaN.\n    deltaThetaSpdc : float\n        The horizontal collection window (rad) of the detector (given in vacuum).\n        It is used to calculate the vertical span of the detector.\n        Only used if ``mode`` is set to ``\"spdc\"``.\n        Default: NaN.\n\n    "), /*tp_doc*/
  __pyx_tp_traverse_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, /*tp_traverse*/
  __pyx_tp_clear_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_paramDict = Py_None; Py_INCREF(Py_None);
  __pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_paramDictSecondOrderNLTMM = Py_None; Py_INCREF(Py_None);
  __pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_waveParamsSet = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor = &__pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor;
  __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor._Init = (PyObject *(*)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor *, TMM::Chi2Tensor *, PyObject *))__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor__Init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor)) __PYX_ERR(0, 91, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor) < (0)) __PYX_ERR(0, 91, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor) < (0)) __PYX_ERR(0, 91, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor, __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor) < (0)) __PYX_ERR(0, 91, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor) < (0)) __PYX_ERR(0, 91, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_Chi2Tensor, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor) < (0)) __PYX_ERR(0, 91, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor) < (0)) __PYX_ERR(0, 91, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_Material_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material)) __PYX_ERR(0, 212, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_Material_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material) < (0)) __PYX_ERR(0, 212, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_Material;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material) < (0)) __PYX_ERR(0, 212, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_Material, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material) < (0)) __PYX_ERR(0, 212, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material) < (0)) __PYX_ERR(0, 212, __pyx_L1_error)
  __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave = &__pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave;
  __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave._Init = (PyObject *(*)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave *, TMM::Wave *))__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave__Init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave)) __PYX_ERR(0, 294, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave) < (0)) __PYX_ERR(0, 294, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave) < (0)) __PYX_ERR(0, 294, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave, __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave) < (0)) __PYX_ERR(0, 294, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave) < (0)) __PYX_ERR(0, 294, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_Wave, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave) < (0)) __PYX_ERR(0, 294, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave) < (0)) __PYX_ERR(0, 294, __pyx_L1_error)
  __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities = &__pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities;
  __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities._Init = (PyObject *(*)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities *, TMM::Intensities *))__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities__Init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities)) __PYX_ERR(0, 556, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities) < (0)) __PYX_ERR(0, 556, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities) < (0)) __PYX_ERR(0, 556, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities, __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities) < (0)) __PYX_ERR(0, 556, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities) < (0)) __PYX_ERR(0, 556, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_Intensities, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities) < (0)) __PYX_ERR(0, 556, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Intensities) < (0)) __PYX_ERR(0, 556, __pyx_L1_error)
  __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM = &__pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM;
  __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM._Init = (PyObject *(*)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM *, TMM::SweepResultNonlinearTMM *, struct __pyx_opt_args_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM__Init *__pyx_optional_args))__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM__Init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM)) __PYX_ERR(0, 594, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM) < (0)) __PYX_ERR(0, 594, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM) < (0)) __PYX_ERR(0, 594, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM, __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM) < (0)) __PYX_ERR(0, 594, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM) < (0)) __PYX_ERR(0, 594, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_SweepResultNonlinearTMM, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM) < (0)) __PYX_ERR(0, 594, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultNonlinearTMM) < (0)) __PYX_ERR(0, 594, __pyx_L1_error)
  __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM = &__pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM;
  __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM._Init = (PyObject *(*)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM *, TMM::WaveSweepResultNonlinearTMM *, struct __pyx_opt_args_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM__Init *__pyx_optional_args))__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM__Init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM)) __PYX_ERR(0, 651, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM) < (0)) __PYX_ERR(0, 651, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM) < (0)) __PYX_ERR(0, 651, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM, __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM) < (0)) __PYX_ERR(0, 651, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM) < (0)) __PYX_ERR(0, 651, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_WaveSweepResultNonlinearTMM, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM) < (0)) __PYX_ERR(0, 651, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultNonlinearTMM) < (0)) __PYX_ERR(0, 651, __pyx_L1_error)
  __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ = &__pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ;
  __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ._Init = (PyObject *(*)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ *, TMM::FieldsZ *))__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ__Init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ)) __PYX_ERR(0, 697, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ) < (0)) __PYX_ERR(0, 697, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ) < (0)) __PYX_ERR(0, 697, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ, __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ) < (0)) __PYX_ERR(0, 697, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ) < (0)) __PYX_ERR(0, 697, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_FieldsZ, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ) < (0)) __PYX_ERR(0, 697, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZ) < (0)) __PYX_ERR(0, 697, __pyx_L1_error)
  __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX = &__pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX;
  __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX._Init = (PyObject *(*)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX *, TMM::FieldsZX *))__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX__Init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX)) __PYX_ERR(0, 732, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX) < (0)) __PYX_ERR(0, 732, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX) < (0)) __PYX_ERR(0, 732, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX, __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX) < (0)) __PYX_ERR(0, 732, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX) < (0)) __PYX_ERR(0, 732, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_FieldsZX, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX) < (0)) __PYX_ERR(0, 732, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__FieldsZX) < (0)) __PYX_ERR(0, 732, __pyx_L1_error)
  __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave = &__pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave;
  __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave._Init = (PyObject *(*)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave *, TMM::HomogeneousWave *))__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave__Init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave)) __PYX_ERR(0, 820, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave) < (0)) __PYX_ERR(0, 820, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave) < (0)) __PYX_ERR(0, 820, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave, __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave) < (0)) __PYX_ERR(0, 820, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave) < (0)) __PYX_ERR(0, 820, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_HomogeneousWave, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave) < (0)) __PYX_ERR(0, 820, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave) < (0)) __PYX_ERR(0, 820, __pyx_L1_error)
  __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer = &__pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer;
  __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer._Init = (PyObject *(*)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer *, TMM::NonlinearLayer *, int, PyObject *))__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer__Init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer)) __PYX_ERR(0, 874, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer) < (0)) __PYX_ERR(0, 874, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer) < (0)) __PYX_ERR(0, 874, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer, __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer) < (0)) __PYX_ERR(0, 874, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer) < (0)) __PYX_ERR(0, 874, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_NonlinearLayer, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer) < (0)) __PYX_ERR(0, 874, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer) < (0)) __PYX_ERR(0, 874, __pyx_L1_error)
  __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM = &__pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM;
  __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM._Init = (PyObject *(*)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM *, TMM::NonlinearTMM *))__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM__Init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM)) __PYX_ERR(0, 964, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM) < (0)) __PYX_ERR(0, 964, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM) < (0)) __PYX_ERR(0, 964, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM) < (0)) __PYX_ERR(0, 964, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM) < (0)) __PYX_ERR(0, 964, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_NonlinearTMM, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM) < (0)) __PYX_ERR(0, 964, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM) < (0)) __PYX_ERR(0, 964, __pyx_L1_error)
  __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities = &__pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities;
  __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities._Init = (PyObject *(*)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities *, TMM::SecondOrderNLIntensities *))__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities__Init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities)) __PYX_ERR(0, 1513, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities) < (0)) __PYX_ERR(0, 1513, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities) < (0)) __PYX_ERR(0, 1513, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities, __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities) < (0)) __PYX_ERR(0, 1513, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities) < (0)) __PYX_ERR(0, 1513, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_SecondOrderNLIntensities, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities) < (0)) __PYX_ERR(0, 1513, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SecondOrderNLIntensities) < (0)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM = &__pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM;
  __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM._Init = (PyObject *(*)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM *, TMM::SweepResultSecondOrderNLTMM *))__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM__Init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM)) __PYX_ERR(0, 1554, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM) < (0)) __PYX_ERR(0, 1554, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM) < (0)) __PYX_ERR(0, 1554, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM, __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM) < (0)) __PYX_ERR(0, 1554, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM) < (0)) __PYX_ERR(0, 1554, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_SweepResultSecondOrderNLTMM, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM) < (0)) __PYX_ERR(0, 1554, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__SweepResultSecondOrderNLTMM) < (0)) __PYX_ERR(0, 1554, __pyx_L1_error)
  __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM = &__pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM;
  __pyx_vtable_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM._Init = (PyObject *(*)(struct __pyx_obj_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM *, TMM::WaveSweepResultSecondOrderNLTMM *))__pyx_f_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM__Init;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM)) __PYX_ERR(0, 1605, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM) < (0)) __PYX_ERR(0, 1605, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM) < (0)) __PYX_ERR(0, 1605, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM, __pyx_vtabptr_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM) < (0)) __PYX_ERR(0, 1605, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM) < (0)) __PYX_ERR(0, 1605, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_WaveSweepResultSecondOrderNLTMM, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM) < (0)) __PYX_ERR(0, 1605, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__WaveSweepResultSecondOrderNLTMM) < (0)) __PYX_ERR(0, 1605, __pyx_L1_error)
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM)) __PYX_ERR(0, 1658, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM_spec, __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM) < (0)) __PYX_ERR(0, 1658, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM = &__pyx_type_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM) < (0)) __PYX_ERR(0, 1658, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount((PyObject*)__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM->tp_dictoffset && __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_SecondOrderNLTMM, (PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM) < (0)) __PYX_ERR(0, 1658, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM) < (0)) __PYX_ERR(0, 1658, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_2_4(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  0, 0,
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_5numpy_dtype = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "dtype",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArray_Descr),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArray_Descr),
  #else
  sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArray_Descr),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 229, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_flatiter = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "flatiter",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayIterObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayIterObject),
  #else
  sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayIterObject),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 274, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_broadcast = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "broadcast",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayMultiIterObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayMultiIterObject),
  #else
  sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayMultiIterObject),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 278, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_ndarray = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "ndarray",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayObject),
  #else
  sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyArrayObject),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 317, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_generic = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "generic",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_generic) __PYX_ERR(1, 826, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_number = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "number",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_number) __PYX_ERR(1, 828, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_integer = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "integer",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_integer) __PYX_ERR(1, 830, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_signedinteger = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "signedinteger",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_signedinteger) __PYX_ERR(1, 832, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "unsignedinteger",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(1, 834, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_inexact = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "inexact",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_inexact) __PYX_ERR(1, 836, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_floating = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "floating",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_floating) __PYX_ERR(1, 838, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_complexfloating = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "complexfloating",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_complexfloating) __PYX_ERR(1, 840, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_flexible = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "flexible",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_flexible) __PYX_ERR(1, 842, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_character = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "character",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #else
  sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_character) __PYX_ERR(1, 844, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_5numpy_ufunc = __Pyx_ImportType_3_2_4(__pyx_t_1, "numpy", "ufunc",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyUFuncObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyUFuncObject),
  #else
  sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_3_2_4(PyUFuncObject),
  #endif
  __Pyx_ImportType_CheckSize_Ignore_3_2_4); if (!__pyx_mstate->__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 908, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__SecondOrderNLTMMCython(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__SecondOrderNLTMMCython},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, __Pyx_FREETHREADING_COMPATIBLE},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "_SecondOrderNLTMMCython",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit__SecondOrderNLTMMCython(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__SecondOrderNLTMMCython(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL && defined(GRAALPY_VERSION_NUM) && GRAALPY_VERSION_NUM > 0x19000000
    PY_INT64_T current_id = GraalPyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && (__PYX_LIMITED_VERSION_HEX < 0x03090000\
      || ((defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)) && __PYX_LIMITED_VERSION_HEX < 0x030A0000))
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__SecondOrderNLTMMCython(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_SecondOrderNLTMMCython' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "_SecondOrderNLTMMCython" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
    PyErr_Clear();
    __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
    if (!__Pyx_RefNanny)
        Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  
__Pyx_RefNannySetupContext("PyInit__SecondOrderNLTMMCython", 0);
  __Pyx_init_runtime_version();
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Library function declarations ---*/
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__pyx_module_is_main_NonlinearTMM___SecondOrderNLTMMCython) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "NonlinearTMM._SecondOrderNLTMMCython")) {
      if (unlikely((PyDict_SetItemString(modules, "NonlinearTMM._SecondOrderNLTMMCython", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_type_init_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code(__pyx_mstate);
  (void)__Pyx_modinit_function_import_code(__pyx_mstate);
  /*--- Execution code ---*/

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1
 * import numpy as np             # <<<<<<<<<<<<<<
 * from CppSecondOrderNLTMM cimport *
 * 
*/
  __pyx_t_1 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_numpy, 0, 0, NULL, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1;
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_np, __pyx_t_2) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":56
 *         raise NotImplementedError()
 * 
 * cdef paramDict = {"wl": PARAM_WL, \             # <<<<<<<<<<<<<<
 *                   "beta": PARAM_BETA, \
 *                   "pol": PARAM_POL, \
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyLong_From_enum__TMM_3a__3a_TMMParam(TMM::TMMParam::PARAM_WL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_wl, __pyx_t_3) < (0)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":57
 * 
 * cdef paramDict = {"wl": PARAM_WL, \
 *                   "beta": PARAM_BETA, \             # <<<<<<<<<<<<<<
 *                   "pol": PARAM_POL, \
 *                   "I0": PARAM_I0, \
*/
  __pyx_t_3 = __Pyx_PyLong_From_enum__TMM_3a__3a_TMMParam(TMM::TMMParam::PARAM_BETA); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_beta, __pyx_t_3) < (0)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":58
 * cdef paramDict = {"wl": PARAM_WL, \
 *                   "beta": PARAM_BETA, \
 *                   "pol": PARAM_POL, \             # <<<<<<<<<<<<<<
 *                   "I0": PARAM_I0, \
 *                   "overrideE0": PARAM_OVERRIDE_E0, \
*/
  __pyx_t_3 = __Pyx_PyLong_From_enum__TMM_3a__3a_TMMParam(TMM::TMMParam::PARAM_POL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_pol, __pyx_t_3) < (0)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":59
 *                   "beta": PARAM_BETA, \
 *                   "pol": PARAM_POL, \
 *                   "I0": PARAM_I0, \             # <<<<<<<<<<<<<<
 *                   "overrideE0": PARAM_OVERRIDE_E0, \
 *                   "E0": PARAM_E0, \
*/
  __pyx_t_3 = __Pyx_PyLong_From_enum__TMM_3a__3a_TMMParam(TMM::TMMParam::PARAM_I0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_I0, __pyx_t_3) < (0)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":60
 *                   "pol": PARAM_POL, \
 *                   "I0": PARAM_I0, \
 *                   "overrideE0": PARAM_OVERRIDE_E0, \             # <<<<<<<<<<<<<<
 *                   "E0": PARAM_E0, \
 *                   "mode": PARAM_MODE,
*/
  __pyx_t_3 = __Pyx_PyLong_From_enum__TMM_3a__3a_TMMParam(TMM::TMMParam::PARAM_OVERRIDE_E0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_overrideE0, __pyx_t_3) < (0)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":61
 *                   "I0": PARAM_I0, \
 *                   "overrideE0": PARAM_OVERRIDE_E0, \
 *                   "E0": PARAM_E0, \             # <<<<<<<<<<<<<<
 *                   "mode": PARAM_MODE,
 *                   "w0": PARAM_WAVE_W0}
*/
  __pyx_t_3 = __Pyx_PyLong_From_enum__TMM_3a__3a_TMMParam(TMM::TMMParam::PARAM_E0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_E0, __pyx_t_3) < (0)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":62
 *                   "overrideE0": PARAM_OVERRIDE_E0, \
 *                   "E0": PARAM_E0, \
 *                   "mode": PARAM_MODE,             # <<<<<<<<<<<<<<
 *                   "w0": PARAM_WAVE_W0}
 * 
*/
  __pyx_t_3 = __Pyx_PyLong_From_enum__TMM_3a__3a_TMMParam(TMM::TMMParam::PARAM_MODE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_mode, __pyx_t_3) < (0)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":63
 *                   "E0": PARAM_E0, \
 *                   "mode": PARAM_MODE,
 *                   "w0": PARAM_WAVE_W0}             # <<<<<<<<<<<<<<
 * 
 * cdef paramDictSecondOrderNLTMM = set(["deltaWlSpdc",
*/
  __pyx_t_3 = __Pyx_PyLong_From_enum__TMM_3a__3a_TMMParam(TMM::TMMParam::PARAM_WAVE_W0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_w0, __pyx_t_3) < (0)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_XGOTREF(__pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_paramDict);
  __Pyx_DECREF_SET(__pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_paramDict, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":65
 *                   "w0": PARAM_WAVE_W0}
 * 
 * cdef paramDictSecondOrderNLTMM = set(["deltaWlSpdc",             # <<<<<<<<<<<<<<
 *                                       "solidAngleSpdc",
 *                                       "deltaThetaSpdc"])
*/
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_deltaWlSpdc) < (0)) __PYX_ERR(0, 65, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_solidAngleSpdc) < (0)) __PYX_ERR(0, 65, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_deltaThetaSpdc) < (0)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_paramDictSecondOrderNLTMM);
  __Pyx_DECREF_SET(__pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_paramDictSecondOrderNLTMM, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":69
 *                                       "deltaThetaSpdc"])
 * 
 * cdef waveParamsSet = set(["waveType",             # <<<<<<<<<<<<<<
 *                 "pwr",
 *                 "overrideE0",
*/
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_waveType) < (0)) __PYX_ERR(0, 69, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_pwr) < (0)) __PYX_ERR(0, 69, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_overrideE0) < (0)) __PYX_ERR(0, 69, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_E0) < (0)) __PYX_ERR(0, 69, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_w0) < (0)) __PYX_ERR(0, 69, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_Ly) < (0)) __PYX_ERR(0, 69, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_a) < (0)) __PYX_ERR(0, 69, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_nPointsInteg) < (0)) __PYX_ERR(0, 69, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_maxX) < (0)) __PYX_ERR(0, 69, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_dynamicMaxX) < (0)) __PYX_ERR(0, 69, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_dynamicMaxXCoef) < (0)) __PYX_ERR(0, 69, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_dynamicMaxXAddition) < (0)) __PYX_ERR(0, 69, __pyx_L1_error)
  if (PySet_Add(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_maxPhi) < (0)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_waveParamsSet);
  __Pyx_DECREF_SET(__pyx_v_12NonlinearTMM_23_SecondOrderNLTMMCython_waveParamsSet, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":114
 *         self._parent = parent # Avoid dealloc of parent
 * 
 *     def Update(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """Update(**kwargs)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_5Update, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Chi2Tensor_Update, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor, __pyx_mstate_global->__pyx_n_u_Update, __pyx_t_2) < (0)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":190
 *             raise ValueError("Unknown kwarg %s" % (k))
 * 
 *     def GetChi2Tensor(self):             # <<<<<<<<<<<<<<
 *         """
 *         Returns rotated 3x3x3 second-order nonlinearity tensor.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_7GetChi2Tensor, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Chi2Tensor_GetChi2Tensor, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Chi2Tensor, __pyx_mstate_global->__pyx_n_u_GetChi2Tensor, __pyx_t_2) < (0)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_9__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Chi2Tensor___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_11_Chi2Tensor_11__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Chi2Tensor___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":251
 *         return self.GetN(wl)
 * 
 *     def GetN(self, wl):             # <<<<<<<<<<<<<<
 *         """GetN(wl)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_7GetN, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Material_GetN, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material, __pyx_mstate_global->__pyx_n_u_GetN, __pyx_t_2) < (0)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":277
 *         return self._thisptr.GetN(wl)
 * 
 *     def IsNonlinear(self):             # <<<<<<<<<<<<<<
 *         """IsNonlinear()
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_9IsNonlinear, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Material_IsNonlinear, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_Material, __pyx_mstate_global->__pyx_n_u_IsNonlinear, __pyx_t_2) < (0)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_11__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Material___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[6])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_8Material_13__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Material___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":389
 *             del self._thisptr
 * 
 *     def SetParams(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """SetParams(**kwargs)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_5SetParams, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Wave_SetParams, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__Wave, __pyx_mstate_global->__pyx_n_u_SetParams, __pyx_t_2) < (0)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Wave___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[9])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_5_Wave_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Wave___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Intensities___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12_Intensities_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_Intensities___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SweepResultNonlinearTMM___reduc, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_24_SweepResultNonlinearTMM_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SweepResultNonlinearTMM___setst, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_WaveSweepResultNonlinearTMM___r, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[15])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_28_WaveSweepResultNonlinearTMM_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_WaveSweepResultNonlinearTMM___s, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[16])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_FieldsZ___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[17])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_8_FieldsZ_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_FieldsZ___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[18])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_FieldsZX___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[19])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_9_FieldsZX_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_FieldsZX___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[20])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":840
 *         self._thisptr = ptr
 * 
 *     def GetMainFields(self, double z):             # <<<<<<<<<<<<<<
 *         """GetMainFields(z)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_3GetMainFields, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_HomogeneousWave_GetMainFields, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[21])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__HomogeneousWave, __pyx_mstate_global->__pyx_n_u_GetMainFields, __pyx_t_2) < (0)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_HomogeneousWave___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[22])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16_HomogeneousWave_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_HomogeneousWave___setstate_cyth, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[23])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":902
 *         self.hw = hw
 * 
 *     def GetIntensity(self, double z):             # <<<<<<<<<<<<<<
 *         """GetIntensity(z)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_3GetIntensity, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearLayer_GetIntensity, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[24])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer, __pyx_mstate_global->__pyx_n_u_GetIntensity, __pyx_t_2) < (0)) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":920
 *         return self._thisptr.GetIntensity(z);
 * 
 *     def GetAbsorbedIntensity(self):             # <<<<<<<<<<<<<<
 *         """GetAbsorbedIntensity()
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_5GetAbsorbedIntensity, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearLayer_GetAbsorbedInten, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[25])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer, __pyx_mstate_global->__pyx_n_u_GetAbsorbedIntensity, __pyx_t_2) < (0)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":932
 *         return self._thisptr.GetAbsorbedIntensity();
 * 
 *     def GetSrcIntensity(self):             # <<<<<<<<<<<<<<
 *         """GetSrcIntensity()
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_7GetSrcIntensity, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearLayer_GetSrcIntensity, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[26])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython__NonlinearLayer, __pyx_mstate_global->__pyx_n_u_GetSrcIntensity, __pyx_t_2) < (0)) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_9__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearLayer___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[27])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_15_NonlinearLayer_11__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearLayer___setstate_cytho, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[28])); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1044
 *     #---------------------------------------------------------------------------
 * 
 *     def AddLayer(self, double d, Material material):             # <<<<<<<<<<<<<<
 *         """AddLayer(d, material)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_5AddLayer, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM_AddLayer, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[29])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_mstate_global->__pyx_n_u_AddLayer, __pyx_t_2) < (0)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1077
 *         self.layers.append(layer)
 * 
 *     def SetParams(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """SetParams(**kwargs)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_7SetParams, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM_SetParams, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[30])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_mstate_global->__pyx_n_u_SetParams, __pyx_t_2) < (0)) __PYX_ERR(0, 1077, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1089
 *             setattr(self, name, value)
 * 
 *     def Solve(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """Solve(**kwargs)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_9Solve, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM_Solve, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[31])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_mstate_global->__pyx_n_u_Solve, __pyx_t_2) < (0)) __PYX_ERR(0, 1089, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1119
 *         self._thisptr.Solve()
 * 
 *     def GetIntensities(self):             # <<<<<<<<<<<<<<
 *         """GetIntensities()
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_11GetIntensities, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM_GetIntensities, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[32])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_mstate_global->__pyx_n_u_GetIntensities, __pyx_t_2) < (0)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1135
 *         return res
 * 
 *     def Sweep(self, str paramStr, np.ndarray[double, ndim = 1] values, int layerNr = 0, double layerZ = 0.0, bool outPwr = True, bool outAbs = False, outEnh = False):             # <<<<<<<<<<<<<<
 *         """Sweep(paramStr, values, layerNr = 0, layerZ = 0, outPwr = True, outAbs = False, outEnh = False)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_13Sweep, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM_Sweep, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[33])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[0]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_mstate_global->__pyx_n_u_Sweep, __pyx_t_2) < (0)) __PYX_ERR(0, 1135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1196
 *         return res
 * 
 *     def GetFields(self, np.ndarray[double, ndim = 1] zs, str dir = "total"):             # <<<<<<<<<<<<<<
 *         """GetFields(zs, dir = "total")
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_15GetFields, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM_GetFields, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[34])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_mstate_global->__pyx_n_u_GetFields, __pyx_t_2) < (0)) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1222
 *         return res
 * 
 *     def GetFields2D(self, np.ndarray[double, ndim = 1] zs, np.ndarray[double, ndim = 1] xs, str dir = "total"):             # <<<<<<<<<<<<<<
 *         """GetFields2D(zs, xs, dir = "total")
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_17GetFields2D, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM_GetFields2D, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[35])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_mstate_global->__pyx_n_u_GetFields2D, __pyx_t_2) < (0)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1250
 *         return res
 * 
 *     def GetAbsorbedIntensity(self):             # <<<<<<<<<<<<<<
 *         """GetAbsorbedIntensity()
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_19GetAbsorbedIntensity, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM_GetAbsorbedIntensit, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[36])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_mstate_global->__pyx_n_u_GetAbsorbedIntensity, __pyx_t_2) < (0)) __PYX_ERR(0, 1250, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1263
 *         return self._thisptr.GetAbsorbedIntensity();
 * 
 *     def GetEnhancement(self, int layerNr, double z = 0.0):             # <<<<<<<<<<<<<<
 *         """GetEnhancement(layerNr, z = 0.0)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_21GetEnhancement, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM_GetEnhancement, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[37])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_2);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_mstate_global->__pyx_tuple[2]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_mstate_global->__pyx_n_u_GetEnhancement, __pyx_t_2) < (0)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1289
 *     # Waves
 * 
 *     def WaveGetPowerFlows(self, int layerNr, double x0 = float("nan"), double x1 = float("nan"), double z = 0.0):             # <<<<<<<<<<<<<<
 *         """WaveGetPowerFlows(layerNr, x0 = float("nan"), x1 = float("nan"), z = 0.0)
 * 
*/
  __pyx_t_4 = __Pyx_PyUnicode_AsDouble(__pyx_mstate_global->__pyx_n_u_nan); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_4, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L1_error)
  __pyx_mstate_global->__pyx_k_ = __pyx_t_4;
  __pyx_t_4 = __Pyx_PyUnicode_AsDouble(__pyx_mstate_global->__pyx_n_u_nan); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_4, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L1_error)
  __pyx_mstate_global->__pyx_k__2 = __pyx_t_4;
  __pyx_t_4 = __Pyx_PyUnicode_AsDouble(__pyx_mstate_global->__pyx_n_u_nan); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_4, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L1_error)
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyUnicode_AsDouble(__pyx_mstate_global->__pyx_n_u_nan); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_4, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L1_error)
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_Pack(3, __pyx_t_2, __pyx_t_3, __pyx_mstate_global->__pyx_float_0_0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_23WaveGetPowerFlows, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM_WaveGetPowerFlows, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[38])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_mstate_global->__pyx_n_u_WaveGetPowerFlows, __pyx_t_3) < (0)) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1323
 *         return (res.first, res.second)
 * 
 *     def WaveGetEnhancement(self, int layerNr, double z = 0.0):             # <<<<<<<<<<<<<<
 *         """WaveGetEnhancement(layerNr, z = 0.0)
 * 
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_25WaveGetEnhancement, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM_WaveGetEnhancement, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[39])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_mstate_global->__pyx_tuple[2]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_mstate_global->__pyx_n_u_WaveGetEnhancement, __pyx_t_3) < (0)) __PYX_ERR(0, 1323, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1347
 *         return res
 * 
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] values, \             # <<<<<<<<<<<<<<
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \
 *             bool outR = False, bool outT = False, bool outEnh = False):
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_27WaveSweep, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM_WaveSweep, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[40])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_mstate_global->__pyx_tuple[3]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_mstate_global->__pyx_n_u_WaveSweep, __pyx_t_3) < (0)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1416
 *         return res
 * 
 *     def WaveGetFields2D(self, np.ndarray[double, ndim = 1] zs, \             # <<<<<<<<<<<<<<
 *             np.ndarray[double, ndim = 1] xs, str dirStr = "total"):
 *         """WaveGetFields2D(zs, xs, dirStr = "total")
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_29WaveGetFields2D, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM_WaveGetFields2D, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[41])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_NonlinearTMM, __pyx_mstate_global->__pyx_n_u_WaveGetFields2D, __pyx_t_3) < (0)) __PYX_ERR(0, 1416, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_31__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[42])); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_3) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_12NonlinearTMM_33__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NonlinearTMM___setstate_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[43])); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_3) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SecondOrderNLIntensities___redu, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[44])); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_3) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_25_SecondOrderNLIntensities_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SecondOrderNLIntensities___sets, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[45])); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_3) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SweepResultSecondOrderNLTMM___r, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[46])); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_3) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_28_SweepResultSecondOrderNLTMM_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SweepResultSecondOrderNLTMM___s, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[47])); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_3) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_WaveSweepResultSecondOrderNLTMM_2, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[48])); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_3) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_32_WaveSweepResultSecondOrderNLTMM_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_WaveSweepResultSecondOrderNLTMM_3, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[49])); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_3) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1730
 *     #---------------------------------------------------------------------------
 * 
 *     def SetParams(self, **kwargs):             # <<<<<<<<<<<<<<
 *         """SetParams(**kwargs)
 * 
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_5SetParams, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SecondOrderNLTMM_SetParams, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[50])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, __pyx_mstate_global->__pyx_n_u_SetParams, __pyx_t_3) < (0)) __PYX_ERR(0, 1730, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1742
 *             setattr(self, name, value)
 * 
 *     def AddLayer(self, double d, Material material):             # <<<<<<<<<<<<<<
 *         """AddLayer(d, material)
 * 
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_7AddLayer, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SecondOrderNLTMM_AddLayer, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[51])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, __pyx_mstate_global->__pyx_n_u_AddLayer, __pyx_t_3) < (0)) __PYX_ERR(0, 1742, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1764
 *         self.materialsCache.append(material)
 * 
 *     def Solve(self):             # <<<<<<<<<<<<<<
 *         """Solve()
 * 
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_9Solve, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SecondOrderNLTMM_Solve, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[52])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, __pyx_mstate_global->__pyx_n_u_Solve, __pyx_t_3) < (0)) __PYX_ERR(0, 1764, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1772
 *         self._thisptr.Solve()
 * 
 *     def UpdateGenParams(self):             # <<<<<<<<<<<<<<
 *         """UpdateGenParams()
 * 
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_11UpdateGenParams, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SecondOrderNLTMM_UpdateGenParams, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[53])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, __pyx_mstate_global->__pyx_n_u_UpdateGenParams, __pyx_t_3) < (0)) __PYX_ERR(0, 1772, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1780
 *         self._thisptr.UpdateGenParams()
 * 
 *     def GetIntensities(self):             # <<<<<<<<<<<<<<
 *         """GetIntensities()
 * 
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_13GetIntensities, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SecondOrderNLTMM_GetIntensities, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[54])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, __pyx_mstate_global->__pyx_n_u_GetIntensities, __pyx_t_3) < (0)) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1799
 *         return res
 * 
 *     def Sweep(self, str paramStr, np.ndarray[double, ndim = 1] valuesP1, np.ndarray[double, ndim = 1] valuesP2, int layerNr = 0, double layerZ = 0.0, bool outPwr = True, bool outAbs = False, bool outEnh = False, bool outP1 = True, bool outP2 = True, bool outGen = True):             # <<<<<<<<<<<<<<
 *         """Sweep(paramStr, valuesP1, valuesP2, layerNr = 0, layerZ = 0.0, outPwr = True, outAbs = False, outEnh = False, outP1 = True, outP2 = True, outGen = True)
 * 
*/
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_15Sweep, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SecondOrderNLTMM_Sweep, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[55])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_3);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_mstate_global->__pyx_tuple[4]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, __pyx_mstate_global->__pyx_n_u_Sweep, __pyx_t_3) < (0)) __PYX_ERR(0, 1799, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1874
 *         return res
 * 
 *     def WaveGetPowerFlows(self, int layerNr, double x0 = float("nan"), double x1 = float("nan"), double z = 0.0):             # <<<<<<<<<<<<<<
 *         """WaveGetPowerFlows(layerNr, x0 = float("nan"), x1 = float("nan"), z = 0.0)
 * 
*/
  __pyx_t_4 = __Pyx_PyUnicode_AsDouble(__pyx_mstate_global->__pyx_n_u_nan); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_4, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 1874, __pyx_L1_error)
  __pyx_mstate_global->__pyx_k__3 = __pyx_t_4;
  __pyx_t_4 = __Pyx_PyUnicode_AsDouble(__pyx_mstate_global->__pyx_n_u_nan); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_4, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 1874, __pyx_L1_error)
  __pyx_mstate_global->__pyx_k__4 = __pyx_t_4;
  __pyx_t_4 = __Pyx_PyUnicode_AsDouble(__pyx_mstate_global->__pyx_n_u_nan); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_4, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 1874, __pyx_L1_error)
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyUnicode_AsDouble(__pyx_mstate_global->__pyx_n_u_nan); if (unlikely(__PYX_CHECK_FLOAT_EXCEPTION(__pyx_t_4, ((double)((double)-1))) && PyErr_Occurred())) __PYX_ERR(0, 1874, __pyx_L1_error)
  __pyx_t_5 = PyFloat_FromDouble(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_Pack(3, __pyx_t_3, __pyx_t_5, __pyx_mstate_global->__pyx_float_0_0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_17WaveGetPowerFlows, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SecondOrderNLTMM_WaveGetPowerFlo, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[56])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, __pyx_mstate_global->__pyx_n_u_WaveGetPowerFlows, __pyx_t_5) < (0)) __PYX_ERR(0, 1874, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1909
 *         return (res.first, res.second)
 * 
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] valuesP1, \             # <<<<<<<<<<<<<<
 *             np.ndarray[double, ndim = 1] valuesP2,
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_19WaveSweep, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SecondOrderNLTMM_WaveSweep, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[57])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_mstate_global->__pyx_tuple[5]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, __pyx_mstate_global->__pyx_n_u_WaveSweep, __pyx_t_5) < (0)) __PYX_ERR(0, 1909, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1994
 *         return res
 * 
 *     def WaveGetFields2D(self, np.ndarray[double, ndim = 1] zs, \             # <<<<<<<<<<<<<<
 *                         np.ndarray[double, ndim = 1] xs, str dirStr = "total"):
 *         """WaveGetFields2D(zs, xs, dirStr = "total")
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_21WaveGetFields2D, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SecondOrderNLTMM_WaveGetFields2D, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[58])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_mstate_global->__pyx_tuple[1]);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_12NonlinearTMM_23_SecondOrderNLTMMCython_SecondOrderNLTMM, __pyx_mstate_global->__pyx_n_u_WaveGetFields2D, __pyx_t_5) < (0)) __PYX_ERR(0, 1994, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_23__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SecondOrderNLTMM___reduce_cython, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[59])); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_5) < (0)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
*/
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_12NonlinearTMM_23_SecondOrderNLTMMCython_16SecondOrderNLTMM_25__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SecondOrderNLTMM___setstate_cyth, NULL, __pyx_mstate_global->__pyx_n_u_NonlinearTMM__SecondOrderNLTMMCy, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[60])); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030E0000
  PyUnstable_Object_EnableDeferredRefcount(__pyx_t_5);
  #endif
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_5) < (0)) __PYX_ERR(2, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1
 * import numpy as np             # <<<<<<<<<<<<<<
 * from CppSecondOrderNLTMM cimport *
 * 
*/
  __pyx_t_5 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_Chi2Tensor_Update_line_114, __pyx_mstate_global->__pyx_kp_u_Update_kwargs_This_method_change) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_Material_GetN_line_251, __pyx_mstate_global->__pyx_kp_u_GetN_wl_Returns_refractive_index) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_NonlinearTMM_AddLayer_line_1044, __pyx_mstate_global->__pyx_kp_u_AddLayer_d_material_Adds_layer_t) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_mstate_global->__pyx_kp_u_NonlinearTMM_Solve_line_1089, __pyx_mstate_global->__pyx_kp_u_Solve_kwargs_Solves_the_structur) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_5) < (0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init NonlinearTMM._SecondOrderNLTMMCython", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init NonlinearTMM._SecondOrderNLTMMCython");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */
/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);

  /* Cached unbound methods */
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_items.method_name = &__pyx_mstate->__pyx_n_u_items;
  __pyx_mstate->__pyx_umethod_PyDict_Type_keys.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_keys.method_name = &__pyx_mstate->__pyx_n_u_keys;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_values.method_name = &__pyx_mstate->__pyx_n_u_values;
  return 0;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1135
 *         return res
 * 
 *     def Sweep(self, str paramStr, np.ndarray[double, ndim = 1] values, int layerNr = 0, double layerZ = 0.0, bool outPwr = True, bool outAbs = False, outEnh = False):             # <<<<<<<<<<<<<<
 *         """Sweep(paramStr, values, layerNr = 0, layerZ = 0, outPwr = True, outAbs = False, outEnh = False)
 * 
*/
  __pyx_mstate_global->__pyx_tuple[0] = PyTuple_Pack(5, __pyx_mstate_global->__pyx_int_0, __pyx_mstate_global->__pyx_float_0_0, Py_True, Py_False, Py_False); if (unlikely(!__pyx_mstate_global->__pyx_tuple[0])) __PYX_ERR(0, 1135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1196
 *         return res
 * 
 *     def GetFields(self, np.ndarray[double, ndim = 1] zs, str dir = "total"):             # <<<<<<<<<<<<<<
 *         """GetFields(zs, dir = "total")
 * 
*/
  __pyx_mstate_global->__pyx_tuple[1] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_n_u_total); if (unlikely(!__pyx_mstate_global->__pyx_tuple[1])) __PYX_ERR(0, 1196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[1]);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1263
 *         return self._thisptr.GetAbsorbedIntensity();
 * 
 *     def GetEnhancement(self, int layerNr, double z = 0.0):             # <<<<<<<<<<<<<<
 *         """GetEnhancement(layerNr, z = 0.0)
 * 
*/
  __pyx_mstate_global->__pyx_tuple[2] = PyTuple_Pack(1, __pyx_mstate_global->__pyx_float_0_0); if (unlikely(!__pyx_mstate_global->__pyx_tuple[2])) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[2]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[2]);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1347
 *         return res
 * 
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] values, \             # <<<<<<<<<<<<<<
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \
 *             bool outR = False, bool outT = False, bool outEnh = False):
*/
  __pyx_mstate_global->__pyx_tuple[3] = PyTuple_Pack(6, __pyx_mstate_global->__pyx_int_0, __pyx_mstate_global->__pyx_float_0_0, Py_True, Py_False, Py_False, Py_False); if (unlikely(!__pyx_mstate_global->__pyx_tuple[3])) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[3]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[3]);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1799
 *         return res
 * 
 *     def Sweep(self, str paramStr, np.ndarray[double, ndim = 1] valuesP1, np.ndarray[double, ndim = 1] valuesP2, int layerNr = 0, double layerZ = 0.0, bool outPwr = True, bool outAbs = False, bool outEnh = False, bool outP1 = True, bool outP2 = True, bool outGen = True):             # <<<<<<<<<<<<<<
 *         """Sweep(paramStr, valuesP1, valuesP2, layerNr = 0, layerZ = 0.0, outPwr = True, outAbs = False, outEnh = False, outP1 = True, outP2 = True, outGen = True)
 * 
*/
  __pyx_mstate_global->__pyx_tuple[4] = PyTuple_Pack(8, __pyx_mstate_global->__pyx_int_0, __pyx_mstate_global->__pyx_float_0_0, Py_True, Py_False, Py_False, Py_True, Py_True, Py_True); if (unlikely(!__pyx_mstate_global->__pyx_tuple[4])) __PYX_ERR(0, 1799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[4]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[4]);

  /* "NonlinearTMM/src/SecondOrderNLTMM.pyx":1909
 *         return (res.first, res.second)
 * 
 *     def WaveSweep(self, str paramStr, np.ndarray[double, ndim = 1] valuesP1, \             # <<<<<<<<<<<<<<
 *             np.ndarray[double, ndim = 1] valuesP2,
 *             int layerNr = 0, double layerZ = 0.0, bool outPwr = True, \
*/
  __pyx_mstate_global->__pyx_tuple[5] = PyTuple_Pack(9, __pyx_mstate_global->__pyx_int_0, __pyx_mstate_global->__pyx_float_0_0, Py_True, Py_False, Py_False, Py_False, Py_True, Py_True, Py_True); if (unlikely(!__pyx_mstate_global->__pyx_tuple[5])) __PYX_ERR(0, 1909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[5]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[5]);
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_tuple;
    for (Py_ssize_t i=0; i<6; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  {
    const struct { const unsigned int length: 11; } index[] = {{564},{29},{548},{24},{33},{30},{37},{179},{761},{16},{21},{1577},{1},{8},{7},{6},{2},{9},{50},{39},{34},{14},{8},{11},{25},{18},{29},{31},{2},{8},{9},{27},{29},{26},{28},{20},{13},{14},{9},{11},{14},{12},{13},{4},{15},{16},{30},{34},{36},{2},{12},{30},{32},{11},{2},{8},{13},{20},{26},{28},{15},{36},{28},{31},{33},{35},{12},{21},{33},{27},{22},{24},{27},{22},{18},{18},{31},{28},{30},{22},{36},{30},{32},{20},{25},{43},{45},{16},{25},{31},{26},{22},{22},{32},{32},{34},{26},{34},{36},{9},{5},{5},{24},{42},{44},{28},{46},{48},{6},{15},{5},{18},{15},{17},{9},{28},{46},{48},{32},{50},{52},{15},{23},{25},{1},{18},{8},{4},{3},{18},{1},{2},{14},{11},{3},{3},{6},{9},{14},{5},{11},{19},{15},{7},{8},{8},{12},{1},{2},{2},{2},{4},{8},{10},{13},{5},{1},{1},{4},{6},{9},{5},{7},{6},{8},{8},{8},{6},{4},{4},{10},{12},{8},{4},{3},{9},{2},{2},{5},{6},{6},{6},{5},{5},{6},{4},{4},{7},{10},{1},{5},{10},{8},{6},{4},{4},{4},{9},{3},{3},{2},{3},{11},{14},{12},{4},{10},{17},{13},{3},{6},{1},{4},{12},{10},{12},{19},{3},{14},{4},{4},{7},{10},{8},{5},{5},{5},{6},{8},{8},{2},{8},{2},{3},{2},{2},{2},{1},{5},{2},{27},{55},{56},{28},{17},{22},{71},{337},{21},{56},{51},{51},{13},{93},{14},{38},{168},{39},{71},{104},{228},{15},{17},{16},{16},{16},{59},{9},{251},{170},{67}};
    #if (CYTHON_COMPRESS_STRINGS) == 2 /* compression: bz2 (3620 bytes) */
const char* const cstring = "BZh91AY&SY\371\035\347\354\000\004\204\177\377\377\177\377\377\377\377\377\377\377\377\377\377\377\377\377\375\300@@@@@@@@@@@@\000@\000`\016C\357*^z\347\225\336\356\272\035\2654DP;\271\300\rj \326\225I\001\332\240\033p\224HI\341J{\020\247\264\320\301\240\324z\206\223&iO\023\322G\250\3321\014\221\246\217SG\2652C\324\323G\243S5\036\214Q\265?A\032\201\241\004\023e\032\247\344\312z\232\215\344SL\206\2322\000h\320\000\000\000\000\000\000\000\000\000\004EO\311\344B\207\251\352\036\220h\r\000\000\006\200\000\000\000\000\000\000\006\200\000h\000H\210\223\010)\243$\311\266\246\243\323B\r12\031\r4\323M\r\r\000\033Hh\017Ph4\r\000\r2z\232\032 \300\0010\000\021\200\000F\000\010\321\211\246\000\230\000\002a\030F\002d\304b\006\004\212\010\002h\3214\304\021\350\232\021\351\250\303L\222i\372\247\244\332\032&\206\200\000\0004\000\000\0004\321\246\237\362\000\351=\267\222L\3111\255\207\006\244h\263*\342\242\2308\316d\302\360?9A\327,^\"ZPPQ5\364\277M\275\226\370\277\276E\266ts\363\215\014:\033\0329\230\333A\322\207\203g\305n\352\324\275\275\367\266\266\340~\007\337y\367\340\336\356\373\235\364\364\227\327\227\3169\"\340:\243\020\352\005\021\025D\374\225\020c\001\260\035\245\222\306<\342*\rp+\027\276\333\004\r\244\014\031cbF\353x!\021hb\246\234\357\275#q\013\363h\220\353\240g%\207!\246AF\r\210\211\236\363\220\343;J\017\333\2610\260h\331\302\010\375~\372=\341\\\231\371BKa\254F\306\222`\320/$\200bNSD\004D\032\027\214\335\354R\212\245R\251:\020\210\025Dot\315]>\314Rhs[\324\313\215a7B/Eb\272\233\346\025\261P\n\205vH\304\260\251\024\227/\205X\"\302\231W\031\004l\256\274\264H\244Y\003\326\2626U\002tW\tq$\t-\250\004{\277\333V\247\026r\265\357\354l\021\261\322=\346\323>>\266\301r\241\017\245\300\231(\350\345\230\244\207&]\3678\352\314A\356\357\241\013Q\0058\222!-\374s\360`\235\347\202&\240\037<\016\005RjQ\r&\217\220|\007:\002\032\026 v\001\246EL\262U\211p\224\243\025y\205\t\202v\206H\033$\0359\020\034)x\341\214u\220\334\341m\243\2117\016\322G q\200-}c\251\250\036#\027 \264\322a\341\n\351|0\251""\241\027\313!@u\002f\006\361\244\245f\006\n)\320i\t\021\221\210@\034\001L\000\031\345\231\224\003`B\265\255P(\246\230\007+X\021t\317Nz$W\013b5'\030\325M\360\256\247d\351\212\246.\027\321\030O\004:\t\220vUL\353\264s)\r\351\324\343a\r\313o\242n\367\013\274\360\357\235\377\026\201``\323\024h\r\0339\255\275\317\361\033'\315\234\363\360@\245\314\004\215\315'\t^9\322\031fi\355\3011\002{tGZq\214\024\030\250\266h\313\273\313h\3003\267\342k2\304\234\262z\270\216j=\214;\037FG\177]e6\301l\r9N\311\320O\200\232${\376\337g\210\366\315\3036\177C\025\254\352\352\214\313g[&}Xa\204b\210\3545J\320\353\230\036|\214\026\320.n\030 w\220'\276\237Z\210\371\375\206\024\234;\276:\343\034M\263\273C\262\363\022\245\242\013\346\201A,DCM\2432\214\301m\314\"m]D\325F\207m\3449%\272A \312m7d\313\002\214\217'Z\333\244\346E\221\234e\027\"t@'\244\364LN)\023z{\237t\351\237\203_\361>\206\243\213\260\303-SS5{\323\207\236TY\351\214\255uE\350\201\333\247r\2519'\231Lv\2025/\025E|^\233|\343'\303<z\341\360S\210x9\317T\2212\035@\210\242\217\035\3720\203Y0#\231\301\214\273\244 \260\002\324\215\031\300kj\032\nM\261$\006\331\nQ\266\276\312\341\277z\2749\312?0\244\310Z\327\006x\360a\t=b6,\345Dc\0222\313\204\3141h\202\302\345\276Eh\327bK\347@\201\247\035M\372\037\023\232\272D\215?\020\005\220\0277\250VGv\214\347\312=\233\203Y\3570\273\242\316M\324\256\r\247\\\336\212\336\230}\305@9\261\007\326\374{\2771\031\233\311\354\365\247\340\3440\301c\334\332\212\376\0054\365\251\311\351\305y\331Q0\2444\324\031\363\366qU\001\343w\212\"y*\362\360\334e,\2679\240\355u;H\205d~\354\243\261\234\271\213\027\237\030\361\233\030\300\240+\023u(f\024\325\031\223\327\3643\230\2210(\363M\021w\036T\342\004\267\342\266\251\225\200\\Pn$\214\210\330\206\236\036Xu\270\304#\334\000LS+\277\313dB\254\374:\315%&H\236c7\024DD\016\"*\205w\031\"\333b4\354\3351\213V\005\201\034\006\202\274\333N\356\3506$\216\003|>\246\216\327\247~-\327b\340\t\030\261\340\340\031\254\243\232\3606(#\214\356s\223\3568\312\210U\232\273u\202\300\205\222\013\327\006\254""\346\375p\337p\254\245\237\316\004\r>\340\226+2t7\270ZX\030\367\3634\271\266\333m\267\ri~\347K9q\300hm\261\2661\266\333\033m\246\330\333m\266\364w\200\314\260yrTF\355\234\025\252s\345E\031\241*\215\375|\262\355\366&_u\326\266\335\2673\006o9\t\316\017'\317\307\310g\370\347!\333$d%\"8\007\270@\247\272RG\020<\014\221\363\250\234\036\002\340\206\350\346\027\334?\353_^V\316c7Cp,\232\217<\303[g\267\346\305v\321\3408tf\245\265R\033\233#\333z\214\340\276\014@U\303\230\202Z0|V\233m\214ci\246\3041\004\r\274\372\234y\246H\203\242\363R\014\\<\211,\331\253\212\211+\265\347!C-\000\006G s\202\033\213@\212\331Q\0173\237\320\274\364E\353\233\241\231r\365\311\236\356\332\362\224\265\203\303\027\246?\220\202\377\335\020^-\207\001\n\201}\276d\006\363$\343\352+\255J\302\372\302\000\323\336\014\260\027\r\225@7f\215\212R\212\253\356\322\260I\007\021\322\r\213\236b\302\222\0250/\213\025\272\246>\0354m\275\204\263\276\350\2250\253rK@\316@+\004c\025\251\260N2P\202_^p\203\330\035\2531[RC`\321U\211\373Ab\352\3120lM\2666?\010\274H\351\361\234\344FJ2\222a\016oc\177\213\220\327z\341\313\317-\217|\031\356\003Xz\265\"\264\327%G\237\t\\\220Y\020\353\331\3062\371\317/r\035i^:{\3071*\344V\204\235j\310`Z\251\276\330U\001\204aP\001\257\325\375\027VT\262\300\202`\330\301\256\333\035\367I^\004\322E3\306Gm\"\0147@\377\331`f\264/\256\316\334\014z9cM\340\226\004\356F5~\262%\300\307%\2674ZcJ\2008Om\020\031+9\326\364\202\245D\225\263W3\257\\\2110\345'\001\210\033CM\004\352$\030\253\207<\366\021\005,\332\305\244A\021dk\243\221\271\344\346\002\035ov<D0\326\241\217(`TF\321\244Cn\222\344\244\350K\030Pm\246\332\033Li\215\2176&!\021\241x\271I\"\245\367r\356\273V\242=\024\033\363\322\247\222v\244\027\331\317& \307\327\307\312\315\256\263\214\272\273%\221 \321\2403\327\274\312\364\203\321 \377\205\021\321\331F@\254\242\244\013\330\002\200\242y0\n\304\027\366sy5:\221t\0225(\244\210\013]\230\224\010$l\031hD(k\020\331(\232#\003\031c8d\335\014\354\250k$M0,\311#`*l[\251\013P4C)\213\206P\036V\001Q*\256\3036q\256_k\3166x{""\336\023\326=\355\237]\361\"\014O\002\273\366\220F\0326\3279\323\005\3104\0369\020\317\322\261;\276\n\016\277\212\325\032\235\230\205\263'\341\200aZ\246(\365\241\345\354\245 \355Y\021\264\034\251\201\006\034\244d\277\350\033H\r(\316>H\305 *\277\305\353\354\363R9\371\340\221\237\272\313\032K\255\327Q\333|\003\231T\273\365P\303\030j\362\266\200\251\006\242{\373\346\316\347E\364\245S\241S\335\222\221rI\027D\366\316\250g\341\315\034\264\321Y\305)\244\216\n\303=T\206_]@y[\021\222\351,Z* \0312J\014\211O\266K&V[\022\343\301\215]\340v=\331\003 \246\323\331\257\211\270S\242:~\231*\362\361\tj8\\\026\353P\210\340#\330kj\221j-\261\255\213\235\242\220k\007\037F{\222T\006\0003\010\342\006=!\010\234\266\275!\353!oU\226g&\235\342\t\016\247 &\224\347a\267w\332(\363k\030VY\311\271\311\272\216\373T\267u\212-\n\274\3010\022\rZ\217\036\334a\205Q\000\356-P+\270\267\340R\230\223{[B\"T\251\0068/\204\222\255!\272\014\025,g\212\251~$b\243P\261\003TP\201H\323\t\030$\\\354C\325)\002h\365! \371h\246+\276@B=b\223\261`\032\332u\344e\304\214m\364\007\010\267Z\241\001\302\300(`\304P\322\202\2724W\247\260\325\314\334au;\210P\356\354:mH\200t\345\3102\367\360\322\257\336\016K\031z\271\254\256\010\001\241E\tU\002\351\243\322\020\rO\266Q\200.\323\310\214X\224\027\305)+j\223\267\226\322\270\"\n\206m\212\263PF\031g\335\357S\351\263\340\366\374&\335\365Y\300\271|\377\201#0\203z\355.\254r!a]\240\031\270\304\271U$9(\232V\240\313x\003\001\313\003\004\256sF\343\033A\004\341@\264Z\025\324\273\276aGz}\254\236X\306H:x\003\322=C\331\232\365wm\250\002\226\252F \222dj\321\221\016d1\330\233C\342\"\010\030g\204@\330\323l\034@\200\205\\\004(Dd\273\314\346:\352\204)\232U+6\323\246\202H\360\244\213\255\254\231E\006\232\006\200\322o*-\272r$\025\210,\2308\215\223\302w\311\201:{H\271\025Z\030\212\\\350\246\001e \366)S\200\23595\324\2640\n\363t\271\361\312\351\333T\212\240J0UAU*%1L\363yJB\312\350\242\230\221E\310\276\204\313w\361\210\235m\010PC\354m\324\272\245R\246\006\214\323\"\221XaC\"\265\0105\270\345@Y\207\233\214\233""\000\351P}\026a\275\035\0376E\215\266M\231^\221\222\265\322`\3003+Oy\371\214\240K\027F\2137\374\235\237'\263\321B)\262\254c$\023L*2\nL\002\314\315\255UX\236\221\t8\333J.b\202\321P\224\236W\"\222\351\256)\256)\260$\362\243\n\3500D\033\360h\246H.\204\205\\l\214\240\200\250\t\310ABL7\036\260\246\204\n{\004\005W\322\024 \205\313!u\024\001\250\025\365U\005M\250\314\320G%v\327\030\341#\252\004)\244\235X\010\035\327NJ\211\010\341P\267\317\026\256\3471~\337ks\264H\265 \312p\346\214Bq\027\024\347\250\245\264'\306c\257\275\273C\\\271\002\304\303If\360E#_\036\242\372B\336\255\356\033t:F\334\006$\306\003aKE\240\342\345\277\325\253\215\025/\013-f=q\005\312+\016\203\276DxUAV\347\243\025!RWn8\241\315\236\220D4Ko\306\375\373J0\312\331\002\027\231L\261>KD\322G\022\315M3\017\013\311\350\224\000\0069\002\223\314<\200\261\001I\027\254\365\321\2113I`\354GI\262Ba/\332\305f\343L\266t\"\245\010\253\373:\202K\251\326c\023\023K\256\03544/\310L\335\220/]P\336\020\222\204h\241\376\242\214Q\016\037\t\0209\314r\336\014K]\2221\032C\004\343Krg\037<$5\320\327O\261\263\350\252 \251\372o\021\277\352\302\331?\264\007\tq\205\001d\304\336\3719\003\034\250\225\034Jv\254\326\254\025*/7\264j\260\303\204\213\326m\013\206\033\252H\213\325\340E*$\\|q\007\274\303\335_\334\303O{\033B5-\207h3\361L\365O\201\300o\230\341\261\243\213mS\034\256lb\327\027R8\201q\264uj\225\203\315\350\014I\225\016\206S\n\253\313;\212\310\322\257\212\201@^-\252\030S\215\t\334yf\326\265\236\317n\216\203)\362\322\207[\365<\235\214\335\322\0260d$<\002\322\362\371ud\223\273\2318\t6\377\230\372.N\326\276\235\274+\336\336\277\312\212W\275E[\236\372\227\245mS\303\020\3414\315-\232\306\234\302\211\360BW&W\007_\315J\377\372\3315\332\302{\316\364\316\251\264\314\212F\2441\314\276\\&\022<\226\220~U\312QGoJfpE`\325\222j\307\261ww\311\230:I\361\247\017|\226}\376\356T\377\014\274\362\004\371\375\233\220\235\336\341\247|U\374h\347\215\200j)AU\214VS1\333\005\\j\362\306\375=]\005\205\246\330+\335\314\270l\233+\032\276\210\3056D\343\220D\230\"\220\202""\222\242\212\212J@Y`E\307\230\316cM\267\364\177\335\223\036C\007\216B\374\031\0212F^\3743\264~\320o\226S\377\213\271\"\234(H|\216\363\366\000";
    PyObject *data = __Pyx_DecompressString(cstring, 3620, 2);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #elif (CYTHON_COMPRESS_STRINGS) != 0 /* compression: zlib (3433 bytes) */
const char* const cstring = "x\332\265Y\313[\333\306\026\307\300\355\007-m \257\333\264i\356\220\276 1\256m\310\273I\353\022H\311\203` )\244\017!\2441V\220%[\017l\323~\367\336%K-\265\364\322K/\275d\311RK/\363'\360'\334sF\017$\313\220W/_,\315\314\2319\317\33793\243\344D\3611_\247\332\204\230$%\336\240\232\304\313\223\037~H\274\277\234(\352D\306\t\304P\211Q\244d\365\311\223\324\021}\211\327\370\022\205ez04\025\374\005C\"\271M\n\262\312\033\301\010\376\255\026%a[\241\272N\324\002c\355\312\221\024\342rL\221yI\323\r\302+\242G\206N\211\212\222Y\"E~\207F\230IJAR$\203\302L\217k*\240\373v\201\026\267\005\340\242\337\336x\342\rmD\230,(\272\301+\002\365\025*R\271\014\032\2615=\226\006k\347j|\251,\323\270\017\202\201{\367\356\021\243T\"w\321\177\023\223\335\343\251\234\037\006\346\246\211\313`\316\345\311$\361e\245V\014\336\220\204\211L\352\332\344\t\213\257\245\351\324\255\370\252tj\232\\%\323/'\337Cn:\264\230\233-J\331U\252\350\252\226zV\026a.\231\220%\205\222Lff\362\0015\026'\252a\014-S\303\324\024\235h\264\240\361\202!\355P\210\226Hk\350\345 4\274A\3642\025\244\202DER\205\350\312T\3312\212o\r\265\252\334\023k\277\004\034C\370\212\251\330\3158\350\013*\306\267\026a9\353\216\305\315\002S\216\014\000\205Ro\013\224\252\254\003P\224r\212\3274\276>\361\353L\332\215\353u\357}\223\275\177O\022\321\250\227)Le\366F\203\253DY\000p\000\002\000\204\227I\222I\335`\355,k\337d\355\314\313\020;\317\332(C\210\024\220|`@\204\365$\010\211\315\231p\225<\032?\222v\344\206\000^\010\026\017;\331k\231\311EU\3016\257a\221\361\241\351c+=3\023\235\260\242\312;\001\362\3227oE\250\337\351\232\360\335\n\025TE|\252\211T[|\214K\312\365\332\242\312\212\004X3[7\212\252B$\235\210T\2266\251\006Z\311u\242\033\232$\030X\360\212\274B\226\346\226\246fn\316\260\"\244\321\227T0t\242\233\233\254\026PV\2716MI6\000T\350<\250Y\013\005RWM\242P\3001\224\3142\226\216\320\002(+\n\321\251\301\352\313\267\274\242\250\230a\252\302\301rI\331\372\226\210\222F]4\301\352y^\326\251k\350\304\225+\333U^\333\322C\231\305\010\214'jm\n\000b\372\036\031\223$j\031u""\341\345\010\320\217I\036\237ZV\221\001\310?f\371\345\362e\242j\344\262~\031\374\014\346\202\225LcX\307k\322.\263\336/\270\2307!\326\233\324\340_\243\334\242\252\225xY\332Eo\363\312\026U\014\254&\310\207\354\200\037\241B\035%_\032xy\330>\206\333\202R6\rR\226y\305\325\205`\272J\206)bn\223\035^0\315R\212<U\000&\246\016\"\245\002\331\330Pw\250\246I\"\235Kol\000\232\242{\234f\322#\r\026\322\257\261&&_R\014\250\263\222QG\371\350\241\002\333\022\331N\031U$\302'\246\224\207\244`\322\021\0314\332TU\371\030\205V \360\010y\210D\225\002r76\030C\314\206\215\215\005h\246\336a\027\364\000\r\300\273K\256MgYMc\241\276\013u\"\035\302\3673e[Q\253\na\300'_\353~?\014\035o\007\352\225\036p\300\320\021\256EU$B\021\321\341\346\312\016/\233n\356\352\254@L\251X!\210\342\227\017t\267\341no\344\247\372\321)\206\026xSv3\327%\243cw\251\246B%T\r\242a*C,\3209\242\244\003\320\005c^\2422\234\240X\302\253\036\032\3362?\205\242\264\266\376\3425\300q\235\240\373z1\013O40I\252E\252Q\262\226$\353I\362\"\302\214\245)%\212Y\202\242\250\223\202\246\226\240\214\303!BP5\215\352e\340\310\222X%\265\251$\251\303ow\212\257I\241\304\025\327\036\275\203\302\000\263:\310P\014\334O\301\221~m\364uE\256\010\221k\344\212\347\225\010K\364\373#\267\234\303Y\020\003\241\337&\031X\221\311\334!Yxg\263w\3104\274\247\247\357\220\031\354O'\3114\214]\303\261\014l\2060~\035\347g\223$\233\tY\023\r\346\211\t\263P\010\007;\351\326{(\375f\211\201\017\266\035\243\252B2c\236\263j\207\211\035\004\206\2245U\200\263k\324.\215\272%\036\300B&hj\313\213*\034\202\265\222\252H\002\201\262\207\273\027h3\3116 \226\354\236lxt\035\224Qt\3015\005Y#ti\305\204\2529\301\303\010\354r\272\217r\257Y\225d\031\022\224\224\240#A=\004\327B\240\262\223\251\010\343\373^z@\000\214H\325+C\250^\203\206U\024\346E\033\342\270%\203\241\032/J<\2340@I\325\004kk\014c\376N\341\271\240\013\330~\332\206e\257\377\r\262\353\357(\373\305\337 {\367md\277E1\3663\260\233I\364X\356\336\275\024\257\270\001Z\215\220\220\370\2710|Q\351\236\226\202\234\315\372\345Z\314\260\314\244S""\230lb\226\245\247\333\301Y,K\261{2\023\206,\254\010\260\214E\232U\207\244\353\371\273d\n\316\265\241\355\340\007^\0249\254l\220\317\374\246\014\333\031>\267\004Iw[\242\242\0065\236\3434*\232\002\3458\"\232\354(\006n\232\202\223\341\016z\205\343\004\274lr\034\224\310r=\305A]\244)\226\035\354\270M\n\274$\273\307?\251TV5#<\315\004#\212\261\031\337\343\251S\331\322US\023\350=\377\350\033\272h\205\357\\pj\356Mp\335\022\036\t\354\020\330a\227\343\242D(U:\3065 \317\2459\267\312\275\360\337kA\243\007\263\020)\306\312'\036\277\254\327*\260-\267\t\312mRq\301?\373D\354\205\316\234R\304[z\t\016{\320s\231\005\215\354}h\372K%\252\207z\310\350\t/)\301\212E\370\255hB@\347~VK*\026S\325\324\361\320\333\335OE\030\304\250qC\3433b\006/\244\271\220\266\341v\017~Qj\234\227\036\\\201\036\327\375\224\214\\\266\202Nhj0\026\223\027\242\304\303{$\212A5\332M\365\212c\2179'\321\"\221\351\"\307]\023\233\020\3238|=\354y\323\214\014\3662\240{B\010\210\335$\027\"\275G\263\367\273\307Cq\215^q\251\301\016\211z\374\342\033\035\201\303y92\202p;AC\217\334S\037\217\266\244V\2516/\253U=F\215\213\343\272\257\332\356\345::\247;f]\3248\304\270\245z\r~\367\341>\316-\322\232\261L\013QA\341\3149\216\320\003+'L\215)\321mX\354\233\202\217\236\030!\032\327\0309\210m\234\202\361\215\217\242\323c\243n\311\177@\225cxu\005\3728\372Q\260{\316\350-;\346\330\0363z\370\3237\233Y\211\2149\366\\f\207\3150&\216\033\357\025\321cg\306Q\025\232\333\255\360I\264\223\245\276\211\351n\254\272\"\306\241\177\343\331\332\025\267X\230\202\250pA\253\207\003O\240\3650\347\344\331qGv\315\2179\3635\364\327k\360&Nek\216\222\211;f'>n\373\345y\275\256\010\222\232\202\303\231j\3025\217\352\233\274\200\237\022D\374$\001\247M\001\235\300I\n\207\327R\212D8F\212T6x8\275\033\374JY\024X\357\027\2315\013[\242\244\301\277\025C\363>\345\341u9|\201d\237Y\305\272\302\227$\341\t_[\0135\241\226Hl\301\321\320\254J\013\005UC\2058\256`*\002\307m\361\246\256\303\035\001Z\2769\234$e\244\2544-\225\370-\220$\211\370!\014\216\250+\354\263 '\351\\`\237d\320\222""\376r{\233\326u\367\213\t\376\277\n\253`r\360Xt_/\330\263lh\034W\202c\017>\335\003A\211\257-\025AVm\255\244\212 \035\236\246\014oeI\225\024C\307\262\267\005=\270\260\272O\005?tzhR\312\212\316\016\304\240\016\356\260&\302\036\236\220\021\360\\\312\340#\213\217**\274\014\277U\370\225x}\373\350\243U\271\214\301f\017\2469k\201\255\360\006\303\361Z\200\327\001\274\006\260oix\321.\253rY-\227\365r\025\254)\303\276\3429\016\233;\006\236\377q\303\301{\260\253\266Fy\371\350\026\020\007\224?@k\330\324\341\337,p\327\251\\`(\363\214\207\226w\245\010a\217\353\201C\275\000\207\177Y\022sx\tD \351\370\253h\206^1)\335\2450\033\332U\311(r\034\334\303\200\235\00172\3310!\216\354\343\211\373\005\305}.e\274w\266\232F\333W\001pU\271*\353\265t-S\323w\361\233\225\276\253;\203\327Z\371\303+}\303\247\366Dk\322\376\2419\336\274\325\312u\206N\355U\234\301[\355sm\336\371a\305YYs\326\326\235u\331\221K\207\367\274\271\316\347S\315|Sj\031\355\231v\24534j\rX7\355i\233?f\331\217\2117X7\026h\342|\236j\362\315z;\301T\371o\3560\3357<\272gX\267\354\271\306Xc\034\006\256\367\r\177\274\367\320\022\354\261\316\320\310\336\202\365\334\316\3747\327\231\177p8\3237|\325\271r\257\315#\3731\353\254\265\356\\J5\305\326d;\323\376y\177\341 \177@\235\225\337\235\337\251C\013\235\241O\366$\013\346}\264w\335\032\263\306}i\213\t\344>\007\213\363V\001EvFF\255A+g\375\332H4\3167\214\3464h\247\267\276l\345;\243c\257F\306\2543@\313[<\260\212v\316Z\363\366W6oW\220\301\260}\016\232Fc\246Q9\374\340\004\001\2376\3176WZ\211\326\231\326l\253\002\002:'\0108I\332\247\340\252\257\032b\363\253&\1778\3247<\342|t\321\236E\342\027\366sT\242\231`\363\234\013W\233\231f\216M\331\273a}i\345;#g\254\353\366Y;o\363\3075\221\377\305F\276Qh\316\267\306{Xs\332\272b\347\234/\322\255\201\026\304\345p\002\303\367\227\235\261\3577\206\233\227Z\225v\002\006/\341\252\331\275\272\235\260\317\3309\014c\244s\037D\214\331_\332y\227\375S\373Fc\274\201\314\316cw\036\340z\035b\001\346\234\3323\255\007v\2463z\336z\331H8\3433\314o\240m\006\374\361M\003E}\362""\366KFp\211\207+\304%v\326\301\271zc\2429\036Q\t;\227\354\325\306\205f\262%\264O\2673\235\241\253\315\031@\357\360\376\330\3768b\374\024x}\350\343\275\347\300\376\241\275\335d\220}`\335\260\307\031sW\222s\356\353F\336\265\356\233\306j\363B\353\273\3751\304\247l\217\007\370\2642\036B;\237M9S?\356':\237^u\256\302\373\3609 \366\242\235y\00533\235\221\177\272\230>\276\205*\023Tjd\357g\353\221]i$\020=\271\316\310\005\364\276\275\006\221\345_\301\310\323F\246\221{\345\246\221\375\262\231h~\326\272\325\316\265\327\367\245\203\277\234_EG\334v\266uG7@Q\347\223+\315\361X&}\3267|\232-V\233\371W8\351\333F\245\2079\267\357\034\246\376\206\274=\3357\374/{\263\361A\203\007t\017@\226\212\255\211v\000\317sn\226\214Z\377\260\236\241\353;\243g\2559\226\223\225\316Yp\212=\000\205\210Bf\324[\247\332\225\375\376\375\313\373\263\373\332\301\351\203\351\203Mgi\331Y^\361\344\274\352\362~\347B\322Ib\024\212n\024:o\030\205\377k\214\3768\270\345\254>s\236m8\033\222#\231\216\371\357\303\276\276\377$\346\373\3415\337\277\210\257\305\376\247\375,rPlc\276<\357W\334\307ML\007\342w\2374\307\232X\200\317z\003\316\305Lk,4?\336O6s\316\340\354~\006\263\350\264\247\345`s\241\225oQ\347\316\323\203\212\263\004jBd\377\300\260\026\255nU\022\235\301\017\367\256Zyg\320\265`\031U_\356\247\370\242\375u|\325\373\027\007\320\236\201M|m\016\230\3702\007\026\006\341\2650\370p\360p\343\375\003\363N\341(4r\215\027\270\217\265/\355\327\235\245\274\223\377\315\371\r\020\\q*\273\240\333\237\211\237P\375\237\372\037\342\353a\377\243\376\336\211\344\014\202\207\212N\361O\230\365Wb\t'/\365o\340k\243\237\357g\333\325k\223\377=\365\277\260?y\360\275\363\013\240i\313\331R\035\265\3144\375\246\301wkZp\n[\230\314n\244e\330}$\250\367C\373\303\007c\007_;O\005G\200\262\001\210\004\027h=\363\370\177\230\005\266\250";
    PyObject *data = __Pyx_DecompressString(cstring, 3433, 1);
    if (unlikely(!data)) __PYX_ERR(0, 1, __pyx_L1_error)
    const char* const bytes = __Pyx_PyBytes_AsString(data);
    #if !CYTHON_ASSUME_SAFE_MACROS
    if (likely(bytes)); else { Py_DECREF(data); __PYX_ERR(0, 1, __pyx_L1_error) }
    #endif
    #else /* compression: none (9360 bytes) */
const char* const bytes = "AddLayer(d, material)\n\n        Adds layer to the TMM.\n\n        Parameters\n        ----------\n        d : float\n            Thickness of the layer in meters. First and the last medium have\n            infinite thickness.\n        material : :class:`Material`\n            Instance of the helper class :class:`Material`\n\n        Examples\n        --------\n        >>> tmm = TMM()\n        >>> tmm.AddLayer(float(\"inf\"), Material.Static(1.5))\n        >>> tmm.AddLayer(50e-9, Material.Static(0.3 + 3j))\n        >>> tmm.AddLayer(float(\"inf\"), Material.Static(1.0))\n        _Chi2Tensor.Update (line 114)GetN(wl)\n\n        Returns refractive index of material at specified wavelength.\n\n        Parameters\n        ----------\n        wl : float\n            Wavelength in meters.\n\n        Returns\n        -------\n        complex\n            Complex refractive index at wavelength wl.\n\n        Examples\n        --------\n        >>> wls = np.array([400e-9, 600e-9, 800e-9], dtype = float)\n        >>> ns = np.array([1.5 + 0.3j, 1.7 + 0.2j, 1.8 + 0.1j], dtype = complex)\n        >>> mat = Material(wls, ns)\n        >>> mat(600e-9)\n        1.7 + 0.2j\n\n        Material.GetN (line 251)NonlinearTMM.AddLayer (line 1044)NonlinearTMM.Solve (line 1089)NonlinearTMM/src/SecondOrderNLTMM.pyxNote that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.Solve(**kwargs)\n\n        Solves the structure.\n\n        Parameters\n        ----------\n        wl : float, optional\n             Wavelength in meters.\n        pol : str, optional\n            \"p\" or \"s\" denoting the polarization of the waves.\n        beta : float, optional\n            Normalized tangential wave vector.\n        E0 : complex, optional\n            Input plane wave amplitude in vacuum. Only used if ``overrideE0`` is\n            True.\n        I0 : float, optional\n            In""put plane wave intensity in the first layer. Only used if\n            ``overrideE0`` is False.\n        overrideE0 : bool, optional\n            Selects between ``E0`` and ``I0``.\n\n        Examples\n        --------\n        >>> tmm.Solve(wl = 532e-9, beta = 0.0)\n\n        Unknown kwarg %sUnknown polarization.Update(**kwargs)\n\n        This method changes the values of second-order nonlinearity tensor. By\n        default the tensor is zero, not rotated and distinctFields set to True.\n\n        Parameters\n        ----------\n        chiXYZ : float, optional\n            Updates tensor value of second-order nonlinearity, where X, Y, Z\n            denote numbers from 1..3 corresponding to x-, y-, z-axis.\n        dXK : float, optional\n            Updates tensor values by contracted notation, where dXK = 0.5 * chiXYZ\n            and K is defined as: 1 = 11; 2 = 22; 3 = 33; 4 = 23, 32; 5 = 31, 13; 6 = 12, 21.\n        distinctFields : bool, optional\n            If set to True, then assumes that two input waves in nonlinear process\n            are the same (e.g second harmonic generation). If False, then the\n            input fields are not equal (as a result the result will be multiplied by 2).\n            Default is true.\n        phiX : float, optional\n            The rotation angle (radians) around x-axis of the second order nonlinear tensor.\n        phiY : float, optional\n            The rotation angle (radians) around y-axis of the second order nonlinear tensor.\n        phiZ : float, optional\n            The rotation angle (radians) around z-axis of the second order nonlinear tensor.\n\n        Examples\n        --------\n        Updates nonlinear tensor of material and then rotates it.\n\n        >>> mat = Material.Static(1.5)\n        >>> mat.chi2.Update(d11 = 1e-12, d22 = 2e-12, chi23 = 3e-12)\n        >>> mat.chi2.Update(phiX = 0.2, phiY = 0.5, phiZ = -0.1)\n\n        ?add_notedisableenablegcisenabledno default __reduce__ due to non-trivial _""_cinit__numpy._core.multiarray failed to importnumpy._core.umath failed to import<stringsource>AddLayer_Chi2Tensor_Chi2Tensor.GetChi2Tensor_Chi2Tensor.Update_Chi2Tensor.__reduce_cython___Chi2Tensor.__setstate_cython__E0_FieldsZ_FieldsZX_FieldsZX.__reduce_cython___FieldsZX.__setstate_cython___FieldsZ.__reduce_cython___FieldsZ.__setstate_cython__GetAbsorbedIntensityGetChi2TensorGetEnhancementGetFieldsGetFields2DGetIntensitiesGetIntensityGetMainFieldsGetNGetSrcIntensity_HomogeneousWave_HomogeneousWave.GetMainFields_HomogeneousWave.__reduce_cython___HomogeneousWave.__setstate_cython__I0_Intensities_Intensities.__reduce_cython___Intensities.__setstate_cython__IsNonlinearLyMaterialMaterial.GetNMaterial.IsNonlinearMaterial.__reduce_cython__Material.__setstate_cython___NonlinearLayer_NonlinearLayer.GetAbsorbedIntensity_NonlinearLayer.GetIntensity_NonlinearLayer.GetSrcIntensity_NonlinearLayer.__reduce_cython___NonlinearLayer.__setstate_cython__NonlinearTMMNonlinearTMM.AddLayerNonlinearTMM.GetAbsorbedIntensityNonlinearTMM.GetEnhancementNonlinearTMM.GetFieldsNonlinearTMM.GetFields2DNonlinearTMM.GetIntensitiesNonlinearTMM.SetParamsNonlinearTMM.SolveNonlinearTMM.SweepNonlinearTMM.WaveGetEnhancementNonlinearTMM.WaveGetFields2DNonlinearTMM.WaveGetPowerFlowsNonlinearTMM.WaveSweepNonlinearTMM._SecondOrderNLTMMCythonNonlinearTMM.__reduce_cython__NonlinearTMM.__setstate_cython____Pyx_PyDict_NextRef_SecondOrderNLIntensities_SecondOrderNLIntensities.__reduce_cython___SecondOrderNLIntensities.__setstate_cython__SecondOrderNLTMMSecondOrderNLTMM.AddLayerSecondOrderNLTMM.GetIntensitiesSecondOrderNLTMM.SetParamsSecondOrderNLTMM.SolveSecondOrderNLTMM.SweepSecondOrderNLTMM.UpdateGenParamsSecondOrderNLTMM.WaveGetFields2DSecondOrderNLTMM.WaveGetPowerFlowsSecondOrderNLTMM.WaveSweepSecondOrderNLTMM.__reduce_cython__SecondOrderNLTMM.__setstate_cython__SetParamsSolveSweep_SweepResultNonlinearTMM_SweepResultNonlinearTMM.__reduce_cython___SweepResultNonlinearTMM.__setstate_cython___SweepResultSecondOr""derNLTMM_SweepResultSecondOrderNLTMM.__reduce_cython___SweepResultSecondOrderNLTMM.__setstate_cython__UpdateUpdateGenParams_WaveWaveGetEnhancementWaveGetFields2DWaveGetPowerFlowsWaveSweep_WaveSweepResultNonlinearTMM_WaveSweepResultNonlinearTMM.__reduce_cython___WaveSweepResultNonlinearTMM.__setstate_cython___WaveSweepResultSecondOrderNLTMM_WaveSweepResultSecondOrderNLTMM.__reduce_cython___WaveSweepResultSecondOrderNLTMM.__setstate_cython___Wave.SetParams_Wave.__reduce_cython___Wave.__setstate_cython__aasyncio.coroutinesbackwardbetachicline_in_tracebackdd_deltaThetaSpdcdeltaWlSpdcdfgdirdirStrdirectiondistinctFieldsdtypedynamicMaxXdynamicMaxXAdditiondynamicMaxXCoefforward__func__gaussian__getstate__ii1i2i3imagincidentinitStruct_is_coroutineitemsjkkeyskwargslastLayerlayerlayerNrlayerZlayerptr__main__materialmaxPhimaxXmode__module__nPointsInteg__name__namenannonlinearnpnsnumpyoutAbsoutEnhoutGenoutP1outP2outPwroutRoutToutmaskoverrideE0pparamparamLayerparamStrparentphiXphiYphiZplanewavepolpoppspwr__pyx_state__pyx_vtable____qualname__real__reduce____reduce_cython____reduce_ex__resresCppsself__set_name__setdefault__setstate____setstate_cython__sfgsolidAngleSpdcspdcsqrtsqueezestartswith__test__totaltukeyvaluevaluesvaluesP1valuesP2w0waveTypewlwlsx0x1xszzeroszs\320\0045\260Q\360*\000\t\017\210d\220)\230?\250!\2509\260A\330\010\017\210q\320\0049\270\025\270a\320?S\320SX\320XY\320Yl\320lm\360>\000\t\017\210d\220)\320\033-\250Q\250i\260t\2704\270q\330\010\020\220\003\2208\2303\230a\320\0049\270\025\270a\320?S\320SX\320XY\320Yl\320lm\360@\001\000\t\017\210d\220)\320\033-\250Q\250i\260t\2704\270q\330\010\020\220\003\2208\2303\230a\320\0049\270\021\360*\000\t\017\210d\220)\320\033.\250a\250y\270\001\330\010\017\210q\200A\3600\000\t\020\210t\2209\230E\240\021\240!\200A\3606\000\t\r\210J\220c\230\021\330\010\014\210I\220V\2301\200A\330FG\3604\000\t+\320*>\270a\270q\330\010\021\220\024\220Y\320\036.\250d\260)\2701\270H\300I\310Q\310e\320S]\320]e\320ef\330\010\016\210i\220q\330\010\013""\2106\220\021\220!\330\010\017\210q\200A\360N\001\000\t\r\210E\220\024\220Q\220f\230E\240\021\330\014\020\220\004\220A\220[\240\001\240\026\240t\2503\250a\250s\260#\260Q\330\020\021\340\014\021\220\023\220A\220Q\220a\220q\330\014\021\220\023\220A\220Q\220a\220q\330\014\024\220F\230$\230a\230q\330\014\020\220\t\230\025\230a\230t\2404\240q\360\006\000\t\r\210E\220\024\220Q\220f\230E\240\021\330\014\020\220\004\220A\220[\240\001\240\030\250\024\250S\260\001\260\023\260C\260q\330\020\021\330\014\021\220\023\220A\220Q\220a\220q\330\014\021\220\023\220A\220Q\220a\220q\330\014\021\220\023\220A\220Q\220a\220q\330\014\024\220F\230$\230a\230q\330\014\020\220\t\230\030\240\021\240$\240d\250$\250a\360\010\000\t\014\320\013\034\230C\230q\330\014\035\230V\2404\240q\250\001\330\014\020\220\t\320\031+\2501\250A\360\010\000\t\014\2107\220#\220Q\330\014\023\2206\230\024\230Q\230a\330\014\023\2206\230\024\230Q\230a\330\014\023\2206\230\024\230Q\230a\330\014\020\220\t\230\034\240Q\240f\250F\260!\360\006\000\t\r\210E\220\024\220Q\220f\230E\240\021\330\014\022\220*\230A\320\0350\260\003\2601\200A\360(\000\t\020\210|\2301\230D\240\t\250\036\260q\270\001\200A\360\036\000\t\r\210C\210y\230\001\230\023\230A\330\010\014\210C\210y\230\001\230\023\230A\330\010\014\210D\220\t\230\021\230#\230Q\360\006\000\t\r\210O\2307\240!\2401\200A\360\026\000\t\r\210F\220)\2306\240\026\240q\330\014\017\210u\220G\2301\330\020\026\220j\240\001\320!4\260C\260q\330\014\023\2201\220F\230&\240\001\200A\360\016\000\t\r\210F\220)\2306\240\026\240q\330\014\017\210u\220G\2301\330\020\026\220j\240\001\320!4\260C\260q\330\014\023\2201\220F\230&\240\001\200A\360\014\000\t\r\210I\220V\2301\200A\360*\000\t\r\210I\220Y\230a\230s\240(\250!\360\006\000\t\r\210O\2307\240!\2401\360\006\000\t\036\230T\240\031\250,\260c\270\022\2701\330\010+\2504\250y\270\t\300\021\300!\330\010\020\220\017\230q\330\010\r\210V\2201\220J\230k\250\021\330\010\014\210G\2207\230!\2301\200A\360\014\000\t\r\210I\320\025%\240Q\200A\360\026\000\t&\240T""\250\031\260/\300\021\330\010\016\210l\230!\330\010\013\2106\220\021\220!\2201\330\010\017\210q\200A\330\032-\320-@\300\001\330\030+\320+@\300\001\360V\001\000\t\034\2301\340\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\360\006\000\t \230q\330\010\014\210H\220K\230q\240\001\330\014\024\220A\330\014\031\230\023\230A\230X\240Q\240a\340\014\024\220O\2401\240A\340\010\021\220\024\220Y\230j\250\001\250\032\2609\270A\270Y\300i\310|\320[d\320dk\320ks\320st\330\010\016\320\016*\250!\330\010\013\2106\220\021\220!\330\010\017\210q\200A\360\032\000\t\022\220\024\220Y\230o\250Q\340\010\016\320\016'\240q\330\010\013\2106\220\021\220!\2201\330\010\017\210q\200A\330:;\360.\000\t+\320*>\270a\270q\330\010\021\220\024\220Y\320\036.\250d\260)\2701\270H\300I\310Q\310e\320S]\320]e\320ef\330\010\016\210i\220q\330\010\013\2106\220\021\220!\330\010\017\210q\200A\360\022\000\t\037\230b\240\006\240a\240q\250\003\2503\250d\260(\270!\360\006\000\t\r\210E\220\025\220a\220q\330\014\020\220\005\220U\230!\2301\330\020\024\220E\230\025\230a\230q\330\024\027\220q\230\003\2303\230e\2404\240y\260\017\270q\300\002\300\"\300C\300r\310\022\3103\310b\320PR\320RS\330\010\017\210q\200A\340\032-\320-@\300\001\330\030+\320+@\300\001\330\031,\320,@\300\001\360h\001\000\t\034\2301\330\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\360\006\000\t \230q\330\010\014\210H\220K\230q\240\001\330\014\024\220A\330\014\031\230\023\230A\230X\240Q\240a\340\014\024\220O\2401\240A\340\010\021\220\024\220Y\230j\250\001\250\032\2609\270A\270^\3109\320TU\320U`\320`i\320iu\320u~\360\000\000\177\001F\002\360\000\000F\002N\002\360\000\000N\002O\002\330\010\016\320\016.\250a\330\010\013\2106\220\021\220!\330\010\017\210q\200A\360\026\000\t\020\210t\2209\230L\250""\001\200A\360 \000\t\020\210t\2209\230M\250\021\250!\200A\360\024\000\t\020\210t\2209\320\0341\260\021\200A\360\026\000\t\020\210t\2209\320\0341\260\021\200A\360\026\000\t\020\210t\2209\320\034,\250A\320\004C\3001\360*\000\t\022\220\024\220Y\230j\250\004\250I\260Q\260e\320;O\310q\320PU\320U]\320]^\330\010\016\210h\220a\330\010\013\2106\220\021\220!\330\010\017\210q\200\001\330\004\n\210+\220Q\320\004~\360\000\000\177\001R\002\360\000\000R\002e\002\360\000\000e\002y\002\360\000\000y\002N\003\360\000\000N\003b\003\360\000\000b\003u\003\360\000\000u\003I\004\360\000\000I\004J\004\360`\001\000\t\034\2301\330\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\360\006\000\t \230q\330\010\014\210H\220K\230q\240\001\330\014\024\220A\330\014\031\230\023\230A\230X\240Q\240a\340\014\024\220O\2401\240A\340\010\021\220\024\220Y\230f\240A\240Z\250y\270\001\270\036\300y\320PQ\320Q\\\320\\e\320eq\320qz\360\000\000{\001B\002\360\000\000B\002J\002\360\000\000J\002K\002\330\010\016\320\016*\250!\330\010\013\2106\220\021\220!\330\010\017\210q\320\004U\320Uh\320h{\360\000\000|\001P\002\360\000\000P\002`\002\360\000\000`\002a\002\360N\001\000\t\034\2301\340\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\330\010\013\2101\330\014\027\220q\360\006\000\t \230q\330\010\014\210H\220K\230q\240\001\330\014\024\220A\330\014\031\230\023\230A\230X\240Q\240a\340\014\024\220O\2401\240A\340\010\021\220\024\220Y\230f\240A\240Z\250y\270\001\270\031\300)\310<\320W`\320`g\320go\320op\330\010\016\320\016&\240a\330\010\013\2106\220\021\220!\330\010\017\210q\320\004f\320fg\360.\000\t\022\220\024\220Y\230l\250$\250i\260q\270\010\300\t\310\021\310%\320Oc\320cd\320di\320iq\320qr\330\010\016\210i\220q\330\010\013\2106\220\021\220!\330\010\017\210q";
    PyObject *data = NULL;
    CYTHON_UNUSED_VAR(__Pyx_DecompressString);
    #endif
    PyObject **stringtab = __pyx_mstate->__pyx_string_tab;
    Py_ssize_t pos = 0;
    for (int i = 0; i < 240; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyUnicode_DecodeUTF8(bytes + pos, bytes_length, NULL);
      if (likely(string) && i >= 22) PyUnicode_InternInPlace(&string);
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
      stringtab[i] = string;
      pos += bytes_length;
    }
    for (int i = 240; i < 271; i++) {
      Py_ssize_t bytes_length = index[i].length;
      PyObject *string = PyBytes_FromStringAndSize(bytes + pos, bytes_length);
      stringtab[i] = string;
      pos += bytes_length;
      if (unlikely(!string)) {
        Py_XDECREF(data);
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    Py_XDECREF(data);
    for (Py_ssize_t i = 0; i < 271; i++) {
      if (unlikely(PyObject_Hash(stringtab[i]) == -1)) {
        __PYX_ERR(0, 1, __pyx_L1_error)
      }
    }
    #if CYTHON_IMMORTAL_CONSTANTS
    {
      PyObject **table = stringtab + 240;
      for (Py_ssize_t i=0; i<31; ++i) {
        #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
        #if PY_VERSION_HEX < 0x030E0000
        if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
        #else
        if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
        #endif
        {
          Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
        }
        #else
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
        #endif
      }
    }
    #endif
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab;
    double const c_constants[] = {0.0};
    for (int i = 0; i < 1; i++) {
      numbertab[i] = PyFloat_FromDouble(c_constants[i]);
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  {
    PyObject **numbertab = __pyx_mstate->__pyx_number_tab + 1;
    int8_t const cint_constants_1[] = {0,2,3};
    for (int i = 0; i < 3; i++) {
      numbertab[i] = PyLong_FromLong(cint_constants_1[i - 0]);
      if (unlikely(!numbertab[i])) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #if CYTHON_IMMORTAL_CONSTANTS
  {
    PyObject **table = __pyx_mstate->__pyx_number_tab;
    for (Py_ssize_t i=0; i<4; ++i) {
      #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      #if PY_VERSION_HEX < 0x030E0000
      if (_Py_IsOwnedByCurrentThread(table[i]) && Py_REFCNT(table[i]) == 1)
      #else
      if (PyUnstable_Object_IsUniquelyReferenced(table[i]))
      #endif
      {
        Py_SET_REFCNT(table[i], _Py_IMMORTAL_REFCNT_LOCAL);
      }
      #else
      Py_SET_REFCNT(table[i], _Py_IMMORTAL_INITIAL_REFCNT);
      #endif
    }
  }
  #endif
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
typedef struct {
    unsigned int argcount : 4;
    unsigned int num_posonly_args : 1;
    unsigned int num_kwonly_args : 1;
    unsigned int nlocals : 5;
    unsigned int flags : 10;
    unsigned int first_line : 11;
} __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        PyObject *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 11, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS), 114};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_kwargs, __pyx_mstate->__pyx_n_u_i1, __pyx_mstate->__pyx_n_u_i2, __pyx_mstate->__pyx_n_u_i3, __pyx_mstate->__pyx_n_u_value, __pyx_mstate->__pyx_n_u_k, __pyx_mstate->__pyx_n_u_distinctFields, __pyx_mstate->__pyx_n_u_phiX, __pyx_mstate->__pyx_n_u_phiY, __pyx_mstate->__pyx_n_u_phiZ};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_Update, __pyx_mstate->__pyx_kp_b_iso88591_AN_E_QfE_A_t3as_Q_AQaq_AQaq_F_aq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 190};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_i, __pyx_mstate->__pyx_n_u_j, __pyx_mstate->__pyx_n_u_k, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_GetChi2Tensor, __pyx_mstate->__pyx_kp_b_iso88591_A_b_aq_3d_E_aq_U_1_E_aq_q_3e4y_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 251};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_wl};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_GetN, __pyx_mstate->__pyx_kp_b_iso88591_A0_t9E, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 277};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_IsNonlinear, __pyx_mstate->__pyx_kp_b_iso88591_A_t9L, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[6] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[6])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[7] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[7])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS), 389};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_kwargs, __pyx_mstate->__pyx_n_u_name_2, __pyx_mstate->__pyx_n_u_value};
    __pyx_mstate_global->__pyx_codeobj_tab[8] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_SetParams, __pyx_mstate->__pyx_kp_b_iso88591_A_F_6_q_uG1_j_4Cq_1F, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[8])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[9] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[9])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[10] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[10])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[11] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[11])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[12] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[12])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[13] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[13])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[14] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[14])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[15] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[15])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[16] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[16])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[17] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[17])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[18] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[18])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[19] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[19])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[20] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[20])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 840};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_z};
    __pyx_mstate_global->__pyx_codeobj_tab[21] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_GetMainFields, __pyx_mstate->__pyx_kp_b_iso88591_A_1D_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[21])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[22] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[22])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[23] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[23])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 902};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_z};
    __pyx_mstate_global->__pyx_codeobj_tab[24] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_GetIntensity, __pyx_mstate->__pyx_kp_b_iso88591_A_t9M, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[24])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 920};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[25] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_GetAbsorbedIntensity, __pyx_mstate->__pyx_kp_b_iso88591_A_t9_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[25])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 932};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[26] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_GetSrcIntensity, __pyx_mstate->__pyx_kp_b_iso88591_A_t9_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[26])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[27] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[27])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[28] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[28])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1044};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_d, __pyx_mstate->__pyx_n_u_material, __pyx_mstate->__pyx_n_u_lastLayer, __pyx_mstate->__pyx_n_u_layerptr, __pyx_mstate->__pyx_n_u_layer};
    __pyx_mstate_global->__pyx_codeobj_tab[29] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_AddLayer, __pyx_mstate->__pyx_kp_b_iso88591_A_IYas_O7_1_T_c_1_4y_q_V1Jk_G7_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[29])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS), 1077};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_kwargs, __pyx_mstate->__pyx_n_u_name_2, __pyx_mstate->__pyx_n_u_value};
    __pyx_mstate_global->__pyx_codeobj_tab[30] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_SetParams, __pyx_mstate->__pyx_kp_b_iso88591_A_F_6_q_uG1_j_4Cq_1F_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[30])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS), 1089};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_kwargs};
    __pyx_mstate_global->__pyx_codeobj_tab[31] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_Solve, __pyx_mstate->__pyx_kp_b_iso88591_A6_Jc_IV1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[31])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1119};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_resCpp, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[32] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_GetIntensities, __pyx_mstate->__pyx_kp_b_iso88591_A_T_l_6_1_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[32])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {8, 0, 0, 13, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1135};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_paramStr, __pyx_mstate->__pyx_n_u_values, __pyx_mstate->__pyx_n_u_layerNr, __pyx_mstate->__pyx_n_u_layerZ, __pyx_mstate->__pyx_n_u_outPwr, __pyx_mstate->__pyx_n_u_outAbs, __pyx_mstate->__pyx_n_u_outEnh, __pyx_mstate->__pyx_n_u_resCpp, __pyx_mstate->__pyx_n_u_outmask, __pyx_mstate->__pyx_n_u_param, __pyx_mstate->__pyx_n_u_paramLayer, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[33] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_Sweep, __pyx_mstate->__pyx_kp_b_iso88591_UUhh_P_P_a_N_1_1_q_1_q_1_q_q_HK, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[33])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1196};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_zs, __pyx_mstate->__pyx_n_u_dir, __pyx_mstate->__pyx_n_u_resCpp, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[34] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_GetFields, __pyx_mstate->__pyx_kp_b_iso88591_C1_Yj_IQe_OqPUU_ha_6_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[34])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1222};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_zs, __pyx_mstate->__pyx_n_u_xs, __pyx_mstate->__pyx_n_u_dir, __pyx_mstate->__pyx_n_u_resCpp, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[35] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_GetFields2D, __pyx_mstate->__pyx_kp_b_iso88591_ffg_Yl_iq_Occddiiqqr_iq_6_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[35])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1250};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[36] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_GetAbsorbedIntensity, __pyx_mstate->__pyx_kp_b_iso88591_A_t9_1_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[36])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1263};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_layerNr, __pyx_mstate->__pyx_n_u_z, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[37] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_GetEnhancement, __pyx_mstate->__pyx_kp_b_iso88591_5Q_d_9A_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[37])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1289};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_layerNr, __pyx_mstate->__pyx_n_u_x0, __pyx_mstate->__pyx_n_u_x1, __pyx_mstate->__pyx_n_u_z, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[38] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_WaveGetPowerFlows, __pyx_mstate->__pyx_kp_b_iso88591_9_a_SSXXYYllm_d_Qit4q_83a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[38])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1323};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_layerNr, __pyx_mstate->__pyx_n_u_z, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[39] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_WaveGetEnhancement, __pyx_mstate->__pyx_kp_b_iso88591_9_d_ay_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[39])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {9, 0, 0, 14, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1347};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_paramStr, __pyx_mstate->__pyx_n_u_values, __pyx_mstate->__pyx_n_u_layerNr, __pyx_mstate->__pyx_n_u_layerZ, __pyx_mstate->__pyx_n_u_outPwr, __pyx_mstate->__pyx_n_u_outR, __pyx_mstate->__pyx_n_u_outT, __pyx_mstate->__pyx_n_u_outEnh, __pyx_mstate->__pyx_n_u_resCpp, __pyx_mstate->__pyx_n_u_outmask, __pyx_mstate->__pyx_n_u_param, __pyx_mstate->__pyx_n_u_paramLayer, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[40] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_WaveSweep, __pyx_mstate->__pyx_kp_b_iso88591_A_V_1_1_q_1_q_1_q_1_q_q_HKq_A_AX, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[40])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1416};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_zs, __pyx_mstate->__pyx_n_u_xs, __pyx_mstate->__pyx_n_u_dirStr, __pyx_mstate->__pyx_n_u_resCpp, __pyx_mstate->__pyx_n_u_direction, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[41] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_WaveGetFields2D, __pyx_mstate->__pyx_kp_b_iso88591_A_aq_Y_d_1HIQeS_eef_iq_6_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[41])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[42] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[42])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[43] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[43])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[44] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[44])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[45] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[45])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[46] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[46])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[47] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[47])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[48] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[48])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[49] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[49])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS), 1730};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_kwargs, __pyx_mstate->__pyx_n_u_name_2, __pyx_mstate->__pyx_n_u_value};
    __pyx_mstate_global->__pyx_codeobj_tab[50] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_SetParams, __pyx_mstate->__pyx_kp_b_iso88591_A_F_6_q_uG1_j_4Cq_1F_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[50])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1742};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_d, __pyx_mstate->__pyx_n_u_material};
    __pyx_mstate_global->__pyx_codeobj_tab[51] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_AddLayer, __pyx_mstate->__pyx_kp_b_iso88591_A_Cy_A_Cy_A_D_Q_O7_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[51])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1764};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[52] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_Solve, __pyx_mstate->__pyx_kp_b_iso88591_A_IV1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[52])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1772};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[53] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_UpdateGenParams, __pyx_mstate->__pyx_kp_b_iso88591_A_I_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[53])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1780};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_resCpp, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[54] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_GetIntensities, __pyx_mstate->__pyx_kp_b_iso88591_A_YoQ_q_6_1_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[54])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {12, 0, 0, 17, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1799};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_paramStr, __pyx_mstate->__pyx_n_u_valuesP1, __pyx_mstate->__pyx_n_u_valuesP2, __pyx_mstate->__pyx_n_u_layerNr, __pyx_mstate->__pyx_n_u_layerZ, __pyx_mstate->__pyx_n_u_outPwr, __pyx_mstate->__pyx_n_u_outAbs, __pyx_mstate->__pyx_n_u_outEnh, __pyx_mstate->__pyx_n_u_outP1, __pyx_mstate->__pyx_n_u_outP2, __pyx_mstate->__pyx_n_u_outGen, __pyx_mstate->__pyx_n_u_resCpp, __pyx_mstate->__pyx_n_u_outmask, __pyx_mstate->__pyx_n_u_param, __pyx_mstate->__pyx_n_u_paramLayer, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[55] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_Sweep, __pyx_mstate->__pyx_kp_b_iso88591_R_R_e_e_y_y_N_N_b_b_u_u_I_I_J_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[55])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {5, 0, 0, 6, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1874};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_layerNr, __pyx_mstate->__pyx_n_u_x0, __pyx_mstate->__pyx_n_u_x1, __pyx_mstate->__pyx_n_u_z, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[56] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_WaveGetPowerFlows, __pyx_mstate->__pyx_kp_b_iso88591_9_a_SSXXYYllm_d_Qit4q_83a_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[56])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {13, 0, 0, 18, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1909};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_paramStr, __pyx_mstate->__pyx_n_u_valuesP1, __pyx_mstate->__pyx_n_u_valuesP2, __pyx_mstate->__pyx_n_u_layerNr, __pyx_mstate->__pyx_n_u_layerZ, __pyx_mstate->__pyx_n_u_outPwr, __pyx_mstate->__pyx_n_u_outR, __pyx_mstate->__pyx_n_u_outT, __pyx_mstate->__pyx_n_u_outEnh, __pyx_mstate->__pyx_n_u_outP1, __pyx_mstate->__pyx_n_u_outP2, __pyx_mstate->__pyx_n_u_outGen, __pyx_mstate->__pyx_n_u_resCpp, __pyx_mstate->__pyx_n_u_outmask, __pyx_mstate->__pyx_n_u_param, __pyx_mstate->__pyx_n_u_paramLayer, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[57] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_WaveSweep, __pyx_mstate->__pyx_kp_b_iso88591_A_h_1_1_q_1_q_1_q_1_q_1_q_1_q_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[57])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {4, 0, 0, 7, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1994};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_zs, __pyx_mstate->__pyx_n_u_xs, __pyx_mstate->__pyx_n_u_dirStr, __pyx_mstate->__pyx_n_u_resCpp, __pyx_mstate->__pyx_n_u_direction, __pyx_mstate->__pyx_n_u_res};
    __pyx_mstate_global->__pyx_codeobj_tab[58] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_NonlinearTMM_src_SecondOrderNLTM, __pyx_mstate->__pyx_n_u_WaveGetFields2D, __pyx_mstate->__pyx_kp_b_iso88591_AFG4_aq_Y_d_1HIQeS_eef_iq_6_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[58])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[59] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[59])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[60] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_mstate->__pyx_kp_b_iso88591_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[60])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CommonTypesMetaclass.init */
  if (likely(__pyx_CommonTypesMetaclass_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
  {
      PyObject *typesModule=NULL;
      typesModule = PyImport_ImportModule("types");
      if (typesModule) {
          __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
          Py_DECREF(typesModule);
      }
  } // error handling follows
  #endif
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CythonFunctionShared.init */
  if (likely(__pyx_CyFunction_init(__pyx_m) == 0)); else
  
  if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetTopmostException (used by SaveResetException) */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C0000
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#elif __PYX_LIMITED_VERSION_HEX > 0x030C0000
    local_value = PyErr_GetRaisedException();
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
#if __PYX_LIMITED_VERSION_HEX > 0x030C0000
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
#else
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
#endif // __PYX_LIMITED_VERSION_HEX > 0x030C0000
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#elif __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    PyErr_SetHandledException(local_value);
    Py_XDECREF(local_value);
    Py_XDECREF(local_type);
    Py_XDECREF(local_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
#if __PYX_LIMITED_VERSION_HEX <= 0x030C0000
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
#endif
}

/* PyObjectCall (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO (used by PyObjectFastCall) */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
          #if CYTHON_COMPILING_IN_LIMITED_API
            return PyObject_Vectorcall(func, args, _nargs, NULL);
          #else
            vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
            if (f) {
                return f(func, args, _nargs, NULL);
            }
          #endif
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyErrFetchRestore (used by RaiseException) */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* BytesEquals (used by UnicodeEquals) */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* DictGetItem */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    if (unlikely(__Pyx_PyDict_GetItemRef(d, key, &value) == 0)) { // no value, no error
        if (unlikely(PyTuple_Check(key))) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args)) {
                PyErr_SetObject(PyExc_KeyError, args);
                Py_DECREF(args);
            }
        } else {
            PyErr_SetObject(PyExc_KeyError, key);
        }
    }
    return value;
}
#endif

/* TupleAndListFromArray (used by fastcall) */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < (0))) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyObjectCallOneArg (used by CallUnboundCMethod0) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetAttrStr (used by UnpackUnboundCMethod) */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* UnpackUnboundCMethod (used by CallUnboundCMethod0) */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod0 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        if (likely(cfunc->flag == METH_NOARGS))
            return __Pyx_CallCFunction(cfunc, self, NULL);
        if (likely(cfunc->flag == METH_FASTCALL))
            return __Pyx_CallCFunctionFast(cfunc, self, NULL, 0);
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, NULL, 0, NULL);
        if (likely(cfunc->flag == (METH_VARARGS | METH_KEYWORDS)))
            return __Pyx_CallCFunctionWithKeywords(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple, NULL);
        if (cfunc->flag == METH_VARARGS)
            return __Pyx_CallCFunction(cfunc, self, __pyx_mstate_global->__pyx_empty_tuple);
        return __Pyx__CallUnboundCMethod0(cfunc, self);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod0(&tmp_cfunc, self);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod0(cfunc, self);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *result;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
    result = __Pyx_PyObject_CallOneArg(cfunc->method, self);
    return result;
}

/* py_dict_items (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_items, d);
}

/* py_dict_values (used by OwnedDictNext) */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_values, d);
}

/* OwnedDictNext (used by RejectKeywords) */
#if CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, PyObject **ppos, PyObject **pkey, PyObject **pvalue) {
    PyObject *next = NULL;
    if (!*ppos) {
        if (pvalue) {
            PyObject *dictview = pkey ? __Pyx_PyDict_Items(p) : __Pyx_PyDict_Values(p);
            if (unlikely(!dictview)) goto bad;
            *ppos = PyObject_GetIter(dictview);
            Py_DECREF(dictview);
        } else {
            *ppos = PyObject_GetIter(p);
        }
        if (unlikely(!*ppos)) goto bad;
    }
    next = PyIter_Next(*ppos);
    if (!next) {
        if (PyErr_Occurred()) goto bad;
        return 0;
    }
    if (pkey && pvalue) {
        *pkey = __Pyx_PySequence_ITEM(next, 0);
        if (unlikely(*pkey)) goto bad;
        *pvalue = __Pyx_PySequence_ITEM(next, 1);
        if (unlikely(*pvalue)) goto bad;
        Py_DECREF(next);
    } else if (pkey) {
        *pkey = next;
    } else {
        assert(pvalue);
        *pvalue = next;
    }
    return 1;
  bad:
    Py_XDECREF(next);
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
    PyErr_FormatUnraisable("Exception ignored in __Pyx_PyDict_NextRef");
#else
    PyErr_WriteUnraisable(__pyx_mstate_global->__pyx_n_u_Pyx_PyDict_NextRef);
#endif
    if (pkey) *pkey = NULL;
    if (pvalue) *pvalue = NULL;
    return 0;
}
#else // !CYTHON_AVOID_BORROWED_REFS
static int __Pyx_PyDict_NextRef(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) {
    int result = PyDict_Next(p, ppos, pkey, pvalue);
    if (likely(result == 1)) {
        if (pkey) Py_INCREF(*pkey);
        if (pvalue) Py_INCREF(*pvalue);
    }
    return result;
}
#endif

/* RejectKeywords */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds) {
    PyObject *key = NULL;
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds))) {
        key = __Pyx_PySequence_ITEM(kwds, 0);
    } else {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *pos = NULL;
#else
        Py_ssize_t pos = 0;
#endif
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
        if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return;
#endif
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL);
#if CYTHON_AVOID_BORROWED_REFS
        Py_XDECREF(pos);
#endif
    }
    if (likely(key)) {
        PyErr_Format(PyExc_TypeError,
            "%s() got an unexpected keyword argument '%U'",
            function_name, key);
        Py_DECREF(key);
    }
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    const char* function_name,
    PyObject *kw)
{
#if CYTHON_COMPILING_IN_PYPY && !defined(PyArg_ValidateKeywordArguments)
    CYTHON_UNUSED_VAR(function_name);
    CYTHON_UNUSED_VAR(kw);
    return 0;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
#if PY_VERSION_HEX >= 0x03090000
        CYTHON_UNUSED_VAR(function_name);
#else
        Py_ssize_t kwsize;
        #if CYTHON_ASSUME_SAFE_SIZE
        kwsize = PyTuple_GET_SIZE(kw);
        #else
        kwsize = PyTuple_Size(kw);
        if (unlikely(kwsize < 0)) return -1;
        #endif
        for (Py_ssize_t pos = 0; pos < kwsize; pos++) {
            PyObject* key = NULL;
            #if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
            #else
            key = PyTuple_GetItem(kw, pos);
            if (unlikely(!key)) return -1;
            #endif
            if (unlikely(!PyUnicode_Check(key))) {
                PyErr_Format(PyExc_TypeError,
                    "%.200s() keywords must be strings", function_name);
                return -1;
            }
        }
#endif
    } else {
        if (unlikely(!PyArg_ValidateKeywordArguments(kw))) return -1;
    }
    return 0;
#endif
}

/* py_dict_keys */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d) {
    return __Pyx_CallUnboundCMethod0(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_keys, d);
}

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS)) {
        return __Pyx_PyList_GetItemRefFast(o, wrapped_i, unsafe_shared);
    } else
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyList_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        return __Pyx_NewRef(PyTuple_GET_ITEM(o, wrapped_i));
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    (void)wraparound;
    (void)boundscheck;
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     int wraparound, int boundscheck, int unsafe_shared) {
    CYTHON_MAYBE_UNUSED_VAR(unsafe_shared);
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)) {
            return __Pyx_PyList_GetItemRefFast(o, n, unsafe_shared);
        } else if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_NewRef(PyList_GET_ITEM(o, n));
        }
    } else
    #if !CYTHON_AVOID_BORROWED_REFS
    if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            return __Pyx_NewRef(PyTuple_GET_ITEM(o, n));
        }
    } else
    #endif
#endif
#if CYTHON_USE_TYPE_SLOTS && !CYTHON_COMPILING_IN_PYPY
    {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (!is_list && mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (is_list || likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    (void)wraparound;
    (void)boundscheck;
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* PyObjectCall2Args (used by CallUnboundCMethod1) */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    int was_initialized =  __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return __Pyx_CallCFunction(cfunc, self, arg);
        } else if (flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, &arg, 1);
        } else if (flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, &arg, 1, NULL);
        }
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod1(&tmp_cfunc, self, arg);
    }
#endif
    PyObject* result = __Pyx__CallUnboundCMethod1(cfunc, self, arg);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *args = PyTuple_New(1);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
    } else
#endif
    {
        result = __Pyx_PyObject_Call2Args(cfunc->method, self, arg);
    }
    return result;
}

/* py_dict_pop */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop)
    PyObject *value;
    if (PyDict_Pop(d, key, &value) == 0) {
        if (default_value) {
            Py_INCREF(default_value);
        } else {
            PyErr_SetObject(PyExc_KeyError, key);
        }
        value = default_value;
    }
    return value;
#elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    return _PyDict_Pop(d, key, default_value);
#else
    if (default_value) {
        return __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, d, key, default_value);
    } else {
        return __Pyx_CallUnboundCMethod1(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, d, key);
    }
#endif
}

/* PyObjectGetAttrStrNoError (used by GetBuiltinName) */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName (used by GetModuleGlobalName) */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* PyDictVersioning (used by GetModuleGlobalName) */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectVectorCallKwBuilder */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* RaiseDoubleKeywords (used by ParseKeywordsImpl) */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* ParseKeywordsImpl (used by ParseKeywords) */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    #if CYTHON_AVOID_BORROWED_REFS
    PyObject *pos = NULL;
    #else
    Py_ssize_t pos = 0;
    #endif
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (
        #if CYTHON_AVOID_BORROWED_REFS
        __Pyx_PyDict_NextRef(kwds, &pos, &key, NULL)
        #else
        PyDict_Next(kwds, &pos, &key, NULL)
        #endif
    ) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        #endif
    }
    __Pyx_END_CRITICAL_SECTION();
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(pos);
    #endif
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}

/* ParseKeywords */
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* ArgTypeTestFunc (used by ArgTypeTest) */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* IsLittleEndian (used by BufferFormatCheck) */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck (used by BufferGetAndValidate) */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              const __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparsable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, int is_complex) {
  CYTHON_UNUSED_VAR(is_complex);
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, int is_complex) {
  CYTHON_UNUSED_VAR(is_complex);
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    const __Pyx_StructField* field = ctx->head->field;
    const __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    const __Pyx_StructField* field = ctx->head->field;
    const __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static int
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return -1;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return -1;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return -1;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i]) {
            PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
            return -1;
        }
        if (*ts != ',' && *ts != ')') {
            PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
            return -1;
        }
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim) {
        PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
        return -1;
    }
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return -1;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return 0;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (__pyx_buffmt_parse_array(ctx, &ts) < 0) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
    if (unlikely(info->buf == NULL)) return;
    if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
    PyBuffer_Release(info);
  }
  static void __Pyx_ZeroBuffer(Py_buffer* buf) {
    buf->buf = NULL;
    buf->obj = NULL;
    buf->strides = __Pyx_zeros;
    buf->shape = __Pyx_zeros;
    buf->suboffsets = __Pyx_minusones;
  }
  static int __Pyx__GetBufferAndValidate(
          Py_buffer* buf, PyObject* obj,  const __Pyx_TypeInfo* dtype, int flags,
          int nd, int cast, __Pyx_BufFmt_StackElem* stack)
  {
    buf->buf = NULL;
    if (unlikely(PyObject_GetBuffer(obj, buf, flags) == -1)) {
      __Pyx_ZeroBuffer(buf);
      return -1;
    }
    if (unlikely(buf->ndim != nd)) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer has wrong number of dimensions (expected %d, got %d)",
                   nd, buf->ndim);
      goto fail;
    }
    if (!cast) {
      __Pyx_BufFmt_Context ctx;
      __Pyx_BufFmt_Init(&ctx, stack, dtype);
      if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
    }
    if (unlikely((size_t)buf->itemsize != dtype->size)) {
      PyErr_Format(PyExc_ValueError,
        "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
        buf->itemsize, (buf->itemsize > 1) ? "s" : "",
        dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
      goto fail;
    }
    if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
    return 0;
  fail:;
    __Pyx_SafeReleaseBuffer(buf);
    return -1;
  }
  
/* IterFinish (used by dict_iter) */
  static CYTHON_INLINE int __Pyx_IterFinish(void) {
      PyObject* exc_type;
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      exc_type = __Pyx_PyErr_CurrentExceptionType();
      if (unlikely(exc_type)) {
          if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
              return -1;
          __Pyx_PyErr_Clear();
          return 0;
      }
      return 0;
  }
  
/* PyObjectCallNoArg (used by PyObjectCallMethod0) */
  static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
      PyObject *arg[2] = {NULL, NULL};
      return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
  }
  
/* PyObjectGetMethod (used by PyObjectCallMethod0) */
  #if !(CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000)))
  static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
      PyObject *attr;
  #if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
      __Pyx_TypeName type_name;
      PyTypeObject *tp = Py_TYPE(obj);
      PyObject *descr;
      descrgetfunc f = NULL;
      PyObject **dictptr, *dict;
      int meth_found = 0;
      assert (*method == NULL);
      if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
          attr = __Pyx_PyObject_GetAttrStr(obj, name);
          goto try_unpack;
      }
      if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
          return 0;
      }
      descr = _PyType_Lookup(tp, name);
      if (likely(descr != NULL)) {
          Py_INCREF(descr);
  #if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
          if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
  #else
          #ifdef __Pyx_CyFunction_USED
          if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
          #else
          if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
          #endif
  #endif
          {
              meth_found = 1;
          } else {
              f = Py_TYPE(descr)->tp_descr_get;
              if (f != NULL && PyDescr_IsData(descr)) {
                  attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                  Py_DECREF(descr);
                  goto try_unpack;
              }
          }
      }
      dictptr = _PyObject_GetDictPtr(obj);
      if (dictptr != NULL && (dict = *dictptr) != NULL) {
          Py_INCREF(dict);
          attr = __Pyx_PyDict_GetItemStr(dict, name);
          if (attr != NULL) {
              Py_INCREF(attr);
              Py_DECREF(dict);
              Py_XDECREF(descr);
              goto try_unpack;
          }
          Py_DECREF(dict);
      }
      if (meth_found) {
          *method = descr;
          return 1;
      }
      if (f != NULL) {
          attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
          Py_DECREF(descr);
          goto try_unpack;
      }
      if (likely(descr != NULL)) {
          *method = descr;
          return 0;
      }
      type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
      PyErr_Format(PyExc_AttributeError,
                   "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                   type_name, name);
      __Pyx_DECREF_TypeName(type_name);
      return 0;
  #else
      attr = __Pyx_PyObject_GetAttrStr(obj, name);
      goto try_unpack;
  #endif
  try_unpack:
  #if CYTHON_UNPACK_METHODS
      if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
          PyObject *function = PyMethod_GET_FUNCTION(attr);
          Py_INCREF(function);
          Py_DECREF(attr);
          *method = function;
          return 1;
      }
  #endif
      *method = attr;
      return 0;
  }
  #endif
  
/* PyObjectCallMethod0 (used by dict_iter) */
  static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
  #if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
      PyObject *args[1] = {obj};
      (void) __Pyx_PyObject_CallOneArg;
      (void) __Pyx_PyObject_CallNoArg;
      return PyObject_VectorcallMethod(method_name, args, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
  #else
      PyObject *method = NULL, *result = NULL;
      int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
      if (likely(is_method)) {
          result = __Pyx_PyObject_CallOneArg(method, obj);
          Py_DECREF(method);
          return result;
      }
      if (unlikely(!method)) goto bad;
      result = __Pyx_PyObject_CallNoArg(method);
      Py_DECREF(method);
  bad:
      return result;
  #endif
  }
  
/* RaiseNeedMoreValuesToUnpack (used by UnpackTuple2) */
  static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
      PyErr_Format(PyExc_ValueError,
                   "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                   index, (index == 1) ? "" : "s");
  }
  
/* RaiseTooManyValuesToUnpack (used by UnpackItemEndCheck) */
  static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
      PyErr_Format(PyExc_ValueError,
                   "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
  }
  
/* UnpackItemEndCheck (used by UnpackTuple2) */
  static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
      if (unlikely(retval)) {
          Py_DECREF(retval);
          __Pyx_RaiseTooManyValuesError(expected);
          return -1;
      }
      return __Pyx_IterFinish();
  }
  
/* RaiseNoneIterError (used by UnpackTupleError) */
  static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
  }
  
/* UnpackTupleError (used by UnpackTuple2) */
  static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
      if (t == Py_None) {
        __Pyx_RaiseNoneNotIterableError();
      } else {
        Py_ssize_t size = __Pyx_PyTuple_GET_SIZE(t);
   #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(size < 0)) return;
   #endif
        if (size < index) {
          __Pyx_RaiseNeedMoreValuesError(size);
        } else {
          __Pyx_RaiseTooManyValuesError(index);
        }
      }
  }
  
/* UnpackTuple2 (used by dict_iter) */
  static CYTHON_INLINE int __Pyx_unpack_tuple2(
          PyObject* tuple, PyObject** value1, PyObject** value2, int is_tuple, int has_known_size, int decref_tuple) {
      if (likely(is_tuple || PyTuple_Check(tuple))) {
          Py_ssize_t size;
          if (has_known_size) {
              return __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple);
          }
          size = __Pyx_PyTuple_GET_SIZE(tuple);
          if (likely(size == 2)) {
              return __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple);
          }
          if (size >= 0) {
              __Pyx_UnpackTupleError(tuple, 2);
          }
          return -1;
      } else {
          return __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple);
      }
  }
  static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
          PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
      PyObject *value1 = NULL, *value2 = NULL;
  #if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
      value1 = __Pyx_PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
      value2 = __Pyx_PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
  #else
      value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
      value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
  #endif
      if (decref_tuple) {
          Py_DECREF(tuple);
      }
      *pvalue1 = value1;
      *pvalue2 = value2;
      return 0;
  #if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
  bad:
      Py_XDECREF(value1);
      Py_XDECREF(value2);
      if (decref_tuple) { Py_XDECREF(tuple); }
      return -1;
  #endif
  }
  static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                         int has_known_size, int decref_tuple) {
      Py_ssize_t index;
      PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
      iternextfunc iternext;
      iter = PyObject_GetIter(tuple);
      if (unlikely(!iter)) goto bad;
      if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
      iternext = __Pyx_PyObject_GetIterNextFunc(iter);
      value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
      value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
      if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
      Py_DECREF(iter);
      *pvalue1 = value1;
      *pvalue2 = value2;
      return 0;
  unpacking_failed:
      if (!has_known_size && __Pyx_IterFinish() == 0)
          __Pyx_RaiseNeedMoreValuesError(index);
  bad:
      Py_XDECREF(iter);
      Py_XDECREF(value1);
      Py_XDECREF(value2);
      if (decref_tuple) { Py_XDECREF(tuple); }
      return -1;
  }
  
/* dict_iter */
  #if CYTHON_COMPILING_IN_PYPY
  #include <string.h>
  #endif
  static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                     Py_ssize_t* p_orig_length, int* p_source_is_dict) {
      is_dict = is_dict || likely(PyDict_CheckExact(iterable));
      *p_source_is_dict = is_dict;
      if (is_dict) {
  #if !CYTHON_COMPILING_IN_PYPY
          *p_orig_length = PyDict_Size(iterable);
          Py_INCREF(iterable);
          return iterable;
  #else
          static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
          PyObject **pp = NULL;
          if (method_name) {
              const char *name = PyUnicode_AsUTF8(method_name);
              if (strcmp(name, "iteritems") == 0) pp = &py_items;
              else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
              else if (strcmp(name, "itervalues") == 0) pp = &py_values;
              if (pp) {
                  if (!*pp) {
                      *pp = PyUnicode_FromString(name + 4);
                      if (!*pp)
                          return NULL;
                  }
                  method_name = *pp;
              }
          }
  #endif
      }
      *p_orig_length = 0;
      if (method_name) {
          PyObject* iter;
          iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
          if (!iterable)
              return NULL;
  #if !CYTHON_COMPILING_IN_PYPY
          if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
              return iterable;
  #endif
          iter = PyObject_GetIter(iterable);
          Py_DECREF(iterable);
          return iter;
      }
      return PyObject_GetIter(iterable);
  }
  #if !CYTHON_AVOID_BORROWED_REFS
  static CYTHON_INLINE int __Pyx_dict_iter_next_source_is_dict(
          PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
          PyObject** pkey, PyObject** pvalue, PyObject** pitem) {
      PyObject *key, *value;
      if (unlikely(orig_length != PyDict_Size(iter_obj))) {
          PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
          return -1;
      }
      if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
          return 0;
      }
      if (pitem) {
          PyObject* tuple = PyTuple_New(2);
          if (unlikely(!tuple)) {
              return -1;
          }
          Py_INCREF(key);
          Py_INCREF(value);
          #if CYTHON_ASSUME_SAFE_MACROS
          PyTuple_SET_ITEM(tuple, 0, key);
          PyTuple_SET_ITEM(tuple, 1, value);
          #else
          if (unlikely(PyTuple_SetItem(tuple, 0, key) < 0)) {
              Py_DECREF(value);
              Py_DECREF(tuple);
              return -1;
          }
          if (unlikely(PyTuple_SetItem(tuple, 1, value) < 0)) {
              Py_DECREF(tuple);
              return -1;
          }
          #endif
          *pitem = tuple;
      } else {
          if (pkey) {
              Py_INCREF(key);
              *pkey = key;
          }
          if (pvalue) {
              Py_INCREF(value);
              *pvalue = value;
          }
      }
      return 1;
  }
  #endif
  static CYTHON_INLINE int __Pyx_dict_iter_next(
          PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
          PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
      PyObject* next_item;
  #if !CYTHON_AVOID_BORROWED_REFS
      if (source_is_dict) {
          int result;
  #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
          Py_BEGIN_CRITICAL_SECTION(iter_obj);
  #endif
          result = __Pyx_dict_iter_next_source_is_dict(iter_obj, orig_length, ppos, pkey, pvalue, pitem);
  #if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
          Py_END_CRITICAL_SECTION();
  #endif
          return result;
      } else if (PyTuple_CheckExact(iter_obj)) {
          Py_ssize_t pos = *ppos;
          Py_ssize_t tuple_size = __Pyx_PyTuple_GET_SIZE(iter_obj);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely(tuple_size < 0)) return -1;
          #endif
          if (unlikely(pos >= tuple_size)) return 0;
          *ppos = pos + 1;
          #if CYTHON_ASSUME_SAFE_MACROS
          next_item = PyTuple_GET_ITEM(iter_obj, pos);
          #else
          next_item = PyTuple_GetItem(iter_obj, pos);
          if (unlikely(!next_item)) return -1;
          #endif
          Py_INCREF(next_item);
      } else if (PyList_CheckExact(iter_obj)) {
          Py_ssize_t pos = *ppos;
          Py_ssize_t list_size = __Pyx_PyList_GET_SIZE(iter_obj);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely(list_size < 0)) return -1;
          #endif
          if (unlikely(pos >= list_size)) return 0;
          *ppos = pos + 1;
          next_item = __Pyx_PyList_GetItemRef(iter_obj, pos);
          if (unlikely(!next_item)) return -1;
      } else
  #endif
      {
          next_item = PyIter_Next(iter_obj);
          if (unlikely(!next_item)) {
              return __Pyx_IterFinish();
          }
      }
      if (pitem) {
          *pitem = next_item;
      } else if (pkey && pvalue) {
          if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
              return -1;
      } else if (pkey) {
          *pkey = next_item;
      } else {
          *pvalue = next_item;
      }
      return 1;
  }
  
/* unicode_tailmatch */
  static int __Pyx_PyUnicode_TailmatchTuple(PyObject* s, PyObject* substrings,
                                            Py_ssize_t start, Py_ssize_t end, int direction) {
      Py_ssize_t i, count = __Pyx_PyTuple_GET_SIZE(substrings);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely(count < 0)) return -1;
      #endif
      for (i = 0; i < count; i++) {
          Py_ssize_t result;
  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          result = PyUnicode_Tailmatch(s, PyTuple_GET_ITEM(substrings, i),
                                       start, end, direction);
  #else
          PyObject* sub = __Pyx_PySequence_ITEM(substrings, i);
          if (unlikely(!sub)) return -1;
          result = PyUnicode_Tailmatch(s, sub, start, end, direction);
          Py_DECREF(sub);
  #endif
          if (result) {
              return (int) result;
          }
      }
      return 0;
  }
  static int __Pyx_PyUnicode_Tailmatch(PyObject* s, PyObject* substr,
                                       Py_ssize_t start, Py_ssize_t end, int direction) {
      if (unlikely(PyTuple_Check(substr))) {
          return __Pyx_PyUnicode_TailmatchTuple(s, substr, start, end, direction);
      }
      return (int) PyUnicode_Tailmatch(s, substr, start, end, direction);
  }
  
/* PyUnicode_Substring */
  static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Substring(
              PyObject* text, Py_ssize_t start, Py_ssize_t stop) {
      Py_ssize_t length;
      #if !CYTHON_COMPILING_IN_LIMITED_API
      if (unlikely(__Pyx_PyUnicode_READY(text) == -1)) return NULL;
      #endif
      length = __Pyx_PyUnicode_GET_LENGTH(text);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely(length < 0)) return NULL;
      #endif
      if (start < 0) {
          start += length;
          if (start < 0)
              start = 0;
      }
      if (stop < 0)
          stop += length;
      else if (stop > length)
          stop = length;
      if (stop <= start)
          return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_unicode);
      if (start == 0 && stop == length)
          return __Pyx_NewRef(text);
  #if CYTHON_COMPILING_IN_LIMITED_API
      return PyUnicode_Substring(text, start, stop);
  #else
      return PyUnicode_FromKindAndData(PyUnicode_KIND(text),
          PyUnicode_1BYTE_DATA(text) + start*PyUnicode_KIND(text), stop-start);
  #endif
  }
  
/* SetStringIndexingError (used by GetItemIntUnicode) */
  static void __Pyx_SetStringIndexingError(const char* message, int has_gil) {
      if (!has_gil) {
          PyGILState_STATE gil_state = PyGILState_Ensure();
          PyErr_SetString(PyExc_IndexError, message);
          PyGILState_Release(gil_state);
      } else
          PyErr_SetString(PyExc_IndexError, message);
  }
  
/* GetItemIntUnicode */
  static CYTHON_INLINE Py_UCS4 __Pyx_GetItemInt_Unicode_Fast(PyObject* ustring, Py_ssize_t i,
                                                             int wraparound, int boundscheck, int has_gil) {
      Py_ssize_t length;
      if (unlikely(__Pyx_PyUnicode_READY(ustring) < 0)) return (Py_UCS4)-1;
      if (wraparound | boundscheck) {
          length = __Pyx_PyUnicode_GET_LENGTH(ustring);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely(length < 0)) return (Py_UCS4)-1;
          #endif
          if (wraparound & unlikely(i < 0)) i += length;
          if ((!boundscheck) || likely(__Pyx_is_valid_index(i, length))) {
              return __Pyx_PyUnicode_READ_CHAR(ustring, i);
          } else {
              __Pyx_SetStringIndexingError("string index out of range", has_gil);
              return (Py_UCS4)-1;
          }
      } else {
          return __Pyx_PyUnicode_READ_CHAR(ustring, i);
      }
  }
  
/* RaiseUnexpectedTypeError */
  static int
  __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
  {
      __Pyx_TypeName obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
      PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                   expected, obj_type_name);
      __Pyx_DECREF_TypeName(obj_type_name);
      return 0;
  }
  
/* AllocateExtensionType */
  static PyObject *__Pyx_AllocateExtensionType(PyTypeObject *t, int is_final) {
      if (is_final || likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
          allocfunc alloc_func = __Pyx_PyType_GetSlot(t, tp_alloc, allocfunc);
          return alloc_func(t, 0);
      } else {
          newfunc tp_new = __Pyx_PyType_TryGetSlot(&PyBaseObject_Type, tp_new, newfunc);
      #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
          if (!tp_new) {
              PyObject *new_str = PyUnicode_FromString("__new__");
              if (likely(new_str)) {
                  PyObject *o = PyObject_CallMethodObjArgs((PyObject *)&PyBaseObject_Type, new_str, t, NULL);
                  Py_DECREF(new_str);
                  return o;
              } else
                  return NULL;
          } else
      #endif
          return tp_new(t, __pyx_mstate_global->__pyx_empty_tuple, 0);
      }
  }
  
/* CallTypeTraverse */
  #if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
  #else
  static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
      #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
      if (__Pyx_get_runtime_version() < 0x03090000) return 0;
      #endif
      if (!always_call) {
          PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
          unsigned long flags = PyType_GetFlags(base);
          if (flags & Py_TPFLAGS_HEAPTYPE) {
              return 0;
          }
      }
      Py_VISIT((PyObject*)Py_TYPE(o));
      return 0;
  }
  #endif
  
/* LimitedApiGetTypeDict (used by SetItemOnTypeDict) */
  #if CYTHON_COMPILING_IN_LIMITED_API
  static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
      PyObject *tp_dictoffset_o;
      Py_ssize_t tp_dictoffset;
      tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
      if (unlikely(!tp_dictoffset_o)) return -1;
      tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
      Py_DECREF(tp_dictoffset_o);
      if (unlikely(tp_dictoffset == 0)) {
          PyErr_SetString(
              PyExc_TypeError,
              "'type' doesn't have a dictoffset");
          return -1;
      } else if (unlikely(tp_dictoffset < 0)) {
          PyErr_SetString(
              PyExc_TypeError,
              "'type' has an unexpected negative dictoffset. "
              "Please report this as Cython bug");
          return -1;
      }
      return tp_dictoffset;
  }
  static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
      static Py_ssize_t tp_dictoffset = 0;
      if (unlikely(tp_dictoffset == 0)) {
          tp_dictoffset = __Pyx_GetTypeDictOffset();
          if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
              tp_dictoffset = 0; // try again next time?
              return NULL;
          }
      }
      return *(PyObject**)((char*)tp + tp_dictoffset);
  }
  #endif
  
/* SetItemOnTypeDict (used by FixUpExtensionType) */
  static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
      int result;
      PyObject *tp_dict;
  #if CYTHON_COMPILING_IN_LIMITED_API
      tp_dict = __Pyx_GetTypeDict(tp);
      if (unlikely(!tp_dict)) return -1;
  #else
      tp_dict = tp->tp_dict;
  #endif
      result = PyDict_SetItem(tp_dict, k, v);
      if (likely(!result)) {
          PyType_Modified(tp);
          if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
              PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
              if (!setNameResult) return -1;
              Py_DECREF(setNameResult);
          }
      }
      return result;
  }
  
/* FixUpExtensionType */
  static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
  #if __PYX_LIMITED_VERSION_HEX > 0x030900B1
      CYTHON_UNUSED_VAR(spec);
      CYTHON_UNUSED_VAR(type);
      CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
  #else
      const PyType_Slot *slot = spec->slots;
      int changed = 0;
  #if !CYTHON_COMPILING_IN_LIMITED_API
      while (slot && slot->slot && slot->slot != Py_tp_members)
          slot++;
      if (slot && slot->slot == Py_tp_members) {
  #if !CYTHON_COMPILING_IN_CPYTHON
          const
  #endif  // !CYTHON_COMPILING_IN_CPYTHON)
              PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
          while (memb && memb->name) {
              if (memb->name[0] == '_' && memb->name[1] == '_') {
                  if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                      assert(memb->type == T_PYSSIZET);
                      assert(memb->flags == READONLY);
                      type->tp_weaklistoffset = memb->offset;
                      changed = 1;
                  }
                  else if (strcmp(memb->name, "__dictoffset__") == 0) {
                      assert(memb->type == T_PYSSIZET);
                      assert(memb->flags == READONLY);
                      type->tp_dictoffset = memb->offset;
                      changed = 1;
                  }
  #if CYTHON_METH_FASTCALL
                  else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                      assert(memb->type == T_PYSSIZET);
                      assert(memb->flags == READONLY);
                      type->tp_vectorcall_offset = memb->offset;
                      changed = 1;
                  }
  #endif  // CYTHON_METH_FASTCALL
  #if !CYTHON_COMPILING_IN_PYPY
                  else if (strcmp(memb->name, "__module__") == 0) {
                      PyObject *descr;
                      assert(memb->type == T_OBJECT);
                      assert(memb->flags == 0 || memb->flags == READONLY);
                      descr = PyDescr_NewMember(type, memb);
                      if (unlikely(!descr))
                          return -1;
                      int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                      Py_DECREF(descr);
                      if (unlikely(set_item_result < 0)) {
                          return -1;
                      }
                      changed = 1;
                  }
  #endif  // !CYTHON_COMPILING_IN_PYPY
              }
              memb++;
          }
      }
  #endif  // !CYTHON_COMPILING_IN_LIMITED_API
  #if !CYTHON_COMPILING_IN_PYPY
      slot = spec->slots;
      while (slot && slot->slot && slot->slot != Py_tp_getset)
          slot++;
      if (slot && slot->slot == Py_tp_getset) {
          PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
          while (getset && getset->name) {
              if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                  PyObject *descr = PyDescr_NewGetSet(type, getset);
                  if (unlikely(!descr))
                      return -1;
                  #if CYTHON_COMPILING_IN_LIMITED_API
                  PyObject *pyname = PyUnicode_FromString(getset->name);
                  if (unlikely(!pyname)) {
                      Py_DECREF(descr);
                      return -1;
                  }
                  int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                  Py_DECREF(pyname);
                  #else
                  CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                  int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                  #endif
                  Py_DECREF(descr);
                  if (unlikely(set_item_result < 0)) {
                      return -1;
                  }
                  changed = 1;
              }
              ++getset;
          }
      }
  #else
      CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
  #endif  // !CYTHON_COMPILING_IN_PYPY
      if (changed)
          PyType_Modified(type);
  #endif  // PY_VERSION_HEX > 0x030900B1
      return 0;
  }
  
/* ValidateBasesTuple (used by PyType_Ready) */
  #if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
  static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
      Py_ssize_t i, n;
  #if CYTHON_ASSUME_SAFE_SIZE
      n = PyTuple_GET_SIZE(bases);
  #else
      n = PyTuple_Size(bases);
      if (unlikely(n < 0)) return -1;
  #endif
      for (i = 1; i < n; i++)
      {
          PyTypeObject *b;
  #if CYTHON_AVOID_BORROWED_REFS
          PyObject *b0 = PySequence_GetItem(bases, i);
          if (!b0) return -1;
  #elif CYTHON_ASSUME_SAFE_MACROS
          PyObject *b0 = PyTuple_GET_ITEM(bases, i);
  #else
          PyObject *b0 = PyTuple_GetItem(bases, i);
          if (!b0) return -1;
  #endif
          b = (PyTypeObject*) b0;
          if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
          {
              __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
              PyErr_Format(PyExc_TypeError,
                  "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
              __Pyx_DECREF_TypeName(b_name);
  #if CYTHON_AVOID_BORROWED_REFS
              Py_DECREF(b0);
  #endif
              return -1;
          }
          if (dictoffset == 0)
          {
              Py_ssize_t b_dictoffset = 0;
  #if CYTHON_USE_TYPE_SLOTS
              b_dictoffset = b->tp_dictoffset;
  #else
              PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
              if (!py_b_dictoffset) goto dictoffset_return;
              b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
              Py_DECREF(py_b_dictoffset);
              if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
  #endif
              if (b_dictoffset) {
                  {
                      __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
                      PyErr_Format(PyExc_TypeError,
                          "extension type '%.200s' has no __dict__ slot, "
                          "but base type '" __Pyx_FMT_TYPENAME "' has: "
                          "either add 'cdef dict __dict__' to the extension type "
                          "or add '__slots__ = [...]' to the base type",
                          type_name, b_name);
                      __Pyx_DECREF_TypeName(b_name);
                  }
  #if !CYTHON_USE_TYPE_SLOTS
                dictoffset_return:
  #endif
  #if CYTHON_AVOID_BORROWED_REFS
                  Py_DECREF(b0);
  #endif
                  return -1;
              }
          }
  #if CYTHON_AVOID_BORROWED_REFS
          Py_DECREF(b0);
  #endif
      }
      return 0;
  }
  #endif
  
/* PyType_Ready */
  CYTHON_UNUSED static int __Pyx_PyType_HasMultipleInheritance(PyTypeObject *t) {
      while (t) {
          PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
          if (bases) {
              return 1;
          }
          t = __Pyx_PyType_GetSlot(t, tp_base, PyTypeObject*);
      }
      return 0;
  }
  static int __Pyx_PyType_Ready(PyTypeObject *t) {
  #if CYTHON_USE_TYPE_SPECS || !CYTHON_COMPILING_IN_CPYTHON || defined(PYSTON_MAJOR_VERSION)
      (void)__Pyx_PyObject_CallMethod0;
  #if CYTHON_USE_TYPE_SPECS
      (void)__Pyx_validate_bases_tuple;
  #endif
      return PyType_Ready(t);
  #else
      int r;
      if (!__Pyx_PyType_HasMultipleInheritance(t)) {
          return PyType_Ready(t);
      }
      PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
      if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
          return -1;
  #if !defined(PYSTON_MAJOR_VERSION)
      {
          int gc_was_enabled;
      #if PY_VERSION_HEX >= 0x030A00b1
          gc_was_enabled = PyGC_Disable();
          (void)__Pyx_PyObject_CallMethod0;
      #else
          PyObject *ret, *py_status;
          PyObject *gc = NULL;
          #if (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400) &&\
                  !CYTHON_COMPILING_IN_GRAAL
          gc = PyImport_GetModule(__pyx_mstate_global->__pyx_kp_u_gc);
          #endif
          if (unlikely(!gc)) gc = PyImport_Import(__pyx_mstate_global->__pyx_kp_u_gc);
          if (unlikely(!gc)) return -1;
          py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_isenabled);
          if (unlikely(!py_status)) {
              Py_DECREF(gc);
              return -1;
          }
          gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
          Py_DECREF(py_status);
          if (gc_was_enabled > 0) {
              ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_disable);
              if (unlikely(!ret)) {
                  Py_DECREF(gc);
                  return -1;
              }
              Py_DECREF(ret);
          } else if (unlikely(gc_was_enabled == -1)) {
              Py_DECREF(gc);
              return -1;
          }
      #endif
          t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
  #if PY_VERSION_HEX >= 0x030A0000
          t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
  #endif
  #else
          (void)__Pyx_PyObject_CallMethod0;
  #endif
      r = PyType_Ready(t);
  #if !defined(PYSTON_MAJOR_VERSION)
          t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
      #if PY_VERSION_HEX >= 0x030A00b1
          if (gc_was_enabled)
              PyGC_Enable();
      #else
          if (gc_was_enabled) {
              PyObject *tp, *v, *tb;
              PyErr_Fetch(&tp, &v, &tb);
              ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_enable);
              if (likely(ret || r == -1)) {
                  Py_XDECREF(ret);
                  PyErr_Restore(tp, v, tb);
              } else {
                  Py_XDECREF(tp);
                  Py_XDECREF(v);
                  Py_XDECREF(tb);
                  r = -1;
              }
          }
          Py_DECREF(gc);
      #endif
      }
  #endif
      return r;
  #endif
  }
  
/* SetVTable */
  static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
      PyObject *ob = PyCapsule_New(vtable, 0, 0);
      if (unlikely(!ob))
          goto bad;
  #if CYTHON_COMPILING_IN_LIMITED_API
      if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
  #else
      if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
  #endif
          goto bad;
      Py_DECREF(ob);
      return 0;
  bad:
      Py_XDECREF(ob);
      return -1;
  }
  
/* GetVTable (used by MergeVTables) */
  static void* __Pyx_GetVtable(PyTypeObject *type) {
      void* ptr;
  #if CYTHON_COMPILING_IN_LIMITED_API
      PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
  #else
      PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
  #endif
      if (!ob)
          goto bad;
      ptr = PyCapsule_GetPointer(ob, 0);
      if (!ptr && !PyErr_Occurred())
          PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
      Py_DECREF(ob);
      return ptr;
  bad:
      Py_XDECREF(ob);
      return NULL;
  }
  
/* MergeVTables */
  static int __Pyx_MergeVtables(PyTypeObject *type) {
      int i=0;
      Py_ssize_t size;
      void** base_vtables;
      __Pyx_TypeName tp_base_name = NULL;
      __Pyx_TypeName base_name = NULL;
      void* unknown = (void*)-1;
      PyObject* bases = __Pyx_PyType_GetSlot(type, tp_bases, PyObject*);
      int base_depth = 0;
      {
          PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
          while (base) {
              base_depth += 1;
              base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
          }
      }
      base_vtables = (void**) PyMem_Malloc(sizeof(void*) * (size_t)(base_depth + 1));
      base_vtables[0] = unknown;
  #if CYTHON_COMPILING_IN_LIMITED_API
      size = PyTuple_Size(bases);
      if (size < 0) goto other_failure;
  #else
      size = PyTuple_GET_SIZE(bases);
  #endif
      for (i = 1; i < size; i++) {
          PyObject *basei;
          void* base_vtable;
  #if CYTHON_AVOID_BORROWED_REFS
          basei = PySequence_GetItem(bases, i);
          if (unlikely(!basei)) goto other_failure;
  #elif !CYTHON_ASSUME_SAFE_MACROS
          basei = PyTuple_GetItem(bases, i);
          if (unlikely(!basei)) goto other_failure;
  #else
          basei = PyTuple_GET_ITEM(bases, i);
  #endif
          base_vtable = __Pyx_GetVtable((PyTypeObject*)basei);
  #if CYTHON_AVOID_BORROWED_REFS
          Py_DECREF(basei);
  #endif
          if (base_vtable != NULL) {
              int j;
              PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
              for (j = 0; j < base_depth; j++) {
                  if (base_vtables[j] == unknown) {
                      base_vtables[j] = __Pyx_GetVtable(base);
                      base_vtables[j + 1] = unknown;
                  }
                  if (base_vtables[j] == base_vtable) {
                      break;
                  } else if (base_vtables[j] == NULL) {
                      goto bad;
                  }
                  base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
              }
          }
      }
      PyErr_Clear();
      PyMem_Free(base_vtables);
      return 0;
  bad:
      {
          PyTypeObject* basei = NULL;
          PyTypeObject* tp_base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
          tp_base_name = __Pyx_PyType_GetFullyQualifiedName(tp_base);
  #if CYTHON_AVOID_BORROWED_REFS
          basei = (PyTypeObject*)PySequence_GetItem(bases, i);
          if (unlikely(!basei)) goto really_bad;
  #elif !CYTHON_ASSUME_SAFE_MACROS
          basei = (PyTypeObject*)PyTuple_GetItem(bases, i);
          if (unlikely(!basei)) goto really_bad;
  #else
          basei = (PyTypeObject*)PyTuple_GET_ITEM(bases, i);
  #endif
          base_name = __Pyx_PyType_GetFullyQualifiedName(basei);
  #if CYTHON_AVOID_BORROWED_REFS
          Py_DECREF(basei);
  #endif
      }
      PyErr_Format(PyExc_TypeError,
          "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
  #if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
  really_bad: // bad has failed!
  #endif
      __Pyx_DECREF_TypeName(tp_base_name);
      __Pyx_DECREF_TypeName(base_name);
  #if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
  other_failure:
  #endif
      PyMem_Free(base_vtables);
      return -1;
  }
  
/* DelItemOnTypeDict (used by SetupReduce) */
  static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k) {
      int result;
      PyObject *tp_dict;
  #if CYTHON_COMPILING_IN_LIMITED_API
      tp_dict = __Pyx_GetTypeDict(tp);
      if (unlikely(!tp_dict)) return -1;
  #else
      tp_dict = tp->tp_dict;
  #endif
      result = PyDict_DelItem(tp_dict, k);
      if (likely(!result)) PyType_Modified(tp);
      return result;
  }
  
/* SetupReduce */
  static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
    int ret;
    PyObject *name_attr;
    name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_mstate_global->__pyx_n_u_name);
    if (likely(name_attr)) {
        ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
    } else {
        ret = -1;
    }
    if (unlikely(ret < 0)) {
        PyErr_Clear();
        ret = 0;
    }
    Py_XDECREF(name_attr);
    return ret;
  }
  static int __Pyx_setup_reduce(PyObject* type_obj) {
      int ret = 0;
      PyObject *object_reduce = NULL;
      PyObject *object_getstate = NULL;
      PyObject *object_reduce_ex = NULL;
      PyObject *reduce = NULL;
      PyObject *reduce_ex = NULL;
      PyObject *reduce_cython = NULL;
      PyObject *setstate = NULL;
      PyObject *setstate_cython = NULL;
      PyObject *getstate = NULL;
  #if CYTHON_USE_PYTYPE_LOOKUP
      getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
  #else
      getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
      if (!getstate && PyErr_Occurred()) {
          goto __PYX_BAD;
      }
  #endif
      if (getstate) {
  #if CYTHON_USE_PYTYPE_LOOKUP
          object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
  #else
          object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
          if (!object_getstate && PyErr_Occurred()) {
              goto __PYX_BAD;
          }
  #endif
          if (object_getstate != getstate) {
              goto __PYX_GOOD;
          }
      }
  #if CYTHON_USE_PYTYPE_LOOKUP
      object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
  #else
      object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
  #endif
      reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
      if (reduce_ex == object_reduce_ex) {
  #if CYTHON_USE_PYTYPE_LOOKUP
          object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
  #else
          object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
  #endif
          reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
          if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_mstate_global->__pyx_n_u_reduce_cython)) {
              reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython);
              if (likely(reduce_cython)) {
                  ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                  ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
              } else if (reduce == object_reduce || PyErr_Occurred()) {
                  goto __PYX_BAD;
              }
              setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate);
              if (!setstate) PyErr_Clear();
              if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_mstate_global->__pyx_n_u_setstate_cython)) {
                  setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython);
                  if (likely(setstate_cython)) {
                      ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                      ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                  } else if (!setstate || PyErr_Occurred()) {
                      goto __PYX_BAD;
                  }
              }
              PyType_Modified((PyTypeObject*)type_obj);
          }
      }
      goto __PYX_GOOD;
  __PYX_BAD:
      if (!PyErr_Occurred()) {
          __Pyx_TypeName type_obj_name =
              __Pyx_PyType_GetFullyQualifiedName((PyTypeObject*)type_obj);
          PyErr_Format(PyExc_RuntimeError,
              "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
          __Pyx_DECREF_TypeName(type_obj_name);
      }
      ret = -1;
  __PYX_GOOD:
  #if !CYTHON_USE_PYTYPE_LOOKUP
      Py_XDECREF(object_reduce);
      Py_XDECREF(object_reduce_ex);
      Py_XDECREF(object_getstate);
      Py_XDECREF(getstate);
  #endif
      Py_XDECREF(reduce);
      Py_XDECREF(reduce_ex);
      Py_XDECREF(reduce_cython);
      Py_XDECREF(setstate);
      Py_XDECREF(setstate_cython);
      return ret;
  }
  
/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType_3_2_4
  #define __PYX_HAVE_RT_ImportType_3_2_4
  static PyTypeObject *__Pyx_ImportType_3_2_4(PyObject *module, const char *module_name, const char *class_name,
      size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_2_4 check_size)
  {
      PyObject *result = 0;
      Py_ssize_t basicsize;
      Py_ssize_t itemsize;
  #if defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API)
      PyObject *py_basicsize;
      PyObject *py_itemsize;
  #endif
      result = PyObject_GetAttrString(module, class_name);
      if (!result)
          goto bad;
      if (!PyType_Check(result)) {
          PyErr_Format(PyExc_TypeError,
              "%.200s.%.200s is not a type object",
              module_name, class_name);
          goto bad;
      }
  #if !( defined(Py_LIMITED_API) || (defined(CYTHON_COMPILING_IN_LIMITED_API) && CYTHON_COMPILING_IN_LIMITED_API) )
      basicsize = ((PyTypeObject *)result)->tp_basicsize;
      itemsize = ((PyTypeObject *)result)->tp_itemsize;
  #else
      if (size == 0) {
          return (PyTypeObject *)result;
      }
      py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
      if (!py_basicsize)
          goto bad;
      basicsize = PyLong_AsSsize_t(py_basicsize);
      Py_DECREF(py_basicsize);
      py_basicsize = 0;
      if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
          goto bad;
      py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
      if (!py_itemsize)
          goto bad;
      itemsize = PyLong_AsSsize_t(py_itemsize);
      Py_DECREF(py_itemsize);
      py_itemsize = 0;
      if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
          goto bad;
  #endif
      if (itemsize) {
          if (size % alignment) {
              alignment = size % alignment;
          }
          if (itemsize < (Py_ssize_t)alignment)
              itemsize = (Py_ssize_t)alignment;
      }
      if ((size_t)(basicsize + itemsize) < size) {
          PyErr_Format(PyExc_ValueError,
              "%.200s.%.200s size changed, may indicate binary incompatibility. "
              "Expected %zd from C header, got %zd from PyObject",
              module_name, class_name, size, basicsize+itemsize);
          goto bad;
      }
      if (check_size == __Pyx_ImportType_CheckSize_Error_3_2_4 &&
              ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
          PyErr_Format(PyExc_ValueError,
              "%.200s.%.200s size changed, may indicate binary incompatibility. "
              "Expected %zd from C header, got %zd-%zd from PyObject",
              module_name, class_name, size, basicsize, basicsize+itemsize);
          goto bad;
      }
      else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_2_4 && (size_t)basicsize > size) {
          if (PyErr_WarnFormat(NULL, 0,
                  "%.200s.%.200s size changed, may indicate binary incompatibility. "
                  "Expected %zd from C header, got %zd from PyObject",
                  module_name, class_name, size, basicsize) < 0) {
              goto bad;
          }
      }
      return (PyTypeObject *)result;
  bad:
      Py_XDECREF(result);
      return NULL;
  }
  #endif
  
/* HasAttr (used by ImportImpl) */
  #if __PYX_LIMITED_VERSION_HEX < 0x030d0000
  static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
      PyObject *r;
      if (unlikely(!PyUnicode_Check(n))) {
          PyErr_SetString(PyExc_TypeError,
                          "hasattr(): attribute name must be string");
          return -1;
      }
      r = __Pyx_PyObject_GetAttrStrNoError(o, n);
      if (!r) {
          return (unlikely(PyErr_Occurred())) ? -1 : 0;
      } else {
          Py_DECREF(r);
          return 1;
      }
  }
  #endif
  
/* ImportImpl (used by Import) */
  static int __Pyx__Import_GetModule(PyObject *qualname, PyObject **module) {
      PyObject *imported_module = PyImport_GetModule(qualname);
      if (unlikely(!imported_module)) {
          *module = NULL;
          if (PyErr_Occurred()) {
              return -1;
          }
          return 0;
      }
      *module = imported_module;
      return 1;
  }
  static int __Pyx__Import_Lookup(PyObject *qualname, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject **module) {
      PyObject *imported_module;
      PyObject *top_level_package_name;
      Py_ssize_t i;
      int status, module_found;
      Py_ssize_t dot_index;
      module_found = __Pyx__Import_GetModule(qualname, &imported_module);
      if (unlikely(!module_found || module_found == -1)) {
          *module = NULL;
          return module_found;
      }
      if (imported_names) {
          for (i = 0; i < len_imported_names; i++) {
              PyObject *imported_name = imported_names[i];
  #if __PYX_LIMITED_VERSION_HEX < 0x030d0000
              int has_imported_attribute = PyObject_HasAttr(imported_module, imported_name);
  #else
              int has_imported_attribute = PyObject_HasAttrWithError(imported_module, imported_name);
              if (unlikely(has_imported_attribute == -1)) goto error;
  #endif
              if (!has_imported_attribute) {
                  goto not_found;
              }
          }
          *module = imported_module;
          return 1;
      }
      dot_index = PyUnicode_FindChar(qualname, '.', 0, PY_SSIZE_T_MAX, 1);
      if (dot_index == -1) {
          *module = imported_module;
          return 1;
      }
      if (unlikely(dot_index == -2)) goto error;
      top_level_package_name = PyUnicode_Substring(qualname, 0, dot_index);
      if (unlikely(!top_level_package_name)) goto error;
      Py_DECREF(imported_module);
      status = __Pyx__Import_GetModule(top_level_package_name, module);
      Py_DECREF(top_level_package_name);
      return status;
  error:
      Py_DECREF(imported_module);
      *module = NULL;
      return -1;
  not_found:
      Py_DECREF(imported_module);
      *module = NULL;
      return 0;
  }
  static PyObject *__Pyx__Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, PyObject *moddict, int level) {
      PyObject *module = 0;
      PyObject *empty_dict = 0;
      PyObject *from_list = 0;
      int module_found;
      if (!qualname) {
          qualname = name;
      }
      module_found = __Pyx__Import_Lookup(qualname, imported_names, len_imported_names, &module);
      if (likely(module_found == 1)) {
          return module;
      } else if (unlikely(module_found == -1)) {
          return NULL;
      }
      empty_dict = PyDict_New();
      if (unlikely(!empty_dict))
          goto bad;
      if (imported_names) {
  #if CYTHON_COMPILING_IN_CPYTHON
          from_list = __Pyx_PyList_FromArray(imported_names, len_imported_names);
          if (unlikely(!from_list))
              goto bad;
  #else
          from_list = PyList_New(len_imported_names);
          if (unlikely(!from_list)) goto bad;
          for (Py_ssize_t i=0; i<len_imported_names; ++i) {
              if (PyList_SetItem(from_list, i, __Pyx_NewRef(imported_names[i])) < 0) goto bad;
          }
  #endif
      }
      if (level == -1) {
          const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
          if (package_sep != (0)) {
              module = PyImport_ImportModuleLevelObject(
                  name, moddict, empty_dict, from_list, 1);
              if (unlikely(!module)) {
                  if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                      goto bad;
                  PyErr_Clear();
              }
          }
          level = 0;
      }
      if (!module) {
          module = PyImport_ImportModuleLevelObject(
              name, moddict, empty_dict, from_list, level);
      }
  bad:
      Py_XDECREF(from_list);
      Py_XDECREF(empty_dict);
      return module;
  }
  
/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *const *imported_names, Py_ssize_t len_imported_names, PyObject *qualname, int level) {
      return __Pyx__Import(name, imported_names, len_imported_names, qualname, __pyx_mstate_global->__pyx_d, level);
  }
  
/* dict_setdefault (used by FetchCommonType) */
  static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value) {
      PyObject* value;
  #if __PYX_LIMITED_VERSION_HEX >= 0x030F0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4)
      PyDict_SetDefaultRef(d, key, default_value, &value);
  #elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x030C0000
      PyObject *args[] = {d, key, default_value};
      value = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_n_u_setdefault, args, 3 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
  #elif CYTHON_COMPILING_IN_LIMITED_API
      value = PyObject_CallMethodObjArgs(d, __pyx_mstate_global->__pyx_n_u_setdefault, key, default_value, NULL);
  #else
      value = PyDict_SetDefault(d, key, default_value);
      if (unlikely(!value)) return NULL;
      Py_INCREF(value);
  #endif
      return value;
  }
  
/* AddModuleRef (used by FetchSharedCythonModule) */
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    static PyObject *__Pyx_PyImport_AddModuleObjectRef(PyObject *name) {
        PyObject *module_dict = PyImport_GetModuleDict();
        PyObject *m;
        if (PyMapping_GetOptionalItem(module_dict, name, &m) < 0) {
            return NULL;
        }
        if (m != NULL && PyModule_Check(m)) {
            return m;
        }
        Py_XDECREF(m);
        m = PyModule_NewObject(name);
        if (m == NULL)
            return NULL;
        if (PyDict_CheckExact(module_dict)) {
            PyObject *new_m;
            (void)PyDict_SetDefaultRef(module_dict, name, m, &new_m);
            Py_DECREF(m);
            return new_m;
        } else {
             if (PyObject_SetItem(module_dict, name, m) != 0) {
                  Py_DECREF(m);
                  return NULL;
              }
              return m;
        }
    }
    static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
        PyObject *py_name = PyUnicode_FromString(name);
        if (!py_name) return NULL;
        PyObject *module = __Pyx_PyImport_AddModuleObjectRef(py_name);
        Py_DECREF(py_name);
        return module;
    }
  #elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
  #else
    static PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
        PyObject *module = PyImport_AddModule(name);
        Py_XINCREF(module);
        return module;
    }
  #endif
  
/* FetchSharedCythonModule (used by FetchCommonType) */
  static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
      return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
  }
  
/* FetchCommonType (used by CommonTypesMetaclass) */
  #if __PYX_LIMITED_VERSION_HEX < 0x030C0000
  static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
      PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
      if (result && metaclass) {
          PyObject *old_tp = (PyObject*)Py_TYPE(result);
      Py_INCREF((PyObject*)metaclass);
  #if __PYX_LIMITED_VERSION_HEX >= 0x03090000
          Py_SET_TYPE(result, metaclass);
  #else
          result->ob_type = metaclass;
  #endif
          Py_DECREF(old_tp);
      }
      return result;
  }
  #else
  #define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
  #endif
  static int __Pyx_VerifyCachedType(PyObject *cached_type,
                                 const char *name,
                                 Py_ssize_t expected_basicsize) {
      Py_ssize_t basicsize;
      if (!PyType_Check(cached_type)) {
          PyErr_Format(PyExc_TypeError,
              "Shared Cython type %.200s is not a type object", name);
          return -1;
      }
      if (expected_basicsize == 0) {
          return 0; // size is inherited, nothing useful to check
      }
  #if CYTHON_COMPILING_IN_LIMITED_API
      PyObject *py_basicsize;
      py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
      if (unlikely(!py_basicsize)) return -1;
      basicsize = PyLong_AsSsize_t(py_basicsize);
      Py_DECREF(py_basicsize);
      py_basicsize = NULL;
      if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
  #else
      basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
  #endif
      if (basicsize != expected_basicsize) {
          PyErr_Format(PyExc_TypeError,
              "Shared Cython type %.200s has the wrong size, try recompiling",
              name);
          return -1;
      }
      return 0;
  }
  static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
      PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
      int get_item_ref_result;
      const char* object_name = strrchr(spec->name, '.');
      object_name = object_name ? object_name+1 : spec->name;
      py_object_name = PyUnicode_FromString(object_name);
      if (!py_object_name) return NULL;
      abi_module = __Pyx_FetchSharedCythonABIModule();
      if (!abi_module) goto done;
      abi_module_dict = PyModule_GetDict(abi_module);
      if (!abi_module_dict) goto done;
      get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
      if (get_item_ref_result == 1) {
          if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
              goto bad;
          }
          goto done;
      } else if (unlikely(get_item_ref_result == -1)) {
          goto bad;
      }
      cached_type = __Pyx_PyType_FromMetaclass(
          metaclass,
          CYTHON_USE_MODULE_STATE ? module : abi_module,
          spec, bases);
      if (unlikely(!cached_type)) goto bad;
      if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
      new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type);
      if (unlikely(new_cached_type != cached_type)) {
          if (unlikely(!new_cached_type)) goto bad;
          Py_DECREF(cached_type);
          cached_type = new_cached_type;
          if (__Pyx_VerifyCachedType(
                  cached_type,
                  object_name,
                  spec->basicsize) < 0) {
              goto bad;
          }
          goto done;
      } else {
          Py_DECREF(new_cached_type);
      }
  done:
      Py_XDECREF(abi_module);
      Py_DECREF(py_object_name);
      assert(cached_type == NULL || PyType_Check(cached_type));
      return (PyTypeObject *) cached_type;
  bad:
      Py_XDECREF(cached_type);
      cached_type = NULL;
      goto done;
  }
  
/* CommonTypesMetaclass (used by CythonFunctionShared) */
  static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
      return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
  }
  #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
  static PyObject* __pyx_CommonTypesMetaclass_call(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *args, CYTHON_UNUSED PyObject *kwds) {
      PyErr_SetString(PyExc_TypeError, "Cannot instantiate Cython internal types");
      return NULL;
  }
  static int __pyx_CommonTypesMetaclass_setattr(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED PyObject *attr, CYTHON_UNUSED PyObject *value) {
      PyErr_SetString(PyExc_TypeError, "Cython internal types are immutable");
      return -1;
  }
  #endif
  static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
      {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
      {0, 0, 0, 0, 0}
  };
  static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
      {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
      #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
      {Py_tp_call, (void*)__pyx_CommonTypesMetaclass_call},
      {Py_tp_new, (void*)__pyx_CommonTypesMetaclass_call},
      {Py_tp_setattro, (void*)__pyx_CommonTypesMetaclass_setattr},
      #endif
      {0, 0}
  };
  static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
      __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
      0,
      0,
      Py_TPFLAGS_IMMUTABLETYPE |
      Py_TPFLAGS_DISALLOW_INSTANTIATION |
      Py_TPFLAGS_DEFAULT,
      __pyx_CommonTypesMetaclass_slots
  };
  static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
      __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
      PyObject *bases = PyTuple_Pack(1, &PyType_Type);
      if (unlikely(!bases)) {
          return -1;
      }
      mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
      Py_DECREF(bases);
      if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
          return -1;
      }
      return 0;
  }
  
/* PyMethodNew (used by CythonFunctionShared) */
  #if CYTHON_COMPILING_IN_LIMITED_API
  static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
      PyObject *result;
      CYTHON_UNUSED_VAR(typ);
      if (!self)
          return __Pyx_NewRef(func);
      #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
      {
          PyObject *args[] = {func, self};
          result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
      }
      #else
      result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
      #endif
      return result;
  }
  #else
  static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
      CYTHON_UNUSED_VAR(typ);
      if (!self)
          return __Pyx_NewRef(func);
      return PyMethod_New(func, self);
  }
  #endif
  
/* PyVectorcallFastCallDict (used by CythonFunctionShared) */
  #if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
  static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
  {
      PyObject *res = NULL;
      PyObject *kwnames;
      PyObject **newargs;
      PyObject **kwvalues;
      Py_ssize_t i;
      #if CYTHON_AVOID_BORROWED_REFS
      PyObject *pos;
      #else
      Py_ssize_t pos;
      #endif
      size_t j;
      PyObject *key, *value;
      unsigned long keys_are_strings;
      #if !CYTHON_ASSUME_SAFE_SIZE
      Py_ssize_t nkw = PyDict_Size(kw);
      if (unlikely(nkw == -1)) return NULL;
      #else
      Py_ssize_t nkw = PyDict_GET_SIZE(kw);
      #endif
      newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
      if (unlikely(newargs == NULL)) {
          PyErr_NoMemory();
          return NULL;
      }
      for (j = 0; j < nargs; j++) newargs[j] = args[j];
      kwnames = PyTuple_New(nkw);
      if (unlikely(kwnames == NULL)) {
          PyMem_Free(newargs);
          return NULL;
      }
      kwvalues = newargs + nargs;
      pos = 0;
      i = 0;
      keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
      while (__Pyx_PyDict_NextRef(kw, &pos, &key, &value)) {
          keys_are_strings &=
          #if CYTHON_COMPILING_IN_LIMITED_API
              PyType_GetFlags(Py_TYPE(key));
          #else
              Py_TYPE(key)->tp_flags;
          #endif
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
          #else
          PyTuple_SET_ITEM(kwnames, i, key);
          #endif
          kwvalues[i] = value;
          i++;
      }
      if (unlikely(!keys_are_strings)) {
          PyErr_SetString(PyExc_TypeError, "keywords must be strings");
          goto cleanup;
      }
      res = vc(func, newargs, nargs, kwnames);
  cleanup:
      #if CYTHON_AVOID_BORROWED_REFS
      Py_DECREF(pos);
      #endif
      Py_DECREF(kwnames);
      for (i = 0; i < nkw; i++)
          Py_DECREF(kwvalues[i]);
      PyMem_Free(newargs);
      return res;
  }
  static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
  {
      Py_ssize_t kw_size =
          likely(kw == NULL) ?
          0 :
  #if !CYTHON_ASSUME_SAFE_SIZE
          PyDict_Size(kw);
  #else
          PyDict_GET_SIZE(kw);
  #endif
      if (kw_size == 0) {
          return vc(func, args, nargs, NULL);
      }
  #if !CYTHON_ASSUME_SAFE_SIZE
      else if (unlikely(kw_size == -1)) {
          return NULL;
      }
  #endif
      return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
  }
  #endif
  
/* CythonFunctionShared (used by CythonFunction) */
  #if CYTHON_COMPILING_IN_LIMITED_API
  static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
      if (__Pyx_CyFunction_Check(func)) {
          return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
      } else if (PyCFunction_Check(func)) {
          return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
      }
      return 0;
  }
  static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
      if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
          int result;
          PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
          if (unlikely(!newFunc)) {
              PyErr_Clear(); // It's only an optimization, so don't throw an error
              return 0;
          }
          result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
          Py_DECREF(newFunc);
          return result;
      }
      return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
  }
  #else
  static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
      if (PyMethod_Check(func)) {
          func = PyMethod_GET_FUNCTION(func);
      }
      return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
  }
  #endif
  static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
  #if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
      __Pyx_Py_XDECREF_SET(
          __Pyx_CyFunction_GetClassObj(f),
              ((classobj) ? __Pyx_NewRef(classobj) : NULL));
  #else
      __Pyx_Py_XDECREF_SET(
          ((PyCMethodObject *) (f))->mm_class,
          (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
  #endif
  }
  static PyObject *
  __Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
  {
      if (unlikely(op->func_doc == NULL)) {
  #if CYTHON_COMPILING_IN_LIMITED_API
          op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
          if (unlikely(!op->func_doc)) return NULL;
  #else
          if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
              op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
              if (unlikely(op->func_doc == NULL))
                  return NULL;
          } else {
              Py_INCREF(Py_None);
              return Py_None;
          }
  #endif
      }
      Py_INCREF(op->func_doc);
      return op->func_doc;
  }
  static PyObject *
  __Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
      PyObject *result;
      CYTHON_UNUSED_VAR(closure);
      __Pyx_BEGIN_CRITICAL_SECTION(op);
      result = __Pyx_CyFunction_get_doc_locked(op);
      __Pyx_END_CRITICAL_SECTION();
      return result;
  }
  static int
  __Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
  {
      CYTHON_UNUSED_VAR(context);
      if (value == NULL) {
          value = Py_None;
      }
      Py_INCREF(value);
      __Pyx_BEGIN_CRITICAL_SECTION(op);
      __Pyx_Py_XDECREF_SET(op->func_doc, value);
      __Pyx_END_CRITICAL_SECTION();
      return 0;
  }
  static PyObject *
  __Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
  {
      if (unlikely(op->func_name == NULL)) {
  #if CYTHON_COMPILING_IN_LIMITED_API
          op->func_name = PyObject_GetAttrString(op->func, "__name__");
  #else
          op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
  #endif
          if (unlikely(op->func_name == NULL))
              return NULL;
      }
      Py_INCREF(op->func_name);
      return op->func_name;
  }
  static PyObject *
  __Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
  {
      PyObject *result = NULL;
      CYTHON_UNUSED_VAR(context);
      __Pyx_BEGIN_CRITICAL_SECTION(op);
      result = __Pyx_CyFunction_get_name_locked(op);
      __Pyx_END_CRITICAL_SECTION();
      return result;
  }
  static int
  __Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
  {
      CYTHON_UNUSED_VAR(context);
      if (unlikely(value == NULL || !PyUnicode_Check(value))) {
          PyErr_SetString(PyExc_TypeError,
                          "__name__ must be set to a string object");
          return -1;
      }
      Py_INCREF(value);
      __Pyx_BEGIN_CRITICAL_SECTION(op);
      __Pyx_Py_XDECREF_SET(op->func_name, value);
      __Pyx_END_CRITICAL_SECTION();
      return 0;
  }
  static PyObject *
  __Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
  {
      CYTHON_UNUSED_VAR(context);
      PyObject *result;
      __Pyx_BEGIN_CRITICAL_SECTION(op);
      Py_INCREF(op->func_qualname);
      result = op->func_qualname;
      __Pyx_END_CRITICAL_SECTION();
      return result;
  }
  static int
  __Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
  {
      CYTHON_UNUSED_VAR(context);
      if (unlikely(value == NULL || !PyUnicode_Check(value))) {
          PyErr_SetString(PyExc_TypeError,
                          "__qualname__ must be set to a string object");
          return -1;
      }
      Py_INCREF(value);
      __Pyx_BEGIN_CRITICAL_SECTION(op);
      __Pyx_Py_XDECREF_SET(op->func_qualname, value);
      __Pyx_END_CRITICAL_SECTION();
      return 0;
  }
  #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
  static PyObject *
  __Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
  {
      CYTHON_UNUSED_VAR(context);
      if (unlikely(op->func_dict == NULL)) {
          op->func_dict = PyDict_New();
          if (unlikely(op->func_dict == NULL))
              return NULL;
      }
      Py_INCREF(op->func_dict);
      return op->func_dict;
  }
  #endif
  static PyObject *
  __Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
  {
      CYTHON_UNUSED_VAR(context);
      Py_INCREF(op->func_globals);
      return op->func_globals;
  }
  static PyObject *
  __Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
  {
      CYTHON_UNUSED_VAR(op);
      CYTHON_UNUSED_VAR(context);
      Py_INCREF(Py_None);
      return Py_None;
  }
  static PyObject *
  __Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
  {
      PyObject* result = (op->func_code) ? op->func_code : Py_None;
      CYTHON_UNUSED_VAR(context);
      Py_INCREF(result);
      return result;
  }
  static int
  __Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
      int result = 0;
      PyObject *res = op->defaults_getter((PyObject *) op);
      if (unlikely(!res))
          return -1;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
      Py_INCREF(op->defaults_tuple);
      op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
      Py_INCREF(op->defaults_kwdict);
      #else
      op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
      if (unlikely(!op->defaults_tuple)) result = -1;
      else {
          op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
          if (unlikely(!op->defaults_kwdict)) result = -1;
      }
      #endif
      Py_DECREF(res);
      return result;
  }
  static int
  __Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
      CYTHON_UNUSED_VAR(context);
      if (!value) {
          value = Py_None;
      } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
          PyErr_SetString(PyExc_TypeError,
                          "__defaults__ must be set to a tuple object");
          return -1;
      }
      PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                   "currently affect the values used in function calls", 1);
      Py_INCREF(value);
      __Pyx_BEGIN_CRITICAL_SECTION(op);
      __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
      __Pyx_END_CRITICAL_SECTION();
      return 0;
  }
  static PyObject *
  __Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
      PyObject* result = op->defaults_tuple;
      if (unlikely(!result)) {
          if (op->defaults_getter) {
              if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
              result = op->defaults_tuple;
          } else {
              result = Py_None;
          }
      }
      Py_INCREF(result);
      return result;
  }
  static PyObject *
  __Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
      PyObject* result = NULL;
      CYTHON_UNUSED_VAR(context);
      __Pyx_BEGIN_CRITICAL_SECTION(op);
      result = __Pyx_CyFunction_get_defaults_locked(op);
      __Pyx_END_CRITICAL_SECTION();
      return result;
  }
  static int
  __Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
      CYTHON_UNUSED_VAR(context);
      if (!value) {
          value = Py_None;
      } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
          PyErr_SetString(PyExc_TypeError,
                          "__kwdefaults__ must be set to a dict object");
          return -1;
      }
      PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                   "currently affect the values used in function calls", 1);
      Py_INCREF(value);
      __Pyx_BEGIN_CRITICAL_SECTION(op);
      __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
      __Pyx_END_CRITICAL_SECTION();
      return 0;
  }
  static PyObject *
  __Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
      PyObject* result = op->defaults_kwdict;
      if (unlikely(!result)) {
          if (op->defaults_getter) {
              if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
              result = op->defaults_kwdict;
          } else {
              result = Py_None;
          }
      }
      Py_INCREF(result);
      return result;
  }
  static PyObject *
  __Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
      PyObject* result;
      CYTHON_UNUSED_VAR(context);
      __Pyx_BEGIN_CRITICAL_SECTION(op);
      result = __Pyx_CyFunction_get_kwdefaults_locked(op);
      __Pyx_END_CRITICAL_SECTION();
      return result;
  }
  static int
  __Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
      CYTHON_UNUSED_VAR(context);
      if (!value || value == Py_None) {
          value = NULL;
      } else if (unlikely(!PyDict_Check(value))) {
          PyErr_SetString(PyExc_TypeError,
                          "__annotations__ must be set to a dict object");
          return -1;
      }
      Py_XINCREF(value);
      __Pyx_BEGIN_CRITICAL_SECTION(op);
      __Pyx_Py_XDECREF_SET(op->func_annotations, value);
      __Pyx_END_CRITICAL_SECTION();
      return 0;
  }
  static PyObject *
  __Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
      PyObject* result = op->func_annotations;
      if (unlikely(!result)) {
          result = PyDict_New();
          if (unlikely(!result)) return NULL;
          op->func_annotations = result;
      }
      Py_INCREF(result);
      return result;
  }
  static PyObject *
  __Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
      PyObject *result;
      CYTHON_UNUSED_VAR(context);
      __Pyx_BEGIN_CRITICAL_SECTION(op);
      result = __Pyx_CyFunction_get_annotations_locked(op);
      __Pyx_END_CRITICAL_SECTION();
      return result;
  }
  static PyObject *
  __Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
      int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
      if (is_coroutine) {
          PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
          fromlist = PyList_New(1);
          if (unlikely(!fromlist)) return NULL;
          Py_INCREF(marker);
  #if CYTHON_ASSUME_SAFE_MACROS
          PyList_SET_ITEM(fromlist, 0, marker);
  #else
          if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
              Py_DECREF(marker);
              Py_DECREF(fromlist);
              return NULL;
          }
  #endif
          module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
          Py_DECREF(fromlist);
          if (unlikely(!module)) goto ignore;
          is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
          Py_DECREF(module);
          if (likely(is_coroutine_value)) {
              return is_coroutine_value;
          }
  ignore:
          PyErr_Clear();
      }
      return __Pyx_PyBool_FromLong(is_coroutine);
  }
  static PyObject *
  __Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
      PyObject *result;
      CYTHON_UNUSED_VAR(context);
      if (op->func_is_coroutine) {
          return __Pyx_NewRef(op->func_is_coroutine);
      }
      result = __Pyx_CyFunction_get_is_coroutine_value(op);
      if (unlikely(!result))
          return NULL;
      __Pyx_BEGIN_CRITICAL_SECTION(op);
      if (op->func_is_coroutine) {
          Py_DECREF(result);
          result = __Pyx_NewRef(op->func_is_coroutine);
      } else {
          op->func_is_coroutine = __Pyx_NewRef(result);
      }
      __Pyx_END_CRITICAL_SECTION();
      return result;
  }
  static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
  #if CYTHON_COMPILING_IN_LIMITED_API
      PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
      if (!py_name) return;
      PyErr_Format(PyExc_TypeError,
          "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
          py_name, message, size);
      Py_DECREF(py_name);
  #else
      const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
      PyErr_Format(PyExc_TypeError,
          "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
          name, message, size);
  #endif
  }
  static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
  #if CYTHON_COMPILING_IN_LIMITED_API
      PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
      if (!py_name) return;
      PyErr_Format(PyExc_TypeError,
          "%.200S() %s",
          py_name, message);
      Py_DECREF(py_name);
  #else
      const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
      PyErr_Format(PyExc_TypeError,
          "%.200s() %s",
          name, message);
  #endif
  }
  #if CYTHON_COMPILING_IN_LIMITED_API
  static PyObject *
  __Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
      CYTHON_UNUSED_VAR(context);
      return PyObject_GetAttrString(op->func, "__module__");
  }
  static int
  __Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
      CYTHON_UNUSED_VAR(context);
      return PyObject_SetAttrString(op->func, "__module__", value);
  }
  #endif
  static PyGetSetDef __pyx_CyFunction_getsets[] = {
      {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
      {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
      {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
      {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
      {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
  #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
      {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
      {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)PyObject_GenericSetDict, 0, 0},
  #else
      {"func_dict", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
      {"__dict__", (getter)PyObject_GenericGetDict, (setter)PyObject_GenericSetDict, 0, 0},
  #endif
      {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
      {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
      {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
      {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
      {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
      {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
      {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
      {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
      {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
      {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
      {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
  #if CYTHON_COMPILING_IN_LIMITED_API
      {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
  #endif
      {0, 0, 0, 0, 0}
  };
  static PyMemberDef __pyx_CyFunction_members[] = {
  #if !CYTHON_COMPILING_IN_LIMITED_API
      {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
  #endif
  #if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
      {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
  #endif
  #if CYTHON_METH_FASTCALL
  #if CYTHON_COMPILING_IN_LIMITED_API
      {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
  #else
      {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
      {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
  #else
      {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
  #endif
  #endif
      {0, 0, 0,  0, 0}
  };
  static PyObject *
  __Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
  {
      PyObject *result = NULL;
      CYTHON_UNUSED_VAR(args);
      __Pyx_BEGIN_CRITICAL_SECTION(m);
      Py_INCREF(m->func_qualname);
      result = m->func_qualname;
      __Pyx_END_CRITICAL_SECTION();
      return result;
  }
  static PyMethodDef __pyx_CyFunction_methods[] = {
      {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
      {0, 0, 0, 0}
  };
  #if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
  #else
  #define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
  #endif
  static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                         PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
  #if !CYTHON_COMPILING_IN_LIMITED_API
      PyCFunctionObject *cf = (PyCFunctionObject*) op;
  #endif
      if (unlikely(op == NULL))
          return NULL;
  #if CYTHON_COMPILING_IN_LIMITED_API
      op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
      if (unlikely(!op->func)) return NULL;
  #endif
      op->flags = flags;
      __Pyx_CyFunction_weakreflist(op) = NULL;
  #if !CYTHON_COMPILING_IN_LIMITED_API
      cf->m_ml = ml;
      cf->m_self = (PyObject *) op;
  #endif
      Py_XINCREF(closure);
      op->func_closure = closure;
  #if !CYTHON_COMPILING_IN_LIMITED_API
      Py_XINCREF(module);
      cf->m_module = module;
  #endif
  #if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
      op->func_dict = NULL;
  #endif
      op->func_name = NULL;
      Py_INCREF(qualname);
      op->func_qualname = qualname;
      op->func_doc = NULL;
  #if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
      op->func_classobj = NULL;
  #else
      ((PyCMethodObject*)op)->mm_class = NULL;
  #endif
      op->func_globals = globals;
      Py_INCREF(op->func_globals);
      Py_XINCREF(code);
      op->func_code = code;
      op->defaults = NULL;
      op->defaults_tuple = NULL;
      op->defaults_kwdict = NULL;
      op->defaults_getter = NULL;
      op->func_annotations = NULL;
      op->func_is_coroutine = NULL;
  #if CYTHON_METH_FASTCALL
      switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
      case METH_NOARGS:
          __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
          break;
      case METH_O:
          __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
          break;
      case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
          __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
          break;
      case METH_FASTCALL | METH_KEYWORDS:
          __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
          break;
      case METH_VARARGS | METH_KEYWORDS:
          __Pyx_CyFunction_func_vectorcall(op) = NULL;
          break;
      default:
          PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
          Py_DECREF(op);
          return NULL;
      }
  #endif
      return (PyObject *) op;
  }
  static int
  __Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
  {
      Py_CLEAR(m->func_closure);
  #if CYTHON_COMPILING_IN_LIMITED_API
      Py_CLEAR(m->func);
  #else
      Py_CLEAR(((PyCFunctionObject*)m)->m_module);
  #endif
  #if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
      Py_CLEAR(m->func_dict);
  #elif PY_VERSION_HEX < 0x030d0000
      _PyObject_ClearManagedDict((PyObject*)m);
  #else
      PyObject_ClearManagedDict((PyObject*)m);
  #endif
      Py_CLEAR(m->func_name);
      Py_CLEAR(m->func_qualname);
      Py_CLEAR(m->func_doc);
      Py_CLEAR(m->func_globals);
      Py_CLEAR(m->func_code);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x030900B1
      Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
  #else
      {
          PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
          ((PyCMethodObject *) (m))->mm_class = NULL;
          Py_XDECREF(cls);
      }
  #endif
  #endif
      Py_CLEAR(m->defaults_tuple);
      Py_CLEAR(m->defaults_kwdict);
      Py_CLEAR(m->func_annotations);
      Py_CLEAR(m->func_is_coroutine);
      Py_CLEAR(m->defaults);
      return 0;
  }
  static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
  {
      if (__Pyx_CyFunction_weakreflist(m) != NULL)
          PyObject_ClearWeakRefs((PyObject *) m);
      __Pyx_CyFunction_clear(m);
      __Pyx_PyHeapTypeObject_GC_Del(m);
  }
  static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
  {
      PyObject_GC_UnTrack(m);
      __Pyx__CyFunction_dealloc(m);
  }
  static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
  {
      {
          int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
          if (e) return e;
      }
      Py_VISIT(m->func_closure);
  #if CYTHON_COMPILING_IN_LIMITED_API
      Py_VISIT(m->func);
  #else
      Py_VISIT(((PyCFunctionObject*)m)->m_module);
  #endif
  #if PY_VERSION_HEX < 0x030C0000 || CYTHON_COMPILING_IN_LIMITED_API
      Py_VISIT(m->func_dict);
  #else
      {
          int e =
  #if PY_VERSION_HEX < 0x030d0000
              _PyObject_VisitManagedDict
  #else
              PyObject_VisitManagedDict
  #endif
                  ((PyObject*)m, visit, arg);
          if (e != 0) return e;
      }
  #endif
      __Pyx_VISIT_CONST(m->func_name);
      __Pyx_VISIT_CONST(m->func_qualname);
      Py_VISIT(m->func_doc);
      Py_VISIT(m->func_globals);
      __Pyx_VISIT_CONST(m->func_code);
  #if !CYTHON_COMPILING_IN_LIMITED_API
      Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
  #endif
      Py_VISIT(m->defaults_tuple);
      Py_VISIT(m->defaults_kwdict);
      Py_VISIT(m->func_is_coroutine);
      Py_VISIT(m->defaults);
      return 0;
  }
  static PyObject*
  __Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
  {
      PyObject *repr;
      __Pyx_BEGIN_CRITICAL_SECTION(op);
      repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                  op->func_qualname, (void *)op);
      __Pyx_END_CRITICAL_SECTION();
      return repr;
  }
  static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
  #if CYTHON_COMPILING_IN_LIMITED_API
      PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
      PyCFunction meth;
      int flags;
      meth = PyCFunction_GetFunction(f);
      if (unlikely(!meth)) return NULL;
      flags = PyCFunction_GetFlags(f);
      if (unlikely(flags < 0)) return NULL;
  #else
      PyCFunctionObject* f = (PyCFunctionObject*)func;
      PyCFunction meth = f->m_ml->ml_meth;
      int flags = f->m_ml->ml_flags;
  #endif
      Py_ssize_t size;
      switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
      case METH_VARARGS:
          if (likely(kw == NULL || PyDict_Size(kw) == 0))
              return (*meth)(self, arg);
          break;
      case METH_VARARGS | METH_KEYWORDS:
          return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
      case METH_NOARGS:
          if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
  #if CYTHON_ASSUME_SAFE_SIZE
              size = PyTuple_GET_SIZE(arg);
  #else
              size = PyTuple_Size(arg);
              if (unlikely(size < 0)) return NULL;
  #endif
              if (likely(size == 0))
                  return (*meth)(self, NULL);
              __Pyx_CyFunction_raise_argument_count_error(
                  (__pyx_CyFunctionObject*)func,
                  "takes no arguments", size);
              return NULL;
          }
          break;
      case METH_O:
          if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
  #if CYTHON_ASSUME_SAFE_SIZE
              size = PyTuple_GET_SIZE(arg);
  #else
              size = PyTuple_Size(arg);
              if (unlikely(size < 0)) return NULL;
  #endif
              if (likely(size == 1)) {
                  PyObject *result, *arg0;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  arg0 = PyTuple_GET_ITEM(arg, 0);
                  #else
                  arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                  #endif
                  result = (*meth)(self, arg0);
                  #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                  Py_DECREF(arg0);
                  #endif
                  return result;
              }
              __Pyx_CyFunction_raise_argument_count_error(
                  (__pyx_CyFunctionObject*)func,
                  "takes exactly one argument", size);
              return NULL;
          }
          break;
      default:
          PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
          return NULL;
      }
      __Pyx_CyFunction_raise_type_error(
          (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
      return NULL;
  }
  static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
      PyObject *self, *result;
  #if CYTHON_COMPILING_IN_LIMITED_API
      self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
      if (unlikely(!self) && PyErr_Occurred()) return NULL;
  #else
      self = ((PyCFunctionObject*)func)->m_self;
  #endif
      result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
      return result;
  }
  static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
      PyObject *result;
      __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
  #if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
       __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
      if (vc) {
  #if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
          return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
  #else
          (void) &__Pyx_PyVectorcall_FastCallDict;
          return PyVectorcall_Call(func, args, kw);
  #endif
      }
  #endif
      if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
          Py_ssize_t argc;
          PyObject *new_args;
          PyObject *self;
  #if CYTHON_ASSUME_SAFE_SIZE
          argc = PyTuple_GET_SIZE(args);
  #else
          argc = PyTuple_Size(args);
          if (unlikely(argc < 0)) return NULL;
  #endif
          new_args = PyTuple_GetSlice(args, 1, argc);
          if (unlikely(!new_args))
              return NULL;
          self = PyTuple_GetItem(args, 0);
          if (unlikely(!self)) {
              Py_DECREF(new_args);
              PyErr_Format(PyExc_TypeError,
                           "unbound method %.200S() needs an argument",
                           cyfunc->func_qualname);
              return NULL;
          }
          result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
          Py_DECREF(new_args);
      } else {
          result = __Pyx_CyFunction_Call(func, args, kw);
      }
      return result;
  }
  #if CYTHON_METH_FASTCALL && CYTHON_VECTORCALL
  static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
  {
      int ret = 0;
      if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
          if (unlikely(nargs < 1)) {
              __Pyx_CyFunction_raise_type_error(
                  cyfunc, "needs an argument");
              return -1;
          }
          ret = 1;
      }
      if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
          __Pyx_CyFunction_raise_type_error(
              cyfunc, "takes no keyword arguments");
          return -1;
      }
      return ret;
  }
  static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
  {
      __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
      Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
      PyObject *self;
  #if CYTHON_COMPILING_IN_LIMITED_API
      PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
      if (unlikely(!meth)) return NULL;
  #else
      PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
  #endif
      switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
      case 1:
          self = args[0];
          args += 1;
          nargs -= 1;
          break;
      case 0:
  #if CYTHON_COMPILING_IN_LIMITED_API
          self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
          if (unlikely(!self) && PyErr_Occurred()) return NULL;
  #else
          self = ((PyCFunctionObject*)cyfunc)->m_self;
  #endif
          break;
      default:
          return NULL;
      }
      if (unlikely(nargs != 0)) {
          __Pyx_CyFunction_raise_argument_count_error(
              cyfunc, "takes no arguments", nargs);
          return NULL;
      }
      return meth(self, NULL);
  }
  static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
  {
      __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
      Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
      PyObject *self;
  #if CYTHON_COMPILING_IN_LIMITED_API
      PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
      if (unlikely(!meth)) return NULL;
  #else
      PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
  #endif
      switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
      case 1:
          self = args[0];
          args += 1;
          nargs -= 1;
          break;
      case 0:
  #if CYTHON_COMPILING_IN_LIMITED_API
          self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
          if (unlikely(!self) && PyErr_Occurred()) return NULL;
  #else
          self = ((PyCFunctionObject*)cyfunc)->m_self;
  #endif
          break;
      default:
          return NULL;
      }
      if (unlikely(nargs != 1)) {
          __Pyx_CyFunction_raise_argument_count_error(
              cyfunc, "takes exactly one argument", nargs);
          return NULL;
      }
      return meth(self, args[0]);
  }
  static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
  {
      __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
      Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
      PyObject *self;
  #if CYTHON_COMPILING_IN_LIMITED_API
      PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
      if (unlikely(!meth)) return NULL;
  #else
      PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
  #endif
      switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
      case 1:
          self = args[0];
          args += 1;
          nargs -= 1;
          break;
      case 0:
  #if CYTHON_COMPILING_IN_LIMITED_API
          self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
          if (unlikely(!self) && PyErr_Occurred()) return NULL;
  #else
          self = ((PyCFunctionObject*)cyfunc)->m_self;
  #endif
          break;
      default:
          return NULL;
      }
      return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
  }
  static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
  {
      __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
      PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
      Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
      PyObject *self;
  #if CYTHON_COMPILING_IN_LIMITED_API
      PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
      if (unlikely(!meth)) return NULL;
  #else
      PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
  #endif
      switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
      case 1:
          self = args[0];
          args += 1;
          nargs -= 1;
          break;
      case 0:
  #if CYTHON_COMPILING_IN_LIMITED_API
          self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
          if (unlikely(!self) && PyErr_Occurred()) return NULL;
  #else
          self = ((PyCFunctionObject*)cyfunc)->m_self;
  #endif
          break;
      default:
          return NULL;
      }
      #if PY_VERSION_HEX < 0x030e00A6
      size_t nargs_value = (size_t) nargs;
      #else
      Py_ssize_t nargs_value = nargs;
      #endif
      return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, nargs_value, kwnames);
  }
  #endif
  static PyType_Slot __pyx_CyFunctionType_slots[] = {
      {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
      {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
      {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
      {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
      {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
      {Py_tp_methods, (void *)__pyx_CyFunction_methods},
      {Py_tp_members, (void *)__pyx_CyFunction_members},
      {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
      {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
      {0, 0},
  };
  static PyType_Spec __pyx_CyFunctionType_spec = {
      __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
      sizeof(__pyx_CyFunctionObject),
      0,
  #ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
      Py_TPFLAGS_METHOD_DESCRIPTOR |
  #endif
  #if CYTHON_METH_FASTCALL
  #if defined(Py_TPFLAGS_HAVE_VECTORCALL)
      Py_TPFLAGS_HAVE_VECTORCALL |
  #elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
      _Py_TPFLAGS_HAVE_VECTORCALL |
  #endif
  #endif // CYTHON_METH_FASTCALL
  #if PY_VERSION_HEX >= 0x030C0000 && !CYTHON_COMPILING_IN_LIMITED_API
      Py_TPFLAGS_MANAGED_DICT |
  #endif
      Py_TPFLAGS_IMMUTABLETYPE | Py_TPFLAGS_DISALLOW_INSTANTIATION |
      Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
      __pyx_CyFunctionType_slots
  };
  static int __pyx_CyFunction_init(PyObject *module) {
      __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
      mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
          mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
      if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
          return -1;
      }
      return 0;
  }
  static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
      __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
      m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
      if (unlikely(!m->defaults))
          return NULL;
      return m->defaults;
  }
  static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
      __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
      m->defaults_tuple = tuple;
      Py_INCREF(tuple);
  }
  static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
      __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
      m->defaults_kwdict = dict;
      Py_INCREF(dict);
  }
  static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
      __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
      m->func_annotations = dict;
      Py_INCREF(dict);
  }
  
/* CythonFunction */
  static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                        PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
      PyObject *op = __Pyx_CyFunction_Init(
          PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
          ml, flags, qualname, closure, module, globals, code
      );
      if (likely(op)) {
          PyObject_GC_Track(op);
      }
      return op;
  }
  
/* pybytes_as_double (used by pyunicode_as_double) */
  static double __Pyx_SlowPyString_AsDouble(PyObject *obj) {
      PyObject *float_value = PyFloat_FromString(obj);
      if (likely(float_value)) {
          double value = __Pyx_PyFloat_AS_DOUBLE(float_value);
          Py_DECREF(float_value);
          return value;
      }
      return (double)-1;
  }
  static const char* __Pyx__PyBytes_AsDouble_Copy(const char* start, char* buffer, Py_ssize_t length) {
      int last_was_punctuation = 1;
      int parse_error_found = 0;
      Py_ssize_t i;
      for (i=0; i < length; i++) {
          char chr = start[i];
          int is_punctuation = (chr == '_') | (chr == '.') | (chr == 'e') | (chr == 'E');
          *buffer = chr;
          buffer += (chr != '_');
          parse_error_found |= last_was_punctuation & is_punctuation;
          last_was_punctuation = is_punctuation;
      }
      parse_error_found |= last_was_punctuation;
      *buffer = '\0';
      return unlikely(parse_error_found) ? NULL : buffer;
  }
  static double __Pyx__PyBytes_AsDouble_inf_nan(const char* start, Py_ssize_t length) {
      int matches = 1;
      char sign = start[0];
      int is_signed = (sign == '+') | (sign == '-');
      start += is_signed;
      length -= is_signed;
      switch (start[0]) {
          #ifdef Py_NAN
          case 'n':
          case 'N':
              if (unlikely(length != 3)) goto parse_failure;
              matches &= (start[1] == 'a' || start[1] == 'A');
              matches &= (start[2] == 'n' || start[2] == 'N');
              if (unlikely(!matches)) goto parse_failure;
              return (sign == '-') ? -Py_NAN : Py_NAN;
          #endif
          case 'i':
          case 'I':
              if (unlikely(length < 3)) goto parse_failure;
              matches &= (start[1] == 'n' || start[1] == 'N');
              matches &= (start[2] == 'f' || start[2] == 'F');
              if (likely(length == 3 && matches))
                  return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
              if (unlikely(length != 8)) goto parse_failure;
              matches &= (start[3] == 'i' || start[3] == 'I');
              matches &= (start[4] == 'n' || start[4] == 'N');
              matches &= (start[5] == 'i' || start[5] == 'I');
              matches &= (start[6] == 't' || start[6] == 'T');
              matches &= (start[7] == 'y' || start[7] == 'Y');
              if (unlikely(!matches)) goto parse_failure;
              return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
          case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
              break;
          default:
              goto parse_failure;
      }
      return 0.0;
  parse_failure:
      return -1.0;
  }
  static CYTHON_INLINE int __Pyx__PyBytes_AsDouble_IsSpace(char ch) {
      return (ch == 0x20) | !((ch < 0x9) | (ch > 0xd));
  }
  CYTHON_UNUSED static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length) {
      double value;
      Py_ssize_t i, digits;
      const char *last = start + length;
      char *end;
      while (__Pyx__PyBytes_AsDouble_IsSpace(*start))
          start++;
      while (start < last - 1 && __Pyx__PyBytes_AsDouble_IsSpace(last[-1]))
          last--;
      length = last - start;
      if (unlikely(length <= 0)) goto fallback;
      value = __Pyx__PyBytes_AsDouble_inf_nan(start, length);
      if (unlikely(value == -1.0)) goto fallback;
      if (value != 0.0) return value;
      digits = 0;
      for (i=0; i < length; digits += start[i++] != '_');
      if (likely(digits == length)) {
          value = PyOS_string_to_double(start, &end, NULL);
      } else if (digits < 40) {
          char number[40];
          last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
          if (unlikely(!last)) goto fallback;
          value = PyOS_string_to_double(number, &end, NULL);
      } else {
          char *number = (char*) PyMem_Malloc((digits + 1) * sizeof(char));
          if (unlikely(!number)) goto fallback;
          last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
          if (unlikely(!last)) {
              PyMem_Free(number);
              goto fallback;
          }
          value = PyOS_string_to_double(number, &end, NULL);
          PyMem_Free(number);
      }
      if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
          return value;
      }
  fallback:
      return __Pyx_SlowPyString_AsDouble(obj);
  }
  
/* CLineInTraceback (used by AddTraceback) */
  #if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
  #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030A0000
  #define __Pyx_PyProbablyModule_GetDict(o) __Pyx_XNewRef(PyModule_GetDict(o))
  #elif !CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  #define __Pyx_PyProbablyModule_GetDict(o) PyObject_GenericGetDict(o, NULL);
  #else
  PyObject* __Pyx_PyProbablyModule_GetDict(PyObject *o) {
      PyObject **dict_ptr = _PyObject_GetDictPtr(o);
      return dict_ptr ? __Pyx_XNewRef(*dict_ptr) : NULL;
  }
  #endif
  static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
      PyObject *use_cline = NULL;
      PyObject *ptype, *pvalue, *ptraceback;
      PyObject *cython_runtime_dict;
      CYTHON_MAYBE_UNUSED_VAR(tstate);
      if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
          return c_line;
      }
      __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
      cython_runtime_dict = __Pyx_PyProbablyModule_GetDict(__pyx_mstate_global->__pyx_cython_runtime);
      if (likely(cython_runtime_dict)) {
          __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
              use_cline, cython_runtime_dict,
              __Pyx_PyDict_SetDefault(cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False))
      }
      if (use_cline == NULL || use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
          c_line = 0;
      }
      Py_XDECREF(use_cline);
      Py_XDECREF(cython_runtime_dict);
      __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
      return c_line;
  }
  #endif
  
/* CodeObjectCache (used by AddTraceback) */
  static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
      int start = 0, mid = 0, end = count - 1;
      if (end >= 0 && code_line > entries[end].code_line) {
          return count;
      }
      while (start < end) {
          mid = start + (end - start) / 2;
          if (code_line < entries[mid].code_line) {
              end = mid;
          } else if (code_line > entries[mid].code_line) {
               start = mid + 1;
          } else {
              return mid;
          }
      }
      if (code_line <= entries[mid].code_line) {
          return mid;
      } else {
          return mid + 1;
      }
  }
  static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
      __Pyx_CachedCodeObjectType* code_object;
      int pos;
      if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
          return NULL;
      }
      pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
      if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
          return NULL;
      }
      code_object = code_cache->entries[pos].code_object;
      Py_INCREF(code_object);
      return code_object;
  }
  static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
      (void)__pyx__find_code_object;
      return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
  #else
      struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
      if (old_count < 0) {
          __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
          return NULL;
      }
  #endif
      __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
  #endif
      return result;
  #endif
  }
  static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
  {
      int pos, i;
      __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
      if (unlikely(!code_line)) {
          return;
      }
      if (unlikely(!entries)) {
          entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
          if (likely(entries)) {
              code_cache->entries = entries;
              code_cache->max_count = 64;
              code_cache->count = 1;
              entries[0].code_line = code_line;
              entries[0].code_object = code_object;
              Py_INCREF(code_object);
          }
          return;
      }
      pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
      if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
          __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
          entries[pos].code_object = code_object;
          Py_INCREF(code_object);
          Py_DECREF(tmp);
          return;
      }
      if (code_cache->count == code_cache->max_count) {
          int new_max = code_cache->max_count + 64;
          entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
              code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
          if (unlikely(!entries)) {
              return;
          }
          code_cache->entries = entries;
          code_cache->max_count = new_max;
      }
      for (i=code_cache->count; i>pos; i--) {
          entries[i] = entries[i-1];
      }
      entries[pos].code_line = code_line;
      entries[pos].code_object = code_object;
      code_cache->count++;
      Py_INCREF(code_object);
  }
  static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
      (void)__pyx__insert_code_object;
      return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
  #else
      struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      __pyx_nonatomic_int_type expected = 0;
      if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
          return;
      }
  #endif
      __pyx__insert_code_object(code_cache, code_line, code_object);
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
      __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
  #endif
  #endif
  }
  
/* AddTraceback */
  #include "compile.h"
  #include "frameobject.h"
  #include "traceback.h"
  #if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
    #ifndef Py_BUILD_CORE
      #define Py_BUILD_CORE 1
    #endif
    #include "internal/pycore_frame.h"
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                         PyObject *firstlineno, PyObject *name) {
      PyObject *replace = NULL;
      if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
      if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
      replace = PyObject_GetAttrString(code, "replace");
      if (likely(replace)) {
          PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
          Py_DECREF(replace);
          return result;
      }
      PyErr_Clear();
      return NULL;
  }
  static void __Pyx_AddTraceback(const char *funcname, int c_line,
                                 int py_line, const char *filename) {
      PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
      PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
      PyObject *exc_type, *exc_value, *exc_traceback;
      int success = 0;
      if (c_line) {
          c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
      }
      PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
      code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
      if (!code_object) {
          code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
          if (unlikely(!code_object)) goto bad;
          py_py_line = PyLong_FromLong(py_line);
          if (unlikely(!py_py_line)) goto bad;
          if (c_line) {
              py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
          } else {
              py_funcname = PyUnicode_FromString(funcname);
          }
          if (unlikely(!py_funcname)) goto bad;
          dict = PyDict_New();
          if (unlikely(!dict)) goto bad;
          {
              PyObject *old_code_object = code_object;
              code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
              Py_DECREF(old_code_object);
          }
          if (unlikely(!code_object)) goto bad;
          __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
      } else {
          dict = PyDict_New();
      }
      getframe = PySys_GetObject("_getframe");
      if (unlikely(!getframe)) goto bad;
      if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
      frame = PyEval_EvalCode(code_object, dict, dict);
      if (unlikely(!frame) || frame == Py_None) goto bad;
      success = 1;
    bad:
      PyErr_Restore(exc_type, exc_value, exc_traceback);
      Py_XDECREF(code_object);
      Py_XDECREF(py_py_line);
      Py_XDECREF(py_funcname);
      Py_XDECREF(dict);
      Py_XDECREF(replace);
      if (success) {
          PyTraceBack_Here(
              (struct _frame*)frame);
      }
      Py_XDECREF(frame);
  }
  #else
  static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
              const char *funcname, int c_line,
              int py_line, const char *filename) {
      PyCodeObject *py_code = NULL;
      PyObject *py_funcname = NULL;
      if (c_line) {
          py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
          if (!py_funcname) goto bad;
          funcname = PyUnicode_AsUTF8(py_funcname);
          if (!funcname) goto bad;
      }
      py_code = PyCode_NewEmpty(filename, funcname, py_line);
      Py_XDECREF(py_funcname);
      return py_code;
  bad:
      Py_XDECREF(py_funcname);
      return NULL;
  }
  static void __Pyx_AddTraceback(const char *funcname, int c_line,
                                 int py_line, const char *filename) {
      PyCodeObject *py_code = 0;
      PyFrameObject *py_frame = 0;
      PyThreadState *tstate = __Pyx_PyThreadState_Current;
      PyObject *ptype, *pvalue, *ptraceback;
      if (c_line) {
          c_line = __Pyx_CLineForTraceback(tstate, c_line);
      }
      py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
      if (!py_code) {
          __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
          py_code = __Pyx_CreateCodeObjectForTraceback(
              funcname, c_line, py_line, filename);
          if (!py_code) {
              /* If the code object creation fails, then we should clear the
                 fetched exception references and propagate the new exception */
              Py_XDECREF(ptype);
              Py_XDECREF(pvalue);
              Py_XDECREF(ptraceback);
              goto bad;
          }
          __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
          __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
      }
      py_frame = PyFrame_New(
          tstate,            /*PyThreadState *tstate,*/
          py_code,           /*PyCodeObject *code,*/
          __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
          0                  /*PyObject *locals*/
      );
      if (!py_frame) goto bad;
      __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
      PyTraceBack_Here(py_frame);
  bad:
      Py_XDECREF(py_code);
      Py_XDECREF(py_frame);
  }
  #endif
  
/* Declarations */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #ifdef __cplusplus
      static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
        return ::std::complex< double >(x, y);
      }
    #else
      static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
        return x + y*(__pyx_t_double_complex)_Complex_I;
      }
    #endif
  #else
      static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
        __pyx_t_double_complex z;
        z.real = x;
        z.imag = y;
        return z;
      }
  #endif
  
/* Arithmetic */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #else
      static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
         return (a.real == b.real) && (a.imag == b.imag);
      }
      static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
          __pyx_t_double_complex z;
          z.real = a.real + b.real;
          z.imag = a.imag + b.imag;
          return z;
      }
      static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
          __pyx_t_double_complex z;
          z.real = a.real - b.real;
          z.imag = a.imag - b.imag;
          return z;
      }
      static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
          __pyx_t_double_complex z;
          z.real = a.real * b.real - a.imag * b.imag;
          z.imag = a.real * b.imag + a.imag * b.real;
          return z;
      }
      #if 1
      static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
          if (b.imag == 0) {
              return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
          } else if (fabs(b.real) >= fabs(b.imag)) {
              if (b.real == 0 && b.imag == 0) {
                  return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
              } else {
                  double r = b.imag / b.real;
                  double s = (double)(1.0) / (b.real + b.imag * r);
                  return __pyx_t_double_complex_from_parts(
                      (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
              }
          } else {
              double r = b.real / b.imag;
              double s = (double)(1.0) / (b.imag + b.real * r);
              return __pyx_t_double_complex_from_parts(
                  (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
          }
      }
      #else
      static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
          if (b.imag == 0) {
              return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
          } else {
              double denom = b.real * b.real + b.imag * b.imag;
              return __pyx_t_double_complex_from_parts(
                  (a.real * b.real + a.imag * b.imag) / denom,
                  (a.imag * b.real - a.real * b.imag) / denom);
          }
      }
      #endif
      static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
          __pyx_t_double_complex z;
          z.real = -a.real;
          z.imag = -a.imag;
          return z;
      }
      static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
         return (a.real == 0) && (a.imag == 0);
      }
      static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
          __pyx_t_double_complex z;
          z.real =  a.real;
          z.imag = -a.imag;
          return z;
      }
      #if 1
          static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
            #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
              return sqrt(z.real*z.real + z.imag*z.imag);
            #else
              return hypot(z.real, z.imag);
            #endif
          }
          static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
              __pyx_t_double_complex z;
              double r, lnr, theta, z_r, z_theta;
              if (b.imag == 0 && b.real == (int)b.real) {
                  if (b.real < 0) {
                      double denom = a.real * a.real + a.imag * a.imag;
                      a.real = a.real / denom;
                      a.imag = -a.imag / denom;
                      b.real = -b.real;
                  }
                  switch ((int)b.real) {
                      case 0:
                          z.real = 1;
                          z.imag = 0;
                          return z;
                      case 1:
                          return a;
                      case 2:
                          return __Pyx_c_prod_double(a, a);
                      case 3:
                          z = __Pyx_c_prod_double(a, a);
                          return __Pyx_c_prod_double(z, a);
                      case 4:
                          z = __Pyx_c_prod_double(a, a);
                          return __Pyx_c_prod_double(z, z);
                  }
              }
              if (a.imag == 0) {
                  if (a.real == 0) {
                      return a;
                  } else if ((b.imag == 0) && (a.real >= 0)) {
                      z.real = pow(a.real, b.real);
                      z.imag = 0;
                      return z;
                  } else if (a.real > 0) {
                      r = a.real;
                      theta = 0;
                  } else {
                      r = -a.real;
                      theta = atan2(0.0, -1.0);
                  }
              } else {
                  r = __Pyx_c_abs_double(a);
                  theta = atan2(a.imag, a.real);
              }
              lnr = log(r);
              z_r = exp(lnr * b.real - theta * b.imag);
              z_theta = theta * b.real + lnr * b.imag;
              z.real = z_r * cos(z_theta);
              z.imag = z_r * sin(z_theta);
              return z;
          }
      #endif
  #endif
  
/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
      __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
  #define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
      __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
  #define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
      {\
          func_type value = func_value;\
          if (sizeof(target_type) < sizeof(func_type)) {\
              if (unlikely(value != (func_type) (target_type) value)) {\
                  func_type zero = 0;\
                  if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                      return (target_type) -1;\
                  if (is_unsigned && unlikely(value < zero))\
                      goto raise_neg_overflow;\
                  else\
                      goto raise_overflow;\
              }\
          }\
          return (target_type) value;\
      }
  
/* FromPy */
  static __pyx_t_double_complex __Pyx_PyComplex_As___pyx_t_double_complex(PyObject* o) {
  #if CYTHON_COMPILING_IN_LIMITED_API
      double real=-1.0, imag=-1.0;
      real = PyComplex_RealAsDouble(o);
      if (unlikely(real == -1.0 && PyErr_Occurred())) goto end;
      imag = PyComplex_ImagAsDouble(o);
    end:
      return __pyx_t_double_complex_from_parts(
          (double)real, (double)imag
      );
  #else
      Py_complex cval;
  #if !CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_GRAAL
      if (PyComplex_CheckExact(o))
          cval = ((PyComplexObject *)o)->cval;
      else
  #endif
          cval = PyComplex_AsCComplex(o);
      return __pyx_t_double_complex_from_parts(
                 (double)cval.real,
                 (double)cval.imag);
  #endif
  }
  
/* Declarations */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #ifdef __cplusplus
      static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
        return ::std::complex< float >(x, y);
      }
    #else
      static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
        return x + y*(__pyx_t_float_complex)_Complex_I;
      }
    #endif
  #else
      static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
        __pyx_t_float_complex z;
        z.real = x;
        z.imag = y;
        return z;
      }
  #endif
  
/* Arithmetic */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #else
      static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
         return (a.real == b.real) && (a.imag == b.imag);
      }
      static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
          __pyx_t_float_complex z;
          z.real = a.real + b.real;
          z.imag = a.imag + b.imag;
          return z;
      }
      static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
          __pyx_t_float_complex z;
          z.real = a.real - b.real;
          z.imag = a.imag - b.imag;
          return z;
      }
      static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
          __pyx_t_float_complex z;
          z.real = a.real * b.real - a.imag * b.imag;
          z.imag = a.real * b.imag + a.imag * b.real;
          return z;
      }
      #if 1
      static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
          if (b.imag == 0) {
              return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
          } else if (fabsf(b.real) >= fabsf(b.imag)) {
              if (b.real == 0 && b.imag == 0) {
                  return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
              } else {
                  float r = b.imag / b.real;
                  float s = (float)(1.0) / (b.real + b.imag * r);
                  return __pyx_t_float_complex_from_parts(
                      (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
              }
          } else {
              float r = b.real / b.imag;
              float s = (float)(1.0) / (b.imag + b.real * r);
              return __pyx_t_float_complex_from_parts(
                  (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
          }
      }
      #else
      static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
          if (b.imag == 0) {
              return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
          } else {
              float denom = b.real * b.real + b.imag * b.imag;
              return __pyx_t_float_complex_from_parts(
                  (a.real * b.real + a.imag * b.imag) / denom,
                  (a.imag * b.real - a.real * b.imag) / denom);
          }
      }
      #endif
      static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
          __pyx_t_float_complex z;
          z.real = -a.real;
          z.imag = -a.imag;
          return z;
      }
      static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
         return (a.real == 0) && (a.imag == 0);
      }
      static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
          __pyx_t_float_complex z;
          z.real =  a.real;
          z.imag = -a.imag;
          return z;
      }
      #if 1
          static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
            #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
              return sqrtf(z.real*z.real + z.imag*z.imag);
            #else
              return hypotf(z.real, z.imag);
            #endif
          }
          static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
              __pyx_t_float_complex z;
              float r, lnr, theta, z_r, z_theta;
              if (b.imag == 0 && b.real == (int)b.real) {
                  if (b.real < 0) {
                      float denom = a.real * a.real + a.imag * a.imag;
                      a.real = a.real / denom;
                      a.imag = -a.imag / denom;
                      b.real = -b.real;
                  }
                  switch ((int)b.real) {
                      case 0:
                          z.real = 1;
                          z.imag = 0;
                          return z;
                      case 1:
                          return a;
                      case 2:
                          return __Pyx_c_prod_float(a, a);
                      case 3:
                          z = __Pyx_c_prod_float(a, a);
                          return __Pyx_c_prod_float(z, a);
                      case 4:
                          z = __Pyx_c_prod_float(a, a);
                          return __Pyx_c_prod_float(z, z);
                  }
              }
              if (a.imag == 0) {
                  if (a.real == 0) {
                      return a;
                  } else if ((b.imag == 0) && (a.real >= 0)) {
                      z.real = powf(a.real, b.real);
                      z.imag = 0;
                      return z;
                  } else if (a.real > 0) {
                      r = a.real;
                      theta = 0;
                  } else {
                      r = -a.real;
                      theta = atan2f(0.0, -1.0);
                  }
              } else {
                  r = __Pyx_c_abs_float(a);
                  theta = atan2f(a.imag, a.real);
              }
              lnr = logf(r);
              z_r = expf(lnr * b.real - theta * b.imag);
              z_theta = theta * b.real + lnr * b.imag;
              z.real = z_r * cosf(z_theta);
              z.imag = z_r * sinf(z_theta);
              return z;
          }
      #endif
  #endif
  
/* Declarations */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #ifdef __cplusplus
      static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
        return ::std::complex< long double >(x, y);
      }
    #else
      static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
        return x + y*(__pyx_t_long_double_complex)_Complex_I;
      }
    #endif
  #else
      static CYTHON_INLINE __pyx_t_long_double_complex __pyx_t_long_double_complex_from_parts(long double x, long double y) {
        __pyx_t_long_double_complex z;
        z.real = x;
        z.imag = y;
        return z;
      }
  #endif
  
/* Arithmetic */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #else
      static CYTHON_INLINE int __Pyx_c_eq_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
         return (a.real == b.real) && (a.imag == b.imag);
      }
      static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_sum_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
          __pyx_t_long_double_complex z;
          z.real = a.real + b.real;
          z.imag = a.imag + b.imag;
          return z;
      }
      static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_diff_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
          __pyx_t_long_double_complex z;
          z.real = a.real - b.real;
          z.imag = a.imag - b.imag;
          return z;
      }
      static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_prod_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
          __pyx_t_long_double_complex z;
          z.real = a.real * b.real - a.imag * b.imag;
          z.imag = a.real * b.imag + a.imag * b.real;
          return z;
      }
      #if 1
      static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
          if (b.imag == 0) {
              return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.real);
          } else if (fabsl(b.real) >= fabsl(b.imag)) {
              if (b.real == 0 && b.imag == 0) {
                  return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
              } else {
                  long double r = b.imag / b.real;
                  long double s = (long double)(1.0) / (b.real + b.imag * r);
                  return __pyx_t_long_double_complex_from_parts(
                      (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
              }
          } else {
              long double r = b.real / b.imag;
              long double s = (long double)(1.0) / (b.imag + b.real * r);
              return __pyx_t_long_double_complex_from_parts(
                  (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
          }
      }
      #else
      static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_quot_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
          if (b.imag == 0) {
              return __pyx_t_long_double_complex_from_parts(a.real / b.real, a.imag / b.real);
          } else {
              long double denom = b.real * b.real + b.imag * b.imag;
              return __pyx_t_long_double_complex_from_parts(
                  (a.real * b.real + a.imag * b.imag) / denom,
                  (a.imag * b.real - a.real * b.imag) / denom);
          }
      }
      #endif
      static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_neg_long__double(__pyx_t_long_double_complex a) {
          __pyx_t_long_double_complex z;
          z.real = -a.real;
          z.imag = -a.imag;
          return z;
      }
      static CYTHON_INLINE int __Pyx_c_is_zero_long__double(__pyx_t_long_double_complex a) {
         return (a.real == 0) && (a.imag == 0);
      }
      static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_conj_long__double(__pyx_t_long_double_complex a) {
          __pyx_t_long_double_complex z;
          z.real =  a.real;
          z.imag = -a.imag;
          return z;
      }
      #if 1
          static CYTHON_INLINE long double __Pyx_c_abs_long__double(__pyx_t_long_double_complex z) {
            #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
              return sqrtl(z.real*z.real + z.imag*z.imag);
            #else
              return hypotl(z.real, z.imag);
            #endif
          }
          static CYTHON_INLINE __pyx_t_long_double_complex __Pyx_c_pow_long__double(__pyx_t_long_double_complex a, __pyx_t_long_double_complex b) {
              __pyx_t_long_double_complex z;
              long double r, lnr, theta, z_r, z_theta;
              if (b.imag == 0 && b.real == (int)b.real) {
                  if (b.real < 0) {
                      long double denom = a.real * a.real + a.imag * a.imag;
                      a.real = a.real / denom;
                      a.imag = -a.imag / denom;
                      b.real = -b.real;
                  }
                  switch ((int)b.real) {
                      case 0:
                          z.real = 1;
                          z.imag = 0;
                          return z;
                      case 1:
                          return a;
                      case 2:
                          return __Pyx_c_prod_long__double(a, a);
                      case 3:
                          z = __Pyx_c_prod_long__double(a, a);
                          return __Pyx_c_prod_long__double(z, a);
                      case 4:
                          z = __Pyx_c_prod_long__double(a, a);
                          return __Pyx_c_prod_long__double(z, z);
                  }
              }
              if (a.imag == 0) {
                  if (a.real == 0) {
                      return a;
                  } else if ((b.imag == 0) && (a.real >= 0)) {
                      z.real = powl(a.real, b.real);
                      z.imag = 0;
                      return z;
                  } else if (a.real > 0) {
                      r = a.real;
                      theta = 0;
                  } else {
                      r = -a.real;
                      theta = atan2l(0.0, -1.0);
                  }
              } else {
                  r = __Pyx_c_abs_long__double(a);
                  theta = atan2l(a.imag, a.real);
              }
              lnr = logl(r);
              z_r = expl(lnr * b.real - theta * b.imag);
              z_theta = theta * b.real + lnr * b.imag;
              z.real = z_r * cosl(z_theta);
              z.imag = z_r * sinl(z_theta);
              return z;
          }
      #endif
  #endif
  
/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum__TMM_3a__3a_TMMParam(enum TMM::TMMParam value) {
  #ifdef __Pyx_HAS_GCC_DIAGNOSTIC
  #pragma GCC diagnostic push
  #pragma GCC diagnostic ignored "-Wconversion"
  #endif
      const enum TMM::TMMParam neg_one = (enum TMM::TMMParam) -1, const_zero = (enum TMM::TMMParam) 0;
  #ifdef __Pyx_HAS_GCC_DIAGNOSTIC
  #pragma GCC diagnostic pop
  #endif
      const int is_unsigned = neg_one > const_zero;
      if (is_unsigned) {
          if (sizeof(enum TMM::TMMParam) < sizeof(long)) {
              return PyLong_FromLong((long) value);
          } else if (sizeof(enum TMM::TMMParam) <= sizeof(unsigned long)) {
              return PyLong_FromUnsignedLong((unsigned long) value);
  #if !CYTHON_COMPILING_IN_PYPY
          } else if (sizeof(enum TMM::TMMParam) <= sizeof(unsigned PY_LONG_LONG)) {
              return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
  #endif
          }
      } else {
          if (sizeof(enum TMM::TMMParam) <= sizeof(long)) {
              return PyLong_FromLong((long) value);
          } else if (sizeof(enum TMM::TMMParam) <= sizeof(PY_LONG_LONG)) {
              return PyLong_FromLongLong((PY_LONG_LONG) value);
          }
      }
      {
          unsigned char *bytes = (unsigned char *)&value;
  #if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
          if (is_unsigned) {
              return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
          } else {
              return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
          }
  #elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
          int one = 1; int little = (int)*(unsigned char *)&one;
          return _PyLong_FromByteArray(bytes, sizeof(enum TMM::TMMParam),
                                       little, !is_unsigned);
  #else
          int one = 1; int little = (int)*(unsigned char *)&one;
          PyObject *from_bytes, *result = NULL, *kwds = NULL;
          PyObject *py_bytes = NULL, *order_str = NULL;
          from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
          if (!from_bytes) return NULL;
          py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum TMM::TMMParam));
          if (!py_bytes) goto limited_bad;
          order_str = PyUnicode_FromString(little ? "little" : "big");
          if (!order_str) goto limited_bad;
          {
              PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
              if (!is_unsigned) {
                  kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                  if (!kwds) goto limited_bad;
                  if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
              }
              result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
          }
          limited_bad:
          Py_XDECREF(kwds);
          Py_XDECREF(order_str);
          Py_XDECREF(py_bytes);
          Py_XDECREF(from_bytes);
          return result;
  #endif
      }
  }
  
/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
  #ifdef __Pyx_HAS_GCC_DIAGNOSTIC
  #pragma GCC diagnostic push
  #pragma GCC diagnostic ignored "-Wconversion"
  #endif
      const int neg_one = (int) -1, const_zero = (int) 0;
  #ifdef __Pyx_HAS_GCC_DIAGNOSTIC
  #pragma GCC diagnostic pop
  #endif
      const int is_unsigned = neg_one > const_zero;
      if (unlikely(!PyLong_Check(x))) {
          int val;
          PyObject *tmp = __Pyx_PyNumber_Long(x);
          if (!tmp) return (int) -1;
          val = __Pyx_PyLong_As_int(tmp);
          Py_DECREF(tmp);
          return val;
      }
      if (is_unsigned) {
  #if CYTHON_USE_PYLONG_INTERNALS
          if (unlikely(__Pyx_PyLong_IsNeg(x))) {
              goto raise_neg_overflow;
          } else if (__Pyx_PyLong_IsCompact(x)) {
              __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
          } else {
              const digit* digits = __Pyx_PyLong_Digits(x);
              assert(__Pyx_PyLong_DigitCount(x) > 1);
              switch (__Pyx_PyLong_DigitCount(x)) {
                  case 2:
                      if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                              return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                          }
                      }
                      break;
                  case 3:
                      if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                              return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                          }
                      }
                      break;
                  case 4:
                      if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                              return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                          }
                      }
                      break;
              }
          }
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
          if (unlikely(Py_SIZE(x) < 0)) {
              goto raise_neg_overflow;
          }
  #else
          {
              int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
              if (unlikely(result < 0))
                  return (int) -1;
              if (unlikely(result == 1))
                  goto raise_neg_overflow;
          }
  #endif
          if ((sizeof(int) <= sizeof(unsigned long))) {
              __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
          } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
              __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
          }
      } else {
  #if CYTHON_USE_PYLONG_INTERNALS
          if (__Pyx_PyLong_IsCompact(x)) {
              __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
          } else {
              const digit* digits = __Pyx_PyLong_Digits(x);
              assert(__Pyx_PyLong_DigitCount(x) > 1);
              switch (__Pyx_PyLong_SignedDigitCount(x)) {
                  case -2:
                      if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                              return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                          }
                      }
                      break;
                  case 2:
                      if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                              return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                          }
                      }
                      break;
                  case -3:
                      if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                              return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                          }
                      }
                      break;
                  case 3:
                      if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                              return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                          }
                      }
                      break;
                  case -4:
                      if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                              return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                          }
                      }
                      break;
                  case 4:
                      if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                              return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                          }
                      }
                      break;
              }
          }
  #endif
          if ((sizeof(int) <= sizeof(long))) {
              __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
          } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
              __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
          }
      }
      {
          int val;
          int ret = -1;
  #if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
          Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
              x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
          if (unlikely(bytes_copied == -1)) {
          } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
              goto raise_overflow;
          } else {
              ret = 0;
          }
  #elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
          int one = 1; int is_little = (int)*(unsigned char *)&one;
          unsigned char *bytes = (unsigned char *)&val;
          ret = _PyLong_AsByteArray((PyLongObject *)x,
                                      bytes, sizeof(val),
                                      is_little, !is_unsigned);
  #else
          PyObject *v;
          PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
          int bits, remaining_bits, is_negative = 0;
          int chunk_size = (sizeof(long) < 8) ? 30 : 62;
          if (likely(PyLong_CheckExact(x))) {
              v = __Pyx_NewRef(x);
          } else {
              v = PyNumber_Long(x);
              if (unlikely(!v)) return (int) -1;
              assert(PyLong_CheckExact(v));
          }
          {
              int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
              if (unlikely(result < 0)) {
                  Py_DECREF(v);
                  return (int) -1;
              }
              is_negative = result == 1;
          }
          if (is_unsigned && unlikely(is_negative)) {
              Py_DECREF(v);
              goto raise_neg_overflow;
          } else if (is_negative) {
              stepval = PyNumber_Invert(v);
              Py_DECREF(v);
              if (unlikely(!stepval))
                  return (int) -1;
          } else {
              stepval = v;
          }
          v = NULL;
          val = (int) 0;
          mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
          shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
          for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
              PyObject *tmp, *digit;
              long idigit;
              digit = PyNumber_And(stepval, mask);
              if (unlikely(!digit)) goto done;
              idigit = PyLong_AsLong(digit);
              Py_DECREF(digit);
              if (unlikely(idigit < 0)) goto done;
              val |= ((int) idigit) << bits;
              tmp = PyNumber_Rshift(stepval, shift);
              if (unlikely(!tmp)) goto done;
              Py_DECREF(stepval); stepval = tmp;
          }
          Py_DECREF(shift); shift = NULL;
          Py_DECREF(mask); mask = NULL;
          {
              long idigit = PyLong_AsLong(stepval);
              if (unlikely(idigit < 0)) goto done;
              remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
              if (unlikely(idigit >= (1L << remaining_bits)))
                  goto raise_overflow;
              val |= ((int) idigit) << bits;
          }
          if (!is_unsigned) {
              if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                  goto raise_overflow;
              if (is_negative)
                  val = ~val;
          }
          ret = 0;
      done:
          Py_XDECREF(shift);
          Py_XDECREF(mask);
          Py_XDECREF(stepval);
  #endif
          if (unlikely(ret))
              return (int) -1;
          return val;
      }
  raise_overflow:
      PyErr_SetString(PyExc_OverflowError,
          "value too large to convert to int");
      return (int) -1;
  raise_neg_overflow:
      PyErr_SetString(PyExc_OverflowError,
          "can't convert negative value to int");
      return (int) -1;
  }
  
/* CIntFromPy */
  static CYTHON_INLINE enum TMM::TMMParam __Pyx_PyLong_As_enum__TMM_3a__3a_TMMParam(PyObject *x) {
  #ifdef __Pyx_HAS_GCC_DIAGNOSTIC
  #pragma GCC diagnostic push
  #pragma GCC diagnostic ignored "-Wconversion"
  #endif
      const enum TMM::TMMParam neg_one = (enum TMM::TMMParam) -1, const_zero = (enum TMM::TMMParam) 0;
  #ifdef __Pyx_HAS_GCC_DIAGNOSTIC
  #pragma GCC diagnostic pop
  #endif
      const int is_unsigned = neg_one > const_zero;
      if (unlikely(!PyLong_Check(x))) {
          enum TMM::TMMParam val;
          PyObject *tmp = __Pyx_PyNumber_Long(x);
          if (!tmp) return (enum TMM::TMMParam) -1;
          val = __Pyx_PyLong_As_enum__TMM_3a__3a_TMMParam(tmp);
          Py_DECREF(tmp);
          return val;
      }
      if (is_unsigned) {
  #if CYTHON_USE_PYLONG_INTERNALS
          if (unlikely(__Pyx_PyLong_IsNeg(x))) {
              goto raise_neg_overflow;
          } else if (__Pyx_PyLong_IsCompact(x)) {
              __PYX_VERIFY_RETURN_INT(enum TMM::TMMParam, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
          } else {
              const digit* digits = __Pyx_PyLong_Digits(x);
              assert(__Pyx_PyLong_DigitCount(x) > 1);
              switch (__Pyx_PyLong_DigitCount(x)) {
                  case 2:
                      if ((8 * sizeof(enum TMM::TMMParam) > 1 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(enum TMM::TMMParam, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(enum TMM::TMMParam) >= 2 * PyLong_SHIFT)) {
                              return (enum TMM::TMMParam) (((((enum TMM::TMMParam)digits[1]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[0]));
                          }
                      }
                      break;
                  case 3:
                      if ((8 * sizeof(enum TMM::TMMParam) > 2 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(enum TMM::TMMParam, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(enum TMM::TMMParam) >= 3 * PyLong_SHIFT)) {
                              return (enum TMM::TMMParam) (((((((enum TMM::TMMParam)digits[2]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[1]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[0]));
                          }
                      }
                      break;
                  case 4:
                      if ((8 * sizeof(enum TMM::TMMParam) > 3 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(enum TMM::TMMParam, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(enum TMM::TMMParam) >= 4 * PyLong_SHIFT)) {
                              return (enum TMM::TMMParam) (((((((((enum TMM::TMMParam)digits[3]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[2]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[1]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[0]));
                          }
                      }
                      break;
              }
          }
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
          if (unlikely(Py_SIZE(x) < 0)) {
              goto raise_neg_overflow;
          }
  #else
          {
              int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
              if (unlikely(result < 0))
                  return (enum TMM::TMMParam) -1;
              if (unlikely(result == 1))
                  goto raise_neg_overflow;
          }
  #endif
          if ((sizeof(enum TMM::TMMParam) <= sizeof(unsigned long))) {
              __PYX_VERIFY_RETURN_INT_EXC(enum TMM::TMMParam, unsigned long, PyLong_AsUnsignedLong(x))
          } else if ((sizeof(enum TMM::TMMParam) <= sizeof(unsigned PY_LONG_LONG))) {
              __PYX_VERIFY_RETURN_INT_EXC(enum TMM::TMMParam, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
          }
      } else {
  #if CYTHON_USE_PYLONG_INTERNALS
          if (__Pyx_PyLong_IsCompact(x)) {
              __PYX_VERIFY_RETURN_INT(enum TMM::TMMParam, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
          } else {
              const digit* digits = __Pyx_PyLong_Digits(x);
              assert(__Pyx_PyLong_DigitCount(x) > 1);
              switch (__Pyx_PyLong_SignedDigitCount(x)) {
                  case -2:
                      if ((8 * sizeof(enum TMM::TMMParam) - 1 > 1 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(enum TMM::TMMParam, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(enum TMM::TMMParam) - 1 > 2 * PyLong_SHIFT)) {
                              return (enum TMM::TMMParam) (((enum TMM::TMMParam)-1)*(((((enum TMM::TMMParam)digits[1]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[0])));
                          }
                      }
                      break;
                  case 2:
                      if ((8 * sizeof(enum TMM::TMMParam) > 1 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(enum TMM::TMMParam, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(enum TMM::TMMParam) - 1 > 2 * PyLong_SHIFT)) {
                              return (enum TMM::TMMParam) ((((((enum TMM::TMMParam)digits[1]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[0])));
                          }
                      }
                      break;
                  case -3:
                      if ((8 * sizeof(enum TMM::TMMParam) - 1 > 2 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(enum TMM::TMMParam, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(enum TMM::TMMParam) - 1 > 3 * PyLong_SHIFT)) {
                              return (enum TMM::TMMParam) (((enum TMM::TMMParam)-1)*(((((((enum TMM::TMMParam)digits[2]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[1]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[0])));
                          }
                      }
                      break;
                  case 3:
                      if ((8 * sizeof(enum TMM::TMMParam) > 2 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(enum TMM::TMMParam, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(enum TMM::TMMParam) - 1 > 3 * PyLong_SHIFT)) {
                              return (enum TMM::TMMParam) ((((((((enum TMM::TMMParam)digits[2]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[1]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[0])));
                          }
                      }
                      break;
                  case -4:
                      if ((8 * sizeof(enum TMM::TMMParam) - 1 > 3 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(enum TMM::TMMParam, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(enum TMM::TMMParam) - 1 > 4 * PyLong_SHIFT)) {
                              return (enum TMM::TMMParam) (((enum TMM::TMMParam)-1)*(((((((((enum TMM::TMMParam)digits[3]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[2]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[1]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[0])));
                          }
                      }
                      break;
                  case 4:
                      if ((8 * sizeof(enum TMM::TMMParam) > 3 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(enum TMM::TMMParam, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(enum TMM::TMMParam) - 1 > 4 * PyLong_SHIFT)) {
                              return (enum TMM::TMMParam) ((((((((((enum TMM::TMMParam)digits[3]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[2]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[1]) << PyLong_SHIFT) | (enum TMM::TMMParam)digits[0])));
                          }
                      }
                      break;
              }
          }
  #endif
          if ((sizeof(enum TMM::TMMParam) <= sizeof(long))) {
              __PYX_VERIFY_RETURN_INT_EXC(enum TMM::TMMParam, long, PyLong_AsLong(x))
          } else if ((sizeof(enum TMM::TMMParam) <= sizeof(PY_LONG_LONG))) {
              __PYX_VERIFY_RETURN_INT_EXC(enum TMM::TMMParam, PY_LONG_LONG, PyLong_AsLongLong(x))
          }
      }
      {
          enum TMM::TMMParam val;
          int ret = -1;
  #if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
          Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
              x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
          if (unlikely(bytes_copied == -1)) {
          } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
              goto raise_overflow;
          } else {
              ret = 0;
          }
  #elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
          int one = 1; int is_little = (int)*(unsigned char *)&one;
          unsigned char *bytes = (unsigned char *)&val;
          ret = _PyLong_AsByteArray((PyLongObject *)x,
                                      bytes, sizeof(val),
                                      is_little, !is_unsigned);
  #else
          PyErr_SetString(PyExc_RuntimeError,
                          "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
          val = (enum TMM::TMMParam) -1;
  #endif
          if (unlikely(ret))
              return (enum TMM::TMMParam) -1;
          return val;
      }
  raise_overflow:
      PyErr_SetString(PyExc_OverflowError,
          "value too large to convert to enum TMM::TMMParam");
      return (enum TMM::TMMParam) -1;
  raise_neg_overflow:
      PyErr_SetString(PyExc_OverflowError,
          "can't convert negative value to enum TMM::TMMParam");
      return (enum TMM::TMMParam) -1;
  }
  
/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
  #ifdef __Pyx_HAS_GCC_DIAGNOSTIC
  #pragma GCC diagnostic push
  #pragma GCC diagnostic ignored "-Wconversion"
  #endif
      const long neg_one = (long) -1, const_zero = (long) 0;
  #ifdef __Pyx_HAS_GCC_DIAGNOSTIC
  #pragma GCC diagnostic pop
  #endif
      const int is_unsigned = neg_one > const_zero;
      if (is_unsigned) {
          if (sizeof(long) < sizeof(long)) {
              return PyLong_FromLong((long) value);
          } else if (sizeof(long) <= sizeof(unsigned long)) {
              return PyLong_FromUnsignedLong((unsigned long) value);
  #if !CYTHON_COMPILING_IN_PYPY
          } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
              return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
  #endif
          }
      } else {
          if (sizeof(long) <= sizeof(long)) {
              return PyLong_FromLong((long) value);
          } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
              return PyLong_FromLongLong((PY_LONG_LONG) value);
          }
      }
      {
          unsigned char *bytes = (unsigned char *)&value;
  #if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
          if (is_unsigned) {
              return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
          } else {
              return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
          }
  #elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
          int one = 1; int little = (int)*(unsigned char *)&one;
          return _PyLong_FromByteArray(bytes, sizeof(long),
                                       little, !is_unsigned);
  #else
          int one = 1; int little = (int)*(unsigned char *)&one;
          PyObject *from_bytes, *result = NULL, *kwds = NULL;
          PyObject *py_bytes = NULL, *order_str = NULL;
          from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
          if (!from_bytes) return NULL;
          py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
          if (!py_bytes) goto limited_bad;
          order_str = PyUnicode_FromString(little ? "little" : "big");
          if (!order_str) goto limited_bad;
          {
              PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
              if (!is_unsigned) {
                  kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                  if (!kwds) goto limited_bad;
                  if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
              }
              result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
          }
          limited_bad:
          Py_XDECREF(kwds);
          Py_XDECREF(order_str);
          Py_XDECREF(py_bytes);
          Py_XDECREF(from_bytes);
          return result;
  #endif
      }
  }
  
/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
  #ifdef __Pyx_HAS_GCC_DIAGNOSTIC
  #pragma GCC diagnostic push
  #pragma GCC diagnostic ignored "-Wconversion"
  #endif
      const int neg_one = (int) -1, const_zero = (int) 0;
  #ifdef __Pyx_HAS_GCC_DIAGNOSTIC
  #pragma GCC diagnostic pop
  #endif
      const int is_unsigned = neg_one > const_zero;
      if (is_unsigned) {
          if (sizeof(int) < sizeof(long)) {
              return PyLong_FromLong((long) value);
          } else if (sizeof(int) <= sizeof(unsigned long)) {
              return PyLong_FromUnsignedLong((unsigned long) value);
  #if !CYTHON_COMPILING_IN_PYPY
          } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
              return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
  #endif
          }
      } else {
          if (sizeof(int) <= sizeof(long)) {
              return PyLong_FromLong((long) value);
          } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
              return PyLong_FromLongLong((PY_LONG_LONG) value);
          }
      }
      {
          unsigned char *bytes = (unsigned char *)&value;
  #if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
          if (is_unsigned) {
              return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
          } else {
              return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
          }
  #elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
          int one = 1; int little = (int)*(unsigned char *)&one;
          return _PyLong_FromByteArray(bytes, sizeof(int),
                                       little, !is_unsigned);
  #else
          int one = 1; int little = (int)*(unsigned char *)&one;
          PyObject *from_bytes, *result = NULL, *kwds = NULL;
          PyObject *py_bytes = NULL, *order_str = NULL;
          from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
          if (!from_bytes) return NULL;
          py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
          if (!py_bytes) goto limited_bad;
          order_str = PyUnicode_FromString(little ? "little" : "big");
          if (!order_str) goto limited_bad;
          {
              PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
              if (!is_unsigned) {
                  kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                  if (!kwds) goto limited_bad;
                  if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
              }
              result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
          }
          limited_bad:
          Py_XDECREF(kwds);
          Py_XDECREF(order_str);
          Py_XDECREF(py_bytes);
          Py_XDECREF(from_bytes);
          return result;
  #endif
      }
  }
  
/* FormatTypeName */
  #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
  static __Pyx_TypeName
  __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
  {
      PyObject *module = NULL, *name = NULL, *result = NULL;
      #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
      name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                                 __pyx_mstate_global->__pyx_n_u_qualname);
      #else
      name = PyType_GetQualName(tp);
      #endif
      if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
      module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                                 __pyx_mstate_global->__pyx_n_u_module);
      if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
      if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
          result = name;
          name = NULL;
          goto done;
      }
      result = PyUnicode_FromFormat("%U.%U", module, name);
      if (unlikely(result == NULL)) goto bad;
    done:
      Py_XDECREF(name);
      Py_XDECREF(module);
      return result;
    bad:
      PyErr_Clear();
      if (name) {
          result = name;
          name = NULL;
      } else {
          result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__5);
      }
      goto done;
  }
  #endif
  
/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
  #ifdef __Pyx_HAS_GCC_DIAGNOSTIC
  #pragma GCC diagnostic push
  #pragma GCC diagnostic ignored "-Wconversion"
  #endif
      const long neg_one = (long) -1, const_zero = (long) 0;
  #ifdef __Pyx_HAS_GCC_DIAGNOSTIC
  #pragma GCC diagnostic pop
  #endif
      const int is_unsigned = neg_one > const_zero;
      if (unlikely(!PyLong_Check(x))) {
          long val;
          PyObject *tmp = __Pyx_PyNumber_Long(x);
          if (!tmp) return (long) -1;
          val = __Pyx_PyLong_As_long(tmp);
          Py_DECREF(tmp);
          return val;
      }
      if (is_unsigned) {
  #if CYTHON_USE_PYLONG_INTERNALS
          if (unlikely(__Pyx_PyLong_IsNeg(x))) {
              goto raise_neg_overflow;
          } else if (__Pyx_PyLong_IsCompact(x)) {
              __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
          } else {
              const digit* digits = __Pyx_PyLong_Digits(x);
              assert(__Pyx_PyLong_DigitCount(x) > 1);
              switch (__Pyx_PyLong_DigitCount(x)) {
                  case 2:
                      if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                              return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                          }
                      }
                      break;
                  case 3:
                      if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                              return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                          }
                      }
                      break;
                  case 4:
                      if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                              return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                          }
                      }
                      break;
              }
          }
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
          if (unlikely(Py_SIZE(x) < 0)) {
              goto raise_neg_overflow;
          }
  #else
          {
              int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
              if (unlikely(result < 0))
                  return (long) -1;
              if (unlikely(result == 1))
                  goto raise_neg_overflow;
          }
  #endif
          if ((sizeof(long) <= sizeof(unsigned long))) {
              __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
          } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
              __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
          }
      } else {
  #if CYTHON_USE_PYLONG_INTERNALS
          if (__Pyx_PyLong_IsCompact(x)) {
              __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
          } else {
              const digit* digits = __Pyx_PyLong_Digits(x);
              assert(__Pyx_PyLong_DigitCount(x) > 1);
              switch (__Pyx_PyLong_SignedDigitCount(x)) {
                  case -2:
                      if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                              return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                          }
                      }
                      break;
                  case 2:
                      if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                              return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                          }
                      }
                      break;
                  case -3:
                      if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                              return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                          }
                      }
                      break;
                  case 3:
                      if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                              return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                          }
                      }
                      break;
                  case -4:
                      if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                              return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                          }
                      }
                      break;
                  case 4:
                      if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                          if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                              __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                          } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                              return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                          }
                      }
                      break;
              }
          }
  #endif
          if ((sizeof(long) <= sizeof(long))) {
              __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
          } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
              __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
          }
      }
      {
          long val;
          int ret = -1;
  #if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
          Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
              x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
          if (unlikely(bytes_copied == -1)) {
          } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
              goto raise_overflow;
          } else {
              ret = 0;
          }
  #elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
          int one = 1; int is_little = (int)*(unsigned char *)&one;
          unsigned char *bytes = (unsigned char *)&val;
          ret = _PyLong_AsByteArray((PyLongObject *)x,
                                      bytes, sizeof(val),
                                      is_little, !is_unsigned);
  #else
          PyObject *v;
          PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
          int bits, remaining_bits, is_negative = 0;
          int chunk_size = (sizeof(long) < 8) ? 30 : 62;
          if (likely(PyLong_CheckExact(x))) {
              v = __Pyx_NewRef(x);
          } else {
              v = PyNumber_Long(x);
              if (unlikely(!v)) return (long) -1;
              assert(PyLong_CheckExact(v));
          }
          {
              int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
              if (unlikely(result < 0)) {
                  Py_DECREF(v);
                  return (long) -1;
              }
              is_negative = result == 1;
          }
          if (is_unsigned && unlikely(is_negative)) {
              Py_DECREF(v);
              goto raise_neg_overflow;
          } else if (is_negative) {
              stepval = PyNumber_Invert(v);
              Py_DECREF(v);
              if (unlikely(!stepval))
                  return (long) -1;
          } else {
              stepval = v;
          }
          v = NULL;
          val = (long) 0;
          mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
          shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
          for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
              PyObject *tmp, *digit;
              long idigit;
              digit = PyNumber_And(stepval, mask);
              if (unlikely(!digit)) goto done;
              idigit = PyLong_AsLong(digit);
              Py_DECREF(digit);
              if (unlikely(idigit < 0)) goto done;
              val |= ((long) idigit) << bits;
              tmp = PyNumber_Rshift(stepval, shift);
              if (unlikely(!tmp)) goto done;
              Py_DECREF(stepval); stepval = tmp;
          }
          Py_DECREF(shift); shift = NULL;
          Py_DECREF(mask); mask = NULL;
          {
              long idigit = PyLong_AsLong(stepval);
              if (unlikely(idigit < 0)) goto done;
              remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
              if (unlikely(idigit >= (1L << remaining_bits)))
                  goto raise_overflow;
              val |= ((long) idigit) << bits;
          }
          if (!is_unsigned) {
              if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                  goto raise_overflow;
              if (is_negative)
                  val = ~val;
          }
          ret = 0;
      done:
          Py_XDECREF(shift);
          Py_XDECREF(mask);
          Py_XDECREF(stepval);
  #endif
          if (unlikely(ret))
              return (long) -1;
          return val;
      }
  raise_overflow:
      PyErr_SetString(PyExc_OverflowError,
          "value too large to convert to long");
      return (long) -1;
  raise_neg_overflow:
      PyErr_SetString(PyExc_OverflowError,
          "can't convert negative value to long");
      return (long) -1;
  }
  
/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
  static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
      while (a) {
          a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
          if (a == b)
              return 1;
      }
      return b == &PyBaseObject_Type;
  }
  static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
      PyObject *mro;
      if (a == b) return 1;
      mro = a->tp_mro;
      if (likely(mro)) {
          Py_ssize_t i, n;
          n = PyTuple_GET_SIZE(mro);
          for (i = 0; i < n; i++) {
              if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                  return 1;
          }
          return 0;
      }
      return __Pyx_InBases(a, b);
  }
  static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
      PyObject *mro;
      if (cls == a || cls == b) return 1;
      mro = cls->tp_mro;
      if (likely(mro)) {
          Py_ssize_t i, n;
          n = PyTuple_GET_SIZE(mro);
          for (i = 0; i < n; i++) {
              PyObject *base = PyTuple_GET_ITEM(mro, i);
              if (base == (PyObject *)a || base == (PyObject *)b)
                  return 1;
          }
          return 0;
      }
      return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
  }
  static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
      if (exc_type1) {
          return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
      } else {
          return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
      }
  }
  static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
      Py_ssize_t i, n;
      assert(PyExceptionClass_Check(exc_type));
      n = PyTuple_GET_SIZE(tuple);
      for (i=0; i<n; i++) {
          if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
      }
      for (i=0; i<n; i++) {
          PyObject *t = PyTuple_GET_ITEM(tuple, i);
          if (likely(PyExceptionClass_Check(t))) {
              if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
          } else {
          }
      }
      return 0;
  }
  static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
      if (likely(err == exc_type)) return 1;
      if (likely(PyExceptionClass_Check(err))) {
          if (likely(PyExceptionClass_Check(exc_type))) {
              return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
          } else if (likely(PyTuple_Check(exc_type))) {
              return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
          } else {
          }
      }
      return PyErr_GivenExceptionMatches(err, exc_type);
  }
  static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
      assert(PyExceptionClass_Check(exc_type1));
      assert(PyExceptionClass_Check(exc_type2));
      if (likely(err == exc_type1 || err == exc_type2)) return 1;
      if (likely(PyExceptionClass_Check(err))) {
          return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
      }
      return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
  }
  #endif
  
/* GetRuntimeVersion */
  #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
  void __Pyx_init_runtime_version(void) {
      if (__Pyx_cached_runtime_version == 0) {
          const char* rt_version = Py_GetVersion();
          unsigned long version = 0;
          unsigned long factor = 0x01000000UL;
          unsigned int digit = 0;
          int i = 0;
          while (factor) {
              while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                  digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                  ++i;
              }
              version += factor * digit;
              if (rt_version[i] != '.')
                  break;
              digit = 0;
              factor >>= 8;
              ++i;
          }
          __Pyx_cached_runtime_version = version;
      }
  }
  #endif
  static unsigned long __Pyx_get_runtime_version(void) {
  #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
      return Py_Version & ~0xFFUL;
  #else
      return __Pyx_cached_runtime_version;
  #endif
  }
  
/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
      const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
      if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
          return 0;
      if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
          return 1;
      {
          char message[200];
          PyOS_snprintf(message, sizeof(message),
                        "compile time Python version %d.%d "
                        "of module '%.100s' "
                        "%s "
                        "runtime version %d.%d",
                         (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                         __Pyx_MODULE_NAME,
                         (allow_newer) ? "was newer than" : "does not match",
                         (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
         );
          return PyErr_WarnEx(NULL, message, 1);
      }
  }
  
/* NewCodeObj */
  #if CYTHON_COMPILING_IN_LIMITED_API
      static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
          PyObject *exception_table = NULL;
          PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
          #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
          PyObject *version_info;
          PyObject *py_minor_version = NULL;
          #endif
          long minor_version = 0;
          PyObject *type, *value, *traceback;
          PyErr_Fetch(&type, &value, &traceback);
          #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
          minor_version = 11;
          #else
          if (!(version_info = PySys_GetObject("version_info"))) goto end;
          if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
          minor_version = PyLong_AsLong(py_minor_version);
          Py_DECREF(py_minor_version);
          if (minor_version == -1 && PyErr_Occurred()) goto end;
          #endif
          if (!(types_module = PyImport_ImportModule("types"))) goto end;
          if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
          if (minor_version <= 7) {
              (void)p;
              result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                            c, n, v, fn, name, fline, lnos, fv, cell);
          } else if (minor_version <= 10) {
              result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                            c, n, v, fn, name, fline, lnos, fv, cell);
          } else {
              if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
              result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                            c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
          }
      end:
          Py_XDECREF(code_type);
          Py_XDECREF(exception_table);
          Py_XDECREF(types_module);
          if (type) {
              PyErr_Restore(type, value, traceback);
          }
          return result;
      }
  #elif PY_VERSION_HEX >= 0x030B0000
    static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                           PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                           PyObject *fv, PyObject *cell, PyObject* fn,
                                           PyObject *name, int fline, PyObject *lnos) {
      PyCodeObject *result;
      result =
        #if PY_VERSION_HEX >= 0x030C0000
          PyUnstable_Code_NewWithPosOnlyArgs
        #else
          PyCode_NewWithPosOnlyArgs
        #endif
          (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
      #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030c00A1
      if (likely(result))
          result->_co_firsttraceable = 0;
      #endif
      return result;
    }
  #elif !CYTHON_COMPILING_IN_PYPY
    #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
            PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #else
    #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
            PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #endif
  static PyObject* __Pyx_PyCode_New(
          const __Pyx_PyCode_New_function_description descr,
          PyObject * const *varnames,
          PyObject *filename,
          PyObject *funcname,
          PyObject *line_table,
          PyObject *tuple_dedup_map
  ) {
      PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL;
      Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
      PyObject *varnames_tuple = PyTuple_New(var_count);
      if (unlikely(!varnames_tuple)) return NULL;
      for (Py_ssize_t i=0; i < var_count; i++) {
          Py_INCREF(varnames[i]);
          if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
      }
      #if CYTHON_COMPILING_IN_LIMITED_API
      varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
      if (!varnames_tuple_dedup) {
          if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
          varnames_tuple_dedup = varnames_tuple;
      }
      #else
      varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
      if (unlikely(!varnames_tuple_dedup)) goto done;
      #endif
      #if CYTHON_AVOID_BORROWED_REFS
      Py_INCREF(varnames_tuple_dedup);
      #endif
      if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL && !CYTHON_COMPILING_IN_GRAAL) {
          Py_ssize_t line_table_length = __Pyx_PyBytes_GET_SIZE(line_table);
          #if !CYTHON_ASSUME_SAFE_SIZE
          if (unlikely(line_table_length == -1)) goto done;
          #endif
          Py_ssize_t code_len = (line_table_length * 2 + 4) & ~3LL;
          code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
          if (unlikely(!code_bytes)) goto done;
          char* c_code_bytes = PyBytes_AsString(code_bytes);
          if (unlikely(!c_code_bytes)) goto done;
          memset(c_code_bytes, 0, (size_t) code_len);
      }
      code_obj = (PyObject*) __Pyx__PyCode_New(
          (int) descr.argcount,
          (int) descr.num_posonly_args,
          (int) descr.num_kwonly_args,
          (int) descr.nlocals,
          0,
          (int) descr.flags,
          code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
          __pyx_mstate_global->__pyx_empty_tuple,
          __pyx_mstate_global->__pyx_empty_tuple,
          varnames_tuple_dedup,
          __pyx_mstate_global->__pyx_empty_tuple,
          __pyx_mstate_global->__pyx_empty_tuple,
          filename,
          funcname,
          (int) descr.first_line,
          (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table) ? line_table : __pyx_mstate_global->__pyx_empty_bytes
      );
  done:
      Py_XDECREF(code_bytes);
      #if CYTHON_AVOID_BORROWED_REFS
      Py_XDECREF(varnames_tuple_dedup);
      #endif
      Py_DECREF(varnames_tuple);
      return code_obj;
  }
  
/* DecompressString */
  static PyObject *__Pyx_DecompressString(const char *s, Py_ssize_t length, int algo) {
      PyObject *module = NULL, *decompress, *compressed_bytes, *decompressed;
      const char* module_name = algo == 3 ? "compression.zstd" : algo == 2 ? "bz2" : "zlib";
      PyObject *methodname = PyUnicode_FromString("decompress");
      if (unlikely(!methodname)) return NULL;
      #if __PYX_LIMITED_VERSION_HEX >= 0x030e0000
      if (algo == 3) {
          PyObject *fromlist = Py_BuildValue("[O]", methodname);
          if (unlikely(!fromlist)) goto bad;
          module = PyImport_ImportModuleLevel("compression.zstd", NULL, NULL, fromlist, 0);
          Py_DECREF(fromlist);
      } else
      #endif
          module = PyImport_ImportModule(module_name);
      if (unlikely(!module)) goto import_failed;
      decompress = PyObject_GetAttr(module, methodname);
      if (unlikely(!decompress)) goto import_failed;
      {
          #ifdef __cplusplus
              char *memview_bytes = const_cast<char*>(s);
          #else
              #if defined(__clang__)
                #pragma clang diagnostic push
                #pragma clang diagnostic ignored "-Wcast-qual"
              #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
                #pragma GCC diagnostic push
                #pragma GCC diagnostic ignored "-Wcast-qual"
              #endif
              char *memview_bytes = (char*) s;
              #if defined(__clang__)
                #pragma clang diagnostic pop
              #elif !defined(__INTEL_COMPILER) && defined(__GNUC__)
                #pragma GCC diagnostic pop
              #endif
          #endif
          #if CYTHON_COMPILING_IN_LIMITED_API && !defined(PyBUF_READ)
          int memview_flags = 0x100;
          #else
          int memview_flags = PyBUF_READ;
          #endif
          compressed_bytes = PyMemoryView_FromMemory(memview_bytes, length, memview_flags);
      }
      if (unlikely(!compressed_bytes)) {
          Py_DECREF(decompress);
          goto bad;
      }
      decompressed = PyObject_CallFunctionObjArgs(decompress, compressed_bytes, NULL);
      Py_DECREF(compressed_bytes);
      Py_DECREF(decompress);
      Py_DECREF(module);
      Py_DECREF(methodname);
      return decompressed;
  import_failed:
      PyErr_Format(PyExc_ImportError,
          "Failed to import '%.20s.decompress' - cannot initialise module strings. "
          "String compression was configured with the C macro 'CYTHON_COMPRESS_STRINGS=%d'.",
          module_name, algo);
  bad:
      Py_XDECREF(module);
      Py_DECREF(methodname);
      return NULL;
  }
  
#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return __Pyx_NewRef(b ? Py_True: Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


  /* MultiPhaseInitModuleState */
  #if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
  #ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
  #if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
    #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
  #else
    #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
  #endif
  #endif
  #if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
  #error "Module state with PEP489 requires atomics. Currently that's one of\
   C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
  #endif
  #if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
  #define __Pyx_ModuleStateLookup_Lock()
  #define __Pyx_ModuleStateLookup_Unlock()
  #elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
  static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
  #define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
  #define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
  #elif defined(__cplusplus) && __cplusplus >= 201103L
  #include <mutex>
  static std::mutex __Pyx_ModuleStateLookup_mutex;
  #define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
  #define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
  #elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
  #include <threads.h>
  static mtx_t __Pyx_ModuleStateLookup_mutex;
  static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
  static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
      mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
  }
  #define __Pyx_ModuleStateLookup_Lock()\
    call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
    mtx_lock(&__Pyx_ModuleStateLookup_mutex)
  #define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
  #elif defined(HAVE_PTHREAD_H)
  #include <pthread.h>
  static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
  #define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
  #define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
  #elif defined(_WIN32)
  #include <Windows.h>  // synchapi.h on its own doesn't work
  static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
  #define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
  #define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
  #else
  #error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
   Requires C standard >= C11, or C++ standard >= C++11,\
   or pthreads, or the Windows 32 API, or Python >= 3.13."
  #endif
  typedef struct {
      int64_t id;
      PyObject *module;
  } __Pyx_InterpreterIdAndModule;
  typedef struct {
      char interpreter_id_as_index;
      Py_ssize_t count;
      Py_ssize_t allocated;
      __Pyx_InterpreterIdAndModule table[1];
  } __Pyx_ModuleStateLookupData;
  #define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
  #if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
  static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
  #endif
  #if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
  static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
  #else
  static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
  #endif
  static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
          __Pyx_InterpreterIdAndModule* table,
          Py_ssize_t count,
          int64_t interpreterId) {
      __Pyx_InterpreterIdAndModule* begin = table;
      __Pyx_InterpreterIdAndModule* end = begin + count;
      if (begin->id == interpreterId) {
          return begin;
      }
      while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
          __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
          if (halfway->id == interpreterId) {
              return halfway;
          }
          if (halfway->id < interpreterId) {
              begin = halfway;
          } else {
              end = halfway;
          }
      }
      for (; begin < end; ++begin) {
          if (begin->id >= interpreterId) return begin;
      }
      return begin;
  }
  static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
      int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
      if (interpreter_id == -1) return NULL;
  #if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
      __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
      {
          __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
          if (likely(data)) {
              __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
              if (likely(data == new_data)) {
                  goto read_finished;
              }
          }
          __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
          __Pyx_ModuleStateLookup_Lock();
          __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
          data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
          __Pyx_ModuleStateLookup_Unlock();
      }
    read_finished:;
  #else
      __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
  #endif
      __Pyx_InterpreterIdAndModule* found = NULL;
      if (unlikely(!data)) goto end;
      if (data->interpreter_id_as_index) {
          if (interpreter_id < data->count) {
              found = data->table+interpreter_id;
          }
      } else {
          found = __Pyx_State_FindModuleStateLookupTableLowerBound(
              data->table, data->count, interpreter_id);
      }
    end:
      {
          PyObject *result=NULL;
          if (found && found->id == interpreter_id) {
              result = found->module;
          }
  #if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
          __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
  #endif
          return result;
      }
  }
  #if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
  static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
      while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
  }
  #else
  #define __Pyx_ModuleStateLookup_wait_until_no_readers()
  #endif
  static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
      Py_ssize_t to_allocate = (*old_data)->allocated;
      while (to_allocate <= interpreter_id) {
          if (to_allocate == 0) to_allocate = 1;
          else to_allocate *= 2;
      }
      __Pyx_ModuleStateLookupData *new_data = *old_data;
      if (to_allocate != (*old_data)->allocated) {
           new_data = (__Pyx_ModuleStateLookupData *)realloc(
              *old_data,
              sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
          if (!new_data) {
              PyErr_NoMemory();
              return -1;
          }
          for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
              new_data->table[i].id = i;
              new_data->table[i].module = NULL;
          }
          new_data->allocated = to_allocate;
      }
      new_data->table[interpreter_id].module = module;
      if (new_data->count < interpreter_id+1) {
          new_data->count = interpreter_id+1;
      }
      *old_data = new_data;
      return 0;
  }
  static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
      __Pyx_InterpreterIdAndModule *read = data->table;
      __Pyx_InterpreterIdAndModule *write = data->table;
      __Pyx_InterpreterIdAndModule *end = read + data->count;
      for (; read<end; ++read) {
          if (read->module) {
              write->id = read->id;
              write->module = read->module;
              ++write;
          }
      }
      data->count = write - data->table;
      for (; write<end; ++write) {
          write->id = 0;
          write->module = NULL;
      }
      data->interpreter_id_as_index = 0;
  }
  static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
      int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
      if (interpreter_id == -1) return -1;
      int result = 0;
      __Pyx_ModuleStateLookup_Lock();
  #if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
      __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
              __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
  #else
      __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
  #endif
      __Pyx_ModuleStateLookupData *new_data = old_data;
      if (!new_data) {
          new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
          if (!new_data) {
              result = -1;
              PyErr_NoMemory();
              goto end;
          }
          new_data->allocated = 1;
          new_data->interpreter_id_as_index = 1;
      }
      __Pyx_ModuleStateLookup_wait_until_no_readers();
      if (new_data->interpreter_id_as_index) {
          if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
              result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
              goto end;
          }
          __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
      }
      {
          Py_ssize_t insert_at = 0;
          {
              __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                  new_data->table, new_data->count, interpreter_id);
              assert(lower_bound);
              insert_at = lower_bound - new_data->table;
              if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                  lower_bound->module = module;
                  goto end;  // already in table, nothing more to do
              }
          }
          if (new_data->count+1 >= new_data->allocated) {
              Py_ssize_t to_allocate = (new_data->count+1)*2;
              new_data =
                  (__Pyx_ModuleStateLookupData*)realloc(
                      new_data,
                      sizeof(__Pyx_ModuleStateLookupData) +
                      (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
              if (!new_data) {
                  result = -1;
                  new_data = old_data;
                  PyErr_NoMemory();
                  goto end;
              }
              new_data->allocated = to_allocate;
          }
          ++new_data->count;
          int64_t last_id = interpreter_id;
          PyObject *last_module = module;
          for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
              int64_t current_id = new_data->table[i].id;
              new_data->table[i].id = last_id;
              last_id = current_id;
              PyObject *current_module = new_data->table[i].module;
              new_data->table[i].module = last_module;
              last_module = current_module;
          }
      }
    end:
  #if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
      __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
  #else
      __Pyx_ModuleStateLookup_data = new_data;
  #endif
      __Pyx_ModuleStateLookup_Unlock();
      return result;
  }
  static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
      int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
      if (interpreter_id == -1) return -1;
      __Pyx_ModuleStateLookup_Lock();
  #if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
      __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
              __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
  #else
      __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
  #endif
      if (data->interpreter_id_as_index) {
          if (interpreter_id < data->count) {
              data->table[interpreter_id].module = NULL;
          }
          goto done;
      }
      {
          __Pyx_ModuleStateLookup_wait_until_no_readers();
          __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
              data->table, data->count, interpreter_id);
          if (!lower_bound) goto done;
          if (lower_bound->id != interpreter_id) goto done;
          __Pyx_InterpreterIdAndModule *end = data->table+data->count;
          for (;lower_bound<end-1; ++lower_bound) {
              lower_bound->id = (lower_bound+1)->id;
              lower_bound->module = (lower_bound+1)->module;
          }
      }
      --data->count;
      if (data->count == 0) {
          free(data);
          data = NULL;
      }
    done:
  #if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
      __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
  #else
      __Pyx_ModuleStateLookup_data = data;
  #endif
      __Pyx_ModuleStateLookup_Unlock();
      return 0;
  }
  #endif
  
/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
